# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.

from types import NoneType

from django.conf import settings
from django.http import HttpResponse
from django.utils.translation import ugettext as _

from sqlalchemy import and_, or_, func
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import join
from sqlalchemy.orm.collections import InstrumentedList
from sqlalchemy.orm.exc import NoResultFound

from basesoft.web.core.util.controllers import AbstractController, AbstractStateController, transaction_decorator, StateManager,\
    LicenseConfigController
from basesoft.web.contrib.license.models import Group, License
from basesoft.web.core.exceptions import PublicUserException, SecurityException, SystemException,\
    PublicSystemException
from basesoft.web.contrib.i18n.models import Country, Locale, DomainCountry
from basesoft.web.contrib.account.controllers import AccountController
from basesoft.web.contrib.contact.controllers import SimpleContactController, ImportContactController
from basesoft.web.contrib.notification.controllers import NotificationController
from basesoft.web.contrib.mailing.controllers import MailBaseController, AbstractMailStateController,\
    BaseMailingStateController
from basesoft.web.contrib.mailing.models import Mail
from basesoft.web.contrib.contact.models import Contact, Person, ContactGroup


from iqportal.tnt.core.controllers import TntController
from iqportal.iqube.core.controllers import DataSourceController
from iqportal.iqube.core.exceptions import IQubeSystemException
from iqportal.tnt.clm.models import TntCLMService, TntCLMWave, TntCLMScaleAnswer, TntCLMScaleAnswerGroup, \
    TntCLMScaleAnswerLocale, TntCLMWaveAttributeLocale, TntCLMWaveAttributeCategoryLocale, TntCLMWaveAnswerLocale, \
    TntCLMWaveScaleAnswerLocale, TntCLMWaveType, TntCLMWaveAttribute, TntCLMWaveLocale, TntCLMWaveLocaleRevision, \
    TntCLMLocale, TntCLMAttribute, TntCLMAttributeLocale, TntCLMAttributeAnswer, \
    TntCLMAttributeAnswerLocale, TntCLMWaveAttributeCategory, TntCLMWaveAnswer, TntCLMWaveScaleAnswer,\
    TntCLMAccount, TntCLMContact, TntCLMWaveMailLocale, TntCLMNotification, TntCLMWaveMailing,\
    TntCLMCustomer, TntCLMCompany
from iqportal.tnt.core.models import TntDepot, TntDepotAlias
from basesoft.web.contrib.file.models import ImageFile

class TntCLMAccountController(AbstractController):

    @transaction_decorator
    def save_account_from_state(self, state, geolevels, locales, account_id=None):
        """
        Save a TNT CLM account including it's linked geolevels & locales.
        """
        ac = AccountController(self.sa_session, account_id=self.get_account().id)
        lc = TntCLMLocaleController(self.sa_session, account_id=self.get_account().id)
        cc = TntController(self.sa_session, account_id=self.get_account().id)

        account = ac.save_account_from_state(state, account_id, polymorphic_type=TntCLMAccount)
        self.sa_session.expire(account)

        if locales is not None and len(locales) > 0:
            account.locales = lc.get_locales(locales)
        else:
            account.locales = []
        if geolevels is not None and len(geolevels) > 0:
            account.geolevels = cc.get_geolevels(geolevels)
        else:
            account.geolevels = []
        self.sa_session.add(account)

        # mark the datasource as stale
        try:
            ds_controller = DataSourceController(self.sa_session, account_id=self.account_id)
            ds_controller.set_stale('%s.TNT-CLM' % settings.SETTINGS_MODULE)
        except IQubeSystemException:
            # A failure on setting stale shouldnt prevent an account from being created.
            # This is already logged in DS controller.
            pass

        return account

    def get_account_state(self, id):
        """
        This returns the tnt account state
        """
        ac = AccountController(self.sa_session, account_id=self.get_account().id)
        if isinstance(id, TntCLMAccount):
            account = id
        else:
            account = ac.get_accounts(id, TntCLMAccount)
            if len(account) == 0:
                raise SecurityException('Invalid tnt account %s' % id)
            account = account[0]
        rv = ac.get_account_state(account)
        rv['account']['locales'] = [l.id for l in account.locales]
        rv['account']['geolevels'] = [l.id for l in account.geolevels]
        return rv

class TntCLMLocaleController(AbstractController):
    """
    Contains methods for all tnt locales objects. Should be extended by all tnt controllers which need to save, load
    object locales.
    """
    ACCOUNT_LOCALES = None

    def set_account_locale_ids(self, ids=None):
        """
        TODO: use this to set the ids from the session if available, so we dont have to query this every time.
        """
        account = self._get_account()
        # only CLM accounts filter locales.
        if not isinstance(account, TntCLMAccount):
            self.ACCOUNT_LOCALES = False
            return

        if ids is not None:
            self.ACCOUNT_LOCALES = self.sa_session.query(TntCLMLocale).filter(TntCLMLocale.id.in_(ids)).all()
            return

        filter = self.get_query_filter_on_locales()
        self.ACCOUNT_LOCALES = self.sa_session.query(TntCLMLocale).filter(filter).all()

    def get_query_filter_on_locales(self, only_active=True):
        """
        Tnt privileges checks on locales.
        """
        account = self._get_account()
        if not isinstance(account, TntCLMAccount):
            if only_active:
                return TntCLMLocale.active==True
            return None

        countries = None
        locales = None
        filter = None

        if len(account.locales) == 0 and len(account.geolevels) == 0 and len(account.services) == 0:
            raise SecurityException('User without any locales or geolevels cannot access requested resource')

        if len(account.geolevels) != 0:
            countries = [c.id for c in account.countries]

        if len(account.locales) != 0:
            locales = [l.id for l in account.locales]

        if len(account.services) != 0:
            if locales is None:
                locales = []
            for service in account.services:
                locales = locales + [l for l in service.locales]

        if countries is not None:
            filter = TntCLMLocale.country_id.in_(countries)
        if locales is not None:
            if filter is not None:
                filter = or_(filter, TntCLMLocale.id.in_(locales))
            else:
                filter = TntCLMLocale.id.in_(locales)

        if not only_active:
            return filter

        if filter is None:
            filter = TntCLMLocale.active==True
        else:
            filter = and_(filter, TntCLMLocale.active==True)
        return filter

    def get_clm_locales_by(self, field=None, values=None, only_active=True):
        """
        """
        if self.ACCOUNT_LOCALES is None:
            self.set_account_locale_ids()

        if values is not None:
            if not isinstance(values, list):
                values = [values]
            if field == 'id':
                values = [int(v) for v in values]

        if self.ACCOUNT_LOCALES is not False:
            if field is None or values is None:
                return self.ACCOUNT_LOCALES
            assert field in ['label', 'id']
            return [l for l in self.ACCOUNT_LOCALES if getattr(l, field) in values]

        if field is not None:
            assert field in ['label', 'id']
            if values is not None:
                filter = getattr(TntCLMLocale, field).in_(values)
                if only_active:
                    filter = and_(filter, TntCLMLocale.active==True)
                return self.sa_session.query(TntCLMLocale).filter(filter).order_by(TntCLMLocale.label).all()
        if only_active:
            return self.sa_session.query(TntCLMLocale).filter(TntCLMLocale.active==True).order_by(TntCLMLocale.label).all()
        return self.sa_session.query(TntCLMLocale).order_by(TntCLMLocale.label).all()


    def get_locale_by_str(self, locale_str):
        """
        Returns a clm Locale object for the given locale_str.
        """
        locs = self.get_clm_locales_by('label', locale_str)
        if len(locs) > 0:
            return locs[0]
        return None

    def get_locales(self, id=None, only_active=True):
        """
        """
        return self.get_clm_locales_by('id', id, only_active)

    def get_domain_locales(self, **kwargs):
        """
        This is so the mailing won't fail (because parent calls this method)
        """
        return self.get_locales()

    def get_services(self, id=None):
        if id is not None:
            return self.get_services_by('id', id)
        return self.get_services_by()

    def get_services_by(self, field=None, values=None):
        if field is not None:
            assert field in ['label', 'id']
            if values is not None:
                if not isinstance(values, list):
                    values = [values]
                filter = getattr(TntCLMService, field).in_(values)
                return self.sa_session.query(TntCLMService).filter(filter).order_by(TntCLMService.label).all()
        return self.sa_session.query(TntCLMService).order_by(TntCLMService.label).all()

    @transaction_decorator
    def save_or_update_locale(self, id, label, description, language_code, direction, country_id, locale_label=None,
                              location=None, location_2015=None, service=None, active=True):

        if id is None:
            self.assert_permissions('create_tnt_clm_locale')
            locale = TntCLMLocale()
        else:
            self.assert_permissions('update_tnt_clm_locale')
            locale = self.get_locales(id, False).pop()
        if location is not None:
            assert location in ['D', 'I']

        locale.label = label
        locale.description = description
        locale.language_code = language_code
        locale.direction = direction
        locale.active = active
        locale.location = location
        locale.location_2015 = location_2015
        locale.country = self.sa_session.query(Country).filter(Country.id==country_id).one()
        if locale_label is not None:
            locale.locale = self.sa_session.query(Locale).filter(Locale.locale==locale_label).one()
        if service is not None:
            locale.service = self._save_service(service)

        # if a locale was set to inactive, need to check if there is any account that has it as config
        if id is not None and not active:
            from basesoft.web.contrib.account.models import AccountConfig, Config
            configs = ['tnt.clm.translate.locale.right', 'tnt.clm.translate.locale.left',
                       'tnt.clm.mail.translate.locale.right', 'tnt.clm.mail.translate.locale.left']
            account_configs = self.sa_session.query(AccountConfig).join(Config).filter(
                    and_(Config.name.in_(configs), AccountConfig.value==id)).all()
            for ac in account_configs:
                self.sa_session.delete(ac)

        self.sa_session.add(locale)
        self.sa_session.flush()
        return locale

    @transaction_decorator
    def _save_service(self, label):
        services = self.get_services_by('label', label)
        if len(services)!=0:
            return services.pop()
        service = TntCLMService()
        service.label = label
        self.sa_session.add(service)
        return service

class AbstractTntStateController(AbstractStateController):
    """
    """

    _locale_name_column = "label"

    def __init__(self, session, manager):
        """
        tnt abstract controller should not use the iqportal locale controller, but the tnt one
        """
        AbstractStateController.__init__(self, session, manager)

        self.locale_controller = TntCLMLocaleController(session, manager.locale, manager.account_id)

class AbstractTntCLMMailStateController(AbstractMailStateController):
    """
    This layer uses the proper locale controller for the tnt clm mails
    """
    MailLocaleClass = TntCLMWaveMailLocale
    _locale_name_column = "label"
    _locale_attributes = ['subject', 'content', 'link_label']

    def __init__(self, session, manager):
        """
        this controller should not use the iqportal locale controller, but the tnt one
        """
        AbstractMailStateController.__init__(self, session, manager)
        self.locale_controller = TntCLMLocaleController(session, manager.locale, manager.account_id)

    def load_state(self, mail):
        """
        Private load method. This is called from a load_by_* method
        This returns a state dict
        """

        assert isinstance(mail, Mail)
        self.assert_permissions(*self.VIEW_ACCESS_PERMISSIONS)

        state = {
            'mail': {
                'objectName': unicode(mail.polymorphic_type),
                'dbID': mail.id,
                'is_wiki': mail.is_wiki,
                'name': mail.name,
                'mail_from_text': mail.mail_from_text,
                'locales': self._load_locales(mail, ['status']),
                'fallbackLocale': {
                    'value': unicode(mail.fallback_locale.label),
                    'name': unicode(mail.fallback_locale.description)
                },
            }
        }

        return state


class TntCLMMailBaseController(MailBaseController):
    """
    Layer on top of the controller for the mailing app, that handles Tnt CLM mails and uses the TntCLMLocale
    """

    MailLocaleClass = TntCLMWaveMailLocale
    LocaleClass = TntCLMLocale

    def mail_is_approved(self, mail_id):
        """
        Checks if a mail has all locales approved
        """
        q = self.sa_session.query(TntCLMWaveMailLocale).filter(and_(TntCLMWaveMailLocale.mail_id==mail_id,\
            TntCLMWaveMailLocale.status!=TntCLMWaveMailLocale.STATUS_APPROVED)).all()
        if len(q) == 0:
            return True
        return False

    @transaction_decorator
    def set_mail_translation_status(self, mail_id, locale_id, status, reason=None):
        """
        Changes status for a wave mail translation
        """
        if status == TntCLMWaveMailLocale.STATUS_OPEN:
            raise PublicUserException(_('Cannot change mail status to OPEN'))
        # status of the last transaction is the one taken into account
        ml = self.get_mail_locales(mail_id, locale_id)
        if len(ml) == 0:
            raise SecurityException("Invalid translation (mail-locale)")
        ml = ml[0]
        if status == TntCLMWaveMailLocale.STATUS_PENDING:
            self.assert_any_permissions('create_tnt_wave_mail', 'update_tnt_wave_mail')
            if ml.status not in [TntCLMWaveMailLocale.STATUS_OPEN, TntCLMWaveMailLocale.STATUS_REJECTED]:
                raise PublicUserException(_('This mail translation is not OPEN or REJECTED'))
            ml.status = status
        if status == TntCLMWaveMailLocale.STATUS_APPROVED or status == TntCLMWaveMailLocale.STATUS_REJECTED:
            self.assert_permissions('approve_tnt_clm_wave_mail_translations')
            # always last status should be pending cuz once that is set than translations cannot be edited anymore
            if status == TntCLMWaveMailLocale.STATUS_APPROVED and \
                ml.status != TntCLMWaveMailLocale.STATUS_PENDING:
                raise PublicUserException(_('There is not request for approval for this mail translation'))
            if status == TntCLMWaveMailLocale.STATUS_REJECTED and \
                ml.status not in [TntCLMWaveMailLocale.STATUS_PENDING,
                                  TntCLMWaveMailLocale.STATUS_APPROVED]:
                raise PublicUserException(_('Only approved or pending mail translations can be rejected'))
            ml.status = status
            ml.rejection_reason = reason
        self.sa_session.add(ml)

    def send_mail_status_notification(self, mail_id, status, locales, rejection_reason=None):
        """
        Send notifications at different stages of mail translation approval process.
        Notifications depending on status as follows:
            * pending -> send notification to all groups that have the permission to approve a
              translation for the provided locales
            * approved -> as above + send notification to all groups that have the permission to translate
              for the provided locales. If a group has both permissions (approve and translate)
              only one notification will be sent to that group.
            * rejected -> as for status approved plus include rejection reason in the notification
        """
        if status == TntCLMWaveMailLocale.STATUS_OPEN:
            return

        mail = self.get_mails(mail_id)[0]

        text = _("Please check Wave Mail Translations Status overview.")
        # send notification to groups that have the following permissions
        permissions = []
        if status == TntCLMWaveMailLocale.STATUS_PENDING:
            permissions = ['approve_tnt_clm_wave_mail_translations']
            subject = mail.name + " - " + _("Request for approval of mail translations")
        if status in [TntCLMWaveMailLocale.STATUS_APPROVED, TntCLMWaveMailLocale.STATUS_REJECTED]:
            permissions = ['approve_tnt_clm_wave_mail_translations', 'create_tnt_wave_mail', 'update_tnt_wave_mail']
            if status == TntCLMWaveMailLocale.STATUS_APPROVED:
                subject = mail.name + " - " + _("Mail translations Approved")
            if status == TntCLMWaveMailLocale.STATUS_REJECTED:
                subject = mail.name + " - " + _("Mail translations Rejected")
                if rejection_reason is not None:
                    text = _("Reason:")+"<br>"+rejection_reason+"<br><br>"+text

        if len(permissions) == 0:
            return

        from basesoft.web.contrib.license.models import GroupPermission, Permission
        groups = self.sa_session.query(Group).join(GroupPermission).join(Permission).filter(
            Permission.name.in_(permissions)
        ).all()

        if len(groups) == 0:
            return

        nc = NotificationController(self.sa_session)
        n = nc.set_notification(subject, text, TntCLMNotification, {
            'js_module': 'tnt_clm_wave_mail_translations_status',
            'parameters': [mail_id]
        })
        nc.create_system_notification(n, groups)

    def export_mail_pdf(self, filename, mail_id, locale_id=None):
        """
        'exports' a pdf filename for the given wave_id and locale_id
        filename can either be a absolute file path but also a HttpResponse object for sending to the client

        permission checking is handled by the mail controller

        returns a filename that can be returned to the client if needed
        """

        assert isinstance(filename, (str, unicode, HttpResponse))
        assert isinstance(mail_id, (int, long))
        assert isinstance(locale_id, (int, long, NoneType))

        lc = TntCLMLocaleController(self.sa_session, account_id=self.get_account().id)

        mail = self.get_mail_by_id(mail_id)
        state = self.load_by_mail_id(mail_id)

        if locale_id is None:
            # get the fallback locale
            locale_id = mail.fallback_locale_id

        locale = lc.get_clm_locales_by('id', locale_id)[0]
        locale_label = locale.label
        current_locale = mail.get_locale_mail(locale, TntCLMWaveMailLocale)

        subject = state['mail']['locales'][locale_label]['subject']
        content = mail.get_html_mail(current_locale, {
            'wave': {
                'label': state['mail']['locales'][locale_label]['link_label'],
                'pdf': True,
            },
            'license': self.sa_session.query(License).filter(License.name == 'TNT CLM').one()
        })

        from iqportal.tnt.clm.pdf.build import TntCLMWaveMailBuilder
        r = TntCLMWaveMailBuilder(self.sa_session)
        build_state = {
            'mail_name': mail.name,
            'locale': locale_id,
            'subject': subject,
            'content': content,
        }
        r.build(filename, build_state)

        # return filename based on locale label
        filename = '%s_%s_E.pdf' % (mail.name.replace(' ', '_'), locale_label)
        return filename

class TntCLMMailingBaseController(BaseMailingStateController):
    """
    Layer on top of the controller for the mailing app, that handles Tnt CLM mailing
    """

    _instance_class = TntCLMWaveMailing
    _mail_controller_class = TntCLMMailBaseController

class WaveController(AbstractController):

    def get_waves(self, id=None):
        self.assert_permissions('view_tnt_clm_wave')
        return self.get_privileged_object_by_id(TntCLMWave, id, TntCLMWave.PRIV_VIEW, order_by=TntCLMWave.code)

    def get_locales(self, locale_id=None):
        # get the locales the current account is linked to
        account_id = None
        account = self.get_account()
        if account:
            account_id = account.id
        lc = TntCLMLocaleController(self.sa_session, account_id=account_id)
        return lc.get_locales(locale_id)

    def get_wave_types(self, id=None):
        return self.get_privileged_object_by_id(TntCLMWaveType, id, TntCLMWaveType.PRIV_VIEW)

    def get_wave_locale_revisions(self, wave_id, locale_id=None):
        acc_locale_ids = [l.id for l in self.get_locales()]
        if locale_id is not None:
            locale_id = int(locale_id)
            assert locale_id in acc_locale_ids
            acc_locale_ids = [locale_id]
        return self.sa_session.query(TntCLMWaveLocaleRevision).filter((TntCLMWaveLocaleRevision.wave_id==wave_id) &
                (TntCLMWaveLocaleRevision.locale_id.in_(acc_locale_ids))).order_by(TntCLMWaveLocaleRevision.id.desc()).all()

    def get_available_wave_locales(self, wave):
        if type(wave) != TntCLMWave:
            wave = self.get_waves(wave)
            if len(wave) == 0:
                return []
            wave = wave[0]
        acc_locale_ids = [l.id for l in self.get_locales()]
        return self.sa_session.query(TntCLMLocale).join(TntCLMWaveLocaleRevision).filter(
                and_(TntCLMWaveLocaleRevision.wave==wave, TntCLMLocale.id.in_(acc_locale_ids))).group_by(TntCLMWaveLocaleRevision.locale_id
                ).order_by(TntCLMLocale.label).all()

    def get_wave_locales_status(self, id):
        """
        returns statuses for last revision of the wave translation
        """
        lrs = self.sa_session.query(TntCLMLocale.id.label('locale_id'), TntCLMWaveLocaleRevision.status.label('status')).join(
                TntCLMWaveLocaleRevision).filter(
                and_(TntCLMWaveLocaleRevision.wave_id==id, TntCLMWaveLocaleRevision.id.in_(
                                        self.sa_session.query(func.max(TntCLMWaveLocaleRevision.id)).
                                        select_from(TntCLMWaveLocaleRevision).

                                        group_by(TntCLMWaveLocaleRevision.wave_id, TntCLMWaveLocaleRevision.locale_id)))).group_by(
                TntCLMWaveLocaleRevision.locale_id).order_by(TntCLMLocale.label).all()
        ret = {}
        for l in lrs:
            ret[l.locale_id] = l.status
        return ret

    def get_wave_default_locale(self, wave_id):
        return self.sa_session.query(TntCLMLocale).join(TntCLMWaveLocaleRevision).join(TntCLMWaveLocale).filter(
                (TntCLMWaveLocaleRevision.wave_id==wave_id) & (TntCLMWaveLocale.is_default==True)).first()

    def get_empty_wave_state(self, id):
        """
        Get an empty wave state
        """
        wave = self.get_privileged_object_by_id(TntCLMWave, id, TntCLMWave.PRIV_VIEW).pop()
        sa = []
        for answer in wave.scale_answers:
            sa.append({'id': answer.id, 'label': ''})

        attributes = []
        for a in wave.attributes:
            config = {'id': a.id, 'label': '', 'type': a.attribute.type}
            # check for categories and it's translation
            if a.category is not None:
                config['category'] = {'id': a.category.id, 'label': ''}

            # if this attribute has answers attached, check for their translations.
            if len(a.answers) > 0:
                answers = []
                for answer in a.answers:
                    answers.append({'id': answer.id, 'label': ''})
                config['answers'] = answers
            attributes.append(config)

        # return the state of this wave.
        return {
            'type': wave.type.module,
            'wave': {
                'opt_out_text': '',
                'closure_text': '',
                'header_text': '',
                'sub_header_text': '',
                'pdf_fontsize': None,
                'scale_answers': sa,
                'attributes': attributes
            }
        }

    def get_wave_schedule_data(self, id):
        """
        Returns wave schedule state
        """
        wave = self.get_waves(id).pop()
        locales_statuses = self.get_wave_locales_status(id)
        app = True
        for locale, status in locales_statuses.items():
            if status != TntCLMWaveLocaleRevision.STATUS_ACTIVE_APPROVED:
                app = False
                break
        mc = TntCLMMailingBaseController(self.sa_session, StateManager(None, self.account_id, None))

        ret = {
            'name': wave.name,
            'status': wave.status,
            'begin_date': wave.begin_date,
            'end_date': wave.end_date,
            'contact_lists': [c.id for c in wave.contact_lists],
            'mailings': [mc.load(m) for m in wave.mailings],
            'all_locales_approved': app
        }
        return ret

    def get_wave_setup(self, id):
        """
        Builds a wave state for sending to the wave designer panel (so structure same as the designer form)
        """
        locale = self.get_wave_default_locale(id)
        if locale is None:
            raise Exception("Wave %s does not have a default locale set" % id)
        wave = self.get_waves(id).pop()
        filter = (TntCLMWaveLocaleRevision.locale==locale) & (TntCLMWaveLocaleRevision.wave==wave)

        default_locale_status = self.get_wave_locales_status(id)[locale.id]

        ret = {
            'name': wave.name,
            'code': wave.code,
            'status': wave.status,
            'type': wave.type.id,
            'default_locale': locale.id,
            'default_locale_status': default_locale_status,
            'wave_attributes': []
        }
        categ = {}
        index = 0;
        for attr in wave.attributes:
            categ_id = 'none'
            categ_name = ''
            cid = None
            if attr.category is not None:
                cid = attr.category.id
                categ_id = attr.category.id
                categ_loc = self.sa_session.query(TntCLMWaveAttributeCategoryLocale).join(TntCLMWaveLocaleRevision
                            ).filter(filter & (TntCLMWaveAttributeCategoryLocale.category==attr.category)
                            ).order_by(TntCLMWaveLocaleRevision.id.desc()).first()
                categ_name = '' if categ_loc is None else categ_loc.label
            if categ.has_key(categ_id):
                ret['wave_attributes'][categ[categ_id]]['attributes'].append(attr.attribute.id)
            else:
                ret['wave_attributes'].append({'id': cid, 'category_id': cid, 'category_name': categ_name,
                                               'attributes': [attr.attribute.id]})
                categ[categ_id] = index
                index = index + 1;
        return ret

    def get_wave_state(self, id, locale_id=None, locale_revision_id=None, load_empty=False, include_license_logo=False):
        """
        Get the wave state for the supplied id and locale.
        """
        if isinstance(id, TntCLMWave):
            wave = id
        else:
            wave = self.get_privileged_object_by_id(TntCLMWave, id, TntCLMWave.PRIV_VIEW).pop()
        id = wave.id

        if locale_id is not None:
            if isinstance(locale_id, TntCLMLocale):
                locale = locale_id
            else:
                locale = self.get_locales(int(locale_id))
                if len(locale) == 0:
                    raise Exception("No access for locale id %s " % locale_id)
                locale = locale[0]
        else:
            locale = self.get_wave_default_locale(id)
            if locale is None:
                raise Exception("Wave %s does not have a default locale set" % id)
            # if user does not have permission on default locale, get state for one of the locales he does have permission to
            locales = self.get_locales()
            lids = [l.id for l in locales]
            if locale.id not in lids:
                locale = locales[0]
        direction = locale.direction
        locale_id = locale.id

        wave_type = wave.type_id
        opt_out_text = ''
        header_text = ''
        sub_header_text = ''
        closure_text = ''
        pdf_fontsize = None

        locale_revision_author = None
        locale_revision_date = None

        filter = (TntCLMWaveLocaleRevision.locale==locale) & (TntCLMWaveLocaleRevision.wave==wave)
        if locale_revision_id is not None:
            filter = filter & (TntCLMWaveLocaleRevision.id<=locale_revision_id)
            locale_revision = self.sa_session.query(TntCLMWaveLocaleRevision).filter(
                        TntCLMWaveLocaleRevision.id==locale_revision_id).first()
        else:
            locale_revision = self.sa_session.query(TntCLMWaveLocaleRevision).filter(
                        (TntCLMWaveLocaleRevision.locale==locale) & (TntCLMWaveLocaleRevision.wave==wave)
                        ).order_by(TntCLMWaveLocaleRevision.id.desc()).first()
        if locale_revision is not None:
            locale_revision_author = locale_revision.account.name
            locale_revision_date = locale_revision.created
            locale_revision_id = locale_revision.id

        revision = self.sa_session.query(TntCLMWaveLocale).join(
                TntCLMWaveLocaleRevision).filter(filter).order_by(TntCLMWaveLocaleRevision.id.desc()).first()
        if revision is not None:
            opt_out_text = revision.opt_out_text
            header_text = revision.header_text
            sub_header_text = revision.sub_header_text
            closure_text = revision.closure_text
            pdf_fontsize = revision.pdf_fontsize


        sa = []
        for answer in wave.scale_answers:
            revision = self.sa_session.query(TntCLMWaveScaleAnswerLocale).join(
                TntCLMWaveLocaleRevision).filter(filter & (TntCLMWaveScaleAnswerLocale.wave_scale_answer==answer)
                ).order_by(TntCLMWaveLocaleRevision.id.desc()).first()
            if revision is None:
                continue
            sa.append({'id': answer.id, 'label': revision.label})

        class MockRevision:
            label = ''
            priority_trigger_at = None

        # loop all the attributes
        attributes = []
        for a in wave.attributes:
            # Check if this attribute has a translation for the locale.

            revision = self.sa_session.query(TntCLMWaveAttributeLocale).join(
                TntCLMWaveLocaleRevision).filter(filter & (TntCLMWaveAttributeLocale.wave_attribute==a)
                ).order_by(TntCLMWaveLocaleRevision.id.desc()).first()
            if revision is None:
                if not load_empty:
                    continue
                revision = MockRevision()

            label = revision.label

            #check for global priority
            if a.priority_trigger_at is not None:
                priority = a.priority_trigger_at
                pr_global = True
            else:
                priority = revision.priority_trigger_at
                pr_global = False


            # can't have an empty attribute.
            if not load_empty and label == '':
                continue

            config = {'id': a.id, 'label': label, 'type': a.attribute.type,
                      'priority_trigger_at': priority,
                      'priority_is_global': pr_global}
            # check for categories and it's translation
            if a.category is not None:
                revision = self.sa_session.query(TntCLMWaveAttributeCategoryLocale).join(
                    TntCLMWaveLocaleRevision).filter(filter & (TntCLMWaveAttributeCategoryLocale.category==a.category)
                    ).order_by(TntCLMWaveLocaleRevision.id.desc()).first()
                if revision is not None:
                    config['category'] = {'id': a.category.id, 'label': revision.label}

            # if this attribute has answers attached, check for their translations.
            if len(a.answers) > 0:
                answers = []
                for answer in a.answers:
                    revision = self.sa_session.query(TntCLMWaveAnswerLocale).join(
                        TntCLMWaveLocaleRevision).filter(filter & (TntCLMWaveAnswerLocale.wave_answer==answer)
                        ).order_by(TntCLMWaveLocaleRevision.id.desc()).first()
                    if revision is None:
                        if a.attribute.type != TntCLMAttribute.TYPE_NPS and not load_empty:
                            continue
                        revision = MockRevision()

                    answers.append({'id': answer.id, 'label': revision.label})

                if len(answers) > 0:
                    config['answers'] = answers
            attributes.append(config)

        locales_statuses = self.get_wave_locales_status(id)



        # return the state of this wave.
        rv = {
            'locale': locale_id,
            'revision': locale_revision_id,
            'revision_author': locale_revision_author,
            'revision_date': locale_revision_date,
            'type': wave.type.module,
            'direction': direction,
            'wave': {
                'type': wave_type,
                'code': wave.code,
                'name': wave.name,
                'status': wave.status,
                'opt_out_text': opt_out_text,
                'closure_text': closure_text,
                'header_text': header_text,
                'sub_header_text': sub_header_text,
                'pdf_fontsize': pdf_fontsize,
                'scale_answers': sa,
                'attributes': attributes,
                'locales_statuses': locales_statuses
            }
        }
        if not include_license_logo:
            return rv

        company_logo = None

        lcc = LicenseConfigController(self.sa_session, scope=wave.groups[0].group.license)
        img = lcc.get('basesoft.license.company.logo')
        if img is not None:
            company_logo = self.sa_session.query(ImageFile).get(img).file_hash

        rv['wave']['company_logo'] = company_logo

        return rv

    def get_wave_translations_values(self, wave_id, locale_id=None, locale_revision_id=None):
        """
        Returns state with all wave translations
        """
        state = self.get_wave_state(wave_id, locale_id, locale_revision_id)
        # single level dictionary
        values = {
            'locale': state['locale'],
            'revision': state['revision'],
            'revision_author': state['revision_author'],
            'revision_date': state['revision_date'],
            'opt_out_text': state['wave']['opt_out_text'],
            'closure_text': state['wave']['closure_text'],
            'header_text': state['wave']['header_text'],
            'sub_header_text': state['wave']['sub_header_text'],
            'pdf_fontsize': state['wave']['pdf_fontsize']
        }
        for answer in state['wave']['scale_answers']:
            values['scale_answer_%s' % answer['id']] = answer['label']

        for attr in state['wave']['attributes']:
            if attr.has_key('category'):
                values['category_%s' % attr['category']['id']] = attr['category']['label']
            values['attribute_%s' % attr['id']] = attr['label']
            if attr.has_key('answers'):
                for a in attr['answers']:
                    values['answer_%s' % a['id']] = a['label']

        return {'direction': state['direction'], 'values': values}

    @transaction_decorator
    def set_wave_status(self, wave, status):
        """
        Sets the status of wave.
        """
        wave.status = status
        self.sa_session.add(wave)

    def schedule_wave(self, id):
        """
        Set status to SCHEDULED if all requirements are ok.
        """
        # scheduling can be done always. However publication of a wave can only happen if all locales are approved.
        self.assert_permissions('schedule_tnt_clm_wave')
        waves = self.get_privileged_object_by_id(TntCLMWave, id, TntCLMWave.PRIV_SCHEDULE)
        if len(waves) == 0:
            raise PublicSystemException('No permission to schedule wave')
        wave = waves.pop()

        if wave.status == TntCLMWave.STATUS_SCHEDULED:
            raise PublicUserException('Wave already SCHEDULED')
        if wave.status != TntCLMWave.STATUS_OPEN:
            raise PublicUserException('Only OPEN waves can be set to SCHEDULED')
        if wave.begin_date is None:
            raise PublicUserException('Wave does not have a begin_date set; it cannot be set to SCHEDULED')
        self.set_wave_status(wave, TntCLMWave.STATUS_SCHEDULED)

    @transaction_decorator
    def set_wave_schedule_data(self, id, begin_date=None, end_date=None, contact_lists=None, mailings=None):
        """
        Sets the schedule information for wave: begin/end date, contact lists, mails.
        """
        self.assert_permissions('update_tnt_clm_wave')
        wave = self.get_waves(id)[0]

        if wave.status == TntCLMWave.STATUS_FINISHED:
            raise SecurityException('Cannot modify schedule information for FINISHED waves')

        if wave.status == TntCLMWave.STATUS_ONLINE and wave.begin_date != begin_date:
            raise PublicUserException(_('Wave is already ONLINE. Cannot change begin date'))

        if begin_date == '':
            begin_date = None
        if end_date == '':
            end_date = None

        if begin_date is None and wave.status == TntCLMWave.STATUS_SCHEDULED:
            raise PublicUserException(_('Cannot remove begin_date for a SCHEDULED wave'))
        wave.begin_date = begin_date
        wave.end_date = end_date

        if isinstance(contact_lists, list) and len(contact_lists) > 0:
            contact_lists = [int(c) for c in contact_lists]
            if wave.status == TntCLMWave.STATUS_ONLINE:
                for c in wave.contact_lists:
                    if c.id not in contact_lists:
                        raise SecurityException('Trying to remove contact list %s from an ONLINE wave' % c.id)
            from basesoft.web.contrib.contact.controllers import ContactController
            cc = ContactController(self.sa_session, account_id=self.get_account().id)
            cls = cc.get_contact_lists(contact_lists)
            wave.contact_lists = cls # invitees for the new contact lists will be set by a task
        else:
            if len(wave.contact_lists) > 0 and wave.status == TntCLMWave.STATUS_ONLINE:
                raise SecurityException('Trying to remove contact lists from an ONLINE wave')
            wave.contact_lists = []

        # mailings contact lists must be in wave contact lists
        if isinstance(mailings, list) and len(mailings) > 0:
            wave_cls = [c.id for c in wave.contact_lists]
            for m in mailings:
                for cl in m['contact_lists']:
                    if int(cl) not in wave_cls:
                        raise SystemException('Mailing contact list %s not linked to wave publication' % cl)
        mc = TntCLMMailingBaseController(self.sa_session, StateManager(None, self.account_id, None))
        wave = mc.save_instance_mailings(wave, 'wave', mailings)

        self.sa_session.add(wave)
        return wave

    @transaction_decorator
    def save_or_update_wave(self, id, name, code, default_locale, type_id, attributes, copy_locales):
        """
        Saves a new wave. All data provided will be saved for default_locale and original translations
        (for scale answers, attributes and attributes answers) will be copied for supplied copy_locales.
        """
        type = self.get_wave_types(type_id).pop()
        wave_attrs = {}
        if id is None:
            self.assert_permissions('create_tnt_clm_wave')
            wave = TntCLMWave()
        else:
            self.assert_permissions('update_tnt_clm_wave')
            wave = self.get_waves(id)[0]
            if wave.status in [TntCLMWave.STATUS_ONLINE, TntCLMWave.STATUS_FINISHED]:
                raise SecurityException('Cannot update ONLINE or FINISHED waves. (id %s)'% wave.id)
            for a in wave.attributes:
                wave_attrs[a.attribute_id] = a

        translations = {}

        wave.name = name
        wave.code = code
        wave.type = type
        new_attrs = []

        if id is not None:
            # cannot change wave default locale
            defl = self.get_wave_default_locale(id)
            if int(default_locale) != defl.id:
                raise SecurityException('Cannot changes wave (%s) default locale' % id)
            # if default locale is pending or approved, cannot add/remove attributes
            default_locale_status = self.get_wave_locales_status(id)[defl.id]
            if default_locale_status in [TntCLMWaveLocaleRevision.STATUS_PENDING, TntCLMWaveLocaleRevision.STATUS_APPROVED,\
                                         TntCLMWaveLocaleRevision.STATUS_ACTIVE_APPROVED]:
                self.sa_session.add(wave)
                return wave.id

        # empty translations for wave default locale: header, subheader, opt-out, closure
        # (pdf_fontsize will be None by default)
        # only when creating
        if id is None:
            for i in ['header_text', 'sub_header_text', 'opt_out_text', 'closure_text']:
                translations[i] = '';

        # wave scale answers will always be the scale answers of the wave's type available at the moment the wave is created
        # and cannot be updated after
        if id is None:
            for sa in self._get_privileged_object(TntCLMScaleAnswer, TntCLMScaleAnswer.type==type, TntCLMScaleAnswer.PRIV_VIEW):
                answer = TntCLMWaveScaleAnswer()
                answer.wave = wave
                answer.scale_answer = sa
                answer.position = sa.position
                self.sa_session.add(answer)
                self.sa_session.flush()
                translations['scale_answer_%s' % answer.id] = sa.default_label

        priorities = {}
        for a in wave.attributes:
            priorities[a.attribute_id] = a.priority_trigger_at
        wave.attributes = []
        position = 1
        for config in attributes:
            category_name = config['category_name']
            category_id = config['id']
            if category_id == '':
                category_id = None
            category = None
            if category_id is not None:
                category = self.sa_session.query(TntCLMWaveAttributeCategory).filter(
                            TntCLMWaveAttributeCategory.id==category_id).first()
            if category_id is None:
                category = TntCLMWaveAttributeCategory()
                category.wave = wave
                self.sa_session.add(category)
                self.sa_session.flush()
            translations['category_%s' % category.id] = category_name

            attribs = config['attributes']
            attrs = self.get_privileged_object_by_id(TntCLMAttribute, attribs, TntCLMAttribute.PRIV_VIEW)
            assert len(attrs) == len(attribs)
            # need to parse the original state in order to keep poritions
            scale_attributes = []
            general_attributes = []
            nps_attributes = []
            text_attributes = []
            for at in attribs:
                attr = [a for a in attrs if a.id == int(at)][0]
                new_attrs.append(attr.id)
                if attr.id in wave_attrs:
                    a = wave_attrs[attr.id]
                else:
                    a = TntCLMWaveAttribute()
                    a.attribute = attr
                a.category = category
                # if priority was already set for this wave-attribute,
                # use previously set priorities, not the ones from the pool
                if priorities.has_key(attr.id):
                    a.priority_trigger_at = priorities[attr.id]
                else:
                    a.priority_trigger_at = attr.priority_trigger_at
                # sort by type
                if a.attribute.type == TntCLMAttribute.TYPE_NPS:
                    nps_attributes.append(a)
                elif a.attribute.type == TntCLMAttribute.TYPE_SCALE:
                    scale_attributes.append(a)
                elif a.attribute.type == TntCLMAttribute.TYPE_TEXT:
                    text_attributes.append(a)
                else:
                    general_attributes.append(a)

            for a in scale_attributes + general_attributes + nps_attributes + text_attributes:
                # attribute answers and attribute default label
                # get saved only once (so only in case attrs are not already linked to this wave)
                a.position = position
                position = position + 1
                self.sa_session.add(a)
                self.sa_session.flush()
                wave.attributes.append(a)
                attr = a.attribute

                if attr not in wave_attrs:
                    translations['attribute_%s' % a.id] = attr.default_label
                    translations['attribute_priority_%s' % a.id] = attr.priority_trigger_at
                    answers = [x.answer for x in a.answers]
                    for attr_answer in attr.answers:

                        # this answer is already added for this attribute in this wave (#3143)
                        if attr_answer in answers:
                            continue

                        answer = TntCLMWaveAnswer()
                        answer.wave_attribute = a
                        answer.answer = attr_answer
                        answer.position = attr_answer.position
                        self.sa_session.add(answer)
                        self.sa_session.flush()
                        translations['answer_%s' % answer.id] = attr_answer.default_label

        # categories of the attributes that will be removed
        categories = {}
        # see if there were any attributes removed
        for key, a in wave_attrs.items():
            if key not in new_attrs:
                if a.category is not None:
                    categories[a.category.id] = a.category
                self.sa_session.delete(a)
        # now delete any category in 'categories' that does not still have any wave attributes attached
        used_categs = [a.category.id for a in wave.attributes if a.category is not None]
        for key, c in categories.items():
            if key not in used_categs:
                self.sa_session.delete(c)

        self.sa_session.add(wave)
        if id is None:
            self.save_default_privilege_object_group(wave)
        self.sa_session.flush()

        # default locale
        self.save_translation(wave.id, default_locale, translations, True)

        if id is not None:
            return wave.id

        for lid in copy_locales:
            tr = self.get_wave_translations_from_source_tables(wave.id, lid)
            if len(tr.keys()) > 0:
                self.save_translation(wave.id, lid, tr)
        return wave.id

    def get_wave_translations_from_source_tables(self, id, locale_id, preserve_translations=True):
        """
        Gets the translations from the source tables for the supplied locale and copies them to the
        wave tables.
        If preserve_translations is True, existing translations on the wave *wont* be overwritten.
        """
        locale = self.get_locales(locale_id)[0]
        wave = self.get_waves(id)[0]
        translations = {}
        if preserve_translations:
            translations = self.get_wave_translations_values(id, locale_id)['values']
            # we only need data related to objects that have original translations (scale answers, attributes and
            # attribute answers), so remove all wave directly related info.
            del translations['locale']
            del translations['opt_out_text']
            del translations['closure_text']
            del translations['header_text']
            del translations['sub_header_text']
            del translations['pdf_fontsize']
            del translations['revision']
            del translations['revision_author']
            del translations['revision_date']

        for sa in wave.scale_answers:
            if translations.has_key('scale_answer_%s' % sa.id):
                continue
            sl = self.sa_session.query(TntCLMScaleAnswerLocale).filter(
                (TntCLMScaleAnswerLocale.scale_answer==sa.scale_answer) & (TntCLMScaleAnswerLocale.locale==locale)).first()
            if sl is None:
                continue
            translations['scale_answer_%s' % sa.id] = sl.label

        for attr in wave.attributes:
            if not translations.has_key('attribute_%s' % attr.id):
                al = self.sa_session.query(TntCLMAttributeLocale).filter(
                    (TntCLMAttributeLocale.attribute==attr.attribute) & (TntCLMAttributeLocale.locale==locale)).first()
                if al is not None:
                    translations['attribute_%s' % attr.id] = al.label
                    translations['attribute_priority_%s' % attr.id] = al.priority_trigger_at
            if len(attr.answers) == 0:
                continue
            for answer in attr.answers:
                if translations.has_key('answer_%s' % answer.id):
                    continue
                al = self.sa_session.query(TntCLMAttributeAnswerLocale).filter(
                    (TntCLMAttributeAnswerLocale.answer==answer.answer) & (TntCLMAttributeAnswerLocale.locale==locale)).first()
                if al is not None:
                    translations['answer_%s' % answer.id] = al.label

        return translations

    def _get_last_attribute_locale(self, wave_attribute_id, locale_id):
        """
        Return last revision of TntCLMWaveAttributeLocale object for locale_id
        """
        q = self.sa_session.query(TntCLMWaveAttributeLocale).join(TntCLMWaveLocaleRevision).filter(
            and_(TntCLMWaveAttributeLocale.wave_attribute_id==wave_attribute_id,
                 TntCLMWaveLocaleRevision.locale_id==locale_id)).order_by(TntCLMWaveLocaleRevision.id.desc()).all()
        if len(q) == 0:
            return False
        return q[0]

    def _can_save_translation_value(self, state, oldstate, key, check_oldstate_for_empty=True):
        """
        Check if for a key we should save value or not.
        @param check_oldstate_for_empty: boolean. In some cases, when translations are '' or None they still
                                                  should be saved no matter what old state is.
        """
        if not state.has_key(key):
            return False
        if state[key] in ['', None]:
            if not check_oldstate_for_empty:
                return True
            if oldstate.has_key(key) and oldstate[key] != state[key]:
                return True
            return False
        if oldstate.has_key(key) and oldstate[key] == state[key]:
            return False
        return True

    @transaction_decorator
    def save_translation(self, wave_id, locale_id, translations, is_default=False):

        self.assert_permissions('translate_tnt_clm_wave')
        wave = self.get_waves(wave_id)[0]
        if wave.status == TntCLMWave.STATUS_FINISHED:
            raise SecurityException('Cannot update translations for FINISHED waves. (id %s)'% wave.id)

        # only PENDING translations cannot be updated
        lr = self.get_wave_locale_revisions(wave_id, locale_id)
        if len(lr) > 0:
            lr = lr[0]
            if lr.status == TntCLMWaveLocaleRevision.STATUS_PENDING:
                raise PublicUserException(_("Cannot update pending translations"))

        # TODO: add permission checks / privilege checks.
        # we need to check for PRIV_TRANSLATE but also use some default TntCLM permission check on the locale.
        controller = TntCLMLocaleController(self.sa_session, account_id=self.get_account().id)
        locale = controller.get_locales(locale_id).pop()

        # default state.
        state = self.get_empty_wave_state(wave_id)
        values = self.get_wave_translations_values(wave_id, locale)['values']

        # save revision
        locale_revision = TntCLMWaveLocaleRevision()
        locale_revision.account = self.get_account()
        locale_revision.locale = locale
        locale_revision.wave = wave
        #self.sa_session.add(locale_revision)

        if translations.has_key('pdf_fontsize') and translations['pdf_fontsize'] is not None \
        and not isinstance(translations['pdf_fontsize'], int):
            if translations['pdf_fontsize'] == '':
                translations['pdf_fontsize'] = None
            else:
                translations['pdf_fontsize'] = int(translations['pdf_fontsize'])
        # Wave locale stuff
        wt = ('header_text', 'sub_header_text', 'opt_out_text', 'closure_text', 'pdf_fontsize')
        tr = {}
        # if is_default is set, then save_translations was called when creating new wave and setting
        # its default locale. In this case all wave translations are '' or None (because wave panel does
        # not contain wave translations related info), but we still need to save them  in order to set default locale
        check_oldstate = not is_default
        for i in wt:
            if self._can_save_translation_value(translations, values, i, check_oldstate):
                tr[i] = translations[i]

        if len(tr) > 0:
            for i in wt:
                if not tr.has_key(i):
                    tr[i] = values[i]

            if not is_default:
                # we could be updating translations for default locale, so need to specify that in the wave locale table
                default_locale = self.get_wave_default_locale(wave_id)
                if locale_id == default_locale.id:
                    is_default = True
            self._save_translation(TntCLMWaveLocale, None, wave_id, locale, tr, locale_revision, is_default)

        # scale answers
        for scale_answers in state['wave']['scale_answers']:
            id = scale_answers['id']
            key = 'scale_answer_%s' % id

            if not self._can_save_translation_value(translations, values, key):
                continue
            self._save_translation(TntCLMWaveScaleAnswerLocale, 'wave_scale_answer_id', id, locale,
                                   translations[key], locale_revision)

        cats = []
        for attributes in state['wave']['attributes']:
            id = attributes['id']
            key = 'attribute_%s' % id

            tr = {}
            if self._can_save_translation_value(translations, values, key):
                tr['label'] = translations[key]

                # having attribute priority in the translations will only happen in case
                # new attributes are linked to a wave
                # otherwise, need to get attribute's priority from the last revision
                # (because a wave attribute's current state is taken from the latest revision always.
                # So even if a global priority is currently set, that might be unset so still need the priority
                # from the latest revision)
                pkey = 'attribute_priority_%s' % id
                if translations.has_key(pkey) and translations[pkey] not in [None, '']:
                    tr['priority_trigger_at'] = translations[pkey]
                else:
                    al = self._get_last_attribute_locale(id, locale_id)
                    if al and al.priority_trigger_at is not None:
                        tr['priority_trigger_at'] = al.priority_trigger_at

                self._save_translation(TntCLMWaveAttributeLocale, 'wave_attribute_id', id, locale,
                                       tr, locale_revision)

            if attributes.has_key('category') and attributes['category']['id'] not in cats:
                id = attributes['category']['id']
                key = 'category_%s' % id
                cats.append(id)

                if self._can_save_translation_value(translations, values, key):
                    self._save_translation(TntCLMWaveAttributeCategoryLocale, 'category_id', id, locale,
                                           translations[key], locale_revision)

            if attributes.has_key('answers'):
                for answer in attributes['answers']:
                    id = answer['id']
                    key = 'answer_%s' % id
                    if not self._can_save_translation_value(translations, values, key):
                        continue
                    self._save_translation(TntCLMWaveAnswerLocale, 'wave_answer_id', id, locale,
                                           translations[key], locale_revision)

    @transaction_decorator
    def _save_translation(self, obj, field, id, locale, label, locale_revision, is_default=False):
        """
        """

        instance = obj()
        if field is not None:
            setattr(instance, field, id)
        instance.locale_revision = locale_revision
        if type(label) == dict:
            for f, t in label.iteritems():
                setattr(instance, f, t)
        else:
            instance.label = label
        if is_default:
            if hasattr(instance, 'is_default'):
                instance.is_default = True
        self.sa_session.add(instance)

    @transaction_decorator
    def set_wave_attributes_priorities(self, id, locale_id, values, priority_global=False):

        if not priority_global:
            self.assert_permissions('translate_tnt_clm_wave')
            # can only set priorities for OPEN or REJECTED translations
            # check status of the last revision
            lr = self.get_wave_locale_revisions(id, locale_id)
            if len(lr) > 0:
                lr = lr[0]
                if lr.status not in [TntCLMWaveLocaleRevision.STATUS_OPEN, TntCLMWaveLocaleRevision.STATUS_REJECTED]:
                    raise PublicUserException(_("Cannot set priority for pending or approved translations"))

            locale = self.get_locales(locale_id)
            if len(locale) == 0:
                raise SecurityException('Invalid locale %s' % locale_id)
            locale = locale[0]
        else:
            self.assert_permissions('set_wave_global_attribute_priority')
            # cant set global priority if there isn't at least one translation that is open or rejected
            oneOk = False
            lrs = self.get_wave_locale_revisions(id)
            for lr in lrs:
                if lr.status in [TntCLMWaveLocaleRevision.STATUS_OPEN, TntCLMWaveLocaleRevision.STATUS_REJECTED]:
                    oneOk = True
                    break
            if not oneOk:
                raise PublicUserException(_("Data for this wave cannot be modified for any locale"))

        wave = self.get_waves(id)
        if len(wave) == 0:
            raise SecurityException('Invalid wave %s' % id)
        wave = wave[0]
        if wave.status in [TntCLMWave.STATUS_ONLINE, TntCLMWave.STATUS_FINISHED]:
            raise SecurityException('Cannot update attributes priorities for ONLINE or FINISHED waves. (id %s)'% wave.id)
        wave_attributes = {}
        for a in wave.attributes:
            wave_attributes[a.id] = a

        for attr_id, priority in values.items():
            attr_id = int(attr_id)
            if attr_id not in wave_attributes.keys():
                raise SecurityException('Invalid attribute id %s for wave id %s' % (attr_id, id))
            if priority is not None and priority not in range(1, 7):
                raise SecurityException('Invalid priority value (%s) for attribute id %s' % (priority, attr_id))

            attribute = wave_attributes[attr_id]
            if priority_global:
                if attribute.priority_trigger_at is None and priority is None:
                    continue
                attribute.priority_trigger_at = priority
                self.sa_session.add(attribute)
                continue

            if attribute.priority_trigger_at is not None:
                # so that users without set_wave_global_attribute_priority permission can set locale priorities
                # (if they don't actually change a global priority)
                if attribute.priority_trigger_at == priority:
                    continue
                if self.has_permissions('set_wave_global_attribute_priority'):
                    raise PublicUserException('Cant set locale priority for attribute %s, locale %s since is already set global' %
                                              (attribute.attribute.id, locale.label))
                else:
                    # if no permission, it's a security exception cuz radios are disabled on the interface
                    raise SecurityException('Cant set priority for attribute %s, locale %s since is already set global' % (attr_id, locale_id))

            al = self._get_last_attribute_locale(attr_id, locale_id)
            if not al:
                raise SystemException('Cant set priority. Attribute id %s does not have revisions for locale id %s' \
                                      % (attr_id, locale_id))
            if al.priority_trigger_at is None and priority is None:
                continue
            al.priority_trigger_at = priority
            self.sa_session.add(al)

    @transaction_decorator
    def set_wave_translation_status(self, wave_id, locale_id, status, reason=None):
        """
        Changes status for a wave translation
        """
        wave = self.get_waves(wave_id)[0]
        if wave.status == TntCLMWave.STATUS_FINISHED:
            raise PublicUserException(_('Invalid request for FINISHED wave'))
        if status == TntCLMWaveLocaleRevision.STATUS_OPEN:
            raise PublicUserException(_('Cannot change status to OPEN'))
        # status of the last transaction is the one taken into account
        lrs = self.get_wave_locale_revisions(wave_id, locale_id)
        if len(lrs) == 0:
            raise SecurityException("Invalid translation (wave-locale)")
        lr = lrs[0]

        # status PENDING can only be set for OPEN revisions
        if status == TntCLMWaveLocaleRevision.STATUS_PENDING:
            self.assert_permissions('translate_tnt_clm_wave')
            if lr.status != TntCLMWaveLocaleRevision.STATUS_OPEN:
                raise PublicUserException(_('This translation is not OPEN'))

        # status APPROVED can only be set for ACTIVE_APPROVED revisions
        if status == TntCLMWaveLocaleRevision.STATUS_APPROVED:
            self.assert_permissions('approve_tnt_clm_translations')
            if lr.status != TntCLMWaveLocaleRevision.STATUS_ACTIVE_APPROVED:
                raise SecurityException('Status APPROVED can only be set for an ACTIVE_APPROVED revision')

        # status ACTIVE_APPROVED or REJECTED can only be set for PENDING revisions
        if status in [TntCLMWaveLocaleRevision.STATUS_ACTIVE_APPROVED, TntCLMWaveLocaleRevision.STATUS_REJECTED]:
            self.assert_permissions('approve_tnt_clm_translations')
            # always last revision should be pending cuz once that is set than translations cannot be edited anymore
            if lr.status != TntCLMWaveLocaleRevision.STATUS_PENDING:
                raise PublicUserException(_('There is no request for approval for this translation'))
            lr.rejection_reason = reason

        lr.status = status

        # if status is set to ACTIVE_APPROVED, need to set all previous ACTIVE_APPROVED revisions to APPROVED
        if status ==  TntCLMWaveLocaleRevision.STATUS_ACTIVE_APPROVED:
            for l in lrs:
                if l.id != lr.id and l.status == TntCLMWaveLocaleRevision.STATUS_ACTIVE_APPROVED:
                    l.status = TntCLMWaveLocaleRevision.STATUS_APPROVED
                    self.sa_session.add(l)

        self.sa_session.add(lr)
        return lr

    def send_status_notification(self, wave_id, status, locales, rejection_reason=None):
        """
        Send notifications at different stages of translation approval process.
        Notifications depending on status as follows:
            * pending -> send notification to all groups that have the permission to approve a
              translation for the provided locales
            * active approved -> as above + send notification to all groups that have the permission to translate
              for the provided locales. If a group has both permissions (approve and translate)
              only one notification will be sent to that group.
            * rejected -> as for status active_approved plus include rejection reason in the notification
        """
        if status in [TntCLMWaveLocaleRevision.STATUS_OPEN, TntCLMWaveLocaleRevision.STATUS_APPROVED]:
            return

        wave = self.get_waves(wave_id)[0]

        text = _("Please check Wave Translations Status overview.")
        # send notification to groups that have the following permissions
        permissions = []
        if status == TntCLMWaveLocaleRevision.STATUS_PENDING:
            permissions = ['approve_tnt_clm_translations']
            subject = wave.name + " - " + _("Request for approval of translations")
        if status in [TntCLMWaveLocaleRevision.STATUS_ACTIVE_APPROVED, TntCLMWaveLocaleRevision.STATUS_REJECTED]:
            permissions = ['approve_tnt_clm_translations', 'translate_tnt_clm_wave']
            if status == TntCLMWaveLocaleRevision.STATUS_ACTIVE_APPROVED:
                subject = wave.name + " - " + _("Translations Approved")
            if status == TntCLMWaveLocaleRevision.STATUS_REJECTED:
                subject = wave.name + " - " + _("Translations Rejected")
                if rejection_reason is not None:
                    text = _("Reason:")+"<br>"+rejection_reason+"<br><br>"+text
        if len(permissions) == 0:
            return

        from basesoft.web.contrib.license.models import GroupPermission, Permission
        groups = self.sa_session.query(Group).join(GroupPermission).join(Permission).filter(
            Permission.name.in_(permissions)
        ).all()
        if len(groups) == 0:
            return

        nc = NotificationController(self.sa_session)
        n = nc.set_notification(subject, text, TntCLMNotification, {
            'js_module': 'tnt_clm_wave_translations_status',
            'parameters': [wave_id]
        })
        nc.create_system_notification(n, groups)

    def get_translations_percentage(self, wave):
        """
        Checks how much was translated for supplied wave,
        per locale and per type (attributes, scale answers, letters, mails)
        """
        from decimal import Decimal

        if not isinstance(wave, TntCLMWave):
            wave = self.get_waves(wave)[0]

        # only check for translations that already have at least one revision for this wave
        # for rest can easily assume they are not translated
        locales = self.sa_session.query(TntCLMWaveLocaleRevision.locale_id).select_from(
                  TntCLMWaveLocaleRevision).filter(TntCLMWaveLocaleRevision.wave==wave).all()
        translations = {}
        # TODO: add letters and mails when implemented
        # should we also add wave translations percentage (includes label, opt-out-text and header when we implement that)?
        for l in locales:
            translations[l[0]] = {'attributes': 0,
                                  'scale_answers': 0}

        # attributes (also count attributes' answers and categories)
        all_attributes = len(wave.attributes)
        all_attributes += len(wave.categories)
        for a in wave.attributes:
            all_attributes += len(a.answers)

        if all_attributes > 0:
            # count translated attributes
            attrs = self.sa_session.query(TntCLMWaveLocaleRevision.locale_id,
                                                 func.count(TntCLMWaveAttributeLocale.wave_attribute_id.distinct())).select_from(
                            join(TntCLMWaveAttributeLocale, TntCLMWaveLocaleRevision,
                                and_(TntCLMWaveLocaleRevision.id==TntCLMWaveAttributeLocale.locale_revision_id,
                                     TntCLMWaveLocaleRevision.wave==wave)
                            )).group_by(TntCLMWaveLocaleRevision.locale_id).all()
            for la in attrs:
                translations[la[0]]['attributes'] += la[1]
            # count translated attributes' answers
            answers = self.sa_session.query(TntCLMWaveLocaleRevision.locale_id,
                                            func.count(TntCLMWaveAnswerLocale.wave_answer_id.distinct())).select_from(
                            join(TntCLMWaveAnswerLocale, TntCLMWaveLocaleRevision,
                                and_(TntCLMWaveLocaleRevision.id==TntCLMWaveAnswerLocale.locale_revision_id,
                                     TntCLMWaveLocaleRevision.wave==wave)
                            )).group_by(TntCLMWaveLocaleRevision.locale_id).all()
            for la in answers:
                translations[la[0]]['attributes'] += la[1]
            # count translated attributes' categories
            categs = self.sa_session.query(TntCLMWaveLocaleRevision.locale_id,
                                           func.count(TntCLMWaveAttributeCategoryLocale.category_id.distinct())).select_from(
                            join(TntCLMWaveAttributeCategoryLocale, TntCLMWaveLocaleRevision,
                                and_(TntCLMWaveLocaleRevision.id==TntCLMWaveAttributeCategoryLocale.locale_revision_id,
                                     TntCLMWaveLocaleRevision.wave==wave)
                            )).group_by(TntCLMWaveLocaleRevision.locale_id).all()
            for la in categs:
                translations[la[0]]['attributes'] += la[1]

            for l in translations:
                translations[l]['attributes'] = 100 * float(translations[l]['attributes'])/float(all_attributes)

        # scale answers
        all_scale_answers = len(wave.scale_answers)
        if all_scale_answers > 0:
            # count translated scale answers
            sa = self.sa_session.query(TntCLMWaveLocaleRevision.locale_id,
                                       func.count(TntCLMWaveScaleAnswerLocale.wave_scale_answer_id.distinct())).select_from(
                            join(TntCLMWaveScaleAnswerLocale, TntCLMWaveLocaleRevision,
                                and_(TntCLMWaveLocaleRevision.id==TntCLMWaveScaleAnswerLocale.locale_revision_id,
                                     TntCLMWaveLocaleRevision.wave==wave)
                            )).group_by(TntCLMWaveLocaleRevision.locale_id).all()
            for la in sa:
                translations[la[0]]['scale_answers'] = 100 * float(la[1])/float(all_scale_answers)

        # round all values to 2 decimals
        precizion = Decimal(10) ** -2
        for l in translations:
            for t in translations[l]:
                translations[l][t] = Decimal(str(translations[l][t])).quantize(precizion)
                # but then, when sent to the client will be converted to float again
                # which won't keep the 2 decimals always
                translations[l][t] = str(translations[l][t])

        return translations

    @transaction_decorator
    def clone_wave(self, id, locales=None):
        wave = self.get_privileged_object_by_id(TntCLMWave, id, TntCLMWave.PRIV_VIEW).pop()

        if locales is None:
            locales = self.get_available_wave_locales(wave)
        else:
            if not isinstance(locales, list):
                locales = [locales]
            if not isinstance(locales[0], TntCLMLocale):
                locales = self.get_locales(locales)
            dl = self.get_wave_default_locale(id)
            locale_ids = [l.id for l in locales]
            if dl.id not in locale_ids:
                raise PublicUserException(_('Wave default locale is not in the copy locales list')+'('+ dl.label+')')
        locs = [{'locale': l, 'revision': None} for l in locales]

        newwave = self._clone(wave, children=(['attributes', 'answers'], ['scale_answers',], ), locales=locs, \
                              categories={}, wave = None)
        newwave.name = '%s-copy' % newwave.name
        newwave.code = '%s-copy' % newwave.code
        self.sa_session.add(newwave)
        self.save_default_privilege_object_group(newwave)
        self.sa_session.flush()
        return newwave

    def _clone(self, obj, children=None, locales=None, locale_revision=None, categories={}, wave=None):
        exclude = ['_sa_instance_state', 'id', 'created', 'last_updated', 'groups', 'wave_id', 'status', 'begin_date',\
                   'end_date']

        cls = obj.__class__
        new = cls()

        for attr, value in obj.__dict__.iteritems():

            if attr in exclude:
                continue
            # special case for locale_revision_id cuz that cannot be cloned
            if attr == 'locale_revision_id':
                if locale_revision is None:
                    raise Exception('Got obj with locale_revision_id, but no locale_revision to set for it')
                new.locale_revision = locale_revision
                continue

            if attr == 'category_id' and hasattr(obj, 'wave_id'):
                if not categories.has_key(value):
                    cat_obj = self.sa_session.query(TntCLMWaveAttributeCategory).filter(TntCLMWaveAttributeCategory.id==value).first()
                    if cat_obj is None:
                        continue
                    cat = self._clone(cat_obj, locales=locales, categories=categories, wave=wave)
                    cat.wave = wave
                    categories[value] = cat
                new.category = categories[value]
                continue

            if not isinstance(value, InstrumentedList):
                setattr(new, attr, value)
                continue

        if type(new) == TntCLMWave:
            wave = new
            for locale in locales:
                if locale['revision'] is None:
                    lr = TntCLMWaveLocaleRevision()
                    lr.wave = new
                    lr.locale = locale['locale']
                    lr.account = self.get_account()
                    locale['revision'] = lr

        if children is not None:
            for child in children:
                attr = child[0]
                value = []
                for i in getattr(obj, attr):
                    if len(child[1:])!=0:
                        value.append(self._clone(i, children=(child[1:],), locales=locales, categories=categories, wave=wave))
                    else:
                        value.append(self._clone(i, locales=locales, categories=categories, wave=wave))
                setattr(new, attr, value)

        if locales is None:
            return new

        locale_obj, field = new.get_locale_object()
        new.locales = []
        for locale in locales:
            filter = (TntCLMWaveLocaleRevision.locale==locale['locale'])
            if type(obj) == TntCLMWave:
                filter = filter & (TntCLMWaveLocaleRevision.wave==obj)
            else:
                filter = filter & (getattr(locale_obj, field)==obj)
            latest = self.sa_session.query(locale_obj).join(TntCLMWaveLocaleRevision).filter(
                    filter).order_by(TntCLMWaveLocaleRevision.id.desc()).first()
            #latest.locale_revision = locale['revision']
            if latest is None:
                continue
            if type(obj) == TntCLMWave:
                self.sa_session.add(self._clone(latest, locales=None, locale_revision=locale['revision'], \
                                                categories=categories, wave=wave))
            else:
                new.locales.append(self._clone(latest, locales=None, locale_revision=locale['revision'], \
                                               categories=categories, wave=wave))

        return new

    def export_wave_pdf(self, filename, wave_id, locale_id=None, locale_revision_id=None):
        """
        'exports' a pdf filename for the given wave_id, locale_id and optionally locale_revision_id
        filename can either be a absolute file path but also a HttpResponse object for sending to the client

        for now permission checking is handled by the get_wave_state() function that is used internally

        returns a filename that can be returned to the client if needed
        """

        assert isinstance(filename, (str, unicode, HttpResponse))
        assert isinstance(wave_id, (int, long))
        assert isinstance(locale_id, (int, long, NoneType))
        assert isinstance(locale_revision_id, (int, long, NoneType))

        state = self.get_wave_state(wave_id, locale_id, locale_revision_id)

        from iqportal.tnt.clm.pdf.build import TntWave38Builder

        r = TntWave38Builder(self.sa_session)
        r.build(filename, state)

        # return filename based on locale and wave-code
        # TODO: include locale_revision_id into filename maybe
        locale_str = self.sa_session.query(TntCLMLocale).filter(TntCLMLocale.id==state['locale']).one().label
        filename = '%s_%s_QI.pdf' % (state['wave']['code'], locale_str)
        return filename

class TntCLMAttributeController(AbstractTntStateController):

    _instance_class = TntCLMAttribute
    _locale_attributes = ['label', 'priority_trigger_at']

    def __init__(self, session, manager):
        """
        tnt abstract controller shoul not use the iqportal locale controller, but the tnt one
        """
        AbstractTntStateController.__init__(self, session, manager)

        self.attribute_answer_controller = TntCLMAttributeAnswerController(session, manager)

    def get_attributes(self, id=None, priv=TntCLMAttribute.PRIV_VIEW):
        """
        """
        self.assert_permissions('view_tnt_clm_attribute')
        return self.get_privileged_object_by_id(TntCLMAttribute, id, priv=priv)

    def get_attributes_by_type(self, type, priv=TntCLMAttribute.PRIV_VIEW):
        self.assert_permissions('view_tnt_clm_attribute')
        return self._get_privileged_object(TntCLMAttribute, TntCLMAttribute.type==type, priv)

    def load_attribute(self, id):
        """
        """
        attr = self.get_attributes(id)
        if len(attr) == 0:
            return {}
        attr = attr[0]

        rv = {'dbID': attr.id,
              'domID': None,
              'type': attr.type,
              'default_label': attr.default_label,
              'priority_trigger_at': attr.priority_trigger_at,
              'locales': self._load_locales(attr),
              'answers': []
              }

        answ = self.sa_session.query(TntCLMAttributeAnswer).filter(
                    TntCLMAttributeAnswer.attribute_id==attr.id).order_by(TntCLMAttributeAnswer.position).all()
        for a in answ:
            rv["answers"].append(self.attribute_answer_controller.load_attribute_answer(a.id))
        return rv

    @transaction_decorator
    def save_attribute(self, state, group=None):
        """
        """
        if group is None:
            group = self._get_default_group()

        assert isinstance(group, Group)

        # save or update
        if state['dbID']:
            self.assert_permissions('update_tnt_clm_attribute')
        else:
            self.assert_permissions('create_tnt_clm_attribute')

        # assert state
        self._assert_state(state, None, dict)
        self._assert_state(state, 'type', int)
        self._assert_state(state, 'default_label', unicode)
        self._assert_state(state, 'priority_trigger_at', (int, None))

        attr = AbstractTntStateController._instance_from_state(self, state)
        if state['dbID']:
            self.assert_privileges_on_object(attr, TntCLMAttribute.PRIV_UPDATE)

        # perform actual updating
        attr.type = state['type']
        attr.default_label = state['default_label']
        attr.priority_trigger_at = state['priority_trigger_at']

        self._save_locale(attr, TntCLMAttributeLocale, state)

        # scale attributes do not have answers
        if state['type'] != TntCLMAttribute.TYPE_SCALE:
            #old answers
            ids = [a.id for a in attr.answers]
            for s in state['answers']:
                if s['dbID'] is not None and s['dbID'] in ids:
                    ids.remove(s['dbID'])
                self.attribute_answer_controller.save_attribute_answer(s, attr)
            # if there's anything left in ids, those have to be removed
            for id in ids:
                self.attribute_answer_controller.delete_attribute_answer(id)
        else: # remove answers if type changed from something else to scale
            if len(attr.answers) > 0:
                for a in attr.answers:
                    self.attribute_answer_controller.delete_attribute_answer(a.id)

        if state['dbID'] is None:
            self.save_default_privilege_object_group(attr)

        self.sa_session.expire(attr)
        # save object
        self.sa_session.add(attr)
        self.sa_session.flush()

        # return the attr's id
        return attr.id

    @transaction_decorator
    def merge(self, keep_attribute_id, *attributes):
        """
        We assume that if we have merge permissions and privs on the attributes, we are allowed to merge
        even if they are attached to waves.
        """
        self.assert_permissions('merge_tnt_clm_attribute')

        attributes = self.get_attributes([id for id in attributes], TntCLMAttribute.PRIV_MERGE)
        keep = self.get_attributes(keep_attribute_id, TntCLMAttribute.PRIV_VIEW).pop()
        type = keep.type
        answers = keep.answers

        for a in attributes:
            if a.id == keep_attribute_id:
                continue
            if a.type != type:
                raise PublicUserException(_('Unable to merge attributes of different types'))
            if answers is None and a.answers is not None:
                raise PublicUserException(_('Unable to merge attributes due to answer difference'))
            if answers is not None and len(answers) != len(a.answers):
                raise PublicUserException(_('Unable to merge attributes due to the different number of answers'))

        wave_attr = self.sa_session.query(TntCLMWaveAttribute).filter(TntCLMWaveAttribute.attribute_id.in_([a.id for a in attributes])).all()
        for attr in wave_attr:
            if attr.attribute == keep:
                continue
            # if attributes have answers we have to merge them also.
            if attr.answers:
                for a, keep_a in zip(attr.answers, keep.answers):
                    a.answer = keep_a
            attr.attribute = keep
            self.sa_session.add(attr)

        for a in attributes:
            if a == keep:
                continue
            self.sa_session.delete(a)
        self.sa_session.flush()

class TntCLMAttributeAnswerController(AbstractTntStateController):

    _instance_class = TntCLMAttributeAnswer
    _locale_attributes = ['label']

    def load_attribute_answer(self, id):
        """
        """
        aa = self._instance_from_id(id)
        return {'dbID': aa.id,
                'domID': None,
                'default_label': aa.default_label,
                'position': aa.position,
                'locales': self._load_locales(aa)
               }

    def save_attribute_answer(self, state, attribute):
        """
        """
        assert isinstance(attribute, TntCLMAttribute)

        self._assert_state(state, None, dict)
        self._assert_state(state, 'position', int)
        self._assert_state(state, 'default_label', unicode)

        aa = AbstractTntStateController._instance_from_state(self, state)

        aa.attribute = attribute
        aa.default_label = state['default_label']
        aa.position = state['position']

        # save locales in tnt locale table
        self._save_locale(aa, TntCLMAttributeAnswerLocale, state)

        # save object
        self.sa_session.add(aa)
        self.sa_session.flush()

        return aa

    def delete_attribute_answer(self, id):
        """
        """
        aa = self._instance_from_id(id)
        try:
            self.sa_session.delete(aa)
            self.sa_session.flush()
        except IntegrityError, e:
            if e.message.find('tblTntCLMWaveAnswer') < 0:
                raise
            raise PublicUserException(_('Cannot remove answers linked to waves.'))

class TntCLMScaleAnswerController(AbstractTntStateController):

    _instance_class = TntCLMScaleAnswer
    _locale_attributes = ['label']

    def get_scale_answers(self, id=None):
        """
        return a list of scale answer objects based on id
        """
        self.assert_permissions('view_tnt_clm_scale_answer')
        return self.get_privileged_object_by_id(TntCLMScaleAnswer, id)

    def get_wave_types(self, id=None):
        """
        return a list of wave type objects based on id
        """
        self.assert_permissions('view_tnt_clm_wave_type')
        return self.get_privileged_object_by_id(TntCLMWaveType, id)

    def load_scale_answer(self, id):
        """
        returns the state of a state answer with a given id
        """
        sa = self.get_scale_answers(id)
        if len(sa) == 0:
            return {}
        sa = sa[0]

        return {
            'dbID': sa.id,
            'domID': None,
            'type': sa.type.id,
            'default_label': sa.default_label,
            'locales': self._load_locales(sa)
        }

    def save_scale_answer(self, state):
        """
        creates/updates a scale answer object based on the given state
        """
        # save or update
        if state['dbID']:
            self.assert_permissions('update_tnt_clm_scale_answer')
        else:
            self.assert_permissions('create_tnt_clm_scale_answer')

        self._assert_state(state, None, dict)
        self._assert_state(state, 'type', int)
        self._assert_state(state, 'default_label', unicode)

        sa = AbstractTntStateController._instance_from_state(self, state)

        wt = self.get_privileged_object_by_id(TntCLMWaveType, state['type'])[0]

        if state['dbID']:
            self.assert_privileges_on_object(sa, TntCLMScaleAnswer.PRIV_UPDATE)

        sa.type = wt
        sa.default_label = state['default_label']

        # save locales in tnt locale table
        self._save_locale(sa, TntCLMScaleAnswerLocale, state)

        # new scale answer
        if not state['dbID']:
            # set the position as the maximum available for this type

            filter, fo = self.get_query_filter_on_privileges(TntCLMScaleAnswer, TntCLMScaleAnswerGroup, TntCLMScaleAnswer.PRIV_VIEW)
            fo = fo.join(TntCLMWaveType, TntCLMScaleAnswer.type_id==TntCLMWaveType.id)
            filter = and_(filter, TntCLMWaveType.id==wt.id)

            q = self.sa_session.query(func.max(TntCLMScaleAnswer.position))
            q = q.select_from(fo).filter(filter)

            try:
                max = int(q.one()[0]) + 1
            except TypeError:
                # we didn't have any answer, so no max...
                max = 1

            sa.position = max
            self.save_default_privilege_object_group(sa)

        self.sa_session.add(sa)
        self.sa_session.flush()

        return sa.id

    def move_scale_answer(self, id, move_up):
        """
        Changes the position of the scale id
        @param id: the id of the scale answer that we want to move
        @param move_up: a boolean that indicates the direction. true if we want to move it up (position decreases),
            false if we want to move it down (position increases)
        @return True if succeded, False otherwise
        """

        sa = self.get_scale_answers(id)
        if len(sa) == 0:
            raise PublicUserException(_('There are no scale answers.'))

        sa = sa[0]

        # we only work with the scale answers of the type that interests us
        filter, fo = self.get_query_filter_on_privileges(TntCLMScaleAnswer, TntCLMScaleAnswerGroup, TntCLMScaleAnswer.PRIV_UPDATE)
        filter = and_(filter, TntCLMScaleAnswer.type==sa.type)
        q = self.sa_session.query(TntCLMScaleAnswer.id, TntCLMScaleAnswer.position).select_from(fo).filter(filter).order_by(TntCLMScaleAnswer.position)

        scale_answers = q.all() # list of tuples that contain the id and position
        n = len(scale_answers)
        try:
            pos = scale_answers.index((sa.id, sa.position))
        except ValueError:
            raise PublicUserException(_('The scale answer was not found.'))

        # notice that 'pos' here is the return value of index() and not the actual .position (thats +1)
        if pos == 0 and move_up:
            return # it's the first already, we don't have to move it up

        if pos == n and not move_up:
            return # it's the last already, we don't have to move it down

        if move_up:
            switcharoo_info = scale_answers[pos - 1]
        else:
            switcharoo_info = scale_answers[pos + 1]

        cp = sa.position
        switcharoo = self.get_scale_answers(switcharoo_info[0])[0] # the answer with which we switch (oh... twisted tongue)
        sa.position = switcharoo_info[1]
        switcharoo.position = cp

        self.sa_session.add(sa)
        self.sa_session.add(switcharoo)
        self.sa_session.flush()

class TntCLMContactController(SimpleContactController):
    """
    Controller for the polymorphic type Contact
    """
    _instance_class = TntCLMContact
    can_set_unique_fields = False
    _entity_map = {
        'E': None, # Means "Express" and can be skipped
        'SpS': 'Special Services',
        'BS': 'Business Solutions',
        'A': 'Associates',
    }

    _locale_controller = None
    _tnt_controller = None

    def __init__(self, sa_session, locale=None, account_id=None):
        SimpleContactController.__init__(self, sa_session, locale=locale, account_id=account_id)
        self._entities = {}
        self._depots = {}
        self._locales = {}
        self._customers = {}
        self._persons = {}

    def set_person_for_import(self, person_data, unique_fields=[]):
        """
        Tnt can't set unique fields, but we need a tnt Person to be unique on email and last name automatically
        """
        person_data = SimpleContactController.set_person_for_import(self, person_data, [])
        if not person_data.has_key('email'):
            return person_data

        last_name = None
        if person_data.has_key('last_name'):
            last_name = person_data['last_name']

        filter, fo = self.get_query_filter_on_privileges(Contact, ContactGroup, Contact.PRIV_UPDATE)
        fo = fo.join(Person, Contact.person_id==Person.id)
        filter = and_(filter, and_(Person.email==person_data['email'], Person.last_name==last_name))

        person = self.sa_session.query(Person).select_from(fo).filter(filter).first()
        if person is not None:
            # so we found a person which matches the tnt unique fields
            # set that as person_object and when saving the contact, it'll know to search for the contact with this person
            person_data['person_object'] = person

        return person_data

    def _set_person(self, id, person_data):
        """
        Store each person map in order to find it if duplicate in the file
        """

        if person_data.has_key('gender') and person_data['gender'] == '99':
            person_data['gender'] = ''

        if not person_data.has_key('email') or person_data['email'] is None:
            return SimpleContactController._set_person(self, id, person_data)

        l = person_data['email']

        last_name = None
        if person_data.has_key('last_name') and person_data['last_name'] is not None:
            last_name = person_data['last_name']
            l = l + '_' + person_data['last_name']

        if self._persons.has_key(l) and isinstance(self._persons[l], Person):
            person_data['person_object'] = self._persons[l]

        person = SimpleContactController._set_person(self, id, person_data)
        self._persons[l] = person
        return person

    def set_import_company_state(self, company_data):

        company = SimpleContactController.set_import_company_state(self, company_data)

        # entity is just provided for extra checks, which were already done in the validation stage,
        # so can disregard that in the import stage

        # if no locale, we check for country - language - location
        #if not company_data.has_key('locale'):
        #    company_data['locale'] = self.__get_locale(company_data['country'], company_data['language'], company_data['entity'], company_data['location'])

        for key in company_data.keys():
            if key in ["subsector", "imp_code"] and str(company_data[key]) == '99':
                company[key] = ''
                continue
            company[key] = company_data[key]

        return company

    def set_company_for_import(self, company_data, unique_fields=[]):
        """
        Tnt can't set unique fields, but we need unique on Customer, depot, locale, revenue_band automatically
        """
        company = SimpleContactController.set_company_for_import(self, company_data, [])

        if not company_data.has_key('country'):
            country, country_key = self.__get_country_from_state(company_data)
        else:
            country = company_data['country']

        if not company.has_key('account_number') or not isinstance(country, Country):
            return company

        locale = self.__get_locale(country, company_data['language'], company_data['entity'], company_data['location'])
        if not locale:
            return company
        company['locale_id'] = locale.id

        if self._customers.has_key('%s_%s' %(company['account_number'], country.id)) and\
        isinstance(self._customers['%s_%s' %(company['account_number'], country.id)], TntCLMCustomer):
            customer = self._customers['%s_%s' %(company['account_number'], country.id)]
        else:
            # get the customer
            customer = self.sa_session.query(TntCLMCustomer).filter(
                and_(
                    TntCLMCustomer.account_number==company['account_number'],
                    TntCLMCustomer.country==country
                )
            ).first()

            # save it to the customers map so it won't be queried again
            self._customers['%s_%s' %(company['account_number'], country.id)] = customer

            if customer is None:
                return company

        company, locale = self._get_company_locale(company)
        company, depot = self._get_company_depot(company, country)

        rb = None
        if company.has_key('revenue_band'):
            rb = company['revenue_band']

        # We check if the company already exists for this combination.
        c = self.sa_session.query(TntCLMCompany).filter(
            and_(
                 TntCLMCompany.customer==customer,
                 TntCLMCompany.depot==depot,
                 TntCLMCompany.locale==locale,
                 TntCLMCompany.revenue_band==rb,
            )
        ).first()
        if c is not None:
            # so we found a company which matches the tnt unique fields
            # set that as company_object and when saving the contact, it'll know to search for the contact with this company
            company['company_object'] = c

        return company

    def _get_company_depot(self, company_data, country):
        """
        Get depot based on company data
        """
        if company_data.has_key('depot') and not company_data.has_key('depot_id'):
            if self._depots.has_key((company_data['depot'], country)):
                d = self._depots[(company_data['depot'], country)]
            else:
                d = self.sa_session.query(TntDepot).filter(and_(
                    TntDepot.default_code==company_data['depot'],
                    TntDepot.country==country
                )).all()
                if len(d) == 0:
                    d = self.sa_session.query(TntDepot).join(TntDepotAlias).filter(and_(
                        TntDepotAlias.code==company_data['depot'],
                        TntDepot.country==country
                    )).all()
                    if len(d) == 0:
                        raise SecurityException("Invalid depot code %s for country %s" % (company_data['depot'], country))
                d = d[0]
                self._depots[(company_data['depot'], country)] = d
            # add as depot_id so that lower it can check for privs
            company_data['depot_id'] = d.id

        if company_data.has_key('depot_id'):
            if company_data['depot_id'] is not None:
                if self._depots.has_key(company_data['depot_id']):
                    depot = self._depots[company_data['depot_id']]
                else:
                    if self._tnt_controller is None:
                        self._tnt_controller = TntController(self.sa_session, account_id=self.get_account().id)
                    depot = self._tnt_controller.get_depots(company_data['depot_id'])
                    if len(depot) == 0:
                        raise SecurityException("Cannot find depot %s. Check privileges. " % company_data['depot_id'])
                    depot = depot[0]
                    self._depots[company_data['depot_id']] = depot
                return company_data, depot
        return company_data, None

    def _get_company_locale(self, company_data):
        """
        Get locale based on company data
        """
        if company_data.has_key('locale'):
            if self._locales.has_key(company_data['locale']):
                locale = self._locales[company_data['locale']]
            else:
                if self._locale_controller is None:
                    self._locale_controller = TntCLMLocaleController(self.sa_session, account_id=self.get_account().id)
                locale = self._locale_controller.get_locale_by_str(company_data['locale'])
                if locale is None:
                    raise SecurityException("Invalid locale %s " % company_data['locale_id'])
                self._locales[company_data['locale']] = locale
            return (company_data, locale)
        if company_data.has_key('locale_id'):
            if company_data['locale_id'] is not None:
                if self._locales.has_key(company_data['locale_id']):
                    locale = self._locales[company_data['locale_id']]
                else:
                    if self._locale_controller is None:
                        self._locale_controller = TntCLMLocaleController(self.sa_session, account_id=self.get_account().id)
                    locale = self._locale_controller.get_locales(company_data['locale_id'])
                    if len(locale) == 0:
                        raise SecurityException("Invalid locale %s " % company_data['locale_id'])
                    locale = locale[0]
                    self._locales[company_data['locale_id']] = locale
                return (company_data, locale)
        return (company_data, None)

    def _set_company(self, id, company_data, instance_class=None):

        # country is set as address_data, but in case of tnt, we need country sparately
        if company_data.has_key('address_data') and company_data['address_data'].has_key('country'):
            company_data['country'] = company_data['address_data']['country']
            # if address_data only has country, remove the address_data so it won't save as address
            if len(company_data['address_data']) == 1:
                del company_data['address_data']

        # we already have the company object (only for import - checked in set_company_for_import)
        if company_data.has_key('company_object') and isinstance(company_data['company_object'], TntCLMCompany):
            company = SimpleContactController._set_company(self, id, company_data)
            # it already has everything set, so just need to update subsector and imp
            for key in ['subsector', 'imp_code']:
                if company_data.has_key(key):
                    if company_data[key] == '':
                        company_data[key] = None
                    company.__setattr__(key, company_data[key])
            return company

        # no company found based on TNT's preset unique fields
        # first search for the customer:

        customer = None
        if company_data.has_key('account_number') and (company_data.has_key('country') or company_data.has_key('locale_id')):
            if company_data.has_key('country'):
                if isinstance(company_data['country'], Country):
                    country = company_data['country']
                else:
                    try:
                        # contact panel gets country id
                        country_id = int(company_data['country'])
                        country = self.sa_session.query(Country).filter(Country.id==country_id).first()
                    except ValueError:
                        country = self.sa_session.query(Country).filter(Country.iso_3166_a2==company_data['country']).first()
            else:
                locale = self.sa_session.query(TntCLMLocale).filter(TntCLMLocale.id==company_data['locale_id']).one()
                country = locale.country

            # we query TntCLMCustomer

            # when updating a TNT company though the panel:
            # even for an already existing company, if we want to change account number for example, we cannot directly
            # update the company's customer because that might also be linked to another company

            # so, if customer with that new account number is not found,
            # a new customer will be created and the company will be linked to that one

            # if indeed a customer with that account number and country will be found, the company will be updated
            # with that customer because we can't have duplicate customers

            if self._customers.has_key('%s_%s' %(company_data['account_number'], country.id)):
                customer = self._customers['%s_%s' %(company_data['account_number'], country.id)]
            else:
                customer = self.sa_session.query(TntCLMCustomer).filter(
                    and_(
                        TntCLMCustomer.account_number==company_data['account_number'],
                        TntCLMCustomer.country==country
                    )
                ).first()
            if customer is None:
                customer = TntCLMCustomer()
                customer.account_number = company_data['account_number']
                customer.country = country

                # save it to the customers map so it'll find it if duplicates in the file
                self._customers['%s_%s' %(company_data['account_number'], country.id)] = customer

        if customer is None:
            raise SecurityException('Customer cannot be None')
        company = SimpleContactController._set_company(self, id, company_data)
        company.customer = customer

        company_data, company.depot = self._get_company_depot(company_data, country)
        company_data, company.locale = self._get_company_locale(company_data)

        if company_data.has_key('revenue_band'):
            if company_data['revenue_band'] == '':
                company.revenue_band = None
            else:
                company.revenue_band = company_data['revenue_band']

        company = self._check_company(company)

        for key in ['subsector', 'imp_code', 'name']:
            if company_data.has_key(key):
                if company_data[key] == '':
                    company_data[key] = None
                company.__setattr__(key, company_data[key])

        return company

    def _check_company(self, company):
        """
        Checks if the company already exists
        """
        c = self.sa_session.query(TntCLMCompany).filter(and_(
            TntCLMCompany.customer==company.customer,
            TntCLMCompany.depot==company.depot,
            TntCLMCompany.locale==company.locale,
            TntCLMCompany.revenue_band==company.revenue_band
        )).first()
        if c is None:
            return company
        return c

    def _load_company_state(self, id):
        """
        Returns dict with company state
        """
        company = self.get_company(id)
        company_state = SimpleContactController._load_company_state(self, company)
        company_state['depot_id'] = company.depot_id
        company_state['revenue_band'] = company.revenue_band
        company_state['subsector'] = company.subsector
        company_state['imp_code'] = company.imp_code
        company_state['locale_id'] = company.locale_id
        company_state['account_number'] = company.customer.account_number
        return company_state

    def get_company_cols(self):
        """
        Used for contact import. Returns a list with this company type columns.
        """
        company = SimpleContactController.get_company_cols(self)
        company_cols = ["revenue_band", "subsector", "imp_code"]
        for col in company_cols:
            company.append({"value": col, "label": TntCLMCompany.cols_i18n[col]})
        company.append({"value": "locale", "label": _("Locale")})
        company.append({"value": "depot", "label": _("Depot")})
        company.append({"value": "entity", "label": _("Entity")})
        company.append({"value": "language", "label": _("Language")})
        company.append({"value": "location", "label": _("Location")})
        company.append({"value": "account_number", "label": _("TNT Account Number")})
        return company

    def company_required_columns(self):
        return [['language', 'location', 'company_country_iso_3166_a2', 'account_number', 'entity']]

    def get_field_plugins(self):
        """
        Used for contact import
        """
        from basesoft.web.core.input.validation import plugins
        fp, gp = SimpleContactController.get_field_plugins(self)
        fp['revenue_band'] = [(plugins.RestrictedLength, {'max': 5})]
        fp['account_number'] = [(plugins.Integer, {'allow_zero': False})]
        return fp, gp

    def __get_country_from_state(self, state):
        country = None
        country_key = None
        o = None
        if state.has_key('country'):
            return state['country'], 'country'
        if state.has_key("company_country_name"):
            o = self.sa_session.query(Country).filter(Country.name==state["company_country_name"]).first()
            if o is None:
                o = self.sa_session.query(DomainCountry).filter(and_(DomainCountry.name==state["company_country_name"],
                    DomainCountry.domain==self.get_default_group().license.domain, DomainCountry.active==1)).first()
                if o is not None:
                    o = o.country
            country_key = 'company_country_name'
        if state.has_key("company_country_iso_3166_a2"):
            o = self.sa_session.query(Country).filter(Country.iso_3166_a2==state["company_country_iso_3166_a2"]).first()
            country_key = 'company_country_iso_3166_a2'
        if state.has_key("company_country_iso_3166_a3"):
            o = self.sa_session.query(Country).filter(Country.iso_3166_a3==state["company_country_iso_3166_a3"]).first()
            country_key = 'company_country_iso_3166_a3'
        if o is not None:
            country = o

        return country, country_key

    def validate_state(self, state):
        """
        Used for contact import
        """
        if not isinstance(state, list):
            states = [state]
        else:
            states = state
        for state in states:
            if state.has_key('gender') and state['gender'] == '99':
                state['gender'] = ''
        returns = SimpleContactController.validate_state(self, states)

        depots = None

        i = 0
        for state in states:
            if not state.has_key('company'):
                continue
            state = state['company']
            ret = returns[i]
            i = i + 1

            country, country_key = self.__get_country_from_state(state)

            # locales
            if not state.has_key('location') or not state.has_key('language') or not state.has_key('entity'):
                ret[country_key] = _("Not all fields for locales are provided")
                continue

            entity = self.__get_entity(state['entity'])
            locale_err = _('Cannot build locale from provided fields')

            locale = self.__get_locale(country, state['language'], entity, state['location'])
            if not locale:
                ret['entity'] = locale_err
                ret['language'] = locale_err
                ret[country_key] = locale_err
                ret['location'] = locale_err

            #depot
            if state.has_key('depot'):
                if depots is None:
                    cc = TntController(self.sa_session, account_id=self.get_account().id)
                    # user rights are also checked here
                    depots = cc.get_depots()
                    codes = []
                    for d in depots:
                        codes.append(d.default_code)
                        for a in d.aliases:
                            codes.append(a.code)
                if state['depot'] not in codes:
                    ret['depot'] = _('Depot does not exist')

        return returns

    def __get_locale(self, country, language, entity, location):
        if not isinstance(entity, TntCLMService):
            entity = self.__get_entity(entity)
        try:
            return self.sa_session.query(TntCLMLocale).filter(
            and_(
                TntCLMLocale.country==country,
                TntCLMLocale.language_code==language,
                TntCLMLocale.location==location,
                TntCLMLocale.service==entity)).one()
        except NoResultFound:
            return False

    def __get_entity(self, entity):

        if entity is None:
            return None

        if self._entities.has_key(entity):
            return self._entities[entity]

        if self._entity_map.has_key(entity):
            return self.__get_entity(self._entity_map[entity])

        entity_obj = self.sa_session.query(TntCLMService).filter(TntCLMService.label==entity).first()
        if entity_obj is None:
            return _('Entity does not exist')

        self._entities[entity] = entity_obj
        return entity_obj

contact_controllers = {'TntCLMContact': TntCLMContactController}

