# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.

from django.utils.translation import ugettext as _
from sqlalchemy import func, case, and_, or_

from basesoft.web.core.util.custom_query import CustomQuery, CustomQueryField, CustomQueryModel
from basesoft.web.contrib.i18n.custom_query import DomainCountryCustomQueryField,\
    DomainCountryCustomQueryModel

from iqportal.tnt.clm.models import TntCLMWave, TntCLMWaveType, TntCLMAttribute, TntCLMScaleAnswer,\
    TntCLMLocale, TntCLMService, TntCLMWaveLocaleRevision
from iqportal.tnt.clm.models import TntCLMWaveMail, TntCLMWaveMailLocale
from iqportal.tnt.response.models import TntCLMWaveParticipant, TntCLMWaveInvitee
from iqportal.tnt.clm.controllers import TntCLMLocaleController
from sqlalchemy.sql.expression import cast
from sqlalchemy.types import Integer

class TntCLMWaveCustomQuery(CustomQuery):
    #permissions = ['wave_manager']

    def tables(self):
        return (
            CustomQueryModel(TntCLMWave),
            CustomQueryModel(TntCLMWaveType, priv=None),
            CustomQueryModel(TntCLMWaveInvitee,
                             onclause=TntCLMWave.id==TntCLMWaveInvitee.wave_id, use_outerjoin=True),
            CustomQueryModel(TntCLMWaveParticipant,
                             onclause=TntCLMWaveInvitee.id==TntCLMWaveParticipant.invitee_id, use_outerjoin=True),
        )
    def fields(self):
        return (
            # latest created waves on top
            CustomQueryField(TntCLMWave.id, _('ID'), key='wave_id', is_sorted=True, sorted_asc=False),
            CustomQueryField(TntCLMWave.name, _('Wave name'), key='wave_name'),
            CustomQueryField(TntCLMWave.id, _('Preview'), key='wave_preview'),
            CustomQueryField(TntCLMWave.id, _('PDF'), key='wave_pdf'),
            CustomQueryField(TntCLMWave.id, _('Translate'), key='wave_translation'),
            CustomQueryField(TntCLMWave.id, _('Translations status'), key='translations_status'),
            CustomQueryField(TntCLMWave.id, _('History'), key='wave_history'),
            CustomQueryField(TntCLMWave.id, _('Schedule'), key='schedule'),
            CustomQueryField(TntCLMWave.id, _('P1'), key='set_p1'),
            CustomQueryField(TntCLMWaveType.label, _('Type'), key='type'),
            CustomQueryField(TntCLMWave.status, _('Status'), key='status', is_filter=True,
                filter_options=[{'value': TntCLMWave.STATUS_OPEN,
                                 'label': TntCLMWave.statuses_i18n[TntCLMWave.STATUS_OPEN]
                                },
                                {'value': TntCLMWave.STATUS_SCHEDULED,
                                 'label': TntCLMWave.statuses_i18n[TntCLMWave.STATUS_SCHEDULED]
                                },
                                {'value': TntCLMWave.STATUS_ONLINE,
                                 'label': TntCLMWave.statuses_i18n[TntCLMWave.STATUS_ONLINE]
                                },
                                {'value': TntCLMWave.STATUS_FINISHED,
                                 'label': TntCLMWave.statuses_i18n[TntCLMWave.STATUS_FINISHED]
                                }]
            ),
            CustomQueryField(cast(func.sum(case([(TntCLMWaveParticipant.is_respondent>0, 1)], else_=0)), Integer),
                             _('Respondents'), key="respondents"),
            # this make is very slow
            #CustomQueryField(func.round(func.sum(case([(TntCLMWaveParticipant.is_respondent>0, 1)], else_=0))
            #                            *100/func.count(TntCLMWaveInvitee.id.distinct()), 2),
            #                 _('Response %'), key="response"),
            CustomQueryField(TntCLMWave.id, _('Export CSV'), key='export_csv'),
        )

    group_by = (TntCLMWave.id, )

class ClmAttributeCustomQuery(CustomQuery):
    permissions = ['view_tnt_clm_attribute']

    def tables(self):
        return (
            CustomQueryModel(TntCLMAttribute),
        )
    def fields(self):
        filters = []
        for k, v in TntCLMAttribute.types.items():
            filters.append({'value': k, 'label': v})

        return (
            CustomQueryField(TntCLMAttribute.id, _('Attribute ID'), key="id"),
            CustomQueryField(TntCLMAttribute.type, _('Type'), key="type", is_filter=True, is_group_by=True,
                             filter_options=filters),
            CustomQueryField(case([(func.length(TntCLMAttribute.default_label)<=60, TntCLMAttribute.default_label)],
                                  else_=func.substring(TntCLMAttribute.default_label, 1, 57)+"..."),
                             _('Default attribute label'), key='default_label', is_filter=True),
            CustomQueryField(TntCLMAttribute.priority_trigger_at, _('Priority trigger at'), key="priority_trigger_at"),
        )

class TntCLMScaleAnswerCustomQuery(CustomQuery):

    def tables(self):
        return (
            CustomQueryModel(TntCLMScaleAnswer),
            CustomQueryModel(TntCLMWaveType),
        )
    def fields(self):
        return (
            CustomQueryField(TntCLMScaleAnswer.id, _('ID'), key='scale_answer_id', sortable=False),
            CustomQueryField(TntCLMScaleAnswer.default_label, _('Default answer label'), key='default_label', sortable=False),
            CustomQueryField(TntCLMWaveType.label, _('Type'), sortable=False),
            CustomQueryField(TntCLMScaleAnswer.position, _('Position'), key='position', sortable=False),
            CustomQueryField(TntCLMScaleAnswer.created, _('Created'), sortable=False),
            CustomQueryField(TntCLMScaleAnswer.last_updated, _('Last updated'), sortable=False),
        )

    # we want to order only by type.id and scale_answer.position, no sorting on the client-side
    # this way we'll see in the datatable the actual order of the scale answers and we'll imediatelly see the result
    # when we reposition them using the arrows from the position column
    order_by = (TntCLMWaveType.id, TntCLMScaleAnswer.position)

class TntCLMLocaleCustomQuery(CustomQuery):
    def tables(self):
        return (
            CustomQueryModel(TntCLMLocale),
            DomainCountryCustomQueryModel(),
            CustomQueryModel(TntCLMService, use_outerjoin=True),
        )
    def fields(self):
        return (
            CustomQueryField(TntCLMLocale.id, _('ID'), key='locale_id'),
            CustomQueryField(TntCLMLocale.label, _('Label'), key='label', is_filter=True),
            CustomQueryField(TntCLMLocale.language_code, _('Language code'), key='language', is_filter=True),
            CustomQueryField(TntCLMLocale.description, _('Description'), key='desc', is_filter=True),
            DomainCountryCustomQueryField(None, _('Country'), is_filter=True, is_group_by=True, key='country'),
            CustomQueryField(TntCLMLocale.location, _('Location'), key='location', is_filter=True),
            CustomQueryField(TntCLMLocale.location_2015, _('Location 2015'), key='location_2015', is_filter=True),
            CustomQueryField(TntCLMService.label, _('Service'), key='service', is_filter=True),
            CustomQueryField(TntCLMLocale.active, _('Active'), key='active', is_filter=True,
                             filter_options=[{'value': 1, 'label': _('Active')},
                                             {'value': 0, 'label': _('Inactive')}])
        )
    order_by = (TntCLMLocale.label,)

    def set_where(self):
        controller = TntCLMLocaleController(self.sa_session, account_id=self.get_account().id)
        self.where = controller.get_query_filter_on_locales(False)

class TntCLMWaveTranslationsStatusCustomQuery(CustomQuery):
    permissions = ['translate_tnt_clm_wave']

    def tables(self):
        return (
            CustomQueryModel(TntCLMWaveLocaleRevision),
            CustomQueryModel(TntCLMLocale,
                             onclause=TntCLMLocale.id==TntCLMWaveLocaleRevision.locale_id),
        )
    def fields(self):
        return (
            CustomQueryField(TntCLMWaveLocaleRevision.locale_id, _('Locale ID'), key='locale_id'),
            CustomQueryField(TntCLMLocale.label, _('Locale'), key='locale', is_filter=True, is_sorted=True),
            CustomQueryField(TntCLMWaveLocaleRevision.locale_id, _('Attributes'), key='attributes'),
            CustomQueryField(TntCLMWaveLocaleRevision.locale_id, _('Scale answers'), key='scale_answers'),
            CustomQueryField(TntCLMWaveLocaleRevision.locale_id, _('Wave preview'), key='wave_preview'),
            CustomQueryField(TntCLMWaveLocaleRevision.locale_id, _('Translate'), key='translate'),
            CustomQueryField(TntCLMWaveLocaleRevision.locale_id, _('Wave PDF'), key='wave_pdf'),
            CustomQueryField(TntCLMWaveLocaleRevision.locale_id, _('P1'), key='set_p1'),
            CustomQueryField(TntCLMWaveLocaleRevision.status, _('Status'), key='status', is_filter=True, is_group_by=True,
                 filter_options=[{'value': TntCLMWaveLocaleRevision.STATUS_OPEN,
                                         'label': TntCLMWaveLocaleRevision.statuses_i18n[TntCLMWaveLocaleRevision.STATUS_OPEN]
                                        },
                                        {'value': TntCLMWaveLocaleRevision.STATUS_PENDING,
                                         'label': TntCLMWaveLocaleRevision.statuses_i18n[TntCLMWaveLocaleRevision.STATUS_PENDING]
                                        },
                                        {'value': TntCLMWaveLocaleRevision.STATUS_APPROVED,
                                         'label': TntCLMWaveLocaleRevision.statuses_i18n[TntCLMWaveLocaleRevision.STATUS_APPROVED]
                                        },
                                        {'value': TntCLMWaveLocaleRevision.STATUS_REJECTED,
                                         'label': TntCLMWaveLocaleRevision.statuses_i18n[TntCLMWaveLocaleRevision.STATUS_REJECTED]
                                        }]
            ),
            CustomQueryField(TntCLMWaveLocaleRevision.rejection_reason, _('Rejection reason'), key='rejection_reason'),
            CustomQueryField(TntCLMWaveLocaleRevision.wave_id, _('Wave ID'), key='wave_id'),
        )
    group_by = (TntCLMWaveLocaleRevision.wave_id, TntCLMWaveLocaleRevision.locale_id, )

    def set_where(self):
        controller = TntCLMLocaleController(self.sa_session, account_id=self.get_account().id)
        self.where = controller.get_query_filter_on_locales()

        self.where = and_(self.where, TntCLMWaveLocaleRevision.id.in_(
                                        self.sa_session.query(func.max(TntCLMWaveLocaleRevision.id)).
                                        select_from(TntCLMWaveLocaleRevision).
                                        group_by(TntCLMWaveLocaleRevision.wave_id, TntCLMWaveLocaleRevision.locale_id)))


class TntCLMWaveMailCustomQuery(CustomQuery):
    permissions = ['view_mail', 'view_tnt_wave_mail']

    def tables(self):
        return (
            CustomQueryModel(TntCLMWaveMail),
            CustomQueryModel(TntCLMLocale, onclause=TntCLMLocale.id==TntCLMWaveMail.fallback_locale_id),
        )
    def fields(self):
        return (
            CustomQueryField(TntCLMWaveMail.id, _('ID'), key='mail_id'),
            CustomQueryField(TntCLMWaveMail.name, _('Name'), key='mail_name'),
            CustomQueryField(TntCLMLocale.label, _('Fallback locale'), key='fallback_locale'),
            CustomQueryField(TntCLMLocale.id, _('Fallback locale ID'), key='fallback_locale_id'),
            CustomQueryField(TntCLMWaveMail.id, _('Preview'), key='mail_preview'),
            CustomQueryField(TntCLMWaveMail.id, _('PDF'), key='mail_pdf'),
            CustomQueryField(TntCLMWaveMail.id, _('Translate'), key='mail_translation'),
            CustomQueryField(TntCLMWaveMail.id, _('Translations status'), key='mail_translations_status'),
            CustomQueryField(TntCLMWaveMail.created, _('Created'), key='created'),
        )

class TntCLMWaveMailTranslationsStatusCustomQuery(CustomQuery):
    permissions = ['view_mail', 'view_tnt_wave_mail']

    def tables(self):
        return (
            CustomQueryModel(TntCLMWaveMailLocale),
            CustomQueryModel(TntCLMLocale,
                             onclause=TntCLMLocale.id==TntCLMWaveMailLocale.locale_id),
        )
    def fields(self):
        return (
            CustomQueryField(TntCLMWaveMailLocale.locale_id, _('Locale ID'), key='locale_id'),
            CustomQueryField(TntCLMLocale.label, _('Locale'), key='locale', is_filter=True, is_sorted=True, sorted_asc=False),
            CustomQueryField(case([(and_(TntCLMWaveMailLocale.subject=='', TntCLMWaveMailLocale.subject==''), '0.00%'),
                                   (and_(TntCLMWaveMailLocale.subject!='', TntCLMWaveMailLocale.subject!=''), '100.00%')
                                   ], else_='50.00%'),
                             _('Percentage'), key='percentage'),
            CustomQueryField(TntCLMWaveMailLocale.locale_id, _('Preview'), key='preview'),
            CustomQueryField(TntCLMWaveMailLocale.locale_id, _('Translate'), key='translate'),
            CustomQueryField(TntCLMWaveMailLocale.locale_id, _('PDF'), key='mail_pdf'),
            CustomQueryField(TntCLMWaveMailLocale.status, _('Status'), key='status', is_filter=True,
                filter_options=[{'value': TntCLMWaveMailLocale.STATUS_OPEN,
                                         'label': TntCLMWaveMailLocale.statuses_i18n[TntCLMWaveMailLocale.STATUS_OPEN]
                                        },
                                        {'value': TntCLMWaveMailLocale.STATUS_PENDING,
                                         'label': TntCLMWaveMailLocale.statuses_i18n[TntCLMWaveMailLocale.STATUS_PENDING]
                                        },
                                        {'value': TntCLMWaveMailLocale.STATUS_APPROVED,
                                         'label': TntCLMWaveMailLocale.statuses_i18n[TntCLMWaveMailLocale.STATUS_APPROVED]
                                        },
                                        {'value': TntCLMWaveMailLocale.STATUS_REJECTED,
                                         'label': TntCLMWaveMailLocale.statuses_i18n[TntCLMWaveMailLocale.STATUS_REJECTED]
                                        }]
            ),
            CustomQueryField(TntCLMWaveMailLocale.rejection_reason, _('Rejection reason'), key='rejection_reason'),
            CustomQueryField(TntCLMWaveMailLocale.mail_id, _('Mail ID'), key='mail_id'),
        )

    def set_where(self):
        controller = TntCLMLocaleController(self.sa_session, account_id=self.get_account().id)
        self.where = controller.get_query_filter_on_locales()

custom_queries = (
    ClmAttributeCustomQuery,
    TntCLMWaveCustomQuery,
    TntCLMScaleAnswerCustomQuery,
    TntCLMLocaleCustomQuery,
    TntCLMWaveTranslationsStatusCustomQuery,
    TntCLMWaveMailCustomQuery,
    TntCLMWaveMailTranslationsStatusCustomQuery,
)
