# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.
from sqlalchemy.orm.session import Session
from basesoft.web.core.models import PrivilegeObject
from basesoft.web.contrib.account.models import Account
from basesoft.web.contrib.notification.models import Notification
from basesoft.web.contrib.contact.models import Company, Contact
from django.utils.translation import ugettext, ugettext_lazy as _
from django.conf import settings
from basesoft.web.contrib.mailing.models import Mail, Mailing
from basesoft.web.core.exceptions import SystemException
from iqportal.tnt.response.models import TntCLMWaveInvitee
from sqlalchemy.sql import and_

class TntCLMAccount(Account):
    """
    @DynamicAttrs
    """
    NAME_I18N = _('CLM Account')

    JS_MODULE = 'tnt_clm_account_details'

    permissions = ['manage_tnt_clm_accounts']

    def __get_countries(self):
        countries = []
        for geolevel in self.geolevels:
            [countries.append(c) for c in geolevel.get_countries() if c not in countries]
        for locale in self.locales:
            if locale.country not in countries:
                countries.append(locale.country)
        return countries

    countries = property(__get_countries)

    def __get_services(self):
        services = []
        [services.append(locale.service) for locale in self.locales if locale.service not in services]
        return services

    services = property(__get_services)

    def __get_geolevels(self):
        # top level geolevels
        gls = [g for g in self.geolevels]

        def get_children(geolevel, childs=[]):
            childs.append(geolevel)
            for g in geolevel.child_geolevels:
                get_children(g, childs)
            return childs
        # append all children of all geolevels
        for gl in self.geolevels:
            if len(gl.child_geolevels) == 0:
                continue
            chg = get_children(gl)
            for c in chg:
                if c not in gls:
                    gls.append(c)
        return gls

    all_geolevels = property(__get_geolevels)

    def __repr__(self):
        return '<TntCLMAccount %s>' % self.name

class TntCLMCustomer(object):
    """
    @DynamicAttrs
    """

    cols_i18n = {
        "account_number": _("TNT Account Number")
    }

    def __repr__(self):
        return '<TntCLMCustomer id="%s" account_number="%s">' % (self.id, self.account_number)

class TntCLMCompany(Company):
    """
    @DynamicAttrs
    """
    NAME_I18N = _('TNT CLM Company')
    STR = 'TntCLMCompany'

    cols_i18n = dict(Company.cols_i18n.items() + {
        "revenue_band": _("Revenue band"),
        "subsector": _("Subsector"),
        "imp_code": _("IMP Code")
    }.items())

    PERMISSIONS = ("tnt_clm_contact_access", )

    JS_MODULE = 'tnt_clm_company_form'

    def __repr__(self):
        return '<TntCLMCompany>'

class TntCLMContact(Contact):
    """
    @DynamicAttrs
    """
    NAME_I18N = _('CLM Contact')
    PERMISSIONS = ("tnt_clm_contact_access", )
    STR = 'TntCLMContact'

    company_instance_class = TntCLMCompany
    can_set_unique_fields = False

    def __repr__(self):
        return '<TntCLMContact>'

class TntCLMAccountLocale(object):
    """
    @DynamicAttrs
    """
    def __repr__(self):
        return '<TntCLMAccountLocale>'

class TntCLMService(object):
    """
    Services like Postal, Express, Associates, Fashion, etc
    @DynamicAttrs
    """
    label = None

    def get_mdx_member_config(self):
        rv = [{
            'datasource': 'TNT-CLM',
            'cube': 'TntCLMResponse',
            'prefix': '[Service]',
            'field': 'label',
        }]
        return rv

    def __repr__(self):
        return '<TntCLMService id=%s>' % self.id


class TntCLMNotification(Notification):
    """
    @DynamicAttrs
    """
    NAME_I18N = _('CLM Notification')

    JS_MODULE = 'tnt_clm_account_details'

    def __repr__(self):
        return '<TntCLMNotification>'


class TntCLMLocale(object):
    """
    @DynamicAttrs
    """

    def __repr__(self):
        return '<TntCLMLocale id="%s", label="%s">' % (self.id, self.label)

    def get_state(self):
        """
        Return the state of the locale (same fields as the regular locale)
        """
        return {
            'id': self.id,
            'name': self.label,
            'locale': self.label,
            'direction': self.direction,
        }

class TntCLMWaveMailLocale(object):
    """
    The mail can be composed in different locales.
    @DynamicAttrs
    """

    mail = None
    locale = None

    subject = None
    content = None

    STATUS_OPEN = 0
    STATUS_PENDING = 1
    STATUS_APPROVED = 2
    STATUS_REJECTED = 3

    statuses_i18n = {STATUS_OPEN: ugettext('Open'),
                     STATUS_PENDING: ugettext('Pending'),
                     STATUS_APPROVED: ugettext('Approved'),
                     STATUS_REJECTED: ugettext('Rejected')
                    }

    def __repr__(self):
        return '<TntCLMWaveMailLocale mail_id="%s", locale_id="%s">' % (self.mail_id, self.locale_id)

class TntCLMWaveMail(Mail):
    """
    This model represents the Wave Mail.
    @DynamicAttrs
    """

    NAME_I18N = _("Tnt Wave Mail")
    JS_MODULE = 'tnt_clm_mail'

    STANDARD_MAIL = False

    def __repr__(self):
        return '<TntCLMWaveMail id="%s">' % self.id

    @staticmethod
    def replacement_fields():
        """
        """
        return {
            ugettext('Wave'): {
                'wave.link': ugettext('Wave link')
            }
        }

    def _replace_field(self, person, field_name, context):
        invitation = context['wave'] # should be a dictionary {'label': 'link label'}

        if field_name == 'wave.link':

            in_preview = False # flag that indicates if we're previewing this content
            in_pdf = False # flag that indicates if we're generating a pdf with this content

            extra = ''
            if invitation.has_key('preview') and invitation['preview']:
                in_preview = True
                extra = 'onclick="return false"'

            if invitation.has_key('pdf') and invitation['pdf']:
                in_pdf = True

            if in_preview or in_pdf:
                link = self._base_url(context) # no link to responses, no hashes, no nothing
            else:
                session = Session.object_session(self)
                q = session.query(TntCLMWaveInvitee).filter(and_(TntCLMWaveInvitee.wave==context['mailing'].wave,\
                                                                 TntCLMWaveInvitee.contact==context['contact'])).first()
                if q is None:
                    return ''

                prefix = self._base_url(context)
                if settings.TNT_EMAIL_URL_FOR_INVITEES is not None:
                    prefix = settings.TNT_EMAIL_URL_FOR_INVITEES
                link = '%siqportal.tnt.response/display/%s/' % (prefix, q.invitee_hash)


            if invitation.has_key('label') and invitation['label']:
                # use the translated label if it's in the context and it has a value
                label = invitation['label']
            else:
                label = link

            return '<a href="%s" %s>%s</a>' % (link, extra, label)

        return Mail._replace_field(self, person, field_name, context)

    def get_html_mail(self, mail_locale, context, html=None):
        """
        Override the original get_html_mail, because person.preferred_locale is not a TNT CLM locale
        @param mail_locale: the tnt mail locale for which we want the content of the mail
        """
        import re
        replacement_fields_pattern = re.compile(self.regexp)

        def replace_field(match_obj):
            """
            Scoped function gets called for every replacement field (<input>) found in the html message with as argument
            the match object found by python's regular expression .sub method. Returns the string the that the value
            needs to be replaced with.
            """
            name = match_obj.group(1)
            return self._replace_field(None, name, context)

        if html is None:
            # get the base mail content
            html = self.get_locale_mail(mail_locale.locale, TntCLMWaveMailLocale).content

        # process replacement fields
        html = re.sub(replacement_fields_pattern, replace_field, html)
        return html

    def clone(self):
        return PrivilegeObject.clone(self, clone_children={
            'locales': {
                'exclude': ['status', 'rejection_reason']
            }
        }, reference_children=['fallback_locale'])

    def mail_return_path(self):
        return '"%s" <%s>' % (settings.EMAIL_FROM_TNT, settings.EMAIL_FROM_TNT)

class TntCLMWaveMainLabel(object):
    """
    @DynamicAttrs
    """


class TntCLMWaveType(PrivilegeObject):
    """
    @DynamicAttrs
    """

    mark_as_deleted_permission = 'delete_tnt_clm_wave'
    mark_as_archived_permission = 'update_tnt_clm_wave'

    _name_property = 'label'

    def get_group_object(self):
        return (TntCLMWaveTypeGroup, 'type')

    def __repr__(self):
        return '<TntCLMWaveType id="%s">' % self.id

class TntCLMWaveTypeGroup(object):
    """
    @DynamicAttrs
    """
    pass

class TntCLMScaleAnswer(PrivilegeObject):
    """
    @DynamicAttrs
    """

    mark_as_deleted_permission = 'delete_tnt_clm_scale_answer'
    mark_as_archived_permission = 'update_tnt_clm_scale_answer'

    _name_property = 'default_label'

    def get_group_object(self):
        return (TntCLMScaleAnswerGroup, 'scale_answer')

    def __repr__(self):
        return '<TntCLMScaleAnswer id="%s">' % self.id

class TntCLMScaleAnswerGroup(object):
    """
    @DynamicAttrs
    """
    pass

class TntCLMScaleAnswerLocale(object):
    """
    @DynamicAttrs
    """
    pass

class TntCLMAttribute(PrivilegeObject):
    """
    @DynamicAttrs
    """
    TYPE_SCALE = 1
    TYPE_SINGLE_CHOICE = 2
    TYPE_TEXT = 3
    TYPE_MULTIPLE_CHOICE = 4
    TYPE_RANGE = 5
    # attribute shown in the general attributes part of the wave, as single choice,
    # but responses are saved for both scale and general answer
    TYPE_SINGLE_SCALE = 6
    # attribute for NPS score
    TYPE_NPS = 7

    types = {
       TYPE_SCALE: _('Scale (1-5)'),
       TYPE_SINGLE_CHOICE: _('Single Choice'),
       TYPE_TEXT: _('Text'),
       TYPE_MULTIPLE_CHOICE: _('Multiple Choice'),
       TYPE_RANGE: _('Range'),
       TYPE_SINGLE_SCALE: _('Single Choice') + '*',
       TYPE_NPS: _('NPS Score')
    }

    PRIV_MERGE = 32

    all_privileges = (PrivilegeObject.PRIV_VIEW, PrivilegeObject.PRIV_UPDATE, \
                      PrivilegeObject.PRIV_ARCHIVE, PrivilegeObject.PRIV_DELEGATE_PRIVILEGES, \
                      PRIV_MERGE,)

    mark_as_deleted_permission = 'delete_tnt_clm_attribute'
    mark_as_archived_permission = 'update_tnt_clm_attribute'

    _name_property = 'default_label'

    def get_privileges(self):
        return PrivilegeObject.get_privileges(self) + [
            (self.PRIV_MERGE, ugettext('Merge privileges')),
        ]

    def get_mdx_member_config(self):
        rv = [{
            'datasource': 'TNT-CLM',
            'cube': 'TntCLMResponse',
            'prefix': '[Attribute]',
            'field': 'default_label',
        }]
        return rv

    def get_group_object(self):
        return (TntCLMAttributeGroup, 'attribute')

    def __repr__(self):
        return '<TntCLMAttribute id="%s" type="%s">' % (self.id, self.type)

class TntCLMAttributeGroup(object):
    """
    @DynamicAttrs
    """
    pass

class TntCLMAttributeLocale(object):
    """
    @DynamicAttrs
    """
    pass

class TntCLMAttributeAnswer(object):
    """
    @DynamicAttrs
    """
    pass

class TntCLMAttributeAnswerLocale(object):
    """
    @DynamicAttrs
    """
    pass

class TntCLMWave(PrivilegeObject):
    """
    @DynamicAttrs
    """
    id = None

    STATUS_OPEN = 0
    STATUS_SCHEDULED = 1
    STATUS_ONLINE = 2
    STATUS_FINISHED = 3

    statuses_i18n = {STATUS_OPEN: ugettext('Offline'),
                     STATUS_SCHEDULED: ugettext('Scheduled'),
                     STATUS_ONLINE: ugettext('Online'),
                     STATUS_FINISHED: ugettext('Finished')
                    }

    PRIV_TRANSLATE = 32
    PRIV_SCHEDULE = 64
    PRIV_VIEW_RESPONSE = 128

    all_privileges = (PrivilegeObject.PRIV_VIEW, PrivilegeObject.PRIV_UPDATE, \
                      PrivilegeObject.PRIV_ARCHIVE, PrivilegeObject.PRIV_DELEGATE_PRIVILEGES, \
                      PRIV_TRANSLATE, PRIV_SCHEDULE, PRIV_VIEW_RESPONSE, )

    mark_as_deleted_permission = 'delete_tnt_clm_wave'
    mark_as_archived_permission = 'update_tnt_clm_wave'

    _name_property = 'name'

    def get_privileges(self):
        return PrivilegeObject.get_privileges(self) + [
            (self.PRIV_TRANSLATE, ugettext('Translate privileges')),
            (self.PRIV_SCHEDULE, ugettext('Schedule privileges')),
            (self.PRIV_VIEW_RESPONSE, ugettext('View response privileges')),
        ]

    def get_mdx_member_config(self):
        rv = [{
            'datasource': 'TNT-CLM',
            'cube': 'TntCLMResponse',
            'prefix': '[Wave].[Code]',
            'field': 'code',
        }]
        return rv

    def get_group_object(self):
        return (TntCLMWaveGroup, 'wave')

    def get_locale_object(self):
        return (TntCLMWaveLocale, None)

    def __repr__(self):
        return '<TntCLMWave id="%s">' % self.id

class TntCLMWaveGroup(object):
    """
    @DynamicAttrs
    """
    pass

class TntCLMWaveLocaleRevision(object):
    """
    @DynamicAttrs
    """

    STATUS_OPEN = 0
    STATUS_PENDING = 1
    STATUS_APPROVED = 2
    STATUS_REJECTED = 3
    STATUS_ACTIVE_APPROVED = 4

    statuses_i18n = {STATUS_OPEN: ugettext('Open'),
                     STATUS_PENDING: ugettext('Pending'),
                     STATUS_APPROVED: ugettext('Approved'),
                     STATUS_REJECTED: ugettext('Rejected'),
                     STATUS_ACTIVE_APPROVED: ugettext('Active Approved')
                    }

class TntCLMWaveLocale(object):
    """
    @DynamicAttrs
    """
    pdf_fontsizes = [{'value': 6, 'label': '6 pt.'},
                     {'value': 7, 'label': '7 pt.'},
                     {'value': 8, 'label': '8 pt.'}]

class TntCLMWaveAttribute(object):
    """
    @DynamicAttrs
    """
    def get_locale_object(self):
        return (TntCLMWaveAttributeLocale, 'wave_attribute')

class TntCLMWaveAttributeLocale(object):
    """
    @DynamicAttrs
    """
    pass

class TntCLMWaveAttributeCategory(object):
    """
    @DynamicAttrs
    """
    def get_locale_object(self):
        return (TntCLMWaveAttributeCategoryLocale, 'category')

class TntCLMWaveAttributeCategoryLocale(object):
    """
    @DynamicAttrs
    """
    pass

class TntCLMWaveScaleAnswer(object):
    """
    @DynamicAttrs
    """
    def get_locale_object(self):
        return (TntCLMWaveScaleAnswerLocale, 'wave_scale_answer')

class TntCLMWaveScaleAnswerLocale(object):
    """
    @DynamicAttrs
    """
    pass

class TntCLMWaveAnswer(object):
    """
    @DynamicAttrs
    """
    def get_locale_object(self):
        return (TntCLMWaveAnswerLocale, 'wave_answer')

class TntCLMWaveAnswerLocale(object):
    """
    @DynamicAttrs
    """
    pass

class TntCLMWaveContactList(object):
    """
    @DynamicAttrs
    """
    STATUS_NOT_PROCESSED = 0
    STATUS_PROCESSING = 1
    STATUS_PROCESSED = 2

class TntCLMWaveMailing(Mailing):
    """
    @DynamicAttrs
    """
    FILTER_RESPONDENTS = 1 # this will filter out respondents; this will send to NON-RESPONDENTS only.
    FILTER_NON_RESPONDENTS = 2 # this will filter out non-respondents; this will send to RESPONDENTS only.

    def get_filters(self):
        return Mailing.get_filters(self) + [
            (self.FILTER_RESPONDENTS, ugettext('Exclude respondents')),
            (self.FILTER_NON_RESPONDENTS, ugettext('Exclude non-respondents')),
        ]

    def approved_for_sending(self):
        """
        Make sure that the wave is online and attached mail has all locales approved, else we do not send the mailing.
        """
        # wave needs to be online
        if self.wave.status != TntCLMWave.STATUS_ONLINE:
            return False

        session = Session.object_session(self)

        # all mail locales have to be approved
        q = session.query(TntCLMWaveMailLocale).filter(and_(TntCLMWaveMailLocale.mail==self.mail,\
            TntCLMWaveMailLocale.status!=TntCLMWaveMailLocale.STATUS_APPROVED)).all()
        if len(q) > 0:
            return False

        # all mailing contact lists need to be processed (invitees created)
        clids = [cl.id for cl in self.contact_lists]
        q = session.query(TntCLMWaveContactList).filter(and_(
            TntCLMWaveContactList.wave==self.wave,\
            TntCLMWaveContactList.contact_list_id.in_(clids),\
            TntCLMWaveContactList.is_processed!=TntCLMWaveContactList.STATUS_PROCESSED
            )).all()
        if len(q) > 0:
            return False
        return True

    def mail_from(self, contact):
        """
        """

        mail_from = settings.EMAIL_FROM_TNT

        if self.mail_from_text not in [None, '']:
            return '"%s" <%s>' % (self.mail_from_text, mail_from)
        return mail_from

    mail_return_path = mail_from