# -*- coding: utf-8 -*-
# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.

import math

from reportlab.platypus import Table, Spacer, Paragraph
from reportlab.lib.units import cm, mm
from reportlab.lib.styles import ParagraphStyle
from reportlab.platypus.flowables import PageBreak

from iqportal.tnt.clm.pdf.template import TntWaveTemplate38, TntCLMWaveMailTemplate

from iqportal.tnt.clm.pdf.style import TntStyle38, TntWaveMailStyle
from iqportal.tnt.clm.pdf.flowable import FormBullet, FormCheckbox, ScaleHeader38, TntParagraph, Html2rml

from iqportal.tnt.clm.models import TntCLMLocale, TntCLMAttribute
from django.http import HttpResponse

from reportlab.platypus.doctemplate import FrameBreak
from datetime import datetime

class TntWaveBuilder:
    """
    abstract tnt wave builder provides a base class for building wave pdf files
    """
    _style = None
    _generalAttrCols = None
    _doc = None

    def __init__(self, sa_session):
        self.sa_session = sa_session

    def build(self, filename, state):
        """
        start the building
        """
        assert isinstance(filename, (unicode, str, HttpResponse))

        self._locale = self.sa_session.query(TntCLMLocale).filter(TntCLMLocale.id==state['locale']).one()
        self._filename = filename

        # dynamic font size from state, fallback to 8
        font_size = state['wave']['pdf_fontsize']
        if font_size is None:
            font_size = 8
        assert font_size in range(6, 10)


        # calculate a 'scale' from the font size with base 8
        self._scale = 1.0 / 8 * font_size

        self._setStyle(self._scale)
        self._build(state)

    def _build(self, state):
        """
        needs to build a self._doc, fill it with elements and finally call self._doc.build()
        """
        raise Exception('abstract')

    def _setStyle(self, scale):
        """
        must be overriden by subclass to set the style object specific to this pdf
        """
        raise Exception('abstract')

    def _build_general_attributes(self, state):
        """
        builds the general attribute table and adds it to self._doc_elements (created in build())
        """

        if self._generalAttrCols is None:
            return

        attributes = []

        for i, attr in enumerate(state['wave']['attributes']):

            # we only want general attributes
            # and when its TYPE_TEXT and its in a category (w39c)
            # http://trac.office.basesoft.com/iqportal/ticket/1706
            if attr['type'] in [TntCLMAttribute.TYPE_SINGLE_CHOICE, TntCLMAttribute.TYPE_SINGLE_SCALE,
                                TntCLMAttribute.TYPE_MULTIPLE_CHOICE] or \
            (attr['type'] == TntCLMAttribute.TYPE_TEXT and attr.has_key('category')):
                attributes.append(attr)
        if len(attributes) == 0:
            return

        # split QA's over x columns (probably 2)
        rows = int(math.ceil(float(len(attributes)) / self._generalAttrCols))

        rowContent = [];
        for i  in range(rows):
            rowContent.append([]);

        # add caption one time and
        captionAdded = False
        for i, v in enumerate(attributes):
            if not captionAdded and attributes[i].has_key('category'):
                caption = attributes[i]['category']['label']
                self._doc_elements.append(TntParagraph(caption, self._style['general_qa_caption']))
                captionAdded = True

            attributes[i]['char'] = "%s." % chr(97 + i)
            rowContent[i % rows].append(attributes[i])

        table_data = []

        # build QA's
        for row in rowContent:
            table_row = []
            table_data.append(table_row)

            for attr in row:

                # value
                value = self._get_attr_value(state, attr['id'])

                # in case of single choice..
                if attr['type'] in [TntCLMAttribute.TYPE_SINGLE_CHOICE, TntCLMAttribute.TYPE_SINGLE_SCALE,
                                    TntCLMAttribute.TYPE_MULTIPLE_CHOICE]:

                    table_style = 'table_general_attr_qa_single'

                    # add char and header to top of table (spanned by style)
                    qa_table_data = [[
                        TntParagraph(attr['char'], self._style['locale-normal']),
                        TntParagraph(attr['label'], self._style['locale-normal'], wrap=7*cm),
                    ]]

                    # basic column widths, 1cm for the bullet, rest for answer
                    colWidths = [1*cm, 8*cm]
                    aRows = len(attr['answers'])

                    # if more than 5 answers split into two columns and give new column widhts (4 in total)
                    if aRows > 5:
                        aRows = int(math.ceil(float(len(attr['answers'])) / 2))
                        colWidths = [1*cm, 3*cm, 1*cm, 4*cm]

                # in case of text
                elif attr['type'] == TntCLMAttribute.TYPE_TEXT:

                    table_style = 'table_general_attr_qa_text'

                    # add char and header to top of table (spanned by style)
                    qa_table_data = [[
                        TntParagraph('%s %s' % (attr['char'], attr['label']), self._style['locale-normal'], wrap=7*cm),
                        TntParagraph('', self._style['locale-normal']),
                    ]]

                    colWidths = [5*cm, 4*cm]
                    if attr.has_key('answers'):
                        aRows = len(attr['answers'])
                    else:
                        aRows = 0

                    # more than 5 answers..
                    if aRows > 5:
                        aRows = int(math.ceil(float(len(attr['answers'])) / 2))
                        colWidths = [2.5*cm, 2*cm, 2.5*cm, 2*cm]

                else:
                    raise Exception('invalid type')

                # add three blank rows if no answers are present
                if not attr.has_key('answers'):

                    qa_table_row = []
                    qa_table_data.append(qa_table_row)
                    spacer = Spacer(colWidths[1], 4.5*mm*self._scale)

                    if value is not None:
                        qa_table_row.append([TntParagraph(value, self._style['locale-normal'], wrap=colWidths[1])])
                    else:
                        qa_table_row.append([spacer])

                    for i in range(3):
                        qa_table_row = []
                        qa_table_data.append(qa_table_row)
                        qa_table_row.append([spacer])

                else:

                    # prepare answer rows
                    aRowContent = [];
                    for i in range(aRows):
                        aRowContent.append([])

                    for i, v in enumerate(attr['answers']):
                        aRowContent[i % aRows].append(v)

                    # add bullets and answers into table data
                    for aRow in aRowContent:
                        qa_table_row = []
                        qa_table_data.append(qa_table_row)
                        for answer in aRow:

                            if attr['type'] in [TntCLMAttribute.TYPE_SINGLE_CHOICE, TntCLMAttribute.TYPE_SINGLE_SCALE]:
                                filled = answer['id'] == value
                                qa_table_row.append(FormBullet(self._scale, filled))
                                qa_table_row.append(TntParagraph(answer['label'], self._style['locale-normal'], wrap=colWidths[1]))

                            elif attr['type'] == TntCLMAttribute.TYPE_MULTIPLE_CHOICE:
                                filled = False
                                if value is not None and answer['id'] in value:
                                    filled = True
                                qa_table_row.append(FormCheckbox(self._scale, filled))
                                qa_table_row.append(TntParagraph(answer['label'], self._style['locale-normal'], wrap=colWidths[1]))

                            elif attr['type'] == TntCLMAttribute.TYPE_TEXT:
                                qa_table_row.append(TntParagraph(answer['label'], self._style['locale-normal'], wrap=colWidths[1]))
                                #qa_table_row.append(FormTextInput(self._scale, colWidths[1] - 0.5*cm, value))

                # render qa table
                qa_table = Table(*self._rtl_table_args(qa_table_data, colWidths), style=self._style[table_style])
                table_row.append(qa_table)

        if len(table_data) == 0:
            return

        table = Table(*self._rtl_table_args(table_data), style=self._style['table_general_attr'])
        self._doc_elements.append(table)

    def _build_nps_attributes(self, state):
        """
        builds the general attribute table and adds it to self._doc_elements (created in build())
        """

        attributes = []

        for i, attr in enumerate(state['wave']['attributes']):
            if attr['type'] == TntCLMAttribute.TYPE_NPS:
                attributes.append(attr)
        if len(attributes) == 0:
            return

        # add caption one time and
        captionAdded = False
        for i, v in enumerate(attributes):
            if not captionAdded and attributes[i].has_key('category'):
                caption = attributes[i]['category']['label']
                self._doc_elements.append(TntParagraph(caption, self._style['general_qa_caption']))
                captionAdded = True

        table_data = []

        # build QA's
        for attr in attributes:

            # attribute label
            table_row = [
                TntParagraph(attr['label'], self._style['locale-normal']),
            ]
            table_data.append(table_row)

            # table with attribute answers
            table_row = []
            table_data.append(table_row)

            value = self._get_attr_value(state, attr['id'])

            qa_table_data = []
            qa_table_row_index = []
            qa_table_data.append(qa_table_row_index)
            qa_table_row_radio = []
            qa_table_data.append(qa_table_row_radio)
            qa_table_row_label = []
            qa_table_data.append(qa_table_row_label)

            colWidths = [1.66*cm, 1.66*cm, 1.66*cm, 1.66*cm, 1.66*cm, 1.66*cm, 1.66*cm, 1.66*cm, 1.66*cm, 1.66*cm, 1.66*cm]

            # add bullets and answers into table data
            index = 0
            for answer in attr['answers']:
                qa_table_row_index.append(TntParagraph(str(index), self._style['locale-normal'], wrap=colWidths[1]))
                filled = answer['id'] == value
                qa_table_row_radio.append(FormBullet(self._scale, filled))
                qa_table_row_label.append(TntParagraph(answer['label'], self._style['locale-normal'], wrap=colWidths[1]))
                index = index + 1

            # render qa table
            qa_table = Table(*self._rtl_table_args(qa_table_data, colWidths), style=self._style['table_nps_attr'])
            table_row.append(qa_table)

        if len(table_data) == 0:
            return

        table = Table(*self._rtl_table_args(table_data), style=self._style['table_general_attr'])
        self._doc_elements.append(table)

    def _build_text_attributes(self, state):
        """
        builds the common text attributes (name, address, ...)
        """

        for attr in state['wave']['attributes']:

            # we only want general attributes
            # and if its a text but with a category neither
            # http://trac.office.basesoft.com/iqportal/ticket/1706
            if attr['type'] != TntCLMAttribute.TYPE_TEXT or (attr['type'] == TntCLMAttribute.TYPE_TEXT and attr.has_key('category')):
                continue

            table_data = [[TntParagraph(attr['label'], self._style['locale-normal'])]]
            value = self._get_attr_value(state, attr['id']) # value if question has been answered or None
            style = self._style['locale-normal']

            # add three blank rows if no answers are present
            if not attr.has_key('answers'):
                spacer = Spacer(4*mm * self._scale, 4*mm * self._scale)

                if value is not None:
                    table_data.append([TntParagraph(value, style)])
                else:
                    table_data.append([spacer])
                table_data.append([spacer])
                table_data.append([spacer])

            else:

                rows = int(math.ceil(float(len(attr['answers'])) / 2))
                rowContent = [[] for i in range(rows)]

                for j, answer in enumerate(attr['answers']):
                    rowContent[j % rows].append(answer)

                for row in rowContent:
                    table_row = []
                    table_data.append(table_row)

                    for answer in row:
                        table_row.append(TntParagraph(answer['label'], style))
                        if value is not None:
                            if value.has_key(answer['id']):
                                table_row.append(TntParagraph(value[answer['id']], style))
                            else:
                                # empty string to keep the same amount of items per row
                                table_row.append(TntParagraph(' ', style))

            rowHeights = [0.65*cm * self._scale] * len(table_data)
            table = Table(*self._rtl_table_args(table_data), style=self._style['table_text_attr'], rowHeights=rowHeights)
            self._doc_elements.append(table)

    def _build_opt_out(self, state):

        # see if we have answers and if the user wanted to opt out. if not, don't fill it
        try:
            opt_out = state['values']['opt_out']
        except:
            opt_out = False

        table_data = [[FormBullet(self._scale, opt_out), TntParagraph(state['wave']['opt_out_text'], self._style['locale-normal'])]]
        colWidths = [1*cm, None]
        table = Table(*self._rtl_table_args(table_data, colWidths), style=self._style['table_opt_out'])
        self._doc_elements.append(table)

    def _rtl_table_args(self, table_data, colWidths=None):
        """
        takes table_data and colWidths and reverses it in case needed for rtl
        """

        if self._locale.direction == 'rtl':

            # reverse cells in rows
            for row in table_data:
                row.reverse()

            # reverse column widths
            if colWidths is not None:
                colWidths.reverse()

        return (table_data, colWidths)


    def _build_closure(self, state):

        # return if no closure
        closure = state['wave']['closure_text']
        if closure is None:
            return

        self._doc_elements.append(TntParagraph(closure, self._style['locale-normal'], wrap=18*cm))

    def _get_attr_value(self, state, attr_id):
        """
        Return the value of the attribute.
        @param state: the state we received for building pdf
        @param attr_id: the id of the attribute we're interested in
        @return: the value if it exists or None if the state doesn't have values (just normal view pdf) or the attribute
            doesn't have an answer
        """
        try:
            return state['values']['responses'][attr_id]['value']
        except:
            return None

class TntWave38Builder(TntWaveBuilder):
    """
    wave style 38 implementation
    """

    scale_icons = ['--', '-', '=', '+', '++', 'X']
    _generalAttrCols = 2

    def _setStyle(self, scale):
        self._style = TntStyle38(self._locale, scale)

    def _build(self, state):
        """
        actualy building happens here. if the state also contains values, they will be shown too
        """

        # text on the final bottom of every page
        wave_ident = '%s/%s' % (self._locale.label, state['wave']['code'])

        # create the reportlab doc from the 38 template
        self._doc = TntWaveTemplate38(self._filename,
            wave_title=state['wave']['header_text'],
            wave_subtitle=state['wave']['sub_header_text'],
            wave_ident=wave_ident,
            locale = self._locale,
            localeFontName = self._style['locale-normal'].fontName
        )

        # begin document build
        self._doc_elements = []

        # scale table has some room before it cuz the 45-degree text goes out
        self._doc_elements.append(Spacer(1*cm, 1*cm))

        # build scale table
        self._build_scale_table(state)

        # second page
        self._doc_elements.append(PageBreak())

        # general attribute table
        self._build_general_attributes(state)

        # nps attribute table
        self._build_nps_attributes(state)

        # some room between
        self._doc_elements.append(Spacer(0.2*cm, 0.2*cm))

        # text attribute table
        self._build_text_attributes(state)

        # opt out text goes into second frame so force a framebreak first
        self._doc_elements.append(FrameBreak())
        self._build_opt_out(state)

        # if we have values, we don't show the last page, since it's a response pdf
        if not state.has_key('values'):
            # third page
            self._doc_elements.append(PageBreak())

            # closure text
            self._build_closure(state)

        # build document
        self._doc.build(self._doc_elements)

    def _build_scale_table(self, state):
        """
        builds the scale table with the design of the 38 wave (45 degree QA texts etc)
        """
        table_data = []

        header1 = ['', '']
        for icon_text in state['wave']['scale_answers']:
            header1.append(ScaleHeader38(self._locale, icon_text['label'], self._style['locale-normal'].fontName))
        table_data.append(header1)

        header2 = ['', '']
        for icon in self.scale_icons:
            header2.append(icon)
        table_data.append(header2)

        for i, attr in enumerate(state['wave']['attributes']):

            # only do scale attributes.. is that 1?
            if attr['type'] != TntCLMAttribute.TYPE_SCALE:
                continue

            row = []
            row.append(TntParagraph("%s." % str(i+1), self._style['question_nr']))
            row.append(TntParagraph(attr['label'], self._style['locale-normal'], wrap=13.5*cm))

            value = self._get_attr_value(state, attr['id'])
            for j in range(1, len(self.scale_icons) + 1): # eg [1, 6] for this wave type
                if (j == value):
                    row.append(FormBullet(self._scale, True)) # show a full circle
                else:
                    row.append(FormBullet(self._scale)) # show the normal empty circle (j will never be None)

            table_data.append(row)

        # first specify all column widhts then calculate the second column (question text)
        # based on the remaining width of the document
        colWidths = [0.5 * cm, 0] + [0.6 * cm] * 6
        #colWidths[1] = self._doc.frame_size[0] - sum(colWidths)
        colWidths[1] = None

        rowHeights = [0.65*cm] * len(table_data)

        # build table but be rtl aware
        table = Table(*self._rtl_table_args(table_data, colWidths), style=self._style['table_scale'], rowHeights=rowHeights)
        self._doc_elements.append(table)

class TntCLMWaveMailBuilder:
    """
    TNT CLM wave mail builder provides a class for building wave mail pdf files
    """
    _style = None
    _generalAttrCols = None
    _doc = None

    def __init__(self, sa_session):
        self.sa_session = sa_session

    def build(self, filename, state):
        """
        start the building
        """
        assert isinstance(filename, (unicode, str, HttpResponse))

        self._locale = self.sa_session.query(TntCLMLocale).filter(TntCLMLocale.id==state['locale']).one()
        self._filename = filename

        # calculate a 'scale' from the font size with base 8
        self._scale = 1.0

        self._setStyle(self._scale)
        self._build(state)

    def _build(self, state):
        """
        actualy building happens here.
        """
        # text on the final bottom of every page
        mail_ident = self._locale.label

        # create the reportlab doc from the 38 template
        self._doc = TntCLMWaveMailTemplate(self._filename,
            mail_name=state['mail_name'],
            mail_ident=mail_ident,
            locale = self._locale,
            localeFontName = self._style['locale-normal'].fontName
        )

        # begin document build
        self._doc_elements = []

        # general attribute table
        self._build_mail(state)

        # build document
        self._doc.build(self._doc_elements)

    def _setStyle(self, scale):
        self._style = TntWaveMailStyle(self._locale, scale)

    def _build_mail(self, state):

        subject = state['subject']
        content = state['content']

        # add subject as title, bigger and bold
        self._doc_elements.append(TntParagraph(subject, self._style['locale-subject'], wrap=18*cm))

        # add the date when the pdf was generated
        creation_date = "%s" % datetime.now().strftime("%m/%d/%Y %H:%M")
        self._doc_elements.append(TntParagraph(creation_date, self._style['locale-normal'], wrap=18*cm))

        # content
        p = Html2rml()
        # get all the parts of the content rml.
        # eg. content1, image, content2, content with alignment right, content with alignment center
        # each content part is a dictionary that contains the rml, alignment, type of the node, extra data
        content_elements = p.parse(content, split_paragraph=True, strip_last_br=True)
        for c in content_elements:
            style = ParagraphStyle('locale-mail', parent=self._style['locale-normal'])
            style.fontName = self._style._locale_font
            if self._style.wrap_words:
                style.wordWrap = 'CJK'

            if c['type'] == 'img':
                # for images add the spaceAfter, or the next paragraph will overlap the image if it's bigger than one line
                style.spaceAfter = c['extra']['h'] - self._style['locale-mail'].fontSize
            else:
                style.spaceAfter = 0
            style.alignment = c['align']

            self._doc_elements.append(Paragraph(c['rml'], style))
