# -*- coding: utf-8 -*-
# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.
import cgi
import re

from django.conf import settings
from reportlab.platypus.flowables import Flowable
from reportlab.lib.units import cm, mm
from reportlab.lib.enums import TA_LEFT, TA_CENTER, TA_RIGHT, TA_JUSTIFY
from reportlab.lib import colors
from reportlab.platypus.paragraph import Paragraph
from reportlab.lib.textsplit import wordSplit

from BeautifulSoup import BeautifulStoneSoup
from xml.dom.minidom import parseString

"""
contains extensions of Flowable that are specific for tnt wave pdfs
"""
# regular expressions for extracting properties from 'style'-like attributes
properties_regex = re.compile("([\w-]+)\s*:\s*([^;]+)")

# regular expression used for getting out the numeric part out of a value like '100px' for example
int_regex = re.compile("[\d]+")

def px2mm(pixels):
    """
    Return the proper resolution (in mm) to be used for images in pdf documents based on the size in pixels
    """
    return int(pixels * 24.5 / 72) # TODO: get the dpi value from somewhere else ?

class Html2rml:
    """
    A parser class that handles conversion for html and spits out rml

    http://docs.python.org/library/xml.dom.minidom.html#dom-example
    http://www.crummy.com/software/BeautifulSoup/documentation.html#Parsing HTML
    http://docs.python.org/library/xml.dom.html
    """
    _debug = False

    def parse(self, html, split_paragraph=False, strip_last_br=False):
        """
        Parse the given html (possibly from a RTE) and return the rml content
        @param html: the content that needs to be converted to rml
        @param split_paragraph: a boolean that indicates if the content should be split by images, paragraphs with
            different alignments, etc
            Eg. 'Test<img ...1><b>etc</b><img ...2><span>aaa</span>' becomes:
            ['Test', '<img ...1>', '<b>etc</b>', '<img ...2>', '<span>aaa</span>']
            each part is a dictionary that has keys: rml, type, align, extra. see method push_content
        @param strip_last_br: a boolean that indicates if the last <br> should be stripped from the list of paragraphs
            (because the paragraphs already add a br). This makes sense only if the split_paragraph flag is True
        @return: rml or list of rml parts
        """
        content = '<root>%s</root>' % html
        self.closing_tags = []
        self.content_parts = []
        self._split_paragraph = split_paragraph
        self._strip_last_br = strip_last_br

        # gets filled while parsing
        self._newContent = ""

        # convert crappy html to correct xml
        prettyContent = unicode(BeautifulStoneSoup(content, convertEntities=BeautifulStoneSoup.ALL_ENTITIES)).encode('utf-8')
        if self._debug: print 'pretty content: %s' % prettyContent

        prettyContent = prettyContent.replace('< ', '&lt; ');
        prettyContent = prettyContent.replace(' >', ' &gt; ');
        prettyContent = prettyContent.replace('& ', '&amp; ');

        # xml dom parser
        dom = parseString(prettyContent)

        # recursive function goes trough the whole document and calling
        # tag corbens
        def processNode(node):

            if self._debug: print 'entering: %s' % node.nodeName

            if hasattr(self, '_tag_open_' + node.nodeName):
                getattr(self, '_tag_open_' + node.nodeName)(node)

            if node.nodeType == node.TEXT_NODE:
                text = node.data
                if text.strip() != '':
                    text = cgi.escape(text)
                    if self._debug: print 'found text node: %s' % text
                    self._newContent += text
            else:
                for child in node.childNodes:
                    processNode(child)

            if hasattr(self, '_tag_close_' + node.nodeName):
                getattr(self, '_tag_close_' + node.nodeName)(node)

            if self._debug: print 'leaving: %s' % node.nodeName

        # start with top tag
        processNode(dom)

        if self._newContent != '':
            self.content_parts.append({'rml': self._newContent, 'type': '', 'align': TA_LEFT})

        if self._split_paragraph:
            return self.content_parts
        else:
            return self.content_parts[0]['rml']

    def push_content(self, type='', align=TA_LEFT, extra=None):
        """
        Check if curently gathered content needs to be pushed in the parts list
        @param type: the node type: img, b, i, etc. '' is mixed rml
        @param align: the alignment for this content part
        @param extra: any useful information for the type of node that we're adding
        """
        content = self._newContent
        if self._split_paragraph and content != '':
            if self._strip_last_br:
                # need to get rid of the last <br> so we won't have double space between paragraphs
                if content[-5:] == '<br/>':
                    content = content[:-5]

            self.content_parts.append({
                'rml': content, # eg: '<b>some text</b>'
                'type': type, # eg: 'img', 'b', ''
                'align': align, # eg: TA_LEFT, TA_RIGHT, TA_CENTER, TA_JUSTIFY
                'extra': extra, # anything useful. eg. the height and width of images
            })
            self._newContent = ""

    def extract_attribute(self, node, attribute, properties=False, extract_int=False):
        """
        Extract the value of the given attribute.
        @param node: a dom node object
        @param attribute: the name of the attribute. Eg. src
        @param properties: a flag that indicates if we want to extract the properties.
            eg: if attribute is 'style', and properties=True, the result would be {'height: '50px', 'width': '100px'}
        @param extract_int: a flag that indicates if the values of the properties should be converted to int
            after ignoring all non-numeric chars (useful for height or width that are 'px' suffixed)
        @return: the attribute/properties value(s) or None
            if there is no attribute with that value or properties in that attribute
        """
        value = None
        for a in node.attributes.items():
            if a[0] == attribute:
                if properties:
                    # need to check if part is in the attribute value (mostly for attribute 'style')
                    attribute_value = a[1] # eg. 'width: 100px; height: 40px;'
                    properties = properties_regex.findall(attribute_value)
                    if len(properties) == 0:
                        return None
                    else:
                        d = {}
                        for p in properties:
                            val = p[1]
                            if extract_int:
                                i = int_regex.search(val)
                                if i is not None:
                                    val = int(i.group())
                                else:
                                    val = None
                            d[p[0]] = val
                        return d
                    return None
                else:
                    # return the attribute value, we don't want the properties
                    return a[1]

        return value

    # tag corbens below get called when they are found during parsing
    def _tag_open_br(self, node):
        self._newContent += '<br/>'

    def _tag_open_img(self, node):
        src = self.extract_attribute(node, 'src')
        if src[0:7] == 'file://':
            pass
        elif src[0:4] != 'http': # check if src is a full url
            if src[0] == '/': # if not, make sure it doesn't start with '/', because BASE_URL already has that
                src = src[1:]
            src = '%s%s' % (settings.BASE_URL, src)

        properties = self.extract_attribute(node, 'style', True, True)

        if properties and properties.has_key('width'):
            width = properties['width']
            w = px2mm(width)
            s_width = 'width="%smm"' % w
        else:
            width = None
            w = 0
            s_width = ''

        if properties and properties.has_key('height'):
            height = properties['height']
            h = px2mm(height)
            s_height = 'height="%smm"' % h
        else:
            height = None
            h = 0
            s_height = ''

        self.push_content()
        self._newContent += ' <img %s %s src="%s" valign="top" /> ' % (s_width, s_height, src)
        self.push_content('img', TA_LEFT, {'w': w*mm, 'h': h*mm})

    def _tag_open_a(self, node):
        url = self.extract_attribute(node, 'href')
        if url is None: # no href attribute
            url = settings.BASE_URL # platypus needs a url when building the pdf, and "" and "#" don't work
        else:
            if url[:4].lower() != 'http':
                url = 'http://' + url
        self._newContent += '<a href="%s" color="blue">' % url

    def _tag_close_a(self, node):
        self._newContent += '</a>'

    def _tag_open_strong(self, node):
        self._newContent += '<b>'

    def _tag_close_strong(self, node):
        self._newContent += '</b>'

    def _tag_open_em(self, node):
        self._newContent += '<i>'

    def _tag_close_em(self, node):
        self._newContent += '</i>'

    def _tag_open_span(self, node):
        properties = self.extract_attribute(node, 'style', True)
        if properties and properties.has_key('text-decoration'):
            text_decoration = properties['text-decoration']

            if text_decoration == 'line-through':
                self._newContent += '<strike>'
                self.closing_tags.append('</strike>')
                return
            if text_decoration == 'underline':
                self._newContent += '<u>'
                self.closing_tags.append('</u>')
                return

        if properties and properties.has_key('color'):
            color = properties['color']
            s_color = 'color="%s"' % color
        else:
            s_color = ''

        if properties and properties.has_key('background-color'):
            bgcolor = properties['background-color']
            s_bgcolor = 'backColor="%s"' % bgcolor
        else:
            s_bgcolor = ''

        self._newContent += '<font %s %s>' % (s_color, s_bgcolor)
        self.closing_tags.append('</font>')

    def _tag_close_span(self, node):
        self._newContent += self.closing_tags.pop()

    def _tag_open_div(self, node):
        self.push_content() # push any content we had until now
        self._newContent += '<p>'

    def _tag_close_div(self, node):
        properties = self.extract_attribute(node, 'style', True)

        map = {
            'right': TA_RIGHT,
            'left': TA_LEFT,
            'center': TA_CENTER,
            'justify': TA_JUSTIFY,
        }

        if properties and properties.has_key('text-align'):
            text_align = properties['text-align']
        else:
            text_align = None

        if map.has_key(text_align):
            align = map[text_align]
        else:
            align = TA_LEFT

        self._newContent += '</p>'
        self.push_content('', align)

class TntParagraph(Paragraph):
    """
    extends paragraph but prevents html to be rendered
    """
    def __init__(self, content, style, wrap=None, *a, **kw):

        # strip white space
        content = content.strip()

        # escape html chars
        content = cgi.escape(content)

        # wrap oriental texts. western (and arabic, hebrew) texts dont need this
        # so exlclude that font..
        # TODO: could be done nicer by checking locale but we dont have that here

        if wrap and style.wrap_words:

            text = ''
            lines = wordSplit(content, wrap, style.fontName, style.fontSize)
            for line in lines:
                text += line[1] + "<br/>"
            content = text

        # convert newlines into br's else they wont be shown
        content = content.replace("\n", "<br/>")

        # super
        Paragraph.__init__(self, content, style, *a, **kw)

class FormBullet(Flowable):
    """
    draws a simple circle
    """

    def __init__(self, font_size, filled=False):

        self.width = 0.3*cm
        self.height = 0.3*cm
        self.filled = filled

    def draw(self):
        canvas = self.canv

        canvas.setLineWidth(0.3*mm)
        canvas.setFillColor(self.filled and colors.HexColor(0x000000) or colors.HexColor(0xFFFFFF))
        canvas.setStrokeColor(colors.HexColor(0x000000))
        canvas.circle(0.08*cm, 0.17*cm, 0.12*cm, stroke=1, fill=1)

class FormCheckbox(Flowable):
    """
    draws a checkbox
    """

    def __init__(self, font_size, filled=False):

        self.width = 0.21*cm
        self.height = 0.21*cm
        self.filled = filled

    def draw(self):
        canvas = self.canv

        canvas.setLineWidth(0.26*mm)
        canvas.setFillColor(self.filled and colors.HexColor(0x000000) or colors.HexColor(0xFFFFFF))
        canvas.setStrokeColor(colors.HexColor(0x000000))
        canvas.rect(-0.01*cm, -0.06*cm, self.width, self.height, fill=1)

class FormTextInput(Flowable):
    """
    draws a square
    """

    def __init__(self, font_size, width=1.5*cm, value=''):

        self.width = width
        self.height = 0.3*cm
        self.text = value

    def draw(self):
        canvas = self.canv

        canvas.setLineWidth(0.1*mm)
        canvas.setFillColor(colors.HexColor(0xFFFFFF))
        canvas.setStrokeColor(colors.HexColor(0x000000))
        canvas.rect(-0.1*cm, 0.1*cm, self.width, 0.3*cm)

        # TODO: print self.text in textbox

class ScaleHeader38(Flowable):
    """
    rotates the text -45 degrees
    used for scale header texts in wave 38
    """

    def __init__(self, locale, text, font_name):
        Flowable.__init__(self)
        self._text=text
        self._font_name = font_name
        self._locale = locale

    def draw(self):
        canvas = self.canv
        canvas.setFont(self._font_name, 6)
        if self._locale.direction == 'rtl':
            canvas.rotate(45)
            canvas.drawString( 0, 0, self._text)
        else:
            canvas.rotate(-45)
            canvas.drawRightString( 1*mm, 1*mm, self._text)
