# -*- coding: utf-8 -*-
# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.

import math

from copy import copy

from reportlab.lib import colors
from reportlab.lib.enums import TA_RIGHT
from reportlab.platypus.tables import TableStyle
from reportlab.lib.units import mm
from reportlab.lib.styles import ParagraphStyle

# import fontconfig module registers the fonts
import basesoft.fonts.fontconfig

from iqportal.tnt.clm.models import TntCLMLocale


class _TntStyle(dict):

    """
    Abstract style class. subclasses of this class gets instanciated by the build.py builders for their specific design.

    For table style explanations and others:
    http://www.reportlab.com/docs/reportlab-userguide.pdf
    """

    # locale font default is sans but can be changed in __init__ to a locale specific font
    _locale_font = 'Sans'
    _locale_font_bold = 'Sans-Bold'
    _locale_italic = 'Sans-Italic'

    _font_size = None

    # when enabled it will use the reportlab wrapwords function for wrapping some asian languages
    wrap_words = False

    def __init__(self, locale, scale):

        assert isinstance(locale, TntCLMLocale)
        assert isinstance(scale, float)

        self._scale = scale
        self._font_size = math.ceil(scale * 8)

        import reportlab.rl_config
        reportlab.rl_config.warnOnMissingFontGlyphs = 0

        # wave contens are shown in the locale font
        # locale fonts can be overriden for specific langauges but default to DejaVuSans
        # add missing fonts for other languages here

        # chinese (hong kong)
        if locale.label == 'HKCNI':
            self._locale_font = self._locale_font_bold = self._locale_italic = 'pmingliu'
            self.wrap_words = True

        # chinese
        if locale.label == 'CNCNI':
            self._locale_font = self._locale_font_bold = self._locale_italic = 'kaitim-gb'
            self.wrap_words = True

        # taiwanese
        if locale.label == 'TWTWI':
            self._locale_font = self._locale_font_bold = self._locale_italic = 'mingti2l-big5'
            self.wrap_words = True

        # japanese
        if locale.label == 'JPJPI':
            self._locale_font = self._locale_font_bold = self._locale_italic = 'msmincho'
            self.wrap_words = True

        # korean
        if locale.label == 'KRKRI':
            self._locale_font = self._locale_font_bold = self._locale_italic = 'gulim'

        # thai
        if locale.label == 'THTHI':
            self._locale_font = self._locale_font_bold = self._locale_italic = 'Garuda'

        # arabic
        if locale.label == 'EGARI':
            self._locale_font = self._locale_font_bold = self._locale_italic = 'corefonts-arial'

        # default style (uses standard Sans font)
        paraNormal = ParagraphStyle('normal', None)
        paraNormal.fontName = 'Sans'
        paraNormal.fontSize = self._font_size
        paraNormal.wrap_words = self.wrap_words

        # rtl
        if locale.direction == 'rtl':
            paraNormal.wordWrap = 'RTL'
            paraNormal.alignment = TA_RIGHT

        self['normal'] = paraNormal

        # locale-normal uses locale specific font
        paraLocale = ParagraphStyle('locale-normal', parent=paraNormal, spaceAfter = 12)
        paraLocale.fontName = self._locale_font
        self['locale-normal'] = paraLocale

        # locale-normal-bold
        paraLocaleBold = ParagraphStyle('locale-normal-bold', parent=paraLocale)
        paraLocaleBold.fontName = self._locale_font_bold
        self['locale-normal-bold'] = paraLocaleBold


class TntTableStyle(TableStyle):
    """
    extends the basic table style and in case the passed locale is rtl it flips the horizontal design
    """

    def __init__(self, locale, commands, *a, **kw):

        if locale.direction == 'rtl':

            commands = copy(commands)

            # convert to list so we can modify
            commands = list(commands)
            for i, row in enumerate(commands):
                commands[i] = list(row)

                # commands[i] looks something like:
                # ['LINEBELOW', (0, 1), (-1, 1), 0.3 * mm, colors.HexColor(0xf18e00)],

                # convert coordinates to lists
                from_coord = commands[i][1] = list(commands[i][1])
                to_coord = commands[i][2] = list(commands[i][2])

                # save old values (we gonnal flipem)
                old_from_x = from_coord[0]
                old_to_x = to_coord[0]

                # flipya
                from_coord[0] = (old_to_x*-1)-1
                to_coord[0] = (old_from_x*-1)-1

        TableStyle.__init__(self, commands, *a, **kw)



class TntStyle38(_TntStyle):

    def __init__(self, locale, scale):

        # super
        _TntStyle.__init__(self, locale, scale)

        # table_scale
        ts = TntTableStyle(locale, [

            # default table font
            ('FONT', (0, 0), (-1, -1), self._locale_font, self._font_size),

            # font color scale columns except last
            ('TEXTCOLOR', (0, 1), (-2, 1), colors.HexColor(0xf18e00)),

            # line under second row
            ('LINEBELOW', (0, 1), (-1, 1), 0.3 * mm, colors.HexColor(0xf18e00)),

            # thin dashed line under all rows but first two
            ('LINEBELOW', (0, 2), (-1, -1), 0.2 * mm, colors.HexColor(0xf6b659), 0, (2 * mm, 1 * mm)), # last tuple is dash

            # background color the last column
            ('BACKGROUND', (-1, 1), (-1, -1), colors.HexColor(0xfbddb3)),

            # everyhthing should align in middle
            ('VALIGN', (0, 0), (-1, -1), 'MIDDLE'),

            # for some reason the VALIGN doesnt work exactly.. give some extra padding depending on scale
            ('TOPPADDING', (0, 2), (-1, -1), 2*mm * self._scale),

            # debug line
            #('LINEBELOW', (2, 0), (4, -1), 0.1 * mm, colors.HexColor(0xff0000)),


        ])
        self['table_scale'] = ts

        # question_nr
        para = ParagraphStyle('question_nr', self['locale-normal'])
        para.alignment = TA_RIGHT
        self['question_nr'] = para

        # general_qa_caption
        para = ParagraphStyle('general_qa_caption', self['locale-normal-bold'])
        para.fontSize = 10
        self['general_qa_caption'] = para

        # general qa table holding all questions (two columns)
        ts = TntTableStyle(locale, [
            ('FONT', (0, 0), (-1, -1), self._locale_font, self._font_size),
            ('LINEABOVE', (0, 0), (-1, -1), 0.1 * mm, colors.HexColor(0xf18e00)),
            ('VALIGN', (0, 0), (-1, -1), 'TOP'),
        ])
        self['table_general_attr'] = ts

        # general qa table holds the question text in the first row and answers in the others
        self['table_general_attr_qa_single'] = TntTableStyle(locale, [
            ('FONT', (0, 0), (-1, -1), self._locale_font, self._font_size),
            ('VALIGN', (0, 0), (-1, -1), 'TOP'),

            # span the question text over two cols
            ('SPAN', (1, 0), (-1, 0)),

            # first row (question text) has more space under it
            ('BOTTOMPADDING', (0, 0), (-1, 0), 2*mm),

            # other rows have no padding
            ('TOPPADDING', (0, 1), (-1, -1), 0),
            ('BOTTOMPADDING', (0, 1), (-1, -1), 0),

        ])

        # nps qa
        self['table_nps_attr'] = TntTableStyle(locale, [
            ('FONT', (0, 0), (-1, 0), self._locale_font_bold, self._font_size),
            ('FONT', (0, 1), (-1, -1), self._locale_font, self._font_size),
            ('VALIGN', (0, 0), (-1, -1), 'TOP'),
            ('BOTTOMPADDING', (0, 0), (-1, -1), 0.5*mm)
        ])

        # text are similar as above but with an extra span
        self['table_general_attr_qa_text'] = TntTableStyle(locale, [
            ('FONT', (0, 0), (-1, -1), self._locale_font, self._font_size),
            ('VALIGN', (0, 0), (-1, -1), 'TOP'),

            # span the question text over two cols
            ('SPAN', (1, 0), (-1, 0)),

            # first row (question text) has more space under it
            ('BOTTOMPADDING', (0, 0), (-1, 0), 2*mm),

            ('LINEBELOW', (0, 1), (-1, -1), 0.1 * mm, colors.HexColor(0xcccccc)),

            # other rows have no padding
            ('TOPPADDING', (0, 1), (-1, -1), 0),
            ('BOTTOMPADDING', (0, 1), (-1, -1), 0),

            # span the question text over all cols
            ('SPAN', (0, 0), (-1, 0)),

        ])


        # text qa table
        ts = TntTableStyle(locale, [

            ('FONT', (0, 0), (-1, -1), self._locale_font, self._font_size),

            # first row has background and bold text
            ('FONT', (0, 0), (-1, 0), self._locale_font_bold, self._font_size),
            ('SPAN', (0, 0), (-1, 0)),
            ('LINEABOVE', (0, 0), (-1, 0), 0.3 * mm, colors.HexColor(0xf18e00)),
            ('LINEBELOW', (0, 0), (-1, 0), 0.3 * mm, colors.HexColor(0xf18e00)),
            ('BACKGROUND', (0, 0), (-1, 0), colors.HexColor(0xfbddb3)),

            # other rows have a black line under them
            ('LINEBELOW', (0, 1), (-1, -1), 0.1 * mm, colors.HexColor(0xcccccc)),

            # valign middle
            ('VALIGN', (0, 0), (-1, -1), 'MIDDLE'),

            # for some reason the VALIGN doesnt work exactly.. give some extra padding depending on scale
            ('TOPPADDING', (0, 0), (-1, -1), 2.3*mm * self._scale),

        ])
        self['table_text_attr'] = ts


        # table_opt_out
        ts = TntTableStyle(locale, [
            ('FONT', (0, 0), (-1, -1), self._locale_italic, self._font_size),
        ])
        self['table_opt_out'] = ts


class TntWaveMailStyle(_TntStyle):

    def __init__(self, locale, scale):

        # super
        _TntStyle.__init__(self, locale, scale)

        # locale-normal uses locale specific font
        paraLocaleSubject = ParagraphStyle('locale-subject', parent=self['locale-normal'], spaceAfter = 12)
        paraLocaleSubject.fontName = self._locale_font_bold
        paraLocaleSubject.fontSize = self._scale * 12
        self['locale-subject'] = paraLocaleSubject

        # locale-normal uses locale specific font
        paraMailLocale = ParagraphStyle('locale-mail', parent=self['locale-normal'], spaceAfter = 12)
        paraMailLocale.fontName = self._locale_font
        if self.wrap_words:
            paraMailLocale.wordWrap = 'CJK'
        self['locale-mail'] = paraMailLocale
