# -*- coding: utf-8 -*-
# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.

import os

from reportlab.pdfgen import canvas
from reportlab.platypus import BaseDocTemplate, PageTemplate
from reportlab.lib.pagesizes import A4
from reportlab.lib.units import mm
from reportlab.lib import colors
from reportlab.lib.units import cm
from reportlab.platypus import Frame

img_path = '%s/img/' % os.path.dirname(__file__)

class TntWaveCanvas38(canvas.Canvas):
    """
    Canvas is the common background for all pages in the pdf.
    This class it draws the specific 38 wave design like the rounded border, logo, page numbering etc
    """

    # set by template in _startBuild
    wave_title = ""
    wave_subtitle = ""
    wave_ident = ""

    # space before border
    border_margin_x = 0.3*cm
    border_margin_top = 0.5*cm
    border_margin_bottom = 2.6*cm

    # border line thickness
    border_width = 0.2*mm
    border_radius = 0.5*cm

    def __init__(self, *args, **kwargs):
        canvas.Canvas.__init__(self, *args, **kwargs)
        self._saved_page_states = []

    def showPage(self):
        self._saved_page_states.append(dict(self.__dict__))
        self._startPage()

    def save(self):
        """
        """

        # see http://code.activestate.com/recipes/576832/
        for state in self._saved_page_states:
            self.__dict__.update(state)

            self.draw_header()
            self.draw_footer()
            self.draw_border()

            canvas.Canvas.showPage(self)

        canvas.Canvas.save(self)

    def draw_border(self):
        """
        draws a rounded rectangle border around the page
        """

        self.setLineWidth(self.border_width)
        self.setStrokeColor(colors.HexColor(0xf18e00))
        self.roundRect(
            x=self.border_margin_x,
            y=self.border_margin_bottom,
            width=self._pagesize[0] - 2*self.border_margin_x,
            height=self._pagesize[1] - self.border_margin_bottom - self.border_margin_top,
            radius=self.border_radius,
            stroke=1,
            fill=0)

    def draw_header(self):
        """
        draws the header of the page.. the title, 'wave' image and subtitle
        """

        # draw the header 'wave' image based on the size of the document and take the border into account
        border_size = self.border_margin_x
        image_width = self._pagesize[0]-2*border_size
        image_height = image_width / 17.72
        self.drawImage('%s/header_36.png' % img_path, x=border_size, y=26*cm, width=image_width, height=image_height)

        # title (header)
        if self.wave_title is not None:
            self.setFont(self.localeFontName, 15)

            x = self.border_margin_x + 1*cm
            y = 27.8*cm

            title_lines = self.wave_title.splitlines()

            for i, line in enumerate(title_lines):

                y2 = y - (i * 0.7*cm)

                if self.locale.direction == 'rtl':
                    x = self._pagesize[0] - x
                    self.drawRightString(x, y2, line)
                else:
                    self.drawString(x, y2, line)

        # subtitle
        if self.wave_subtitle is not None:
            self.setFont(self.localeFontName, 10)
            y = 25*cm
            x = self.border_margin_x + 1*cm
            if self.locale.direction == 'rtl':
                x = self._pagesize[0] - x
                self.drawRightString(x, y, self.wave_subtitle)
            else:
                self.drawString(x, y, self.wave_subtitle)


    def draw_footer(self):

        # tnt logo
        image_width = 4.13*cm
        image_height = image_width / 1.95
        self.drawImage('%s/tnt_logo_36.png' % img_path, 17*cm, y=0.25*cm, width=image_width, height=image_height)

        # wave ident
        if self.wave_ident is not None:
            self.setFont('Sans', 8)
            self.drawString(2*cm, 1.2*cm, self.wave_ident)

        # page nr
        page_count = len(self._saved_page_states)
        self.setFont("Sans", 6)
        self.drawRightString(2*cm, 0.5*cm, "%d/%d" % (self._pageNumber, page_count))


class TntWaveTemplate38(BaseDocTemplate):
    """
    In reportlab pdf , a template can be assigned to all or one or more pages. For example the first page of the document can be
    another template than the second and rest.

    For the tnt wave 38 there is one template for all pages. Here we set the frame size (the are for the flowables) and also reference
    the canvas object used to render the background.

    Overriding template also allows us to pass in variables from the builder to the canvas (header etc)
    """

    def __init__(self, filename, wave_title, wave_subtitle, wave_ident, locale, localeFontName, **kw):

        BaseDocTemplate.__init__(self, filename, pagesize=A4, showBoundary=False, **kw)

        self.wave_title = wave_title
        self.wave_subtitle = wave_subtitle
        self.wave_ident = wave_ident
        self.locale = locale
        self.localeFontName = localeFontName

    def build(self, flowables):

        self._calc()

        margin_x = 1.5*cm

        # main frame is space between the rounded borders.
        frameMain = Frame(
            x1=margin_x,
            y1=2.6*cm,
            width= A4[0] - 2*margin_x,
            height = 22 * cm,
            id='normal',
            leftPadding=0,
            topPadding=0,
            rightPadding=0,
            bottomPadding=0
        )

        # small frame below rounded borders used for opt-out input on second page
        frameOptout = Frame(
            x1=margin_x,
            y1=0.5*cm,
            width= A4[0] - 2*margin_x,
            height = 2 * cm,
            id='normal',
            leftPadding=0,
            topPadding=0,
            rightPadding=0,
            bottomPadding=0
        )

        self.addPageTemplates([PageTemplate(id='TntWave', frames=[frameMain, frameOptout], pagesize=self.pagesize)])

        BaseDocTemplate.build(self, flowables, canvasmaker=TntWaveCanvas38)


    def _startBuild(self, *args, **kwargs):
        BaseDocTemplate._startBuild(self, *args, **kwargs)

        # pass variables from template to canvas
        self.canv.wave_title = self.wave_title
        self.canv.wave_subtitle = self.wave_subtitle
        self.canv.wave_ident = self.wave_ident
        self.canv.locale = self.locale
        self.canv.localeFontName = self.localeFontName


class TntCLMWaveMailCanvas(canvas.Canvas):
    """
    Canvas is the common background for all pages in the pdf.
    This class it draws the tnt wave mail design like the right orange border, logo, page numbering etc
    """
    # set by template in _startBuild
    mail_name = ""
    wave_ident = ""

    # space before border
    border_margin_x = 0.3*cm
    border_margin_top = 0.5*cm
    border_margin_bottom = 2.6*cm

    # border line thickness
    border_width = 0.2*mm
    border_radius = 0.5*cm

    def __init__(self, *args, **kwargs):
        canvas.Canvas.__init__(self, *args, **kwargs)
        self._saved_page_states = []

    def showPage(self):
        self._saved_page_states.append(dict(self.__dict__))
        self._startPage()

    def save(self):
        """
        """

        # see http://code.activestate.com/recipes/576832/
        for state in self._saved_page_states:
            self.__dict__.update(state)

            self.draw_footer()
            canvas.Canvas.showPage(self)

        canvas.Canvas.save(self)

    def draw_footer(self):

        # wave ident
        if self.mail_ident is not None:
            self.setFont('Sans', 8)
            self.drawString(1.5*cm, 1.2*cm, self.mail_ident)

        # page nr (show only if there are multiple pages)
        page_count = len(self._saved_page_states)
        if page_count > 1:
            self.setFont("Sans", 6)
            self.drawString(1.5*cm, 0.5*cm, "%d/%d" % (self._pageNumber, page_count))


class TntCLMWaveMailTemplate(BaseDocTemplate):
    """
    In reportlab pdf , a template can be assigned to all or one or more pages. For example the first page of the document can be
    another template than the second and rest.

    Overriding template also allows us to pass in variables from the builder to the canvas (header etc)
    """

    def __init__(self, filename, mail_name, mail_ident, locale, localeFontName, **kw):

        BaseDocTemplate.__init__(self, filename, pagesize=A4, showBoundary=False, **kw)

        self.mail_name = mail_name
        self.mail_ident = mail_ident
        self.locale = locale
        self.localeFontName = localeFontName

    def build(self, flowables):

        self._calc()

        margin_x = 1.5*cm

        # main frame is space between the rounded borders.
        frameMain = Frame(
            x1=margin_x,
            y1=2.6*cm,
            width= A4[0] - 2*margin_x,
            height = 24 * cm,
            id='normal',
            leftPadding=0,
            topPadding=0,
            rightPadding=0,
            bottomPadding=0
        )

        self.addPageTemplates([PageTemplate(id='TntCLMWaveMail', frames=[frameMain], pagesize=self.pagesize)])

        BaseDocTemplate.build(self, flowables, canvasmaker=TntCLMWaveMailCanvas)

    def _startBuild(self, *args, **kwargs):
        BaseDocTemplate._startBuild(self, *args, **kwargs)

        # pass variables from template to canvas
        self.canv.mail_name = self.mail_name
        self.canv.mail_ident = self.mail_ident
        self.canv.locale = self.locale
        self.canv.localeFontName = self.localeFontName

