# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.

from django.utils.translation import ugettext as _

from basesoft.web.contrib.notification.rpc import NotificationRPCHandler
from basesoft.web.contrib.mailing.rpc import MailingRPCHandler

from basesoft.web.core.util.controllers import StateManager

from iqportal.tnt.clm.controllers import WaveController, TntCLMLocaleController,\
    TntCLMAttributeController, TntCLMAccountController, TntCLMMailBaseController,\
    TntCLMMailingBaseController
from iqportal.tnt.clm.models import TntCLMWave, TntCLMWaveMailLocale, TntCLMLocale, TntCLMWaveLocaleRevision,\
    TntCLMAttribute

class AccountRPCHandler:
    """
    Used for creation and updates of CLM Accounts. Linking of Geolevels and/or locales.
    """
    def __init__(self, request, **kwargs):
        self.sa_session = request.Session()
        self.controller = TntCLMAccountController(request.Session(), account_id=request.session['user_id'])

    def save_account(self, account_id, state, geolevels, locales):
        self.controller.save_account_from_state(state, geolevels, locales, account_id=account_id)

    def get_account_state(self, id):
        """
        """
        return self.controller.get_account_state(id)

class LocaleRPCHandler:
    def __init__(self, request, **kwargs):
        self.sa_session = request.Session()
        self.controller = TntCLMLocaleController(request.Session(), account_id=request.session['user_id'])
        if request.session.has_key('clm_locales'):
            self.controller.set_account_locale_ids(request.session['clm_locales'])
        else:
            request.session['clm_locales'] = [l.id for l in self.controller.get_locales()]

    def get_locales(self):
        rv = []
        for l in self.controller.get_locales():
            label = l.label
            if l.service is not None:
                label = '%s (%s)' % (label, l.service.label)
            rv.append(label)
        return rv

    def get_locales_state(self):
        return [{'id': l.id,
                 'label': l.label,
                 'description': l.description,
                 'direction': l.direction,
                 'active': l.active,
                 } for l in self.controller.get_locales()]

    def get_locale_state(self, id):
        locale = self.controller.get_locales(id, False).pop()
        locale_label = None
        if locale.locale is not None:
            locale_label = locale.locale.locale
        service = None
        if locale.service is not None:
            service = locale.service.label
        return {
            'label': locale.label,
            'description': locale.description,
            'language_code': locale.language_code,
            'country': locale.country.id,
            'locale': locale_label,
            'location': locale.location,
            'location_2015': locale.location_2015,
            'service': service,
            'direction': locale.direction,
            'active': locale.active
        }

    def get_locales_dicts(self, set_default=None, only_active=True):
        ls = self.controller.get_locales(None, only_active)
        if set_default is None or not set_default:
            return [{'value': l.id, 'label': l.label} for l in ls]
        ret = []
        default_set = False
        for l in ls:
            d = {'value': l.id, 'label': l.label}
            if l.label == "XXENI":
                d['default'] = True
                default_set = True
            ret.append(d)
        # in case user does not have permission on XXENI locale, just use the first one as default
        if not default_set and len(ret) > 0:
            ret[0]['default'] = True
        return ret

    def get_services(self):
        return [{'value': s.id, 'label': s.label} for s in self.controller.get_services()]

    def save_locale(self, id, label, desc, language_code, direction, country_id, locale_id, location, location_2015, service, active):
        if id is None:
            locale = self.controller.get_clm_locales_by('label', label, False)
            if len(locale) > 0:
                return locale[0].id
        self.controller.save_or_update_locale(id, label, desc, language_code, direction, country_id, locale_id, location, location_2015, service, active)
        return True

class WaveRPCHandler:
    def __init__(self, request, **kwargs):
        self.sa_session = request.Session()
        self.request = request
        self.account_id = request.session['user_id']
        self.controller = WaveController(request.Session(), account_id=request.session['user_id'])

    def get_waves(self, value_attr='id', label_attr='name'):
        return [{'value': getattr(w, value_attr), 'label': getattr(w, label_attr)} for w in self.controller.get_waves()]

    def get_wave_types(self):
        return [{'value': wt.id, 'label': wt.label} for wt in self.controller.get_wave_types()]

    def get_locales_dicts(self, wave_id, read_only=False):
        wave_default_locale = self.controller.get_wave_default_locale(wave_id)
        available = self.controller.get_available_wave_locales(wave_id)
        ret = []
        default_set = False
        if read_only:
            for l in available:
                d = {'value': l.id, 'label': l.label}
                if l.id == wave_default_locale.id:
                    d['default'] = True
                    default_set = True
                ret.append(d)
            # in case user does not have permission on wave's default locale, just use the first one as default
            if not default_set and len(ret) > 0:
                ret[0]['default'] = True
            return ret

        avls = [l.id for l in available]

        controller = TntCLMLocaleController(self.sa_session, account_id=self.account_id)
        if self.request.session.has_key('clm_locales'):
            controller.set_account_locale_ids(self.request.session['clm_locales'])
        else:
            self.request.session['clm_locales'] = [l.id for l in controller.get_locales()]

        ls = controller.get_locales()
        statuses = self.controller.get_wave_locales_status(wave_id)
        default_set = False
        for l in ls:
            lab = l.label
            if l.id in statuses and  statuses[l.id] in [TntCLMWaveLocaleRevision.STATUS_PENDING,
                                                        TntCLMWaveLocaleRevision.STATUS_APPROVED,
                                                        TntCLMWaveLocaleRevision.STATUS_REJECTED]:
                lab = lab + " (" + TntCLMWaveLocaleRevision.statuses_i18n[statuses[l.id]] + ")"
            if l.id not in avls:
                lab = lab + " (" + _("not translated") + ")"
            d = {'value': l.id, 'label': lab}
            if l.id == wave_default_locale.id:
                d['default'] = True
                default_set = True
            ret.append(d)
        # in case user does not have permission on wave's default locale, just use the first one as default
        if not default_set and len(ret) > 0:
            ret[0]['default'] = True
        return ret

    def get_available_wave_locales_dicts(self, wave_id, locale_id=None):
        if locale_id is None:
            default_locale = self.controller.get_wave_default_locale(wave_id)
        else:
            default_locale = self.controller.get_locales(int(locale_id))[0]
        ret = []
        ls = self.controller.get_available_wave_locales(wave_id)
        default_set = False
        for l in ls:
            d = {'value': l.id, 'label': l.label}
            if l.id == default_locale.id:
                d['default'] = True
                default_set = True
            ret.append(d);
        # in case user does not have permission on wave's default locale, just use the first one as default
        if not default_set and len(ret) > 0:
            ret[0]['default'] = True
        return ret

    def get_wave_locale_revisions(self, wave_id, locale_id):
            revs = self.controller.get_wave_locale_revisions(wave_id, locale_id)
            ret = []
            if len(revs) == 0:
                return ret
            #for i in range(0, len(revs)):
            i = len(revs);
            for r in revs:
                d = {'value': r.id, 'label': _("revision") + " " + unicode(i)}
                # select latest revision (revs is ordered desc)
                if i == len(revs):
                    d['default'] = True
                ret.append(d)
                i = i - 1
            return ret

    def get_wave_setup(self, wave_id):
        return  self.controller.get_wave_setup(wave_id)

    def get_wave_schedule_data(self, wave_id):
        return  self.controller.get_wave_schedule_data(wave_id)

    def get_wave_mailing_status(self, wave_id):
        mailings = self.get_wave_schedule_data(wave_id)['mailings']
        if len(mailings) == 0:
            return True
        mc = TntCLMMailBaseController(self.sa_session, account_id=self.account_id)
        for mailing in mailings:
            if mailing['status'] == 0:
                if not mc.mail_is_approved(mailing['mail_id']):
                    return False
        return True

    def get_wave_state(self, wave_id, locale_id=None, load_empty=False):
        if locale_id == '':
            locale_id = None
        return self.controller.get_wave_state(wave_id, locale_id, load_empty=load_empty, include_license_logo=True)

    def get_wave_translation(self, wave_id, locale_id, locale_revision_id=None, include_previous_revision=None):
        if locale_id == '':
            locale_id = None
        if locale_revision_id == '':
            locale_revision_id = None

        r1 = self.controller.get_wave_translations_values(wave_id, locale_id, locale_revision_id)

        if include_previous_revision is None or not include_previous_revision:
            return r1

        revs = self.controller.get_wave_locale_revisions(wave_id, locale_id)
        r2 = None
        if len(revs) >= 2:
            #revs is ordered desc
            prev_r = None
            if locale_revision_id is None:
                prev_r = revs[1]
            else:
                i = 0
                for r in revs:
                    if r.id == locale_revision_id:
                        break;
                    i = i + 1;
                try:
                    prev_r = revs[i + 1];
                except IndexError:
                    pass
            if prev_r is not None:
                r2 = self.controller.get_wave_translations_values(wave_id, locale_id, prev_r.id)

        if r2 is None:
            return [r1, r1]
        else:
            return [r1, r2]

    def get_available_pdf_fontsizes(self):
        from iqportal.tnt.clm.models import TntCLMWaveLocale
        return TntCLMWaveLocale.pdf_fontsizes

    def save_wave_translation(self, wave_id, values):
        locale_id = values['locale']
        self.controller.save_translation(wave_id, locale_id, values)

    def set_wave_attributes_priorities(self, wave_id, locale_id, values, priority_global=False):
        if len(values) == 0:
            return
        self.controller.set_wave_attributes_priorities(wave_id, locale_id, values, priority_global)

    def save_wave(self, wave_id, values):
        cl = []
        if 'copy_locales' in values:
            cl = values['copy_locales']

        # convert to category_id to id; related to changes for ticket #3285
        for i in values['wave_attributes']:
            i['id'] = i['category_id']

        return self.controller.save_or_update_wave(wave_id, values['name'], values['code'], values['default_locale'],
                                                   values['type'], values['wave_attributes'], cl)

    def schedule_wave(self, wave_ids):
        if not isinstance(wave_ids, list):
            wave_ids = [wave_ids]
        for wid in wave_ids:
            self.controller.schedule_wave(wid)

    def set_wave_schedule_data(self, wave_id, begin_date, end_date, contact_lists, mailings, schedule=False):
        wave = self.controller.set_wave_schedule_data(wave_id, begin_date, end_date, contact_lists, mailings)
        if wave.status != TntCLMWave.STATUS_SCHEDULED and schedule:
            self.schedule_wave(wave_id)

    def copy_wave(self, wave_id, locales=None):
        wave = self.controller.clone_wave(wave_id, locales)
        return wave.id

    def set_wave_translation_status(self, wave_locale_list, status, reason=None):
        assert isinstance(wave_locale_list, list)
        waves = {}
        if len(wave_locale_list) > 0:
            for wl in wave_locale_list:
                self.controller.set_wave_translation_status(wl['wave_id'], wl['locale_id'], status, reason)
                if waves.has_key(wl['wave_id']):
                    waves[wl['wave_id']].append(wl['locale_id'])
                else:
                    waves[wl['wave_id']] = [wl['locale_id']]
        for wave_id, locales in waves.items():
            self.controller.send_status_notification(wave_id, status, locales, reason)
        return True

    def get_translations_percentage(self, wave_id):
        wave = self.controller.get_waves(wave_id)[0]
        tr = self.controller.get_translations_percentage(wave)
        return {'wave': {'id': wave_id, 'name': wave.name},
                'translations': tr}


class AttributeRPCHandler:
    def __init__(self, request, **kwargs):
        self.sa_session = request.Session()
        self.controller = TntCLMAttributeController(request.Session(), StateManager(None, request.session['user_id'], None))

    def _pretty_label(self, a):
        label = unicode(a.id) + " - "
        if len(a.default_label) <= 30:
            label = label + a.default_label
        else:
            label = label + a.default_label[0:14] + "..." +\
                    a.default_label[len(a.default_label)-14:len(a.default_label)-1]
        return label

    def get_attributes(self):
        return [{'value': a.id, 'label': a.default_label} for a in self.controller.get_attributes()]

    def get_attributes_recognizable_labels(self):
        ret = []
        for a in self.controller.get_attributes():
            d = {'value': a.id, 'title': a.default_label, 'label': self._pretty_label(a)}
            ret.append(d)
        return ret

    def get_attributes_filter_by_type(self, attribute_id=None):
        if attribute_id is None or attribute_id == '':
            return []

        attribute = self.controller.get_attributes(attribute_id).pop()
        return [{'value': a.id, 'label': a.default_label} for a in self.controller.get_attributes_by_type(attribute.type)]

    def get_scale_attributes(self):
        return [{'value': a.id, 'title': a.default_label, 'label': self._pretty_label(a)} for a in self.controller.get_attributes_by_type(TntCLMAttribute.TYPE_SCALE)]

    def merge(self, keep_attribute_id, attributes):
        self.controller.merge(keep_attribute_id, *attributes)


class TntCLMMailingRPCHandler(MailingRPCHandler):

    def __init__(self, request, **kwargs):
        self.sa_session = request.Session()
        self.controller = TntCLMMailingBaseController(self.sa_session, StateManager(None, request.session['user_id'], None))


class TntCLMMailRPCHandler:

    def __init__(self, request, **kwargs):
        self.sa_session = request.Session()
        self.locale = request.session['config']['basesoft.web.contrib.locale']
        self.controller = TntCLMMailBaseController(self.sa_session, self.locale, account_id=request.session['user_id'])
        self.locale_controller = TntCLMLocaleController(self.sa_session, account_id=request.session['user_id'])

    def mail_is_approved(self, mail_id):
        return self.controller.mail_is_approved(mail_id)

    def set_mail_translation_status(self, mail_locale_list, status, reason=None):
        assert isinstance(mail_locale_list, list)
        mails = {}
        if len(mail_locale_list) > 0:
            for ml in mail_locale_list:
                self.controller.set_mail_translation_status(ml['mail_id'], ml['locale_id'], status, reason)
                if mails.has_key(ml['mail_id']):
                    mails[ml['mail_id']].append(ml['locale_id'])
                else:
                    mails[ml['mail_id']] = [ml['locale_id']]
        for mail_id, locales in mails.items():
            self.controller.send_mail_status_notification(mail_id, status, locales, reason)
        return True

    def get_mail_locales(self, mail_type, mail_id, readOnly=False):
            """
            Returns locales dicts for Wave Mail polymorphic type.
            @param readOnly: only return mail locales (so locales for which mail has translations)
            """

            if mail_id is None and readOnly:
                return []
            locales = self.locale_controller.get_locales()
            if mail_id is None and not readOnly:
                ret = []
                for l in locales:
                    state = l.get_state()
                    # I just hope this key doesn't get changed
                    state['locale'] = state['locale'] + ' (' + _('not translated') + ')'
                    ret.append(state)
                return ret
            mail_locales = self.controller.get_mail_locales(mail_id, locales, TntCLMLocale.label)
            if readOnly:
                return [l.locale.get_state() for l in mail_locales]
            ml_ids = {}
            for l in mail_locales:
                ml_ids[l.locale.id] = l
            ret = []
            for l in locales:
                state = l.get_state()
                if l.id in ml_ids:
                    status = ml_ids[l.id].status
                    if status in [TntCLMWaveMailLocale.STATUS_PENDING,
                                  TntCLMWaveMailLocale.STATUS_APPROVED,
                                  TntCLMWaveMailLocale.STATUS_REJECTED]:
                        state['locale'] = state['locale'] + ' (' + TntCLMWaveMailLocale.statuses_i18n[status] + ')'
                else:
                    state['locale'] = state['locale'] + ' (' + _('not translated') + ')'
                ret.append(state)
            return ret

    def load_mail_with_locales(self, mail_id, locale_id):
        """
        Load the requested mail and also the list of all the locales (with their info: id, direction, etc)
        This method is called from the MailDetailPanel on load
         or when requested
        @param mail_id: {int} the id of the mail we want to load
        @param locale_id: {int} the id of the tnt mail locale we want to preview
        @return: the state of the given mail
        """
        mail_id = int(mail_id)
        locale_id = int(locale_id)

        mail = self.controller.get_mail_by_id(mail_id)
        state = self.controller.load_by_mail_id(mail_id)

        using_fallback = False # a flag that indicates if the fallback is used
        direction = 'ltr' # direction of the locale used for fallback

        locale = self.locale_controller.get_clm_locales_by('id', locale_id)[0]
        current_locale = mail.get_locale_mail(locale, TntCLMWaveMailLocale)
        locale_label = locale.label

        postfix = ''
        if mail.is_wiki:
            postfix = '_wiki'
        if not state['mail']['locales'].has_key(locale_label):
            locale_label = state['mail']['fallbackLocale']['locale']
            using_fallback = True
            fl = self.locale_controller.get_clm_locales_by('label', locale_label)[0]
            direction = fl.direction

        subject = state['mail']['locales'][locale_label]['subject%s' % postfix]
        content = mail.get_html_mail(current_locale, {
            'license': mail.groups[0].group.license,
            'wave': {
                'label': state['mail']['locales'][locale_label]['link_label%s' % postfix],
                'preview': True,
            }
        })

        return {
            'direction': direction,
            'using_fallback': using_fallback,
            'subject': subject,
            'content': content,
        }

handlers = (
    AccountRPCHandler,
    LocaleRPCHandler,
    WaveRPCHandler,
    AttributeRPCHandler,
    TntCLMMailRPCHandler,
    TntCLMMailingRPCHandler,
)
