# Copyright (c) 2011 BaseSoft B.V.
# See LICENSE for details.

from datetime import datetime
from iqportal.iqube.core.controllers import ProjectController
from sqlalchemy.sql import and_, func
from django.conf import settings

from basesoft.scheduler.task import ScheduleTask

from iqportal.tnt.clm.models import TntCLMWave, TntCLMWaveLocaleRevision, TntCLMWaveMailing, TntCLMWaveMailLocale
from iqportal.tnt.response.controllers import ResponseController
from iqportal.tnt.clm.controllers import WaveController
from iqportal.tnt.report.controllers import TntCLMTntCLMWaveAggregator

from basesoft.web.contrib.mailing.schedule import MailingTask, HtmlMailTask
from basesoft.web.contrib.mailing.models import Mailing, MailResult

from basesoft import logging
logger = logging.getLogger("iqportal.tnt.clm.schedule")

class PublishWaveTask(ScheduleTask):
    INTERVAL = 600 # every ten minutes

    def execute(self):

        rv = []

        # get all scheduled waves in the current period
        result = self.sa_session.query(TntCLMWave).filter(
            (TntCLMWave.status==TntCLMWave.STATUS_SCHEDULED) &
            (TntCLMWave.archived==0) &
            (TntCLMWave.deleted==0) &
            (TntCLMWave.begin_date <= datetime.now()) &
                ((TntCLMWave.end_date > datetime.now()) |
                (TntCLMWave.end_date==None))).all()

        if len(result) > 0:
            rc = ResponseController(self.sa_session)
            wc = WaveController(self.sa_session)

        agg = TntCLMTntCLMWaveAggregator()
        # since multiple waves could be already in the query result, update the status for all of them now
        # so they won't be selected next time
        for wave in result:
            if not self.is_approved_for_publication(wave):
                logger.info('PublishWaveTask: Cannot publish Wave [%s] due to unapproved locales' % wave.id)
                continue
            logger.info('PublishWaveTask: Changing Wave [%s] status to STATUS_ONLINE' % wave.id)
            rv.append({'wave_id': wave.id, 'msg': 'changed status to STATUS_ONLINE'})

            wc.set_wave_status(wave, TntCLMWave.STATUS_ONLINE)
            # add this wave to the agg table
            agg.run(wave.id)

        for wave in result:
            rc.make_all_static(wave)
            logger.info('PublishWaveTask: Made locales state static for wave [%s]' % wave.id)

        if len(result) > 0:
            # Flush the datasource after the waves are set to online to make them available in the report widgets
            pc = ProjectController(self.sa_session)
            ds = pc.get_datasource_by('name', '%s.TNT-CLM' % settings.SETTINGS_MODULE).pop()
            pc.flush_schema(ds.id)

        # immediately send mails to invitees
        # this will also send reminder emails for online waves
        WavePublicationMailing().execute()

        return rv

    def is_approved_for_publication(self, wave):
        revisions = self.sa_session.query(func.max(TntCLMWaveLocaleRevision.id)).select_from(TntCLMWaveLocaleRevision).filter(TntCLMWaveLocaleRevision.wave_id==wave.id).group_by(TntCLMWaveLocaleRevision.wave_id, TntCLMWaveLocaleRevision.locale_id)

        result = self.sa_session.query(TntCLMWaveLocaleRevision).filter(
            and_(TntCLMWaveLocaleRevision.wave_id==wave.id,
                 TntCLMWaveLocaleRevision.id.in_(revisions),
                 TntCLMWaveLocaleRevision.status!=TntCLMWaveLocaleRevision.STATUS_ACTIVE_APPROVED)
            ).group_by(TntCLMWaveLocaleRevision.locale_id).all()
        return len(result)==0

class CreateWaveInviteesTask(ScheduleTask):
    INTERVAL = 90

    def execute(self):

        # get all waves that are online
        result = self.sa_session.query(TntCLMWave).filter(
            and_(TntCLMWave.status==TntCLMWave.STATUS_ONLINE,
                 TntCLMWave.archived==0,
                 TntCLMWave.deleted==0,
            )).all()

        if len(result) > 0:
            rc = ResponseController(self.sa_session)

        for w in result:
            # now we add invitees to the wave.
            invitees = rc.save_or_update_invitees(w)
            if invitees > 0:
                logger.info('PublishWaveTask: Added [%s] Invitees to Wave [%s]' % (invitees, w.id))

class FinishWaveTask(ScheduleTask):
    INTERVAL = 3600 # every hour

    def execute(self):
        rv = []

        # get all waves that are supposed to finish
        result = self.sa_session.query(TntCLMWave).filter(
            and_(TntCLMWave.status==TntCLMWave.STATUS_ONLINE,
                 TntCLMWave.end_date <= datetime.now()
            )
        ).all()
        if len(result) > 0:
            wc = WaveController(self.sa_session)
        for wave in result:
            logger.info('PublishWaveTask: Changing Wave [%s] status to STATUS_FINISHED' % wave.id)
            rv.append({'wave_id': wave.id, 'msg': 'changed status to STATUS_FINISHED'})

            wc.set_wave_status(wave, TntCLMWave.STATUS_FINISHED)

            # all attached mailings will be set to status archived
            wave_mailings = wave.mailings
            for mailing in wave_mailings:
                mailing.status = Mailing.STATUS_ARCHIVED
                mailing.number_of_recipients = self.sa_session.query(MailResult).filter(MailResult.mailing==mailing).count()
                self.sa_session.add(mailing)

        self.sa_session.flush()

        return rv

class WavePublicationMailing(MailingTask):
    """
    Implements a MailingTask that takes care of sending out mails to invitees.
    This will be ran immediately after the PublishWaveTask has finished (it will not run automaticly).
    """
    INTERVAL = 600
    mailing_model = TntCLMWaveMailing
    email_host = settings.EMAIL_HOST_TNT
    email_port = settings.EMAIL_PORT_TNT

    def _filter_contact(self, mailing, contact):
        """
        This filters the contacts based on the applied filter.
        """
        if mailing.filter == TntCLMWaveMailing.FILTER_NONE:
            return False

        rc = ResponseController(self.sa_session)
        is_respondent = rc.is_respondent(contact, mailing.wave)
        if mailing.filter == TntCLMWaveMailing.FILTER_NON_RESPONDENTS:
            return not is_respondent
        else:
            return is_respondent

    def _build_message(self, contact, mailing, mail_from=None, license=None):
        """
        Mailing overrides _build_message cuz it deals with mailing objects instead of mail.. by default call parent
        with context but subclasses can extend this behaviour.
        """
        company = contact.company
        # if there isn't a company attached to this contact, use wave default locale for the mail
        locale = None
        if company is not None:
            if hasattr(company, 'locale'):
                locale = contact.company.locale
        if locale is None:
            logger.warn('Using default wave locale for contact id %s' % contact.id)
            wc = WaveController(self.sa_session)
            locale = wc.get_wave_default_locale(mailing.wave.id)

        mail = mailing.mail
        wave = {}
        if self.mail_locales.has_key(locale.id):
            mail_locale = self.mail_locales[locale.id]
        else:
            mail_locale = mail.get_locale_mail(locale, TntCLMWaveMailLocale)
            self.mail_locales[locale.id] = mail_locale
        if mail_locale:
            wave['label'] = mail_locale.link_label

        context = {
            'mailing': mailing,
            'contact': contact,
            'wave': wave,
            'license': license
        }

        subject = mail_locale.subject
        html = mail.get_html_mail(mail_locale, context, mail_locale.content).encode('utf-8')

        return HtmlMailTask._build_message(self, contact.person, mail, subject=subject, html=html,
                                           context=context, mail_from=mail_from)

# publish tasks so the scheduler can pick them up
tasks = (PublishWaveTask, CreateWaveInviteesTask, FinishWaveTask, )
