# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.
from datetime import date, datetime

from django.http import HttpResponse

from basesoft.web.core.jsonrpc.dispatcher import JsonRpcDispatcher
from basesoft.web.contrib.account.decorators import login_required
from basesoft.web.core.forms.handler import FormValidator
from basesoft.web.core.input.validation import plugins
from basesoft.web.contrib.i18n.models import Country

from iqportal.tnt.clm.models import TntCLMWave
from iqportal.tnt.response.controllers import ResponseController
from iqportal.tnt.clm.controllers import TntCLMAttributeController, TntCLMScaleAnswerController,\
                                         WaveController, TntCLMMailBaseController
from basesoft.web.core.util.controllers import StateManager
from basesoft.web.contrib.license.views import LicenseRpc

def clm_company_form_rpc(request):
    """
    """
    form_handler = FormValidator({
        'subsector': [(plugins.Number, {})],
        'imp_code': [(plugins.Number, {})],
        'revenue_band': [(plugins.RestrictedLength, {'max': 5})]
    })
    dispatcher = JsonRpcDispatcher()
    dispatcher.register_instance(form_handler)
    return dispatcher.handle_request(request)

def wave_form_rpc(request):
    """
    """
    form_handler = FormValidator({
        'name': [(plugins.Required, {})],
        'type': [(plugins.Required, {})],
        'default_locale': [(plugins.Required, {})],
        'wave_label': [(plugins.Required, {})]
    })
    dispatcher = JsonRpcDispatcher()
    dispatcher.register_instance(form_handler)
    return dispatcher.handle_request(request)

def locale_form_rpc(request):
    """
    """
    form_handler = FormValidator({
        'label': [(plugins.Required, {}), (plugins.RestrictedLength, {'max': 6, 'min': 4})],
        'description': [(plugins.Required, {})],
        'country': [(plugins.Required, {})],
        'direction': [(plugins.Required, {})],
        'location': [(plugins.Required, {})],
        'location_2015': [(plugins.Required, {})]
    })
    dispatcher = JsonRpcDispatcher()
    dispatcher.register_instance(form_handler)
    return dispatcher.handle_request(request)

def attribute_rpc(request):

    controller = TntCLMAttributeController(request.Session(), StateManager(None, request.session['user_id'], None))

    class Handler:

        def get_clm_locales(self):
            if request.session.has_key('clm_locales'):
                controller.locale_controller.set_account_locale_ids(request.session['clm_locales'])
            else:
                controller.locale_controller.set_account_locale_ids()

            locales = []
            rv = []
            for l in controller.locale_controller.get_locales():
                rv.append(l.label)
                locales.append(l.id)

            request.session['clm_locales'] = locales
            return rv

        def save_attribute(self, state):
            id = controller.save_attribute(state)
            return controller.load_attribute(id)

        def load_attribute(self, id):
            return controller.load_attribute(id)


    dispatcher = JsonRpcDispatcher()
    dispatcher.register_instance(Handler())
    return dispatcher.handle_request(request)

@login_required
def scale_answer_rpc(request):

    controller = TntCLMScaleAnswerController(request.Session(), StateManager(None, request.session['user_id'], None))

    class Handler:

        def get_wave_types(self):
            types = [{'value': e.id, 'label': e.label, 'default': False} for e in controller.get_wave_types()]
            if len(types) > 0:
                types[0]['default'] = True

            return types

        def save_scale_answer(self, state):
            id = controller.save_scale_answer(state)
            return controller.load_scale_answer(id)

        def load_scale_answer(self, id):
            return controller.load_scale_answer(id)

        def move_scale_answer(self, id, move_up):
            """
            Changes the position of the scale answer.
            @param id: the id of the scale answer that we want to move up or down
            @param move_up: a boolean that indicates if we want to move the scale answer up. true if so, false if we
                want to move it down
            @return: an int that indicates the result
            """
            return controller.move_scale_answer(id, move_up)

    dispatcher = JsonRpcDispatcher()
    dispatcher.register_instance(Handler())
    return dispatcher.handle_request(request)

@login_required
def download_wave_pdf(request, wave_id, locale_id=None, locale_revision_id=None):
    """
    returns a response with pdf data in it
    """

    wave_id = int(wave_id)

    try: locale_id = int(locale_id)
    except ValueError: locale_id = None

    try: locale_revision_id = int(locale_revision_id)
    except ValueError: locale_revision_id = None

    # create response object which will be passed into the builder to write the pdf data into
    response = HttpResponse(mimetype='application/pdf')

    # build pdf
    controller = WaveController(request.Session(), account_id=request.session['user_id'])
    filename = controller.export_wave_pdf(response, wave_id, locale_id)

    # set filename in response
    response['Content-Disposition'] = 'attachment; filename=%s' % filename

    return response

@login_required
def download_mail_pdf(request, mail_id, locale_id=None):
    """
    returns a response with pdf data in it
    """

    mail_id = int(mail_id)

    try: locale_id = int(locale_id)
    except ValueError: locale_id = None

    # create response object which will be passed into the builder to write the pdf data into
    response = HttpResponse(mimetype='application/pdf')

    # build pdf
    controller = TntCLMMailBaseController(request.Session(), account_id=request.session['user_id'])
    filename = controller.export_mail_pdf(response, mail_id, locale_id)

    # set filename in response
    response['Content-Disposition'] = 'attachment; filename=%s' % filename

    return response

@login_required
def export_wave_csv(request, wave_id, hash, country_id=''):

    sa_session = request.Session()

    c = ResponseController(sa_session, account_id=request.session['user_id'])
    file = c.download_response_csv(hash)

    wave = sa_session.query(TntCLMWave).filter(TntCLMWave.id==wave_id).one()

    country = ''
    if country_id != '':
        country = '_%s' % sa_session.query(Country).filter(Country.id==country_id).one().iso_3166_a2

    response = HttpResponse(file, mimetype='application/excel')

    wave_name = wave.name.replace(' ', '_')
    response['Content-Disposition'] = 'attachment; filename=%s%s_%s.xls' % (wave_name, country, date.strftime(datetime.now(), '%d-%m-%y_%H-%I'))

    return response

class TntPrivsDelegation(LicenseRpc):

    def _set_datasource_stale(self):
        from iqportal.iqube.core.controllers import DataSourceController
        from django.conf import settings
        ds_controller = DataSourceController(self.session, account_id=self.account_id)
        ds_controller.set_stale('%s.TNT-CLM' % settings.SETTINGS_MODULE)

    def save_privilege_delegation(self, object, id, group_id, privileges, group_has_privilege, license_has_privilege=None, domain_has_privilege=None):
        privileges = [int(p) for p in privileges]
        self.controller.save_privilege_delegation(object, id, int(group_id), privileges, group_has_privilege, license_has_privilege, domain_has_privilege)

        # the tnt datasource needs to be set as stale
        self._set_datasource_stale()
        return True

    def delete_group_privileges(self, object, id, group_ids):
        self.controller.delete_group_privileges(object, id, group_ids)

        # the tnt datasource needs to be set as stale
        self._set_datasource_stale()
        return True

@login_required
def privs_delegation_rpc(request):
    """
    """
    user_id = request.session['user_id']
    sa_session = request.Session()

    dispatcher = JsonRpcDispatcher()
    dispatcher.register_instance(TntPrivsDelegation(sa_session, user_id))
    return dispatcher.handle_request(request)
