# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.

"""
TNT Core Controller module
"""
import re
import datetime

from sqlalchemy import and_
from sqlalchemy.sql.expression import distinct
from django.conf import settings
from django.utils.translation import ugettext as _

from basesoft.web.core.util.controllers import AbstractController, transaction_decorator
from basesoft.web.contrib.i18n.controllers import I18nController
from basesoft.web.core.exceptions import SecurityException, PublicUserException

from iqportal.tnt.clm.models import TntCLMService, TntCLMCompany, TntCLMAccount
from iqportal.iqube.core.controllers import DataSourceController
from iqportal.iqube.core.exceptions import IQubeSystemException
from iqportal.tnt.core.models import TntGeolevel, TntGeolevelGroup, TntDepot, TntDepotAlias, TntGeolevelType

class TntController(AbstractController):
    """
    Contains methods for geolevel manipulation
    """

    def get_geolevels(self, id=None, type_id=None, privs=TntGeolevel.PRIV_VIEW):
        """
        Return a list with TntGeolevel.
        """

        self.assert_permissions('view_geolevel')
        account = self.get_account()

        if hasattr(account, 'geolevels'):
            glids = [g.id for g in account.all_geolevels]
            if id is not None:
                if not isinstance(id, list):
                    id = [id]
                for x in id:
                    if int(x) not in glids:
                        raise SecurityException('Trying to access geolevel without the proper rights [id=%s]' % x)
            elif id is None:
                id = glids

        filter, fo = self.get_query_filter_on_privileges(TntGeolevel, TntGeolevelGroup, privs)

        if id is not None:
            if not isinstance(id, list):
                id = [id]
            filter = and_(filter, TntGeolevel.id.in_(id))

        if type_id is not None:
            if not isinstance(type_id, list): type_id = [type_id]
            filter = and_(filter, TntGeolevel.type_id.in_(type_id))

        filter = and_(filter, and_(TntGeolevel.deleted==False, TntGeolevel.archived==False))

        return self.sa_session.query(TntGeolevel).select_from(fo).filter(filter).all()

    def get_geolevel_types_by(self, field=None, values=None):
        """
        Return a list of TntGeolevelType objects.
        """
        if field is not None:
            assert field in ['name', 'id']
            if values is not None:
                return self.sa_session.query(TntGeolevelType).filter(getattr(TntGeolevelType, field).in_(values)).all()
        return self.sa_session.query(TntGeolevelType).all()

    def get_services(self):
        return self.sa_session.query(TntCLMService).all()

    def get_revenue_bands(self):
        """
        gets all possible revenue bands
        """
        return self.sa_session.query(distinct(TntCLMCompany.revenue_band)).all()


    def isChildOf(self, gid, geolevel):
        """
        Recursively checks if gid is child of geolevel
        """
        if len(geolevel.child_geolevels) == 0:
            return False
        gids = [g.id for g in geolevel.child_geolevels]
        if gid in gids:
            return True
        for child in geolevel.child_geolevels:
            return self.isChildOf(gid, child)
        return False

    @transaction_decorator
    def save_geolevel_state(self, state):
        """
        Saves a geolevel from state;
        State example:
            {
                dbID: None,
                name: u'Benelux',
                typeName: u'TntBusinessUnit',
                countries: [22, 155, 128], # ids of the countries: Belgium, Netherlands, Luxembourg
                accounts: [1,4] #ids of the accounts geolevel will be linked to
                parent_geolevels: [] #ids of the geolevels this will be attached to
            }
        """
        assert isinstance(state['countries'], list)
        assert isinstance(state['accounts'], list)
        assert isinstance(state['parent_geolevels'], list)

        name = state['name'].strip()

        if state['dbID'] not in [None, '']:
            self.assert_permissions('update_geolevel')
            g = self.sa_session.query(TntGeolevel).filter(TntGeolevel.id==state["dbID"]).one()
            self.assert_privileges_on_object(g, TntGeolevel.PRIV_UPDATE)
            # need to check if countries or parent geolevels have changed in order to set correct last_updated
            # so that the aggregators know that there was a modification
            # only if name and type have not changed, else last_updated will be set automatically
            if g.name == name and g.type.name == state['typeName']:
                cids = [int(c) for c in state['countries']]
                gcids = [c.id for c in g.countries]
                pids = [int(c) for c in state['parent_geolevels']]
                gpids = [c.id for c in g.parent_geolevels]
                if cids != gcids or pids != gpids:
                    g.last_updated = datetime.datetime.now()
        else:
            self.assert_permissions('create_geolevel')
            g = TntGeolevel()

        # type
        type = self.get_geolevel_types_by('name', [state['typeName']])
        if len(type) == 0:
            type = self.save_or_update_geolevel_type(None, state['typeName'])
        else:
            type = type[0]
        g.type = type

        # check valid name for mdx
        p = re.compile('^[\w\s\d\.\-\(\)\']+$')
        if len(name) == 0 or not p.match(name):
            raise PublicUserException(_("Not a valid name"))
        g.name = name

        # countries
        if len(state['countries']) > 0:
            i18n = I18nController(self.sa_session)
            g.countries = i18n.get_countries(id = state['countries'])
        else:
            g.countries = []

        # accounts
        if len(state['accounts']) > 0:
            from basesoft.web.contrib.account.controllers import AccountController
            acc = AccountController(self.sa_session, self.locale, self.account_id)
            g.accounts = acc.get_accounts(id = state['accounts'])
        else:
            g.accounts = []

        # attached to
        parent_geolevels = []
        if len(state['parent_geolevels']):
            all_geolevels = self.get_geolevels()
            all_geolevel_objs = {} # exclude the ones that are children
            for gl in all_geolevels:
                if not self.isChildOf(gl.id, g):
                    all_geolevel_objs[gl.id] = gl
            for parent in state['parent_geolevels']:
                if not all_geolevel_objs.has_key(int(parent)):
                    raise SecurityException('Invalid parent geolevel %s' % parent)
                parent_geolevels.append(all_geolevel_objs[int(parent)])

        # check all new parent geolevels: one should not be already child of another
        for pg in parent_geolevels:
            for p in parent_geolevels:
                if p.id == pg.id: # don't check with itself
                    continue
                if self.isChildOf(p.id, pg):
                    raise SecurityException('Invalid parent geolevels. %s is child of %s' % (p.id, pg.id))
        g.parent_geolevels = parent_geolevels

        self.sa_session.add(g)

        try:
            # mark the datasource as stale
            ds_controller = DataSourceController(self.sa_session, account_id=self.account_id)
            ds_controller.set_stale('%s.TNT-CLM' % settings.SETTINGS_MODULE)
        except IQubeSystemException:
            pass

        if state['dbID'] is None:
            self.save_default_privilege_object_group(g)
            # all new geolevels will be available for everybody in tnt license
            # (so adding View priv to license level)
            self.grant_privilege_to(g, TntGeolevel.PRIV_VIEW, True, True)

        return g

    @transaction_decorator
    def save_or_update_geolevel_type(self, id, name):
        """
        """
        if id is None:
            gt = TntGeolevelType()
        else:
            gt = self.get_geolevel_types_by('id', [id])
        gt.name = name
        self.sa_session.add(gt)
        return gt

    def load_geolevel_state(self, id):
        """
        Returns a geolevel state
        @param id: the id of the geolevel object that we want
        @return: the state of the object or an empty dict if the geolevel was not found

        State example:
            {
                dbID: None,
                name: u'Benelux',
                typeName: u'TntBusinessUnit',
                countries: [22, 155, 128], # ids of the countries: Belgium, Netherlands, Luxembourg
                accounts: [],
                parent_geolevels: [],
                child_geolevels: []
            }
        """
        if not isinstance(id, TntGeolevel):
            results = self.get_geolevels(id)

            if len(results) == 0:
                return {}

            g = results[0]
        else:
            g = id
        state = {
             'dbID': g.id,
             'name': g.name,
             'typeName': g.type.name,
             'countries': [c.id for c in g.countries],
             'accounts': [a.id for a in g.accounts],
             'parent_geolevels': [a.id for a in g.parent_geolevels],
             'child_geolevels': [a.id for a in g.child_geolevels]
        }
        return state

    @transaction_decorator
    def delete_geolevels(self, id):
        """
        Delete the specified geolevel(s)

        @param id: {int} or {List} of ints that represent the ids of the geolevels we want to delete
        """

        self.assert_permissions('delete_geolevel')

        if not isinstance(id, list):
            id = [id]

        for i in id:
            self.mark_as_deleted(TntGeolevel, i)

    def get_depots(self, id=None, country_id=None, priv=TntDepot.PRIV_VIEW):
        """
        Returns a list with TntDepot's
        """

        self.assert_permissions('view_depot')

        where = None

        if id is not None:
            if not isinstance(id, list):
                id = [id]
            where = TntDepot.id.in_(id)

        if country_id is not None:
            if not isinstance(country_id, list):
                country_id = [country_id]

        account = self._get_account()
        # only CLM accounts filter on country.
        if isinstance(account, TntCLMAccount):
            ac = [c.id for c in account.countries]
            if country_id is None:
                country_id = ac
            else:
                country_id = country_id + ac

        if country_id is not None:
            if where is None:
                where = TntDepot.country_id.in_(country_id)
            else:
                where = where & TntDepot.country_id.in_(country_id)

        return self._get_privileged_object(TntDepot, where, priv)

    @transaction_decorator
    def save_depot_state(self, state):
        """
        Saves a depot from state;
        State example:
            {
                dbID: None,
                default_code: u'Home Depot',
                country: 22, # id of the country: Belgium
                service: 1, # id of the service
                new_aliases: [{'alias': 'Alias 1'}, {'alias': 'Alias 2'}]
            }
        """
        if state['dbID']:
            self.assert_permissions('update_depot')
            d = self.sa_session.query(TntDepot).filter(TntDepot.id==state["dbID"]).one()
            self.assert_privileges_on_object(d, TntDepot.PRIV_UPDATE)
        else:
            self.assert_permissions('create_depot')
            d = TntDepot()

        i18n = I18nController(self.sa_session)

        # TODO: #3284 this should not be depending on CLM.
        # import here, else we get a looping import
        from iqportal.tnt.clm.controllers import TntCLMLocaleController
        lc = TntCLMLocaleController(self.sa_session)

        countries = i18n.get_countries(id = int(state["country"]))
        assert len(countries) == 1 # we can only have one country
        d.country = countries[0]
        d.default_code = state['default_code']
        d.label = state['label']

        if state["service"]:
            d.service = lc.get_services(state["service"]).pop()

        if state['dbID'] in [None, '']:
            self.save_default_privilege_object_group(d)

        if self.has_permissions('depot_alias_manager'):
            for a in d.aliases:
                self.sa_session.delete(a)
            d.aliases = []
            # update aliases
            for new_alias in state['aliases']:
                if new_alias['alias'] is None:
                    continue
                alias_code = new_alias['alias'].strip()
                if alias_code == '':
                    continue
                na = TntDepotAlias()
                na.code = alias_code
                d.aliases.append(na)
        self.sa_session.add(d)
        return d

    def load_depot_state(self, id):
        """
        Returns a depot state
        @param id: the id of the depot object that we want
        @return: the state of the object or an empty dict if the depot was not found

        State example:
            {
                dbID: 3,
                default_code: u'Home Depot',
                countries: 22, # id of the country: Belgium
                existing_aliases: 'Alias 1, Alias 2' # comma separated list of the aliases names
            }
        """
        results = self.get_depots(id)

        if len(results) == 0:
            return {}

        d = results[0]
        service = None
        if d.service is not None:
            service = d.service.id

        state = {
             'dbID': d.id,
             'default_code': d.default_code,
             'label': d.label,
             'country': d.country.id,
             'service': service,
             'aliases': [{'alias': a.code } for a in d.aliases]
        }

        return state

    @transaction_decorator
    def delete_depots(self, id):
        """
        Delete the specified depot(s)

        @param id: {int} or {List} of ints that represent the ids of the depots we want to delete
        """

        self.assert_permissions('delete_depot')

        if not isinstance(id, list):
            id = [id]

        for i in id:
            self.mark_as_deleted(TntDepot, i)
