# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.

from sqlalchemy import func
from django.utils.translation import ugettext as _

from basesoft.web.core.util.custom_query import CustomQuery, CustomQueryField, CustomQueryModel
from basesoft.web.contrib.i18n.custom_query import DomainCountryCustomQueryModel,\
    DomainCountryCustomQueryField

from iqportal.tnt.core.models import TntGeolevel, TntGeolevelCountry, TntGeolevelAccount, TntDepot, TntDepotAlias, TntGeolevelType
from iqportal.tnt.clm.models import TntCLMService

class GeolevelCustomQuery(CustomQuery):
    permissions = ['geolevel_manager']

    def tables(self):
        return (
            CustomQueryModel(TntGeolevel),
            CustomQueryModel(TntGeolevelCountry, onclause=TntGeolevel.id==TntGeolevelCountry.geolevel_id, use_outerjoin=True),
            CustomQueryModel(TntGeolevelAccount, onclause=TntGeolevel.id==TntGeolevelAccount.geolevel_id, use_outerjoin=True),
            CustomQueryModel(TntGeolevelType, onclause=TntGeolevel.type_id==TntGeolevelType.id)
        )
    def fields(self):
        return (
            CustomQueryField(TntGeolevel.id, _('ID'), key='id'),
            CustomQueryField(TntGeolevel.name, _('Name'), is_filter=True),
            CustomQueryField(TntGeolevelType.name, _('Type'), key='geolevel_type', is_filter=True, is_group_by=True),
            CustomQueryField(func.count(TntGeolevelCountry.country_id.distinct()), _('Countries'), key='countries'),
            CustomQueryField(func.count(TntGeolevelAccount.account_id.distinct()), _('Accounts'), key='accounts'),
        )
    group_by = (TntGeolevel.id, )

    def set_where(self):
        account = self.get_account()
        if not hasattr(account, 'geolevels'):
            return None
        self.where = TntGeolevel.id.in_([g.id for g in account.geolevels])


class DepotCustomQuery(CustomQuery):
    permissions = ['depot_manager']

    def tables(self):
        return (
            CustomQueryModel(TntDepot),
            DomainCountryCustomQueryModel(),
            # TODO: #3284 service is part of CLM whereas depot is part of the Core app. Service should be part of the core app (future, if ever we get more than one app for TNT)
            CustomQueryModel(TntCLMService, onclause=TntCLMService.id==TntDepot.service_id, use_outerjoin=True),
            CustomQueryModel(TntDepotAlias, onclause=TntDepotAlias.depot_id==TntDepot.id, use_outerjoin=True),
        )

    def fields(self):
        return (
            CustomQueryField(TntDepot.id, _('ID')),
            CustomQueryField(TntDepot.default_code, _('Default code'), key='tntdepot_default_code', is_filter=True),
            CustomQueryField(TntDepot.label, _('Label'), key='tntdepot_label', is_filter=True),
            CustomQueryField(TntCLMService.label, _('Service'), is_filter=True),
            CustomQueryField(func.group_concat(' ', TntDepotAlias.code), _('Aliases')),
            DomainCountryCustomQueryField(None, _('Country'), key="country", is_filter=True, is_group_by=True),
            CustomQueryField(TntDepotAlias.code, _('Alias'), in_select=False, is_filter=True),
        )

    group_by = (TntDepot.id, )


custom_queries = (GeolevelCustomQuery, DepotCustomQuery, )
