# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.

"""
TNT Report Controller module
"""
from django.conf import settings
from django.utils import simplejson

from basesoft.web.contrib.dashboard.controllers import DashboardWidgetStateController
from basesoft.web.core.util.controllers import AbstractController, transaction_decorator
from basesoft.web.core.jsonrpc.dispatcher import JsonRpcJSONEncoder
from basesoft.web.core.jsonrpc.dispatcher import json_object_hook

from iqportal.iqube.core.controllers import DataSourceController, MdxMemberController, IQubeAggregator
from iqportal.tnt.report.models import TntCLMDashboardWidget, TntCLMDashboardWidgetFilter
from iqportal.iqube.compose.controllers import MdxCalculatedMemberController

from basesoft import logging

logger = logging.getLogger('iqportal.tnt.report')

class TntBusinessFunctionController(AbstractController):

    def get_business_functions(self, _id=None):
        self.assert_permissions('tnt_view_business_function')
        cmc = MdxCalculatedMemberController(self.sa_session, account_id=self.get_account().id)
        return cmc.get_calculated_member(_id)

    def get_business_functions_by_dimension(self, dimension):
        self.assert_permissions('tnt_view_business_function')
        cmc = MdxCalculatedMemberController(self.sa_session, account_id=self.get_account().id)
        ds_controller = DataSourceController(self.sa_session, account_id=self.account_id)
        datasource = ds_controller.get_datasource_by('mondrian_project', '%s.TNT-CLM' % settings.SETTINGS_MODULE).pop()

        return cmc.get_calculated_members_for(datasource.id, 'TntCLMResponse', dimension=dimension)

class TntResponseFullAggregator(IQubeAggregator):

    _table = "aggTntCLMFullWaveResponse"
    _columns = [
        'id',
        'wave_id',
        'participant_id',
        'attribute_id',
        'answer_id',
        'scale_answer_id',
        'participant_session_id',
        'response_text',
        'invitee_id',
        'depot_id',
        'locale_id',
        'service_id',
        'location',
        'location_2015',
        'revenue_band',
        'is_respondent',
        'is_processed',
        'created',
        'last_updated',
    ]

    _select_sql = """
        SELECT  r.id,
                i.wave_id,
                p.id as participant_id,
                attr.attribute_id,
                ans.answer_id,
                sc.scale_answer_id,
                r.participant_session_id,
                r.response_text,
                i.id as invitee_id,
                co.depot_id,
                p.locale_id,
                l.service_id,
                l.location,
                l.location_2015,
                co.revenue_band,
                p.is_respondent,
                p.is_processed,
                r.created,
                r.last_updated
        FROM tblTntCLMWaveResponse r
        INNER JOIN tblTntCLMWaveParticipant p ON r.participant_id=p.id
        INNER JOIN tblTntCLMWaveInvitee i ON p.invitee_id=i.id
        INNER JOIN tblContact c ON i.contact_id=c.id
        INNER JOIN tblTntCLMCompany co ON c.company_id=co.company_id
        INNER JOIN tblTntCLMLocale l ON co.locale_id = l.id
        INNER JOIN tblTntCLMWaveAttribute attr ON r.wave_attribute_id = attr.id
        LEFT JOIN tblTntCLMWaveAnswer ans ON r.wave_answer_id = ans.id
        LEFT JOIN tblTntCLMWaveScaleAnswer sc ON r.wave_scale_answer_id = sc.id
        WHERE response_text is NULL AND r.last_updated > %(max_last_updated)s
        ORDER BY r.id
    """

    def run(self, loop_validator=None):

        rows = self.db_engine_ro.execute("""SELECT MAX(last_updated) FROM aggTntCLMFullWaveResponse""")
        for max_last_updated, in rows:
            break
        rows.close()

        if max_last_updated is None: max_last_updated = 0

        return self._work({
            'max_last_updated': max_last_updated,
        }, loop_validator)

class TntTrueResponseCubeAggregator(IQubeAggregator):

    _table = "aggTntCLMTrueWaveResponse"
    _columns = [
        'answer_id',
        'attribute_type',
        'attribute_id',
        'scale_answer_id',
        'country_id',
        'depot_id',
        'service_id',
        'is_main',
        'wave_id',
        'revenue_band',
        'location',
        'location_2015',
        'is_respondent',
        'is_processed',
        'responses',
        'fact_count',
        'created',
    ]

    _select_sql = """
        select
            `tblTntCLMAttributeAnswer`.`id` as `answer_id`,
            `tblTntCLMAttribute`.`type` as `attribute_type`,
            `tblTntCLMAttribute`.`id` as `attribute_id`,
            `tblTntCLMScaleAnswer`.`id` as `scale_answer_id`,

            `tblTntDepot`.`country_id` as `country_id`,
            `tblTntDepot`.`id` as `depot_id`,

            `tblTntCLMService`.`id` as `service_id`,
            `aggTntCLMTntCLMWave`.`wave_is_main_id` as `is_main`,
            `aggTntCLMTntCLMWave`.`wave_id` as `wave_id`,
            `aggTntCLMFullWaveResponse`.`revenue_band` as `revenue_band`,
            `aggTntCLMFullWaveResponse`.`location` as `location`,
            `aggTntCLMFullWaveResponse`.`location_2015` as `location_2015`,

            `tblTntCLMWaveRespondentType`.`id` as `is_respondent`,
            `aggTntCLMFullWaveResponse`.`is_processed` as `is_processed`,

            count(`aggTntCLMFullWaveResponse`.`id`) as `responses`,

            count(*) as `fact_count`,
            now() as `created`
        from
            `aggTntCLMFullWaveResponse` as `aggTntCLMFullWaveResponse`

            inner join `aggTntCLMTntCLMWave` as `aggTntCLMTntCLMWave` on `aggTntCLMFullWaveResponse`.`wave_id` = `aggTntCLMTntCLMWave`.`wave_id` and `aggTntCLMFullWaveResponse`.`wave_id` = %(wave_id)s

            inner join `tblTntDepot` as `tblTntDepot` on `aggTntCLMFullWaveResponse`.`depot_id` = `tblTntDepot`.`id` and `tblTntDepot`.`id` = %(depot_id)s

            inner join `tblTntCLMService` as `tblTntCLMService` on `aggTntCLMFullWaveResponse`.`service_id` = `tblTntCLMService`.`id`
            inner join `tblTntCLMWaveRespondentType` as `tblTntCLMWaveRespondentType` on `aggTntCLMFullWaveResponse`.`is_respondent` = `tblTntCLMWaveRespondentType`.`id`

            INNER JOIN `tblTntCLMAttribute` AS `tblTntCLMAttribute` ON `aggTntCLMFullWaveResponse`.`attribute_id` = `tblTntCLMAttribute`.`id`
            LEFT JOIN `tblTntCLMAttributeAnswer` AS `tblTntCLMAttributeAnswer` ON `aggTntCLMFullWaveResponse`.`answer_id` = `tblTntCLMAttributeAnswer`.`id`
            LEFT JOIN `tblTntCLMScaleAnswer` AS `tblTntCLMScaleAnswer` ON `aggTntCLMFullWaveResponse`.`scale_answer_id` = `tblTntCLMScaleAnswer`.`id`

        group by
            `tblTntCLMAttributeAnswer`.`id`,
            `tblTntCLMAttribute`.`type`,
            `tblTntCLMAttribute`.`id`,
            `tblTntCLMScaleAnswer`.`id`,
            `tblTntDepot`.`country_id`,
            `tblTntDepot`.`id`,
            `tblTntCLMService`.`id`,
            `aggTntCLMTntCLMWave`.`wave_is_main_id`,
            `aggTntCLMTntCLMWave`.`wave_id`,
            `aggTntCLMFullWaveResponse`.`revenue_band`,
            `aggTntCLMFullWaveResponse`.`location`,
            `aggTntCLMFullWaveResponse`.`location_2015`,
            `tblTntCLMWaveRespondentType`.`id`,
            `aggTntCLMFullWaveResponse`.`is_processed`;
    """



    def run(self, loop_validator=None):

        data_modified = False

        rows = self.db_engine_ro.execute("""
            SELECT i.wave_id, company.depot_id, UNIX_TIMESTAMP(MAX(r.last_updated))
            FROM aggTntCLMFullWaveResponse AS r
            INNER JOIN tblTntCLMWaveParticipant AS p
                ON r.participant_id = p.id
            INNER JOIN tblTntCLMWaveParticipantSession AS ps
                ON p.id = ps.participant_id
            INNER JOIN tblTntCLMWaveInvitee AS i
                ON p.invitee_id = i.id
            INNER JOIN tblContact contact
                ON i.contact_id = contact.id
            INNER JOIN `tblTntCLMCompany` company
                ON contact.company_id = company.company_id
            GROUP BY i.wave_id, company.depot_id
        """)

        for wave_id, depot_id, last_response in rows:

            last_response_agg, = self.db_engine_ro.execute("""
                SELECT UNIX_TIMESTAMP(MAX(created))
                FROM """+self._table+"""
                WHERE `wave_id` = %s
                AND `depot_id`= %s
            """, wave_id, depot_id).fetchone()

            if last_response_agg is None or (last_response > last_response_agg):

                self._work({
                    'wave_id': wave_id,
                    'depot_id': depot_id,
                    'last_response': last_response,
                }, loop_validator)

                data_modified = True

        rows.close()

        return data_modified

    # code below can be used when we want to aggregate per wave/depot/imp?/sic?
    """
    def _work(self, where_args, loop_validator=None):

        ds_controller = DataSourceController(Session())
        datasource = ds_controller.get_datasource_by('mondrian_project', '%s.TNT-CLM' % settings.SETTINGS_MODULE).pop()
        #self._client = datasource.connect()

        cubes = [
            #'TntCLMResponse',
            'TntCLMVirtualCube',
            'TntCLMTrueResponseCube',
            'TntCLMTrueRespondentsCube'
        ]

        for cube in cubes:
            ic = IQubeController()

            members = [
                ['Wave', 'Code', unicode(where_args['wave_id']) ],
                #['Depot', unicode(where_args['depot_id']) ]
            ]

            print 'clearing members: %s' % (members)
            ic.clear_mondrian_member_cache(datasource.mondrian_project, cube, members)

        super(TntTrueResponseCubeAggregator, self)._work(where_args, loop_validator)
    """


    def _delete(self, where_args):
        logger.info(('%s: deleting from '+self._table+' for %s') % (self, str(where_args)))
        self.db_engine.execute("DELETE FROM "+self._table+" WHERE `wave_id` = %(wave_id)s and `depot_id` = %(depot_id)s", where_args)




# ---


class TntOldResponseCubeAggregator(IQubeAggregator):

    _table = "aggTntCLMWaveResponse"
    _columns = [
        'wave_id',
        'locale_id',
        'country_id',
        'service_id',
        'location',
        'location_2015',
        'depot_id',
        'revenue_band',
        'respondent_type',
        'attribute_id',
        'attribute_position',
        'attribute_type',
        'answer_id',
        'wave_scale_answer_id',
        'attribute_text',
        'answer_text',
        'response_int',
        'row_count',
        'created',

    ]

    _select_sql = """
        SELECT
            i.wave_id,
            c.locale_id,
            l.country_id,
            l.service_id,
            l.location,
            l.location_2015,
            c.depot_id,
            c.revenue_band,
            CASE p.is_respondent WHEN 1 THEN 'I' WHEN 2 THEN 'P' ELSE NULL END as respondent_type,
            a.attribute_id,
            a.position as attribute_position,
            CASE attr.type WHEN 1 THEN 'Scale' WHEN 2 THEN 'Single' WHEN 6 THEN 'Scale' ELSE 'N/A' END as attribute_type,
            ans.answer_id,
            wsa.scale_answer_id as wave_scale_answer_id,
            CASE a.position WHEN 30 THEN 'Country question' ELSE attr.default_label END as attribute_text,
            IF(ISNULL(wsa.id), ansl.default_label, CASE wsa.position WHEN 1 THEN 'Much lower' WHEN 2 THEN 'Lower' WHEN 3 THEN 'As' WHEN 4 THEN 'Higher' WHEN 5 THEN 'Much higher' ELSE 'N/A' END) AS answer_text,
            wsa.position as response_int,
            COUNT(DISTINCT p.id) as row_count,
            MAX(r.last_updated) as created

            FROM tblTntCLMWaveResponse AS r
            INNER JOIN tblTntCLMWaveParticipant AS p
                ON r.participant_id = p.id

            INNER JOIN tblTntCLMWaveInvitee AS i
                ON p.invitee_id = i.id
            INNER JOIN tblContact as contact
                ON i.contact_id = contact.id
            INNER JOIN tblTntCLMCompany as c
                ON contact.company_id = c.company_id
            INNER JOIN tblTntCLMWave as w
                ON i.wave_id = w.id
            INNER JOIN tblTntCLMLocale AS l
                ON c.locale_id = l.id
            INNER JOIN tblTntCLMWaveAttribute AS a
                ON r.wave_attribute_id = a.id
            INNER JOIN tblTntCLMAttribute AS attr
                ON a.attribute_id = attr.id
            LEFT JOIN tblTntCLMWaveAnswer AS ans
                ON r.wave_attribute_id = ans.wave_attribute_id
                    AND r.wave_answer_id = ans.id
            LEFT JOIN tblTntCLMAttributeAnswer AS ansl
                ON ans.answer_id = ansl.id
            LEFT JOIN tblTntCLMWaveScaleAnswer AS wsa
                ON r.wave_scale_answer_id = wsa.id
            WHERE i.wave_id = %(wave_id)s AND c.depot_id = %(depot_id)s
            GROUP BY i.wave_id, l.country_id, l.service_id, l.location, l.location_2015, c.locale_id, c.revenue_band, c.depot_id, p.is_respondent, a.attribute_id, ans.answer_id, wsa.scale_answer_id
        """



    def run(self, loop_validator=None):

        data_modified = False

        rows = self.db_engine_ro.execute("""
            SELECT i.wave_id, company.depot_id, UNIX_TIMESTAMP(MAX(r.last_updated))
            FROM tblTntCLMWaveResponse AS r
            INNER JOIN tblTntCLMWaveParticipant AS p
                ON r.participant_id = p.id
            INNER JOIN tblTntCLMWaveParticipantSession AS ps
                ON p.id = ps.participant_id
            INNER JOIN tblTntCLMWaveInvitee AS i
                ON p.invitee_id = i.id
            INNER JOIN tblContact contact
                ON i.contact_id = contact.id
            INNER JOIN `tblTntCLMCompany` company
                ON contact.company_id = company.company_id
            GROUP BY i.wave_id, company.depot_id
        """)

        for wave_id, depot_id, last_response in rows:

            last_response_agg, = self.db_engine_ro.execute("""
                SELECT UNIX_TIMESTAMP(MAX(created))
                FROM """+self._table+"""
                WHERE `wave_id` = %s
                AND `depot_id`= %s
            """, wave_id, depot_id).fetchone()

            if last_response_agg is None or (last_response > last_response_agg):

                self._work({
                    'wave_id': wave_id,
                    'depot_id': depot_id,
                    'last_response': last_response,
                }, loop_validator)

                data_modified = True

        rows.close()

        return data_modified


    def _delete(self, where_args):
        logger.info(('%s: deleting from '+self._table+' for %s') % (self, str(where_args)))
        self.db_engine.execute("DELETE FROM "+self._table+" WHERE `wave_id` = %(wave_id)s and `depot_id` = %(depot_id)s", where_args)

class TntCLMGeolevelHierarchyAggregator(IQubeAggregator):
    """
    Refills aggTntCLMGeolevelHierarchy if a change was made on Geolevels.
    """

    def run(self, loop_validator=None):
        rows = self.db_engine_ro.execute("""SELECT MAX(a.created) AS agg_date, MAX(t.last_updated) AS tbl_date
                                            FROM aggTntCLMGeolevelHierarchy a, tblTntGeolevel t""")
        for agg_date, tbl_date in rows:
            break

        rows.close()

        if agg_date is not None and agg_date >= tbl_date:
            return 0

        self.db_engine.execute("""
            DELETE FROM `aggTntCLMGeolevelHierarchy`
        """)
        self.db_engine.execute("""

            INSERT INTO `aggTntCLMGeolevelHierarchy` SELECT

            IFNULL(bu.id, IFNULL(bu2.id, IFNULL(bu3.id, IFNULL(bu4.id, 0)))) AS business_unit_id,
            IFNULL(bu.name, IFNULL(bu2.name, IFNULL(bu3.name, IFNULL(bu4.name, 'No business unit')))) AS business_unit,

            IFNULL(r.id, IFNULL(r2.id, 0)) AS region_id,
            IFNULL(r.name, IFNULL(r2.name, 'No region')) AS region,

            IFNULL(sr.id, 0) AS subregion_id,
            IFNULL(sr.name, 'No subregion') AS subregion,

            c.country_id AS country_id,
            c.name AS country,
            NOW()

            FROM viewTntCountry AS c
            LEFT JOIN tblTntGeolevelCountry AS gc
            ON c.country_id = gc.country_id

            # join directly with country
            LEFT JOIN tblTntGeolevel AS sr
            ON gc.geolevel_id = sr.id
                AND sr.type_id = 3
                AND sr.deleted = 0
            LEFT JOIN tblTntGeolevel AS r
            ON gc.geolevel_id = r.id
                AND r.type_id = 2
                AND r.deleted = 0
            LEFT JOIN tblTntGeolevel AS bu
            ON gc.geolevel_id = bu.id
                AND bu.type_id = 1
                AND bu.deleted = 0

            # join with parent geolevel
            LEFT JOIN tblTntGeolevelGeolevel AS ggr
            ON sr.id = ggr.geolevel_id
            LEFT JOIN tblTntGeolevel AS r2
            ON ggr.parent_geolevel_id = r2.id
                AND r2.type_id = 2
                AND r2.deleted = 0

            LEFT JOIN tblTntGeolevelGeolevel AS ggbu
            ON r.id = ggbu.geolevel_id
            LEFT JOIN tblTntGeolevel AS bu2
            ON ggbu.parent_geolevel_id = bu2.id
                AND bu2.type_id = 1
                AND bu2.deleted = 0

            LEFT JOIN tblTntGeolevelGeolevel AS ggbu2
            ON r2.id = ggbu2.geolevel_id
            LEFT JOIN tblTntGeolevel AS bu3
            ON ggbu2.parent_geolevel_id = bu3.id
                AND bu3.type_id = 1
                AND bu3.deleted = 0

            LEFT JOIN tblTntGeolevelGeolevel AS ggbu3
            ON sr.id = ggbu3.geolevel_id
            LEFT JOIN tblTntGeolevel AS bu4
            ON ggbu3.parent_geolevel_id = bu4.id
                AND bu4.type_id = 1
                AND bu4.deleted = 0
        """)

        # we know for sure something was changed, so just return >0
        return 1

class TntCLMTntCLMWaveAggregator(IQubeAggregator):
    """
    Adds a new wave to aggTntCLMTntCLMWave. This table only contains online waves for which updates are not possible
    anymore, so we don't need an update. As soon as a wave goes online, this aggregator is called.
    """

    def run(self, wave_id):
        self.db_engine.execute("""
            INSERT INTO `aggTntCLMTntCLMWave`
            SELECT w.id AS wave_id, w.code AS wave_code, l.id AS wave_is_main_id, l.label AS wave_is_main_label
                FROM tblTntCLMWave w INNER JOIN tblTntCLMWaveMainLabel l ON w.is_main = l.id
                WHERE w.id=%s;
        """, wave_id)

class TntWidgetStateController(DashboardWidgetStateController):
    """
    Controller for Tnt report widgets
    """

    _instance_class = TntCLMDashboardWidget

    def __init__(self, session, manager):
        DashboardWidgetStateController.__init__(self, session, manager)
        self.ds_controller = DataSourceController(session, account_id=manager.account_id)
        self.mm_controller = MdxMemberController(session, account_id=manager.account_id)
        self.cm_controller = MdxCalculatedMemberController(session, account_id=manager.account_id)

    @transaction_decorator
    def save(self, column, state, position):
        widget = DashboardWidgetStateController._instance_from_state(self, state)

        # general widget attributes
        widget.show_table = state['showTable']

        if state.has_key('tableAttributes'):
            widget.table_attributes = simplejson.dumps(state['tableAttributes'], ensure_ascii=False, cls=JsonRpcJSONEncoder)
        else:
            widget.table_attributes = '{}'

        widget.show_graph = state['showGraph']

        if state.has_key('graphAttributes'):
            widget.graph_attributes = simplejson.dumps(state['graphAttributes'], ensure_ascii=False, cls=JsonRpcJSONEncoder)
        else:
            widget.graph_attributes = '{}'

        if state.has_key('splitby'):
            widget.split_by = state['splitby']
        if state.has_key('onRows'):
            widget.on_rows = state['onRows']
        if state.has_key('rankby'):
            widget.rank_by = state['rankby']

        misc = {}
        if state.has_key('number_of_items') and state['number_of_items'] is not None:
            misc['number_of_items'] = state['number_of_items']
        if state.has_key('direction') and state['direction'] is not None:
            misc['direction'] = state['direction']

        # FIXME - next misc attributes should be placed in graphAttributes
        if state.has_key('graphType') and state['graphType'] is not None:
            misc['graphType'] = state['graphType']
        if state.has_key('stackedBar') and state['stackedBar'] is not None:
            misc['stackedBar'] = state['stackedBar']
        if state.has_key('horizontalBar') and state['horizontalBar'] is not None:
            misc['horizontalBar'] = state['horizontalBar']

        if len(misc) > 0:
            widget.misc = simplejson.dumps(misc, ensure_ascii=False, cls=JsonRpcJSONEncoder)

        widget.show_sample_size_warning = state['showSampleSizeWarning']
        if state.has_key('geolevelSelectionExpanded') and state['geolevelSelectionExpanded'] is not None:
            widget.show_geolevel_selection = state['geolevelSelectionExpanded']
        widget.locked = state['lockWidget']

        DashboardWidgetStateController.save(self, column, state, position, widget)

        # filters as they are sent from client (dict with filter_category - filter list)
        fs = state['tableConfig']

        datasource = self.ds_controller.get_datasource_by('mondrian_project', '%s.TNT-CLM' % settings.SETTINGS_MODULE).pop()
        cube = 'TntCLMResponse'

        categs = {}
        for k, c in TntCLMDashboardWidgetFilter.filter_categories.items():
            categs[c] = k

        current_filters = widget.filters
        new_filters = []
        index = 0
        for category, filters in fs.items():


            if not categs.has_key(category):
                # disabled for now, can become an exception after we ran the dashboard convert script
                #raise SecurityException('Invalid filter category %s' % category)
                continue

            for filter in filters:

                try:
                    filter_obj = current_filters[index]
                except IndexError:
                    filter_obj = TntCLMDashboardWidgetFilter()

                filter_obj.category = categs[category]
                filter_obj.widget = widget

                try:
                    bf_id = int(filter)
                    filter_obj.calculated_member = self.cm_controller.get_calculated_member(bf_id)[0]
                    filter_obj.mdx_member = None
                except ValueError:
                    mm = self.mm_controller.save_from_str(datasource, cube, filter)
                    filter_obj.mdx_member = mm
                    filter_obj.calculated_member = None
                new_filters.append(filter_obj)
                index = index + 1

        new_wfids = [f.id for f in new_filters]

        for f in current_filters:
            if f.id not in new_wfids:
                self.sa_session.delete(f)

        widget.filters = new_filters

    def load(self, id):
        widget = self._instance_from_id(id)

        filters = {}
        for f in widget.filters:

            categ = TntCLMDashboardWidgetFilter.filter_categories[f.category]
            if not filters.has_key(categ):
                filters[categ] = []
            if f.calculated_member is None:
                filters[categ].append(self.mm_controller.load_str(f.mdx_member))
            else:
                filters[categ].append(f.calculated_member.id)

        state = {
            "objectName": widget.object_name,
            "loaderName": widget.object_loader,
            "dbID": widget.id,
            "position": widget.position,
            "title": widget.title,
            "show_header": widget.show_header,
            "showTable": widget.show_table,
            "tableAttributes": simplejson.loads(widget.table_attributes, object_hook=json_object_hook),
            "showGraph": widget.show_graph,
            "graphAttributes": simplejson.loads(widget.graph_attributes, object_hook=json_object_hook),
            "splitby": widget.split_by,
            "onRows": widget.on_rows,
            "rankby": widget.rank_by,
            "showSampleSizeWarning": widget.show_sample_size_warning,
            "geolevelSelectionExpanded": widget.show_geolevel_selection,
            "lockWidget": widget.locked,
            "tableConfig": filters,
            "graphConfig": filters
        }

        if widget.misc is not None:
            misc = simplejson.loads(widget.misc, object_hook=json_object_hook)
            for key, value in misc.items():
                state[key] = value

        return state
