# Copyright (c) 2011 BaseSoft B.V.
# See LICENSE for details.

from xml.dom.minidom import Document

from django.conf import settings

from basesoft.web.core.util.controllers import AbstractController
from basesoft.web.contrib.account.models import Account, GroupAccount
from basesoft.web.contrib.license.models import Group, License, Domain
from basesoft.web.core.management.commands.iqubesync import ProjectBase, CubeBase, RoleBase

from iqportal.iqube.core.models import DataSource
from iqportal.tnt.clm.models import TntCLMWave
from iqportal.tnt.core.models import TntGeolevel
from iqportal.tnt.core.controllers import TntController
from iqportal.tnt.report.controllers import TntBusinessFunctionController
from sqlalchemy.orm.exc import NoResultFound

class TntCLMProject(ProjectBase):
    project_name = 'TNT-CLM'
    group_privileges = DataSource.PRIV_VIEW + DataSource.PRIV_DELEGATE_PRIVILEGES
    description = 'Customer Loyalty Measurement'
    additional_graphs = ('tnt_graph_datatablecube',)
    client_module = 'iqportal.tnt.report.client'
    role_level = DataSource.ROLE_LEVEL_ACCOUNT

    def __init__(self, sa_session, controller):
        ProjectBase.__init__(self, sa_session, controller)
        # the TNT-CLM data is directly drawn from the IQPortal database.
        self._db_from_connstr(settings.DB_CONNSTR_READONLY)

    def execute(self):

        tnt_domain = None
        try:
            tnt_domain = self.sa_session.query(Domain).filter(Domain.name=='TNT').one()
        except NoResultFound:
            pass

        if tnt_domain is None:
            print 'skipping iqubesync for tnt (tnt domain not found)'
        else:
            ProjectBase.execute(self)


    def get_cubes(self):
        return (TntCLMResponseCube, TntCLMTrueResponseCube, TntCLMTrueRespondentsCube, TntCLMVirtualCube, )

    def get_roles(self):
        return (TntCLMRole, )


class DefaultTntResponseCube(CubeBase):
    def _xml(self, fact_table, respondents=False):

        rsp = """<Measure name="Responses" column="responses" aggregator="sum" datatype="Integer" />"""

        if respondents:
            rsp = """
                <Dimension name="Customer" foreignKey="invitee_id" highCardinality="true">
                    <Hierarchy hasAll="true" allMemberName="All Customers" primaryKey="id" primaryKeyTable="tblTntCLMWaveInvitee">
                        <Join leftKey="contact_id" rightKey="id">
                            <Table name="tblTntCLMWaveInvitee" />
                            <Join leftKey="company_id" rightKey="company_id">
                                <Table name="tblContact" />
                                <Join leftKey="customer_id" rightKey="id">
                                    <Table name="tblTntCLMCompany" />
                                    <Table name="tblTntCLMCustomer" />
                                </Join>
                            </Join>
                        </Join>
                        <Level name="Account Number" table="tblTntCLMCustomer" column="account_number" type="Numeric" approxRowCount="328241" />
                    </Hierarchy>
                </Dimension>
                <Dimension name="Participant">
                    <Hierarchy hasAll="true" allMemberName="All Participants" >
                        <Level name="Participant" column="participant_id" />
                    </Hierarchy>
                </Dimension>

                <Measure name="Respondents" column="participant_id" aggregator="distinct-count" datatype="Integer" formatString="#,###"/>
                <Measure name="Responses" column="id" aggregator="distinct-count" datatype="Integer" formatString="#,###"/>
                """

        return """
    <Cube name="%(name)s" defaultMeasure="Responses">

        <Table name="%(fact_table)s" />

        <Dimension name="Answer" foreignKey="answer_id" >
            <Hierarchy hasAll="true" allMemberName="All Answers" primaryKey="id" primaryKeyTable="tblTntCLMAttributeAnswer">
                <Table name="tblTntCLMAttributeAnswer" />
                <Level name="Answer" table="tblTntCLMAttributeAnswer" column="id" type="Integer" uniqueMembers="true" ordinalColumn="position">
                    <CaptionExpression>
                        <SQL dialect="mysql">
                            IF(`tblTntCLMAttributeAnswer`.`default_label`!='',
                                `tblTntCLMAttributeAnswer`.`default_label`,
                                CAST((`tblTntCLMAttributeAnswer`.`position`-1) AS CHAR)
                            )
                        </SQL>
                    </CaptionExpression>
                </Level>
            </Hierarchy>
        </Dimension>

        <Dimension name="Attribute" foreignKey="attribute_id" >
            <Hierarchy hasAll="true" allMemberName="All Attributes" primaryKey="id" primaryKeyTable="tblTntCLMAttribute">
                <Table name="tblTntCLMAttribute" />
                <Level name="Type" table="tblTntCLMAttribute" column="type" type="Integer" uniqueMembers="true">
                    <CaptionExpression>
                        <SQL dialect="mysql">
                            CASE `type`
                                WHEN 1 THEN 'Scale'
                                WHEN 2 THEN 'Single choice'
                                WHEN 3 THEN 'Text'
                                WHEN 4 THEN 'Multiple choice'
                                WHEN 5 THEN 'Range'
                                WHEN 6 THEN 'Single scale'
                                WHEN 7 THEN 'NPS Score'
                                ELSE 'Undefined' END
                        </SQL>
                    </CaptionExpression>
                </Level>
                <Level name="Attribute" table="tblTntCLMAttribute" column="id" type="Integer" uniqueMembers="true">
                    <CaptionExpression>
                        <SQL dialect="mysql">
                            CONCAT(CAST(`tblTntCLMAttribute`.`id` AS CHAR), ' - ', `tblTntCLMAttribute`.`default_label`)
                        </SQL>
                    </CaptionExpression>
                </Level>
            </Hierarchy>
        </Dimension>

        <Dimension name="KPI" foreignKey="scale_answer_id" >
            <Hierarchy hasAll="true" allMemberName="All KPIs" primaryKey="id" primaryKeyTable="tblTntCLMScaleAnswer">
                <Table name="tblTntCLMScaleAnswer" />
                <Level name="KPI" table="tblTntCLMScaleAnswer" column="id" type="Integer" captionColumn="default_label" uniqueMembers="true"/>
            </Hierarchy>
        </Dimension>

        <Dimension name="Geolevel" foreignKey="depot_id" >
            <Hierarchy hasAll="true" allMemberName="All Geolevels" primaryKey="id" primaryKeyTable="tblTntDepot">
                <Join leftKey="country_id" rightKey="id">
                    <Table name="tblTntDepot" />
                    <Join leftKey="id" rightKey="country_id">
                        <Table name="tblCountry" />
                        <Table name="aggTntCLMGeolevelHierarchy" />
                    </Join>
                </Join>
                <Level name="Business Unit" table="aggTntCLMGeolevelHierarchy" column="business_unit_id" type="Integer" captionColumn="business_unit" ordinalColumn="business_unit" />
                <Level name="Region" table="aggTntCLMGeolevelHierarchy" column="region_id" type="Integer" captionColumn="region" ordinalColumn="region" />
                <Level name="Sub Region" table="aggTntCLMGeolevelHierarchy" column="subregion_id" type="Integer" captionColumn="subregion" ordinalColumn="subregion" />
                <Level name="Country" table="aggTntCLMGeolevelHierarchy" column="country_id" type="Integer" captionColumn="country" ordinalColumn="country" />
                <Level name="Depot" table="tblTntDepot" column="id" type="Integer" captionColumn="default_code" ordinalColumn="default_code" />
            </Hierarchy>
        </Dimension>

        <Dimension name="Service" foreignKey="service_id">
            <Hierarchy hasAll="true" allMemberName="All Services" primaryKey="id" primaryKeyTable="tblTntCLMService">
                <Table name="tblTntCLMService" />
                <Level name="Service" column="id" type="Integer" captionColumn="label" />
            </Hierarchy>
        </Dimension>

        <Dimension name="Wave" foreignKey="wave_id" >
            <Hierarchy hasAll="true" allMemberName="All Waves" primaryKey="wave_id" primaryKeyTable="aggTntCLMTntCLMWave">
                <Table name="aggTntCLMTntCLMWave" />
                <Level name="Main" column="wave_is_main_id" type="Integer" captionColumn="wave_is_main_label" ordinalColumn="wave_is_main_label" />
                <Level name="Code" column="wave_id" type="Integer" captionColumn="wave_code" ordinalColumn="wave_code" />
            </Hierarchy>
        </Dimension>

        <Dimension name="Revenue Band"  >
            <Hierarchy hasAll="true" allMemberName="All Revenue Bands" >
                <Level name="Revenue Band" column="revenue_band" />
            </Hierarchy>
        </Dimension>

        <Dimension name="Location">
            <Hierarchy hasAll="true" allMemberName="All Locations">
                <Level name="Location" column="location" />
            </Hierarchy>
        </Dimension>

        <Dimension name="Location_2015">
            <Hierarchy hasAll="true" allMemberName="All Locations_2015">
                <Level name="Location_2015" column="location_2015" />
            </Hierarchy>
        </Dimension>

        <Dimension name="Respondent Type" foreignKey="is_respondent">
            <Hierarchy hasAll="true" allMemberName="All Respondent Types" primaryKey="id" primaryKeyTable="tblTntCLMWaveRespondentType">
                <Table name="tblTntCLMWaveRespondentType" />
                <Level name="Respondent Type" column="id" nameColumn="label" type="Integer" />
            </Hierarchy>
        </Dimension>

        <Dimension name="P1">
            <Hierarchy hasAll="true" allMemberName="All P1s">
                <Level name="P1" column="is_processed" >
                    <CaptionExpression>
                        <SQL dialect="mysql">
                            CASE `is_processed`
                                WHEN 0 THEN 'P1 not processed yet'
                                WHEN 1 THEN 'No P1'
                                WHEN 2 THEN 'P1'
                                WHEN 3 THEN 'P1 not processed'
                                ELSE 'Undefined' END
                        </SQL>
                    </CaptionExpression>
                </Level>
            </Hierarchy>
        </Dimension>

        %(respondents)s

    </Cube>
        """  % {'name': self.name, 'fact_table': fact_table, 'respondents': rsp}


class TntCLMTrueResponseCube(DefaultTntResponseCube):
    name = 'TntCLMTrueResponseCube'

    def __xml(self):
        return self._xml('aggTntCLMTrueWaveResponse')

    xml = property(__xml)

class TntCLMTrueRespondentsCube(DefaultTntResponseCube):
    name = 'TntCLMTrueRespondentsCube'

    def __xml(self):
        return self._xml('aggTntCLMFullWaveResponse', True)

    xml = property(__xml)

class TntCLMVirtualCube(CubeBase):
    name = 'TntCLMVirtualCube'

    def _get_xml(self):

        return """
    <VirtualCube name="TntCLMVirtualCube">

        <CubeUsages>
            <CubeUsage cubeName="TntCLMTrueResponseCube" />
            <CubeUsage cubeName="TntCLMTrueRespondentsCube" />
        </CubeUsages>

        <VirtualCubeDimension cubeName="TntCLMTrueRespondentsCube" name="Answer" />
        <VirtualCubeDimension cubeName="TntCLMTrueRespondentsCube" name="Attribute" />
        <VirtualCubeDimension cubeName="TntCLMTrueRespondentsCube" name="KPI" />
        <VirtualCubeDimension cubeName="TntCLMTrueRespondentsCube" name="Geolevel" />
        <VirtualCubeDimension cubeName="TntCLMTrueRespondentsCube" name="Service" />
        <VirtualCubeDimension cubeName="TntCLMTrueRespondentsCube" name="Wave" />
        <VirtualCubeDimension cubeName="TntCLMTrueRespondentsCube" name="Revenue Band" />
        <VirtualCubeDimension cubeName="TntCLMTrueRespondentsCube" name="Location" />
        <VirtualCubeDimension cubeName="TntCLMTrueRespondentsCube" name="Location_2015" />
        <VirtualCubeDimension cubeName="TntCLMTrueRespondentsCube" name="Respondent Type" />
        <VirtualCubeDimension cubeName="TntCLMTrueRespondentsCube" name="P1" />

        <VirtualCubeDimension cubeName="TntCLMTrueRespondentsCube" name="Customer" />
        <VirtualCubeDimension cubeName="TntCLMTrueRespondentsCube" name="Participant" />

        <VirtualCubeMeasure cubeName="TntCLMTrueRespondentsCube" name="[Measures].[Responses]" />
        <VirtualCubeMeasure cubeName="TntCLMTrueRespondentsCube" name="[Measures].[Respondents]" />

        <CalculatedMember
           name="No. of Answers"
           dimension="KPI"
           formula="[KPI].[1] + [KPI].[2] + [KPI].[3] + [KPI].[4] + [KPI].[5]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Integer"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###"/>
       </CalculatedMember>

       <CalculatedMember
           name="% Much lower"
           dimension="KPI"
           formula="100 * ([KPI].[1]) / [KPI].[No. of Answers]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

       <CalculatedMember
           name="% Lower"
           dimension="KPI"
           formula="100 * ([KPI].[2]) / [KPI].[No. of Answers]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

       <CalculatedMember
           name="% As"
           dimension="KPI"
           formula="100 * ([KPI].[3]) / [KPI].[No. of Answers]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

       <CalculatedMember
           name="% Higher"
           dimension="KPI"
           formula="100 * ([KPI].[4]) / [KPI].[No. of Answers]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

       <CalculatedMember
           name="% Much higher"
           dimension="KPI"
           formula="100 * ([KPI].[5]) / [KPI].[No. of Answers]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

       <CalculatedMember
           name="Mean Score"
           dimension="KPI"
           formula="(([KPI].[1]*1)+([KPI].[2]*2)+([KPI].[3]*3)+([KPI].[4]*4)+([KPI].[5]*5)) / [KPI].[No. of Answers]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

       <CalculatedMember
           name="OES"
           dimension="KPI"
           formula="[KPI].[Mean Score]*20"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

       <CalculatedMember
           name="Dissatisfied %"
           dimension="KPI"
           formula="100 * ([KPI].[2] + [KPI].[1]) / [KPI].[No. of Answers]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

       <CalculatedMember
           name="Meeting Cust. Exp. %"
           dimension="KPI"
           formula="100 * ([KPI].[3] + [KPI].[4] + [KPI].[5]) / [KPI].[No. of Answers]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

        <CalculatedMember
           name="Exceeding Cust. Exp. %"
           dimension="KPI"
           formula="100 * ([KPI].[4] + [KPI].[5]) / [KPI].[No. of Answers]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

       <CalculatedMember
           name="Dissatisfied"
           dimension="KPI"
           formula="[KPI].[1] + [KPI].[2]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###"/>
       </CalculatedMember>

       <CalculatedMember
           name="Meeting Cust. Exp."
           dimension="KPI"
           formula="[KPI].[3] + [KPI].[4] + [KPI].[5]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###"/>
       </CalculatedMember>

       <CalculatedMember
           name="Exceeding Cust. Exp."
           dimension="KPI"
           formula="[KPI].[4] + [KPI].[5]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###"/>
       </CalculatedMember>

        <CalculatedMember
           name="NCS Score"
           dimension="KPI"
           formula="[KPI].[Exceeding Cust. Exp. %] - [KPI].[Dissatisfied %]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

        <CalculatedMember
           name="Overall satisfaction"
           dimension="Attribute"
           formula="AGGREGATE({[Attribute].[1].[25], [Attribute].[1].[27], [Attribute].[1].[29]})"
           visible="true">
           <CalculatedMemberProperty name="SOLVE_ORDER" value="1"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###"/>
       </CalculatedMember>

        <CalculatedMember
           name="Score"
           dimension="Answer"
           formula="[Answer].[4]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Integer"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="3"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###"/>
       </CalculatedMember>

        <CalculatedMember
           name="Score %"
           dimension="Answer"
           formula="100 * [Answer].[4] / [Answer].[All Answers]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="3"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

       <CalculatedMember
           name="All NPS"
           dimension="Answer"
           formula="[Answer].[55] + [Answer].[56] + [Answer].[57] + [Answer].[58] + [Answer].[59] + [Answer].[60] +
           [Answer].[61] + [Answer].[62] + [Answer].[63] + [Answer].[64] + [Answer].[65]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Integer"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="3"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###"/>
       </CalculatedMember>

       <CalculatedMember
           name="% NPS"
           dimension="Answer"
           formula="100 * (([Answer].[64] + [Answer].[65]) - ([Answer].[55] + [Answer].[56] + [Answer].[57] +
           [Answer].[58] + [Answer].[59] + [Answer].[60] + [Answer].[61]))/[Answer].[All NPS]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="3"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

       <CalculatedMember
           name="NPS KPI Detractor"
           dimension="Answer"
           formula="[Answer].[55] + [Answer].[56] + [Answer].[57] + [Answer].[58] + [Answer].[59] + [Answer].[60] + [Answer].[61]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Integer"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="3"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###"/>
       </CalculatedMember>

       <CalculatedMember
           name="NPS KPI Passives"
           dimension="Answer"
           formula="[Answer].[62] + [Answer].[63]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Integer"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="3"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###"/>
       </CalculatedMember>

       <CalculatedMember
           name="NPS KPI Promotors"
           dimension="Answer"
           formula="[Answer].[64] + [Answer].[65]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Integer"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="3"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###"/>
       </CalculatedMember>

       <CalculatedMember
           name="NPS KPI Detractor %"
           dimension="Answer"
           formula="100 * ([Answer].[55] + [Answer].[56] + [Answer].[57] + [Answer].[58] + [Answer].[59] +
           [Answer].[60] + [Answer].[61])/[Answer].[All NPS]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Integer"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="3"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

       <CalculatedMember
           name="NPS KPI Passives %"
           dimension="Answer"
           formula="100 * ([Answer].[62] + [Answer].[63])/[Answer].[All NPS]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Integer"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="3"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

       <CalculatedMember
           name="NPS KPI Promotors %"
           dimension="Answer"
           formula="100 * ([Answer].[64] + [Answer].[65])/[Answer].[All NPS]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Integer"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="3"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

       <CalculatedMember
           name="NPS 0 - Not At All Likely %"
           dimension="Answer"
           formula="100*[Answer].[55]/[Answer].[All NPS]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Integer"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="3"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>
       <CalculatedMember
           name="NPS 1 %"
           dimension="Answer"
           formula="100*[Answer].[56]/[Answer].[All NPS]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Integer"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="3"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>
       <CalculatedMember
           name="NPS 2 %"
           dimension="Answer"
           formula="100*[Answer].[57]/[Answer].[All NPS]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Integer"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="3"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>
       <CalculatedMember
           name="NPS 3 %"
           dimension="Answer"
           formula="100*[Answer].[58]/[Answer].[All NPS]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Integer"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="3"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>
       <CalculatedMember
           name="NPS 4 %"
           dimension="Answer"
           formula="100*[Answer].[59]/[Answer].[All NPS]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Integer"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="3"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>
       <CalculatedMember
           name="NPS 5 - Neutral %"
           dimension="Answer"
           formula="100*[Answer].[60]/[Answer].[All NPS]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Integer"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="3"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>
       <CalculatedMember
           name="NPS 6 %"
           dimension="Answer"
           formula="100*[Answer].[61]/[Answer].[All NPS]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Integer"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="3"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>
       <CalculatedMember
           name="NPS 7 %"
           dimension="Answer"
           formula="100*[Answer].[62]/[Answer].[All NPS]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Integer"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="3"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>
       <CalculatedMember
           name="NPS 8 %"
           dimension="Answer"
           formula="100*[Answer].[63]/[Answer].[All NPS]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Integer"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="3"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>
       <CalculatedMember
           name="NPS 9 %"
           dimension="Answer"
           formula="100*[Answer].[64]/[Answer].[All NPS]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Integer"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="3"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>
       <CalculatedMember
           name="NPS 10 - Extremely Likely %"
           dimension="Answer"
           formula="100*[Answer].[65]/[Answer].[All NPS]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Integer"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="3"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>
    </VirtualCube>
       """
    xml = property(_get_xml)


class TntCLMResponseCube(CubeBase):
    name = 'TntCLMResponse'
    _xml = """
    <Cube name="TntCLMResponse" defaultMeasure="No. of Answers">
        <Table name="aggTntCLMWaveResponse" />

        <Dimension name="KPI">
            <Hierarchy hasAll="true" allMemberName="All KPIs">
                <Level name="KPI" column="answer_text" uniqueMembers="true"/>
            </Hierarchy>
        </Dimension>


        <Dimension name="Answer">
            <Hierarchy hasAll="true" allMemberName="All Answers">
                <Level name="Answer" column="answer_id" nameColumn="answer_text" />
            </Hierarchy>
        </Dimension>

        <Dimension name="Attribute">
            <Hierarchy hasAll="true">
                <Level name="Type" column="attribute_type" uniqueMembers="true"/>
                <Level name="Attribute" column="attribute_id" nameColumn="attribute_text" uniqueMembers="true"/>
            </Hierarchy>
        </Dimension>

        <Dimension name="Position">
            <Hierarchy hasAll="true">
                <Level name="Position" column="attribute_id" uniqueMembers="true"/>
            </Hierarchy>
        </Dimension>

        <Dimension name="Pool">
            <Hierarchy hasAll="true">
                <Level name="Pool" column="attribute_id" uniqueMembers="false"/>
            </Hierarchy>
        </Dimension>

        <Dimension name="Locale">
            <Hierarchy hasAll="true" allMemberName="All Locales">
                <Level name="Locale" column="locale_id" uniqueMembers="true"/>
            </Hierarchy>
        </Dimension>

        <Dimension name="Country" foreignKey="country_id">
            <Hierarchy hasAll="true" primaryKey="country_id" allMemberName="All Countries">
                <Table name="viewTntCountry"/>
                <Level name="Country" column="code" table="viewTntCountry" uniqueMembers="true"/>
            </Hierarchy>
        </Dimension>

        <Dimension name="Service" foreignKey="service_id">
            <Hierarchy hasAll="true" primaryKey="id" allMemberName="All Services">
                <Table name="tblTntCLMService"/>
                <Level name="Service" column="label" table="tblTntCLMService" uniqueMembers="true"/>
            </Hierarchy>
        </Dimension>

        <Dimension name="Depot" foreignKey="depot_id">
            <Hierarchy hasAll="true" primaryKey="id" allMemberName="All Depots">
                <Table name="tblTntDepot"/>
                <Level name="Depot" column="default_code" table="tblTntDepot" uniqueMembers="true"/>
            </Hierarchy>
        </Dimension>

        <Dimension name="Wave" foreignKey="wave_id">
            <Hierarchy hasAll="true" primaryKey="id" allMemberName="All Waves">
                <Table name="tblTntCLMWave"/>

                    <Level name="Main" column="is_main">
                        <KeyExpression>
                            <SQL dialect="mysql">
                                CASE is_main WHEN 1 THEN 'Main Waves' ELSE 'Other Waves' END
                            </SQL>
                        </KeyExpression>
                    </Level>

                    <Level name="Code" column="code" uniqueMembers="true"/>

            </Hierarchy>
        </Dimension>

        <Dimension name="Revenue Band">
            <Hierarchy hasAll="true" allMemberName="All Revenue Bands">
                <Level name="Revenue Band" column="revenue_band" uniqueMembers="true"/>
            </Hierarchy>
        </Dimension>

        <Dimension name="Location">
            <Hierarchy hasAll="true" allMemberName="All Locations">
                <Level name="Location" column="location" uniqueMembers="true"/>
            </Hierarchy>
        </Dimension>

        <Dimension name="Location_2015">
            <Hierarchy hasAll="true" allMemberName="All Locations_2015">
                <Level name="Location_2015" column="location_2015" uniqueMembers="true"/>
            </Hierarchy>
        </Dimension>

        <Dimension name="Respondent Type">
            <Hierarchy hasAll="true" allMemberName="All Respondent Types">
                <Level name="Respondent Type" column="respondent_type" uniqueMembers="true"/>
            </Hierarchy>
        </Dimension>

        <Measure name="No. of Answers" column="row_count" aggregator="sum" datatype="Integer" formatString="#,###" />

        <CalculatedMember
           name="No. of Answers"
           dimension="KPI"
           formula="[KPI].[Much lower] + [KPI].[Lower] + [KPI].[As] + [KPI].[Higher] + [KPI].[Much higher]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Integer"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###"/>
       </CalculatedMember>

       <CalculatedMember
           name="%% Much lower"
           dimension="KPI"
           formula="100 * ([KPI].[Much lower]) / [KPI].[No. of Answers]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

       <CalculatedMember
           name="%% Lower"
           dimension="KPI"
           formula="100 * ([KPI].[Lower]) / [KPI].[No. of Answers]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

       <CalculatedMember
           name="%% As"
           dimension="KPI"
           formula="100 * ([KPI].[As]) / [KPI].[No. of Answers]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

       <CalculatedMember
           name="%% Higher"
           dimension="KPI"
           formula="100 * ([KPI].[Higher]) / [KPI].[No. of Answers]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

       <CalculatedMember
           name="%% Much higher"
           dimension="KPI"
           formula="100 * ([KPI].[Much higher]) / [KPI].[No. of Answers]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

       <CalculatedMember
           name="Mean Score"
           dimension="KPI"
           formula="(([KPI].[Much lower]*1)+([KPI].[Lower]*2)+([KPI].[As]*3)+([KPI].[Higher]*4)+([KPI].[Much higher]*5)) / [KPI].[No. of Answers]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

       <CalculatedMember
           name="Dissatisfied"
           dimension="KPI"
           formula="100 * ([KPI].[Lower] + [KPI].[Much lower]) / [KPI].[No. of Answers]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

       <CalculatedMember
           name="Meeting Cust. Exp."
           dimension="KPI"
           formula="100 * ([KPI].[As] + [KPI].[Higher] + [KPI].[Much higher]) / [KPI].[No. of Answers]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

        <CalculatedMember
           name="Exceeding Cust. Exp."
           dimension="KPI"
           formula="100 * ([KPI].[Higher] + [KPI].[Much higher]) / [KPI].[No. of Answers]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

        <CalculatedMember
           name="NCS Score"
           dimension="KPI"
           formula="[KPI].[Exceeding Cust. Exp.] - [KPI].[Dissatisfied]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

        <CalculatedMember
           name="Overall satisfaction"
           dimension="Position"
           formula="[Position].[25] + [Position].[27] + [Position].[29]"
           visible="true">
           <CalculatedMemberProperty name="SOLVE_ORDER" value="1"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###"/>
       </CalculatedMember>

        <CalculatedMember
           name="Score"
           dimension="KPI"
           formula="[KPI].[strongly]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Integer"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###"/>
       </CalculatedMember>

        <CalculatedMember
           name="Score %%"
           dimension="KPI"
           formula="100 * [KPI].[strongly] / [KPI].[All KPIs]"
           visible="true">
           <CalculatedMemberProperty name="DATATYPE" value="Numeric"/>
           <CalculatedMemberProperty name="SOLVE_ORDER" value="2"/>
           <CalculatedMemberProperty name="FORMAT_STRING" value="#,###.00"/>
       </CalculatedMember>

        %(dynamic_cm)s

    </Cube>
    """

    def __get_xml(self):



        dynamic_cm = ''

        # geolevels

        for geol in self.sa_session.query(TntGeolevel).all():
            countries = geol.get_countries()
            if len(countries) == 0:
                continue
            geo_cm = '+'.join(['[Country].[%s]' % country.iso_3166_a2 for country in countries])
            dynamic_cm += """
                <CalculatedMember
                   name="%(name)s"
                   dimension="Country"
                   formula="%(formula)s"
                   visible="true">
               </CalculatedMember>
            """ % {
                'name': geol.name,
                'formula': geo_cm
            }



        xml = self._xml % { 'dynamic_cm': dynamic_cm }

        return xml

    xml = property(__get_xml)

class AccountRole():

    def __init__(self, session, xmlDoc):
        self.sa_session = session
        self.xmlDoc = xmlDoc

    def get_all(self):
        return (
            self.TntCLMTrueResponse_account_role,
            self.TntCLMResponseCube_account_role,
        )

    def TntCLMResponseCube_account_role(self, account):
        xmlDoc = self.xmlDoc
        xmlCg = xmlDoc.createElement('CubeGrant')
        xmlCg.setAttribute('cube', 'TntCLMResponse')
        xmlCg.setAttribute('access', 'all')

        if account.polymorphic_type == 'TntCLMAccount':
            xmlHg = xmlDoc.createElement('HierarchyGrant')
            xmlHg.setAttribute('hierarchy', '[Country]')
            xmlHg.setAttribute('access', 'custom')
            xmlHg.setAttribute('rollupPolicy', 'partial')
            xmlCg.appendChild(xmlHg)

            for geolevel in account.all_geolevels:
                xmlMg = xmlDoc.createElement('MemberGrant')
                xmlMg.setAttribute('member', '[Country].[%s]' % geolevel.name)
                xmlMg.setAttribute('access', 'all')
                xmlHg.appendChild(xmlMg)

            for country in account.countries:
                xmlMg = xmlDoc.createElement('MemberGrant')
                xmlMg.setAttribute('member', '[Country].[%s]' % country.iso_3166_a2)
                xmlMg.setAttribute('access', 'all')
                xmlHg.appendChild(xmlMg)

            # TODO: if we don't have any services linked, we get rights to none
            xmlHg = xmlDoc.createElement('HierarchyGrant')
            xmlHg.setAttribute('hierarchy', '[Service]')
            xmlHg.setAttribute('access', 'custom')
            xmlHg.setAttribute('rollupPolicy', 'partial')
            xmlCg.appendChild(xmlHg)

            for service in account.services:
                xmlMg = xmlDoc.createElement('MemberGrant')
                xmlMg.setAttribute('member', '[Service].[%s]' % service.label)
                xmlMg.setAttribute('access', 'all')
                xmlHg.appendChild(xmlMg)

        xmlHg = xmlDoc.createElement('HierarchyGrant')
        xmlHg.setAttribute('hierarchy', '[Wave]')
        xmlHg.setAttribute('access', 'custom')
        xmlHg.setAttribute('rollupPolicy', 'partial')
        xmlCg.appendChild(xmlHg)

        # get all the waves to which this account has VIEW_RESPONSES privs to.
        ac = AbstractController(self.sa_session, account_id=account.id, check_active=False)
        waves = ac.get_privileged_object_by_id(TntCLMWave, None, priv=TntCLMWave.PRIV_VIEW_RESPONSE)
        for wave in waves:
            xmlMg = xmlDoc.createElement('MemberGrant')
            xmlMg.setAttribute('member', '[Wave].[Code].[%s]' % wave.code)
            xmlMg.setAttribute('access', 'all')
            xmlHg.appendChild(xmlMg)

        return xmlCg

    def TntCLMTrueResponse_account_role(self, account):
        """
        Method used for cubes with same dimensions and hierarchies
        """
        xmlDoc = self.xmlDoc
        xmlCg = xmlDoc.createElement('CubeGrant')
        xmlCg.setAttribute('cube', 'TntCLMVirtualCube')
        xmlCg.setAttribute('access', 'all')

        if account.polymorphic_type == 'TntCLMAccount':
            xmlHg = xmlDoc.createElement('HierarchyGrant')
            xmlHg.setAttribute('hierarchy', '[Geolevel]')
            xmlHg.setAttribute('access', 'custom')
            xmlHg.setAttribute('rollupPolicy', 'partial')
            xmlCg.appendChild(xmlHg)

            # we need to get minimum of type ids (cant just use max here because that might actually be the min one)
            topLevel = None
            types = {None: 'Country'}
            # just get top geolevels; that will include their children also
            geolevels = []
            #gc = TntController(self.sa_session)
            for geolevel in account.geolevels:
                # check if current geolevel was not already added
                if geolevel in geolevels:
                    continue
                # check if current geolevel is child of any of the geolevels we already added
                """is_child = False
                for g in geolevels:
                    if gc.isChildOf(geolevel.id, g):
                        is_child = True
                        break
                if is_child:
                    continue"""
                geolevels.append(geolevel)
                xmlMg = xmlDoc.createElement('MemberGrant')

                dimension_name = None
                if geolevel.type.name == 'BusinessUnit':
                    dimension_name = 'Business Unit'
                if geolevel.type.name == 'Region':
                    dimension_name = 'Region'
                if geolevel.type.name == 'Subregion':
                    dimension_name = 'Sub Region'

                if dimension_name == None:
                    print 'WARNING: member is None, skipping geolevel. account: %s, geolevel.id: %s, geolevel.type.name: %s' % (account.id, geolevel.id, geolevel.type.name)
                    continue

                if topLevel is None or geolevel.type.id < topLevel:
                    topLevel = geolevel.type.id

                types[geolevel.type.id] = dimension_name
                member = '[Geolevel].[%s].[%s]' % (dimension_name, geolevel.id)

                xmlMg.setAttribute('member', member)
                xmlMg.setAttribute('access', 'all')
                xmlHg.appendChild(xmlMg)

            # get all countries this account is directly linked to (so not also geolevels countries)
            countries = []
            for country in account.countries:

                if country in countries:
                    continue
                # or it's not part of any of the already added geolevels
                """for g in geolevels:
                    if country in g.countries:
                        continue"""
                countries.append(country)
                xmlMg = xmlDoc.createElement('MemberGrant')
                xmlMg.setAttribute('member', '[Geolevel].[Country].[%s]' % country.id)
                xmlMg.setAttribute('access', 'all')
                xmlHg.appendChild(xmlMg)

            xmlHg.setAttribute('topLevel', '[Geolevel].[%s]' % types[topLevel])

            # TODO: if we don't have any services linked, we get rights to none
            xmlHg = xmlDoc.createElement('HierarchyGrant')
            xmlHg.setAttribute('hierarchy', '[Service]')
            xmlHg.setAttribute('access', 'custom')
            xmlHg.setAttribute('rollupPolicy', 'partial')
            xmlCg.appendChild(xmlHg)

            for service in account.services:
                xmlMg = xmlDoc.createElement('MemberGrant')
                xmlMg.setAttribute('member', '[Service].[%s]' % service.id)
                xmlMg.setAttribute('access', 'all')
                xmlHg.appendChild(xmlMg)

        xmlHg = xmlDoc.createElement('HierarchyGrant')
        xmlHg.setAttribute('hierarchy', '[Wave]')
        xmlHg.setAttribute('access', 'custom')
        xmlHg.setAttribute('rollupPolicy', 'partial')
        xmlCg.appendChild(xmlHg)

        # get all online and finished waves to which this account has VIEW_RESPONSES privs to.
        ac = AbstractController(self.sa_session, account_id=account.id, check_active=False)
        waves = ac.get_privileged_object_by_id(TntCLMWave, None, priv=TntCLMWave.PRIV_VIEW_RESPONSE)
        for wave in waves:
            if wave.status in [TntCLMWave.STATUS_ONLINE, TntCLMWave.STATUS_FINISHED]:
                xmlMg = xmlDoc.createElement('MemberGrant')
                xmlMg.setAttribute('member', '[Wave].[Code].[%s]' % wave.id)
                xmlMg.setAttribute('access', 'all')
                xmlHg.appendChild(xmlMg)
        return xmlCg

class TntCLMRole(RoleBase):

    def _get_xml(self):

        """
        Creates xml for all tnt accounts in the following format:

            <Role name="MyRole">
                <SchemaGrant access="none">
                    <CubeGrant cube="TntCLMResponse" access="all">

                        <HierarchyGrant hierarchy="[Country]" access="custom" rollupPolicy="partial">
                                <MemberGrant member="[Country].[BE]" access="all"/>
                                <MemberGrant member="[Country].[NL]" access="all"/>
                        </HierarchyGrant>

                        <!-- other dimensions here -->

                    </CubeGrant>
                </SchemaGrant>
            </Role>
        """

        # get all active CLM accounts
        # FIXME should we check for view_wave_responses permission?
        accounts = self.sa_session.query(Account).join(GroupAccount).join(Group).join(License).filter(
                    License.name == 'TNT CLM').all()

        xmlDoc = Document()
        xmlRoot = xmlDoc.createElement('Roles')
        xmlDoc.appendChild(xmlRoot)

        account_roles = AccountRole(self.sa_session, xmlDoc)

        for account in accounts:

            xmlRole = xmlDoc.createElement('Role');
            xmlRole.setAttribute('name', 'iqportal_account_%s' % account.id)
            xmlRoot.appendChild(xmlRole)

            xmlSg = xmlDoc.createElement('SchemaGrant')
            xmlSg.setAttribute('access', 'all')
            xmlRole.appendChild(xmlSg)

            for f in account_roles.get_all():
                xmlSg.appendChild(f(account))


        xml = ''
        for c in xmlRoot.childNodes:
            xml += c.toprettyxml()

        return xml

    xml = property(_get_xml)

    example_role = """
    """

projects = (TntCLMProject,)
