# Copyright (c) 2011 BaseSoft B.V.
# See LICENSE for details.
from django.utils.translation import ugettext as _
from django.conf import settings

from basesoft.web.core.exceptions import PublicUserException

from iqportal.iqube.core.reporters import DefaultIQubeReporter, ConfigurableWidgetReporter,\
    AbstractRankConfigurableWidgetReporter, DefaultCellClickReporter
from iqportal.iqube.xmla.mdx import MdxBuilder
from iqportal.iqube.core.models import CalculatedMember, IncorrectCalculatedMember
from iqportal.iqube.compose.controllers import MdxCalculatedMemberController
from iqportal.iqube.core.controllers import DataSourceController
from iqportal.iqube.xmla.result import MdxException

from basesoft import logging

logger = logging.getLogger("iqportal.tnt.report")

class TntCLMCellClickReporter(DefaultCellClickReporter):
    def get_result(self):
        # For now we only can do this if there is no query based calculated members used.
        try:
            result = self._get_mdx([{'hierarchy_unique_name': 'Participant', 'member_unique_name': '[Participant].Children','depth': 0}])
        # If something goes wrong it most likely is related to a query based calculated member.
        except MdxException, e:
            # log the exception
            logger.warn('MdxException: %s' % e.message);
            return None

        res = result.get_dict()
        rv = []
        for i in res['columns']:
            [rv.append(int(c['caption'])) for c in i]
        return {'slicers': res['slicers'], 'participant_ids': rv}

class DefaultTntCLMReporter(DefaultIQubeReporter):
    """
    This reporter can be used to set configs on the global and local filtering interfaces.
    See ticket #1880
    """

    def __init__(self, sa_session, account_id=None):

        ds_controller = DataSourceController(sa_session, account_id=account_id)
        datasource = ds_controller.get_datasource_by('mondrian_project', '%s.TNT-CLM' % settings.SETTINGS_MODULE).pop()

        DefaultIQubeReporter.__init__(self, sa_session, datasource, account_id=account_id)

        self.cm_controller = MdxCalculatedMemberController(self.sa_session, account_id=self.account_id)
        self._defaults = {}
        self._set_defaults()

    def _set_defaults(self):
        pass

    def get_default(self, dimension):
        if self._defaults.has_key(dimension):
            return self._defaults[dimension]
        return []

    def graph_config(self, target, config):
        if target == 'table':
            isT = False
            se = True
            if config.has_key('tableAttributes'):
                if config['tableAttributes'].has_key('isTransposed'):
                    isT = config['tableAttributes']['isTransposed']
                if config['tableAttributes'].has_key('slicersExpanded'):
                    se = config['tableAttributes']['slicersExpanded']
            return {
                'module': 'tnt_graph_datatablecube',
                'name': 'tnt_graph_datatablecube',
                'id': None,
                'config': {'isTransposed': isT, 'slicersExpanded': se, 'showSettingsIcon': False},
                'graph_type_id': 6,
            }
        else:
            return {
                'module': 'iqreport_fusion',
                'name': 'iqreport_fusion',
                'id': None,
                'config': {u'bar_horizontal': True, u'show_legend': True, u'overlap_legend': False, u'legend_outside_location': 'bottom left', u'aspect_ratio': 6.5, u'min_value': 0, u'max_value': 100},
                'graph_type_id': 3,
            }

    def _apply_config(self, rv, config, definitions, order=None):

        apply = {'rows': [], 'columns': [], 'slicers': []}

        if not rv['mdx_config'].has_key('mdx_with'):
            rv['mdx_config']['mdx_with'] = []

        if order is None:
            order = definitions.keys()

        for key in order:
            options = self.get_default(key)
            if config.has_key(key) and len(config[key]) > 0:
                options = config[key]
            if not definitions.has_key(key):
                continue
            direction =definitions[key]
            for cm in options:
                try:
                    _id = int(cm)
                    cm = self.cm_controller.get_calculated_member(_id)[0]
                    name = '[%s]' % cm.label
                    formula, _type, dimension = self.cm_controller.get_parsed(cm.id, [], True)
                    name, _with, dimension = self.cm_controller.get_with_statement(name, cm.type, formula, 'Position', None)
                    apply[direction].append({'hierarchy_unique_name': dimension, 'member_unique_name': name})
                    rv['mdx_config']['mdx_with'] += _with
                    continue
                except ValueError:
                    pass
                try:
                    apply[direction].append(self._make_mdx_from_cm(cm))
                    rv['mdx_config']['mdx_with'].append(str(cm))
                except IncorrectCalculatedMember:
                    apply[direction].append(self._make_mdx(cm))

        if len(apply['rows']) > 0:
            rv['mdx_config']['rows'] = apply['rows']
        if len(apply['columns']) > 0:
            rv['mdx_config']['columns'] = apply['columns']
        if len(apply['slicers']) > 0:
            rv['mdx_config']['slicers'] = apply['slicers']

        return rv

    def set_config(self, config):
        rv = self.get_default_config()
        rv['datasource_id'] = self._datasource.id
        rv = self.apply_config(rv, config)

        return rv

    def get_mdx_config(self, config, target):
        cfg = self.set_config(config)
        cfg['graph'] = self.graph_config(target, config)
        return cfg

    def get_default_config(self):
        return {
            'mdx_config': {
                'columns': [],
                'rows': [],
                'cube': 'TntCLMResponse',
                'slicers': [],
                'nonempty': True,
                'mdx_with': [],
            }
        }

class TntCLMSatisfactionReporter(DefaultTntCLMReporter):

    """
    table version of the statisfaction widget
    """

    def _set_defaults(self):
        self._defaults['answers'] = ['[KPI].[%s]' % a for a in ['No. of Answers', 'Dissatisfied', 'Meeting Cust. Exp.', 'Exceeding Cust. Exp.', 'NCS Score', 'Mean Score']]

    def apply_config(self, rv, config):

        definitions = {
            'attributes': 'rows',
            'waves': 'rows',
            'answers': 'columns',
            'geolevels': 'rows',
            'depots': 'rows',
            'services': 'rows',
            'revenue_band': 'rows',
            'location': 'rows',
            'respondent_type': 'rows'
        }


        # ticket 2052: if no waves are selected then take all main waves and put them on slicers
        if config.has_key('waves') and len(config['waves']) == 1 and config['waves'][0] == '[Wave].[Main Waves]':
            definitions['waves'] = 'slicers'

        # service 'express' is default service
        if config.has_key('services') and len(config['services']) == 1 and config['services'][0] == '[Service].[Express]':
            definitions['services'] = 'slicers'

        # when overal
        if config.has_key('attributes') and len(config['attributes']) == 1 and config['attributes'][0] == '[Position].[Overall Satisfaction]':

            slice = False
            for key in definitions:
                if key == 'attributes' or definitions[key] == 'slicers':
                    print 'continueing for %s' % definitions[key]
                    continue

                if config.has_key(key) and len(config[key]):
                    print 'found config for %s' % config[key]
                    slice = True
                    break

            if slice:
                definitions['attributes'] = 'slicers'

        return self._apply_config(rv, config, definitions)


class TntCLMSatisfactionGraphsReporter(DefaultTntCLMReporter):

    def _set_defaults(self):
        self._defaults['answers'] = ['[KPI].[%s]' % a for a in ['No. of Answers', 'Dissatisfied', 'Meeting Cust. Exp.', 'Exceeding Cust. Exp.', 'NCS Score', 'Mean Score']]

    def get_mdx_config(self, config, target):
        config = DefaultTntCLMReporter.get_mdx_config(self, config, target)
        config['mdx_config']['nonempty'] = MdxBuilder.ROW

        if self.splitby:
            return config

        # make dissatisfied negative
        graph_config = config['graph']['config']
        if graph_config.has_key('type') and graph_config['type'] == 'stcolumn3dline':
            config['mdx_config']['mdx_with'] = [ 'MEMBER [KPI].[Dissatisfied] as (-100 * ([KPI].[Lower] + [KPI].[Much lower]) / [KPI].[No. of Answers]), SOLVE_ORDER=3' ]

        return config

    def graph_config(self, target, config):

        cfg = DefaultTntCLMReporter.graph_config(self, target, config)

        if target == 'graph' and hasattr(self, 'splitby'):
            cfg['config'] = {u'show_legend': True, u'aspect_ratio': 2.5, 'fusion_chart.legendPosition': 'BOTTOM'}
            if self.splitby:
                cfg['config']['type'] ='msline2d'
            else:

                cfg['config']['type'] = 'mscolumn3d'

                if config.has_key('horizontalBar') and config['horizontalBar']:
                    cfg['config']['type'] = 'msbar3d'

                if config.has_key('stackedBar') and config['stackedBar']:

                    # 'answers': [u'[KPI].[Exceeding Cust. Exp.]', u'[KPI].[Dissatisfied]', u'[KPI].[NCS Score]']

                    # color map makes sure KPI keep having the same colors
                    color_map = {
                       '[KPI].[Exceeding Cust. Exp.]': '#4684EE',
                       '[KPI].[NCS Score]': '#fe6403',
                       '[KPI].[Dissatisfied]': '#DC3912',
                    }
                    colors = []

                    # line index always is NCS score
                    line_index = None

                    for i in range(len(config['answers'])):
                        member = config['answers'][i]
                        colors.append(color_map[member])
                        if member == '[KPI].[NCS Score]':
                            line_index = i+1


                    cfg['config']['aspect_ratio'] = 1.4
                    cfg['config']['type'] = 'stcolumn3dline'
                    cfg['config']['fusion_chart.valuePosition'] = 'BELOW'

                    if line_index is not None:
                        cfg['config']['render_line_indexes'] = unicode(line_index)
                    cfg['config']['fusion_chart.showPercentValues'] = False
                    cfg['config']['override_colors'] = ','.join(colors)

                    if config.has_key('horizontalBar') and config['horizontalBar']:
                        cfg['config']['type'] = 'stbar3d'


        return cfg

    def apply_config(self, rv, config):

        definitions = {
            'attributes': 'slicers',
            'waves': 'rows',
            'answers': 'columns',
            'geolevels': 'slicers',
            'depots': 'slicers',
            'services': 'slicers',
            'revenue_band': 'slicers',
            'location': 'slicers',
            'respondent_type': 'slicers'
        }

        dimensions_default = {
            'waves': ['[Wave].[Main Waves].Children'],
            'attributes': ['[Pool].Children'],
            'geolevels': ['[Country].Children'],
            'depots': ['[Depot].Children'],
            'services': ['[Service].Children'],
        }

        rows_dimensions_default = {
            'waves': ['[Wave].[Main Waves].Children'],
            'attributes': ['[Position].[Overall Satisfaction]'],
            'geolevels': ['[Country].[All Countries]'],
            'depots': ['[Depot].[All Depots]'],
            'services': ['[Service].[All Services]'],
            'revenue_band': ['[Revenue Band].[All Revenue Bands]'],
            'location': ['[Location].[All Locations]'],
            'respondent_type': ['[Respondent Type].[All Respondent Types]'],
        }

        # express to slicers if its the only service
        if config.has_key('services') and len(config['services']) == 1 and config['services'][0] == '[Service].[Express]':
            definitions['services'] = 'slicers'

        if config.has_key('onRows') and config['onRows'] not in [None, '', 'waves']:
            definitions[config['onRows']] = 'rows'
            if config['onRows'] == 'answers':
                definitions['waves'] = 'columns'
                if not config.has_key('waves') or len(config['waves']) == 0:
                    config['waves'] = rows_dimensions_default['waves']
            else:
                definitions['waves'] = 'slicers'

                if not config.has_key(config['onRows']) or len(config[config['onRows']]) == 0:
                    if not rows_dimensions_default.has_key(config['onRows']):
                        raise Exception('Invalid dimension for onRows')
                    config[config['onRows']] = rows_dimensions_default[config['onRows']]

        self.splitby = False
        self._defaults['answers'] = ['[KPI].[%s]' % a for a in ['Exceeding Cust. Exp.', 'NCS Score', 'Dissatisfied']]

        if config.has_key('splitby') and config['splitby'] not in [None, '']:

            # cannot split by a dimension on rows
            if definitions[config['splitby']] == 'rows':
                raise PublicUserException(_('Cannot split by row dimension'))

            # need values in order to split by
            if not config.has_key(config['splitby']) or len(config[config['splitby']]) == 0:
                if not dimensions_default.has_key(config['splitby']):
                    raise Exception('Invalid dimension for splitby')
                config[config['splitby']] = dimensions_default[config['splitby']]

            self.splitby = True
            self._defaults['answers'] = ['[KPI].[%s]' % a for a in ['NCS Score']]
            definitions[config['splitby']] = 'columns'
            definitions['answers'] = 'slicers'

        return self._apply_config(rv, config, definitions)

class TntCLMLoyaltyScoreReporter(DefaultTntCLMReporter):

    def _set_defaults(self):
        self._defaults['answers'] = ['[KPI].[%s]' % a for a in ['No. of Answers', 'Score', 'Score %']]

    def apply_config(self, rv, config):

        definitions = {
            'waves': 'rows',
            'answers': 'columns',
            'geolevels': 'rows',
            'depots': 'rows',
            'services': 'rows',
            'revenue_band': 'rows',
            'location': 'rows',
            'respondent_type': 'rows',
            'attributes': 'slicers',
        }

        # put services on slicers if thats the onkly one
        if config.has_key('services') and len(config['services']) == 1 and config['services'][0] == '[Service].[Express]':
            definitions['services'] = 'slicers'

        # override the total answers column as this is a non-scale attribute
        # TODO: maybe move this to XMLAClient level?
        cm = CalculatedMember('[KPI]', 'No. of Answers', '[KPI].[All KPIs]')
        if not rv['mdx_config'].has_key('mdx_with'):
            rv['mdx_config']['mdx_with'] = []
        rv['mdx_config']['mdx_with'].append(str(cm))

        if config.has_key('answers'):
            answers = []
            [answers.append(i) for i in config['answers'] if i in ['[KPI].[%s]' % j for j in ['No. of Answers', 'Score', 'Score %']]]
            config['answers'] = answers

        return self._apply_config(rv, config, definitions)


class TntCLMLeagueTableReport(DefaultTntCLMReporter):
    """
        WITH
        MEMBER [KPI].[NCS Rank] AS
           Rank(
              [Country].CurrentMember,
              Order(
                 [Country].[All Countries].Children,
                 [KPI].[NCS Score],
                 BDESC
              ),
              [KPI].[NCS Score]
           )
        SELECT CROSSJOIN({[Wave].[Code].Members}, {[KPI].[NCS Rank], [KPI].[NCS Score]}) ON AXIS(0)
        , NON EMPTY {Order([Country].[All Countries].Children,
        ([KPI].[NCS Score],
        [Wave].[All Waves].LastChild), BDESC)} ON AXIS(1)
        FROM [TntCLMResponse]
        WHERE ([Position].[Overall Satisfaction])
    """

    def apply_config(self, rv, config):
        definitions = {}

        setup = {
            'positions': {
                'name': '[Position]',
                'members': ['[Position].[All Positions].Children'],
            },

            'waves': {
                'name': '[Wave]',
                'members': ['[Wave].[Code].Members'],
            },

            'attributes': {
                'name': '[Attribute]',
                'members': ['[Attribute].[Scale].Children'],
            },

            'geolevels': {
                'name': '[Country]',
                'members': ['[Country].[All Countries].Children'],
            },

            'depots': {
                'name': '[Depot]',
                'members': ['[Depot].[All Depots].Children'],
            },

            'services': {
                'name': '[Services]',
                'members': ['[Services].[All Services].Children'],
            },

            'location': {
                'name': '[Location]',
                'members': ['[Location].[All Locations].Children'],
            },

            'revenue_band': {
                'name': '[Revenue Band]',
                'members': ['[Revenue Band].[All Revenue Bands].Children'],
            },

            'respondent_type': {
                'name': '[Respondent Type]',
                'members': ['[Respondent Type].[All Respondent Types].Children'],
            }

        }

        # process
        for key in setup.keys():

            if config.has_key(key) and not len(config[key]) == 0:
                members = config[key]

                # check all members if they are calculated members and add them to
                new_members = []
                for member in members:
                    try:
                        cm = CalculatedMember(member_string=member)
                        new_member = '%s.[%s]' % (cm.dimension, cm.name)
                        new_members.append(new_member)
                        rv['mdx_config']['mdx_with'].append(member)
                    except IncorrectCalculatedMember:
                        new_members.append(member)

                setup[key]['members'] = new_members


        if not config.has_key('answers') or not len(config['answers']):
            config['answers'] = ['[KPI].[%s]' % self.RANK_MEMBER]
        config['answers'].append('[KPI].[Rank]')

        pos = self.POSITION
        if type(pos) != list:
            pos = [self.POSITION]

        config['positions'] = ["[Position].[%s]" % p for p in pos]

        if not rv['mdx_config'].has_key('mdx_with'):
            rv['mdx_config']['mdx_with'] = []

        rankhierarchy = setup['geolevels']['name']
        rankmembers = ','.join(setup['geolevels']['members'])

        if config.has_key('rankby') and config['rankby'] is not None:
            rankvalue = '[KPI].[%s]' % config['rankby']
        else:
            rankvalue = '[KPI].[%s]' % self.RANK_MEMBER

        extra_with = ["""
            MEMBER [KPI].[Rank] AS Rank(
                %(rankhierarchy)s.CurrentMember,
                Order(
                    {%(rankmembers)s},
                    %(rankvalue)s,
                    BDESC
                ),
                %(rankvalue)s
            )""" % {
                'rankhierarchy': rankhierarchy,
                'rankmembers': rankmembers,
                'rankvalue': rankvalue
            }]

        if not config.has_key('splitby'):
            config['splitby'] = None

        if config['splitby'] is None or not setup[config['splitby']].has_key('members'):
            extra_with.append("""
                SET [Ordered Set] AS Filter(
                    Order({%(rankmembers)s}, (%(rankvalue)s), BDESC),
                    NOT ISEMPTY(%(rankvalue)s))
            """ % {
                #'rankhierarchy': rankhierarchy,
                'rankmembers': rankmembers,
                'rankvalue': rankvalue
            })
        else:
            extra_with.append("""
                SET [Ordered Set] AS Filter(
                    Order({%(rankmembers)s}, (%(rankvalue)s, {%(splitby)s}.Item(0) ), BDESC),
                    NOT ISEMPTY(%(rankvalue)s))
            """ % {
                #'rankhierarchy': rankhierarchy,
                'rankmembers': rankmembers,
                'rankvalue': rankvalue,
                'splitby': ','.join(setup[config['splitby']]['members'])
            })

        rv['mdx_config']['rows'] = [{
            "member_unique_name":"[Ordered Set]",
            "hierarchy_unique_name":"[Ordered Set]",
            "depth": "0"
        }]
        rv['mdx_config']['mdx_with'].extend(extra_with)

        # each dimension that is not in splitby or rankby goes into slicers
        for key in setup.keys():
            if config['splitby'] is not None and key in (config['splitby'], ):
                definitions[key] = 'columns'
                continue
            definitions[key] = 'slicers'

        # force these dimensions not into slicers
        definitions['answers'] = 'columns'
        definitions['geolevels'] = 'rows'

        return self._apply_config(rv, config, definitions, order=['waves', 'attributes', 'geolevels', 'location', 'revenue_band', 'respondent_type', 'answers', 'depots', 'services', 'positions'])


class TntCLMNCSLeagueTableReport(TntCLMLeagueTableReport):

    RANK_MEMBER = 'NCS Score'
    POSITION = ['25','27','29']

class TntCLMLoyaltyLeagueTableReport(TntCLMLeagueTableReport):

    RANK_MEMBER = 'Score %'
    POSITION = '31'

    def apply_config(self, rv, config):
        cm = CalculatedMember('[KPI]', 'No. of Answers', '[KPI].[All KPIs]')
        if not rv['mdx_config'].has_key('mdx_with'):
            rv['mdx_config']['mdx_with'] = []
        rv['mdx_config']['mdx_with'].append(str(cm))

        if config.has_key('answers'):
            answers = []
            [answers.append(i) for i in config['answers'] if i in ['[KPI].[%s]' % j for j in ['No. of Answers', 'Score', 'Score %']]]
            config['answers'] = answers

        return TntCLMLeagueTableReport.apply_config(self, rv, config)



class TntCLMActionPlanningReport(DefaultTntCLMReporter):

    def _set_defaults(self):
        self._defaults['answers'] = ['[KPI].[NCS Score]']

    def apply_config(self, rv, config):

        definitions = {}

        setup = {
            'waves': {
                'name': '[Wave]',
                'members': '[Wave].[Code].Members',
            },

            'attributes': {
                'name': '[Attribute]',
                'members': '[Attribute].[Scale].Children',
            },

            'geolevels': {
                'name': '[Country]',
                'members': '[Country].[All Countries].Children',
            },

            'depots': {
                'name': '[Depots]',
                'members': '[Depot].[All Depots].Children',
            },

            'services': {
                'name': '[Services]',
                'members': '[Services].[All Services].Children',
            },

            'location': {
                'name': '[Location]',
                'members': ['[Location].[All Locations].Children'],
            },

            'revenue_band': {
                'name': '[Revenue Band]',
                'members': ['[Revenue Band].[All Revenue Bands].Children'],
            },

            'respondent_type': {
                'name': '[Respondent Type]',
                'members': ['[Respondent Type].[All Respondent Types].Children'],
            }
        }

        # process
        for key in setup.keys():

            if config.has_key(key) and not len(config[key]) == 0:
                members = config[key]

                # we can only do a topcount on one answer member so just take the first
                if key == 'answers':
                    members = [members[0]]

                # check all members if they are calculated members and add them to
                new_members = []
                for member in members:
                    try:
                        cm = CalculatedMember(member_string=member)
                        new_member = '%s.[%s]' % (cm.dimension, cm.name)
                        new_members.append(new_member)
                        rv['mdx_config']['mdx_with'].append(member)
                    except IncorrectCalculatedMember:
                        new_members.append(member)

                setup[key]['members'] = ','.join(new_members)

        if not config.has_key('direction'):
            config['direction'] = 'bottom'

        if config['direction'] == 'bottom':
            method = 'BottomCount'
        else:
            method = 'TopCount'

        if not config.has_key('rankby'):
            config['rankby'] = 'attributes'

        rank_by = setup[config['rankby']]['members']

        if config.has_key('attributes') and len(config['attributes']):
            rank_by = """
                EXTRACT(
                    FILTER(
                        CROSSJOIN(
                            {%(rank_by)s},
                            {[Attribute].[Scale].Children}
                        ),
                        NOT ISEMPTY([KPI].[No. of Answers])
                    ),
                    [Attribute]
                )
            """  % {'rank_by': rank_by}

        if not config.has_key('number_of_items'):
            config['number_of_items'] = 5

        if not config.has_key('splitby'):
            config['splitby'] = None

        if config['splitby'] is None:
            extra_with = """
                SET [set1] as '%(method)s(
                    %(rank_by)s,
                    %(number_of_items)s.0,
                    %(answers)s
                )'
                """  % {
                'method': method,
                'rank_by': rank_by,
                'answers': '[KPI].[NCS Score]',
                'number_of_items': config['number_of_items']
            }
        else:

            extra_with = """
                SET [set1] as 'Generate(
                    {%(split_by)s},
                    %(method)s(Crossjoin(
                        {%(split_by_dimension)s.CurrentMember},
                        {%(rank_by)s}
                    ),
                    %(number_of_items)s.0, %(answers)s)
                )'
                """  % {
                'method': method,
                'split_by': setup[config['splitby']]['members'],
                'split_by_dimension': setup[config['splitby']]['name'],
                'rank_by': rank_by,
                'answers': '[KPI].[NCS Score]',
                'number_of_items': config['number_of_items']
            }

        rv['mdx_config']['mdx_with'].append(extra_with)


        # only answers get treated normally
        definitions['answers'] = 'columns'

        # each dimension that is not in splitby or rankby goes into slicers
        for key in setup.keys():
            if key in (config['rankby'], config['splitby']):
                continue
            definitions[key] = 'slicers'

        return self._apply_config(rv, config, definitions)

    def get_default_config(self):

        config = DefaultTntCLMReporter.get_default_config(self)
        config['mdx_config']['rows'] = [{
            "member_unique_name": "[set1]",
            "hierarchy_unique_name" :"[set1]",
            "depth": "0"
        }]

        return config

class TntCLMGeneralQuestionsReporter(DefaultTntCLMReporter):

    """
    with member [Measures].[No. of Answers %] as '(([Answer].CurrentMember / [Answer].[All Answers]) * 100)'
    select {[Measures].[No. of Answers], [Measures].[No. of Answers %]} ON COLUMNS,
      NON EMPTY Crossjoin({[Attribute].[Single].Children}, {[Answer].Children}) ON ROWS
    from [TntCLMResponse]
    """

    GRAPH_TABLE_ONLY = 1
    GRAPH_PIE = 2
    GRAPH_BAR_V = 3
    GRAPH_BAR_H = 4

    def graph_config(self, target, config):

        cfg = DefaultTntCLMReporter.graph_config(self, target, config)

        if target == 'graph':
            cfg['config'] = {
                'show_legend': True,
                'aspect_ratio': 2.5,
                'fusion_chart.legendPosition': 'BOTTOM',
            }

            print self.graphConfig

            if self.graphConfig == self.GRAPH_PIE:
                cfg['config']['type'] = 'pie3d'
                cfg['config']['fusion_chart.showLabels'] = False
                cfg['config']['fusion_chart.showPercentValues'] = False
                cfg['config']['fusion_chart.showPercentInToolTip'] = True
            else:
                cfg['config']['type'] = 'stcolumn3d'
                if self.graphConfig == self.GRAPH_BAR_H:
                    cfg['config']['type'] = 'stbar3d'

        return cfg

    def apply_config(self, rv, config):

        definitions = {
            'attributes': 'rows',
            'answers': 'columns',
            'waves': 'columns',
            'geolevels': 'columns',
            'depots': 'columns',
            'revenue_band': 'columns',
            'location': 'columns',
            'respondent_type': 'columns',
            'pool': 'slicers',
            'qa_answers': 'rows',
            'services': 'slicers',
        }

        config['qa_answers'] = ['[Answer].Children']

        extra_with = "MEMBER [Measures].[No. of Answers %] as '(([Answer].CurrentMember / [Answer].[All Answers]) * 100)'"
        rv['mdx_config']['mdx_with'].append(extra_with)

        # the first two attributes are always '[Attribute].[Single].Children}, {[Answer].Children}'. the rest can be selected
        # tru the interface and are Pool members. Put those on slicers
        #if len(config['attributes']) > 2:
        #    config['pool'] = config['attributes'][2:]
        #    config['attributes'] = config['attributes'][:2]

        self.graphConfig = int(config['graphType'])
        if self.graphConfig in (self.GRAPH_PIE, self.GRAPH_BAR_H, self.GRAPH_BAR_V):

            # force absolute measure
            #config['answers'] = ['[Measures].[No. of Answers]']

            # put everything else in slicers
            for d in definitions.keys():

                if d not in ('answers', 'qa_answers'):
                    definitions[d] = 'slicers'

        return self._apply_config(rv, config, definitions, order=['pool', 'waves', 'attributes', 'geolevels', 'location', 'revenue_band', 'services', 'respondent_type', 'answers', 'depots', 'positions', 'qa_answers'])

def GeolevelChildrenProcessor(selected, datasource, account):
    """
    Due to the setup of bogus Hierarchy called Geolevel, we have to filter out duplicates in case
    children or members are used on any Geolevel member.
    """
    if type(selected) != dict:
        return selected
    if selected['dimension_unique_name'] != '[Geolevel]':
        return selected
    if selected['member_unique_name'][-8:] != 'Children' and selected['member_unique_name'][-7:] != 'Members':

        name = selected['member_unique_name']
        depth = name.count(']')
        id = name[name.rfind('[')+1:-1]
        if depth == 2:
            # Business Unit
            selected['member_unique_name'] = '[Geolevel].[Business Unit].[%s]' % id
            return selected
        if depth == 3:
            # Region
            selected['member_unique_name'] = '[Geolevel].[Region].[%s]' % id
            return selected
        if depth == 4:
            # Subregion
            selected['member_unique_name'] = '[Geolevel].[Sub Region].[%s]' % id
            return selected
        if depth == 5:
            # Countrye
            selected['member_unique_name'] = '[Geolevel].[Country].[%s]' % id
            return selected

        return selected
    client = datasource.connect(account_id = account.id)

    unique_name = 'MEMBER_UNIQUE_NAME'
    member = selected['member_unique_name'][:-9]

    if selected['member_unique_name'][-7:] == 'Members':
        unique_name = 'LEVEL_UNIQUE_NAME'
        member = selected['member_unique_name'][:-8]

    members = client.get_member_tree_op('TntCLMVirtualCube', member, client.MDTREEOP_CHILDREN, unique_name_handle=unique_name)
    rv = []
    dupe = []
    for m in members:
        if m['member_caption'] in dupe:
            continue
        rv.append(m)
        dupe.append(m['member_caption'])

    return sorted(rv, key=lambda k: k['member_caption'])

class TntCLMLoyaltyLeagueTableWidgetReporter(AbstractRankConfigurableWidgetReporter):

    RANK_MEMBER_NAME = ('Answer', 'Rank',)
    RANK_MEMBER_VALUE = ('Answer', 'Score %',)
    RANK_HIERARCHY = 'Geolevel'
    RANK_AXIS = 'columns'

    def deserialize(self, x):
        return GeolevelChildrenProcessor(ConfigurableWidgetReporter.deserialize(self, x), self._datasource, self.get_account())

    def get_mdx_config(self, config, applied_config):

        applied_config['dimensions']['Attribute'] = len(applied_config['dimensions'])
        index = applied_config['dimensions']['Attribute']
        applied_config['axes_%s' % index] = "slicers"
        applied_config['dimension_%s' % index] = [{"member_unique_name": "[Attribute].[2].[31]"}]

        index = applied_config['dimensions']['Geolevel']

        self.RANK_MEMBERS = ','.join([x['member_unique_name'] for x in applied_config['dimension_%s' % index]])
        applied_config['dimension_%s' % index] = [{"member_unique_name": "[Ordered Set]"}]

        rank_value = '[%s].[%s]' % (self.RANK_MEMBER_VALUE[0], self.RANK_MEMBER_VALUE[1])

        order_set = []
        # we have to add all used dimensions from the opposite axis to order the set.
        if applied_config['axes_%s' % index] == 'rows':
            axis = 'columns'
        else:
            axis = 'rows'
        # loop all the dimensions and see if they are on the opposite axis
        for a, d in applied_config['dimensions'].iteritems():
            if d == index:
                continue
            if applied_config['axes_%s' % d] != axis:
                continue
            # we already have the rank value included in the order, we dont want it twice.
            if len(applied_config['dimension_%s' % d]) == 0 or applied_config['dimension_%s' % d][0]['member_unique_name'] == rank_value:
                continue
            # We use only the first item from a certain set.
            order_set.append('{%s}.Item(0)' % applied_config['dimension_%s' % d][0]['member_unique_name'])
        if len(order_set) > 0:
            order_set = ', %s' % ','.join(order_set)
        else:
            order_set = ''

        number = "MEMBER [Answer].[No. of Answers] AS '[Answer].[All Answers]'"
        index = applied_config['dimensions']['Answer']
        applied_config['dimension_%s' % index].append(number)

        split_by = []
        for a in applied_config['dimensions'].values():
            if applied_config.has_key('split_by_%s' % a) and applied_config['split_by_%s' % a]:
                split_by = split_by + [x['member_unique_name'] for x in applied_config['dimension_%s' % a]]

        if len(split_by) > 0:
            order_set = '%s, {%(splitby)s}.Item(0)' % (order_set, ','.join(split_by))
        self._extra_with = ["""
                SET [Ordered Set] AS Filter(
                    Order({%(rankmembers)s}, (%(rankvalue)s %(order_set)s), BDESC),
                    NOT ISEMPTY(%(rankvalue)s))
            """ % {
                'rankmembers': self.RANK_MEMBERS,
                'rankvalue': rank_value,
                'order_set': order_set,
            }]

        return AbstractRankConfigurableWidgetReporter.get_mdx_config(self, config, applied_config)


class TntCLMDefaultConfigurableWidgetReporter(ConfigurableWidgetReporter):
    def deserialize(self, x):
        return GeolevelChildrenProcessor(ConfigurableWidgetReporter.deserialize(self, x), self._datasource, self.get_account())

class DiffandAvgReporter(TntCLMDefaultConfigurableWidgetReporter):

    """

    Shows the difference as the last column and under each column the average. Query should be something like this:

    WITH
        set [set_avg] as {[Revenue Band].Children}

        MEMBER [Revenue Band].[Avg] as '
            avg([set_avg])
        )'

        set [set_diff] as {[Wave].[Main].Members}

        MEMBER [Wave].[Diff] as '
            IIf([set_diff].Item([set_diff].Count - 1) > 0,
            [set_diff].Item([set_diff].Count - 1) - [set_diff].Item([set_diff].Count - 2),
            null
        )'

        SELECT NON EMPTY {[set_diff], [Wave].[Diff]} ON COLUMNS,
            NON EMPTY {[set_avg], [Revenue Band].[Avg]} ON ROWS
        FROM [TntCLMTrueResponseCube]
        WHERE {[KPI].[Dissatisfied]}

    """

    def __init__(self, *args, **kwargs):
        self._extra_with = [

            # avg set
            "set [set_avg] as {%s}",

            """
                MEMBER [%s].[Avg] as '
                    avg([set_avg])
                )
                ', SOLVE_ORDER=3
            """,

            # FIXME: check for !null instead of >0
            "set [set_diff] as {%s}",

            """
                MEMBER [%s].[Diff] as '
                IIf([set_diff].Item([set_diff].Count - 1) > 0,
                    [set_diff].Item([set_diff].Count - 1) - [set_diff].Item([set_diff].Count - 2),
                    null
                )
                ', SOLVE_ORDER=3
            """,

        ]
        super(DiffandAvgReporter, self).__init__(*args, **kwargs)

    def get_mdx_config(self, config, applied_config):

        # avg
        dim = applied_config['row_order'][-1]
        dim_id = applied_config['dimensions'][dim]
        members = [x['member_unique_name'] for x in applied_config['dimension_%s' % dim_id]]

        if len(members) == 0:
            raise PublicUserException(_("Empty 'row' selection"))

        self._extra_with[0] = self._extra_with[0] % ",".join(members)
        self._extra_with[1] = self._extra_with[1] % dim
        applied_config['dimension_%s' % dim_id] = ["[set_avg], [%s].[Avg]" % dim]

        # diff
        dim = applied_config['column_order'][-1]
        dim_id = applied_config['dimensions'][dim]
        members = [x['member_unique_name'] for x in applied_config['dimension_%s' % dim_id]]

        if len(members) == 0:
            raise PublicUserException(_("Empty 'column' selection"))

        self._extra_with[2] = self._extra_with[2] % ",".join(members)
        self._extra_with[3] = self._extra_with[3] % dim
        applied_config['dimension_%s' % dim_id] = ["[set_diff], [%s].[Diff]" % dim]

        return TntCLMDefaultConfigurableWidgetReporter.get_mdx_config(self, config, applied_config)

class CrossReferenceReporter(TntCLMDefaultConfigurableWidgetReporter):

    def get_mdx_config(self, config, applied_config):

        # add avg if it's a table
        if applied_config.has_key('show_graph') and applied_config['show_graph']:
            return ConfigurableWidgetReporter.get_mdx_config(self, config, applied_config)
        for dim in ['Answer', 'AnswerFrom', 'KPI', 'KPIFrom']:
            if applied_config['dimensions'].has_key(dim) and\
            applied_config['axes_%s' % applied_config['dimensions'][dim]] == 'rows':
                index = applied_config['dimensions'][dim]
                self._extra_with = ["MEMBER [%s].[Average] AS AVG({%s})" % (dim,
                                    ','.join([x['member_unique_name'] for x in applied_config['dimension_%s' % index]]))
                                    ]
                applied_config['dimension_%s' % index].append('[%s].[Average]' % dim)

        return TntCLMDefaultConfigurableWidgetReporter.get_mdx_config(self, config, applied_config)

class GeneralAtttributesWidgetReporter(TntCLMDefaultConfigurableWidgetReporter):
    pass

class GAWidgetReporter(TntCLMDefaultConfigurableWidgetReporter):

    def get_mdx_config(self, config, applied_config):

        applied_config = self.modify_applied_config(config, applied_config)
        # we need both Answer and Attribute dimensions on same axis, rows or columns
        if applied_config['dimensions'].has_key("Answer") and applied_config['dimensions'].has_key("Attribute"):
            answer_index = applied_config['dimensions']["Answer"]
            attribute_index = applied_config['dimensions']["Attribute"]

            if applied_config['axes_%s' % answer_index] == applied_config['axes_%s' % attribute_index] and\
            applied_config['axes_%s' % answer_index] in ['rows', 'columns']:
                self._extra_with = ["""set [set1] as '
                    Extract((
                        {%(answers)s} *
                        {%(attributes)s}
                        ), [Answer])'""" % {
                            'answers': ','.join([x['member_unique_name'] for x in applied_config['dimension_%s' % answer_index]]),
                            'attributes': ','.join([x['member_unique_name'] for x in applied_config['dimension_%s' % attribute_index]])
                        }
                ]
                axis = applied_config['axes_%s' % answer_index]

                applied_config['dimension_%s' % answer_index] = ['[set1]']
                applied_config['dimension_%s' % attribute_index] = []

        return TntCLMDefaultConfigurableWidgetReporter.get_mdx_config(self, config, applied_config)

class NpsKpiWidgetReporter(TntCLMDefaultConfigurableWidgetReporter):

    def get_mdx_config(self, config, applied_config):
        # TODO do we need anything extra here?
        return TntCLMDefaultConfigurableWidgetReporter.get_mdx_config(self, config, applied_config)

reporters = (TntCLMDefaultConfigurableWidgetReporter, TntCLMLoyaltyLeagueTableWidgetReporter, DiffandAvgReporter,
             CrossReferenceReporter, GAWidgetReporter, NpsKpiWidgetReporter, GeneralAtttributesWidgetReporter,)
