# Copyright (c) 2011 BaseSoft B.V.
# See LICENSE for details.

import random
import csv
import string
import smtplib
import os

from types import NoneType
from hashlib import sha256 as sha
from datetime import datetime, timedelta
from subprocess import call

from email import Encoders
from email.mime.text import MIMEText
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.Utils import formatdate

from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.exc import IntegrityError
from sqlalchemy.sql import and_, desc
from sqlalchemy import func
from sqlalchemy.orm import join
from sqlalchemy import create_engine

from django.conf import settings
from django.utils.translation import ugettext as _
from django.http import HttpResponse
from django.utils import simplejson

from basesoft.web.core.jsonrpc.dispatcher import JsonRpcJSONEncoder, json_object_hook
from basesoft.web.contrib.i18n.models import Country
from basesoft.web.contrib.contact.models import Contact, Person
from basesoft.web.core.exceptions import SystemException, PublicUserException, SecurityException
from basesoft.web.core.util.controllers import AbstractController, LicenseConfigController, transaction_decorator
from basesoft.web.contrib.file.controllers import AnonymousFileController, ProcessFileController
from basesoft.web.contrib.file.models import ImportFile, ImportFileType, File, ImageFile

from iqportal.tnt.clm.controllers import WaveController, TntCLMLocaleController
from iqportal.tnt.core.controllers import TntController
from iqportal.tnt.core.models import TntGeolevel, TntGeolevelCountry, TntDepot
from iqportal.tnt.response.models import TntCLMWaveInvitee, TntCLMWaveParticipant, TntCLMWaveResponse,\
    TntCLMWaveParticipantSession, TntCLMWavePriorityCSVFile, TntCLMWavePriorityCSVFileParticipant
from iqportal.tnt.clm.models import TntCLMAttribute, TntCLMWave, TntCLMLocale, TntCLMCompany, TntCLMWaveAttribute, TntCLMWaveAnswer,\
    TntCLMWaveLocaleRevision, TntCLMWaveContactList, TntCLMWaveType, TntCLMCustomer, TntCLMWaveScaleAnswer

from basesoft import logging
logger = logging.getLogger("iqportal.tnt.response")

class ResponseController(AbstractController):
    """
    Tnt CLM Response controller
    """

    def save_or_update_invitees(self, wave):
        """
        Creates/updates invitees.
        This can also be used after the publication has been set online and some contacts are missing.
        """

        # check for online status of wave.
        if wave.status != TntCLMWave.STATUS_ONLINE:
            raise SystemException('Cannot add invitees for Wave which is not online.')

        contact_lists = self.sa_session.query(TntCLMWaveContactList).filter(
                        and_(TntCLMWaveContactList.wave==wave,
                             TntCLMWaveContactList.is_processed==TntCLMWaveContactList.STATUS_NOT_PROCESSED)).all()
        if len(contact_lists) == 0:
            return []

        # since multiple contact lists could be already in the query result, update the status for all of them now
        # (because saving invitees could take a while) so they won't be selected next time
        self.set_wave_contact_lists_status(contact_lists, TntCLMWaveContactList.STATUS_PROCESSING)

        rv = 0
        for cl in contact_lists:
            rv = rv + self._create_invitees(cl)
            self.set_wave_contact_lists_status(cl, TntCLMWaveContactList.STATUS_PROCESSED)
        return rv

    def _create_invitees(self, wave_contact_list):
        self.db_engine = create_engine(settings.DB_CONNSTR, echo=False).connect()
        wave = wave_contact_list.wave
        # list of already invited contacts.
        sql = "SELECT contact_id FROM tblTntCLMWaveInvitee WHERE wave_id = %s" % wave.id
        contacts = [cid for cid, in self.db_engine.execute(sql)]

        rv = 0

        cl = wave_contact_list.contact_list

        insert_sql = "INSERT INTO tblTntCLMWaveInvitee (contact_id, wave_id, invitee_hash, created, last_updated) VALUES (%s, "+str(wave.id)+", '%s', NOW(), NOW())"

        sql = """SELECT clc.contact_id FROM tblContactListContact clc
                 INNER JOIN tblContact c on c.id=clc.contact_id
                 WHERE contact_list_id = %s AND c.archived=0 AND c.deleted=0"""
        result = self.db_engine.execute(sql % cl.id)
        for cid, in result:
            if cid in contacts:
                continue
            hash = sha(str(random.random())).hexdigest()[:8]
            while True:
                try:
                    s = insert_sql % (cid, hash)
                    self.db_engine.execute(s)
                    break
                except IntegrityError, e: # hash must be unique
                    if e.message.find("Duplicate entry '"+hash+"' for key") < 0:
                        raise
                    hash = sha(str(random.random())).hexdigest()[:8]
            rv = rv + 1
            contacts.append(cid)
        return rv

    @transaction_decorator
    def set_wave_contact_lists_status(self, wave_contact_lists, status):
        if not isinstance(wave_contact_lists, list):
            wave_contact_lists = [wave_contact_lists]
        for wcl in wave_contact_lists:
            wcl.is_processed = status
            self.sa_session.add(wcl)

    def get_invitee_by_account_number(self, account_no, country_id):
        """
        This is used when the invitee uses the login page.
        TODO: selection in case of multiple ONLINE waves.
        """
        fo = join(TntCLMWaveInvitee,
                TntCLMWave, TntCLMWave.id==TntCLMWaveInvitee.wave_id
            ).join(
                Contact, Contact.id==TntCLMWaveInvitee.contact_id
            ).join(
                TntCLMCompany, Contact.company_id==TntCLMCompany.company_id
            ).join(
                TntCLMCustomer, TntCLMCompany.customer_id==TntCLMCustomer.id
            ).join(
                Country, TntCLMCustomer.country_id==Country.id
            )
        invites = self.sa_session.query(TntCLMWaveInvitee).select_from(fo).filter(
            and_(TntCLMCustomer.account_number==account_no, Country.id==country_id)).order_by(TntCLMWave.id.desc()).all()

        # no invites, then incorrect Account No + Country.
        if len(invites) == 0:
            return None

        # We check if there are any online waves.
        for invite in invites:
            if invite.wave.status==TntCLMWave.STATUS_ONLINE:
                return invite

        return invites[0]

    def is_respondent(self, contact, wave, minimum_responses=10):
        """
        Checks if the contact is a respondent in the wave. Respondent when number of responses >= minimum
        """
        responses = self.sa_session.query(TntCLMWaveResponse).join(TntCLMWaveParticipant).join(TntCLMWaveInvitee).filter(
            and_(TntCLMWaveInvitee.wave==wave, TntCLMWaveInvitee.contact==contact)).all()
        return len(responses) >= minimum_responses

    @transaction_decorator
    def make_all_static(self, wave):
        wc = WaveController(self.sa_session)
        locales = self.sa_session.query(TntCLMLocale).join(TntCLMWaveLocaleRevision).filter(
            TntCLMWaveLocaleRevision.wave==wave).group_by(TntCLMWaveLocaleRevision.locale).all()

        for locale in locales:
            state = wc.get_wave_state(wave, locale)
            self.make_static(wave, state)

        self.sa_session.flush()

    @transaction_decorator
    def make_static(self, wave, state):
        revision = self.sa_session.query(TntCLMWaveLocaleRevision).filter(
            and_(TntCLMWaveLocaleRevision.wave==wave, TntCLMWaveLocaleRevision.id==state['revision'])).first()
        revision.state = simplejson.dumps(state, ensure_ascii=False, cls=JsonRpcJSONEncoder)
        self.sa_session.add(revision)
        return revision

    def export_response_pdf(self, filename, participant_id, locale_id=None):
        """
        'exports' a pdf filename for the given wave_id, participant_id and optionally locale_id
        filename can either be an absolute file path but also a HttpResponse object for sending to the client
        returns a filename that can be returned to the client if needed
        """

        self.assert_permissions('view_wave_responses')

        assert isinstance(filename, (str, unicode, HttpResponse))
        assert isinstance(locale_id, (int, long, NoneType))

        s = self.sa_session.query(TntCLMWaveParticipantSession).filter(
            TntCLMWaveParticipantSession.participant_id==participant_id).order_by(
            desc(TntCLMWaveParticipantSession.created)).all()
        if len(s) == 0:
            raise SystemException('Invalid respondent (no session id for supplied participant id %s)' % participant_id)

        pc = ParticipantController(self.sa_session, None, s[0].id)
        state = pc.get_wave_state_for_invitee(locale_id, wave_status_check=False)

        self.assert_privileges_on_object(pc.participant.invitee.wave, TntCLMWave.PRIV_VIEW_RESPONSE)

        from iqportal.tnt.clm.pdf.build import TntWave38Builder
        r = TntWave38Builder(self.sa_session)
        r.build(filename, state)

        lc = TntCLMLocaleController(self.sa_session, account_id=self.get_account().id)
        if len(lc.get_locales(state['locale'], only_active=False)) == 0:
            raise SecurityException('Unauthorized access to CLM response for participant `%s` and locale `%s`' % (participant_id, state['locale']))

        # return filename based on wave-code, locale and participant
        locale_str = self.sa_session.query(TntCLMLocale).filter(TntCLMLocale.id==state['locale']).one().label
        filename = '%s_%s_%s_QI.pdf' % (state['wave']['code'], locale_str, participant_id)
        return filename

    def export_response_csv(self, wave_id, country_ids=None, no_texts=False):
        """
        Exports wave respodent data in csv format to a file in the writable dir and returns the hash to load this file again
        http://trac.office.basesoft.com/iqportal/ticket/1850

        @param wave_id
        @param country_ids
        @param no_texts: export without text responses
        """

        self.assert_permissions('view_wave_responses')

        logger.info('export_response_csv: starting')

        wave_type_module = self.sa_session.query(TntCLMWaveType).join(TntCLMWave).filter(TntCLMWave.id==wave_id).one().module

        scale_length = self.sa_session.query(TntCLMWaveAttribute) \
            .join(TntCLMAttribute) \
            .filter(TntCLMWaveAttribute.wave_id==wave_id) \
            .filter(TntCLMAttribute.type==TntCLMAttribute.TYPE_SCALE) \
            .count()

        choice_length = self.sa_session.query(TntCLMWaveAttribute) \
            .join(TntCLMAttribute) \
            .filter(TntCLMWaveAttribute.wave_id==wave_id) \
            .filter(TntCLMAttribute.type.in_([TntCLMAttribute.TYPE_SINGLE_CHOICE, TntCLMAttribute.TYPE_SINGLE_SCALE,
                                              TntCLMAttribute.TYPE_MULTIPLE_CHOICE, TntCLMAttribute.TYPE_NPS])) \
            .count()

        # query all text attributes used in the wave
        text_wave_attrs = self.sa_session.query(TntCLMWaveAttribute) \
            .join(TntCLMAttribute) \
            .filter(TntCLMWaveAttribute.wave_id==wave_id) \
            .filter(TntCLMAttribute.type==TntCLMAttribute.TYPE_TEXT) \
            .order_by(TntCLMWaveAttribute.attribute_id, TntCLMWaveAttribute.position)

        # text_titles are the headers that get written to the sheet
        text_titles = [];

        # we cant rely on positions so make a mapping that holds the wave_attr.position and wave_answer.id
        text_positions = {};

        for w_attr in text_wave_attrs:

            text_positions[w_attr.position] = {}

            attr = w_attr.attribute

            # questions without answers just use 0 for answer mapping
            if len(attr.answers) == 0:
                text_titles.append(attr.default_label)

                text_positions[ w_attr.position ][0] = len(text_titles)

            # questions with answers
            else:

                for answer in attr.answers:

                    text_titles.append(answer.default_label)

                    answer_id = self.sa_session.query(TntCLMWaveAnswer).filter(TntCLMWaveAnswer.wave_attribute == w_attr).filter(TntCLMWaveAnswer.answer == answer).one().id

                    text_positions[ w_attr.position][ answer_id ] = len(text_titles)

        text_length = len(text_titles)

        #print 'text_titles: %s' % text_titles
        #print 'text_positions: %s' % text_positions

        import xlwt

        country_where = ''
        if country_ids is not None:

            if type(country_ids) == int:
                country_ids = [country_ids]

            for i in country_ids:
                assert type(i) == int
            country_where = """
                AND country.id IN (%s)
            """ % ','.join([str(i) for i in country_ids])

        db_engine = create_engine(settings.DB_CONNSTR_READONLY, echo=False)

        logger.debug('export_response_csv: starting SELECT query')
        rows = db_engine.execute("""
            SELECT
                country.iso_3166_a2   AS COUNTRY,
                depot.default_code  AS DEPOT,
                p.id as PARTICIPANT_ID,
                comp2.name as COMPANY_NAME,
                SUBSTR(LPAD(customer.account_number, 10, '0'), -10) AS ACCOUNT_NR,
                IFNULL(c.email, person.email) AS EMAIL,
                comp.revenue_band AS REVENUE,
                IF(p.opt_out=0, 'N', 'Y') AS ANONYMOUS,
                CASE WHEN p.is_respondent = 1 THEN 'Internet' WHEN p.is_respondent = 2 THEN 'Paper' ELSE '' END AS CHANNEL,
                l.location AS LOCATION,
                s.label as SERVICE,
                wave_attr.position AS WAVE_ATTR_POSITION,
                attr.type AS ATTR_TYPE,
                wave_attr.id AS WAVE_ATTR_ID,
                wave_answer.id as wave_answer_id,
                CASE
                    WHEN attr.type = 1 THEN wave_scale_answer.position
                    WHEN attr.type = 2 THEN wave_answer.position
                    WHEN attr.type = 3 THEN r.response_text
                    WHEN attr.type = 4 THEN wave_answer.position
                    WHEN attr.type = 6 THEN wave_answer.position
                    WHEN attr.type = 7 THEN wave_answer.position - 1
                END as ANSWER

            FROM `tblTntCLMWaveResponse` `r`
                JOIN tblTntCLMWaveAttribute wave_attr ON r.wave_attribute_id = wave_attr.id
                JOIN tblTntCLMAttribute attr ON wave_attr.attribute_id = attr.id

                LEFT JOIN tblTntCLMWaveAnswer wave_answer ON r.wave_answer_id = wave_answer.id
                LEFT JOIN tblTntCLMWaveScaleAnswer wave_scale_answer ON r.wave_scale_answer_id = wave_scale_answer.id

                JOIN `tblTntCLMWaveParticipant` `p` ON `r`.`participant_id` = `p`.`id`
                JOIN `tblTntCLMWaveInvitee` `i` ON `p`.`invitee_id` = `i`.`id`
                JOIN `tblContact` `c` ON `i`.`contact_id` = `c`.`id`
                JOIN tblPerson person ON c.person_id = person.id
                JOIN `tblTntCLMCompany` `comp` ON `c`.`company_id` = `comp`.`company_id`
                JOIN `tblCompany` `comp2` ON `comp`.`company_id` = `comp2`.`id`
                JOIN `tblTntCLMCustomer` `customer` ON `comp`.`customer_id` = `customer`.`id`
                JOIN tblTntDepot depot ON comp.depot_id = depot.id
                JOIN `tblTntCLMLocale` `l` ON `comp`.`locale_id` = `l`.`id`
                LEFT JOIN tblTntCLMService s on l.service_id = s.id
                JOIN tblCountry country ON l.country_id = country.id
            WHERE wave_attr.wave_id = %(wave_id)s
                """+country_where+"""

            ORDER BY p.id, wave_attr.position, wave_answer.position

        """, {'wave_id': wave_id})
        logger.debug('export_response_csv: done SELECT query')


        workbook = xlwt.Workbook(encoding='utf-8')
        sheet = workbook.add_sheet('respondents')

        header = [
            'COUNTRY',
            'DEPOT',
            'PARTICIPANT_ID',
            'COMPANY_NAME',
            'ACCOUNTNR',
            'EMAIL',
            'REVENUE',
            'ANONYMOUS',
            'CHANNEL',
            'LOCATION',
            'SERVICE',
        ]

        header += ['VR%s' % str(i+1) for i in range(scale_length)]
        header += ['VR%s' % string.ascii_uppercase[i] for i in range(choice_length)]

        if not no_texts:
            header += text_titles

        for i, col in enumerate(header):
            sheet.write(0, i, col)

        data = []

        def write_xls(data):

            if not hasattr(write_xls, 'row_count'):
                write_xls.row_count = 1

            if not len(data):
                return

            if no_texts:
                clength = scale_length + choice_length
            else:
                clength = scale_length + choice_length + text_length

            csv_row = [
                data[0]['country'],
                data[0]['depot'],
                data[0]['participant_id'],
                data[0]['company_name'],
                data[0]['account_nr'],
                data[0]['email'],
                data[0]['revenue'],
                data[0]['anonymous'],
                data[0]['channel'],
                data[0]['location'],
                data[0]['service'],
            ]

            offset = len(csv_row)-1

            csv_row = csv_row + ['99'] * clength



            multi_at = {}
            for data_row in data:

                if data_row['attr_type'] in (TntCLMAttribute.TYPE_SCALE, TntCLMAttribute.TYPE_SINGLE_CHOICE,
                                             TntCLMAttribute.TYPE_MULTIPLE_CHOICE, TntCLMAttribute.TYPE_NPS,
                                             TntCLMAttribute.TYPE_SINGLE_SCALE, TntCLMAttribute.TYPE_TEXT):

                    answer = data_row['answer']

                    if wave_type_module == 'clm_wave_36' and data_row['attr_type'] == TntCLMAttribute.TYPE_SCALE and answer == 6:
                        answer = '88'

                    # when answer is not text
                    if data_row['attr_type'] != TntCLMAttribute.TYPE_TEXT:
                        if data_row['attr_type'] != TntCLMAttribute.TYPE_MULTIPLE_CHOICE:
                            csv_row[offset + int(data_row['wave_attr_position'])] = answer
                        else:
                            if not multi_at.has_key(data_row['wave_attr_id']):
                                multi_at[data_row['wave_attr_id']] = {'position': int(data_row['wave_attr_position']),
                                                                      'answers': []}
                            multi_at[data_row['wave_attr_id']]['answers'].append(str(answer))

                    # when answer is text...
                    else:

                        if not no_texts:

                            wave_attr_position = int(data_row['wave_attr_position'])

                            # wave_answer is null for questions without answers, then use 0 like in the mapping
                            try:
                                wave_answer_id = int(data_row['wave_answer_id'])
                            except TypeError:
                                wave_answer_id = 0

                            # get the position from the map and set the answer in that column
                            text_position = text_positions[ wave_attr_position] [ wave_answer_id ]
                            csv_row[offset + scale_length + choice_length + text_position] = answer

            for at in multi_at:
                csv_row[offset + multi_at[at]['position']] = ','.join(multi_at[at]['answers'])

            # if the participant only has text responses and no_texts is true, then don't export it
            if no_texts:
                has_resp = False
                for i in range(10, len(csv_row)):
                    if csv_row[i] != '99':
                        has_resp = True
                        break
                if not has_resp:
                    return

            for i, col in enumerate(csv_row):
                sheet.write(write_xls.row_count, i, col)

            write_xls.row_count = write_xls.row_count + 1

        for row in rows:

            fields = ['country', 'depot', 'participant_id', 'company_name', 'account_nr', 'email', 'revenue', 'anonymous', 'channel',
                      'location', 'service', 'wave_attr_position', 'attr_type', 'wave_attr_id', 'wave_answer_id', 'answer']
            row_obj = dict(zip(fields, row))

            if len(data) and data[0]['participant_id'] != row_obj['participant_id']:
                write_xls(data,)
                data = []

            data.append(row_obj)

        write_xls(data)

        logger.debug('export_response_csv: writing excel file')
        hash = sha(str(random.random())).hexdigest()[:8]
        workbook.save(os.path.join(settings.TMP_FILE_REPOSITORY, 'tnt_wave_%s.xls' % hash))

        logger.info('export_response_csv: done.')

        return hash

    def download_response_csv(self, hash):

        file = open(os.path.join(settings.TMP_FILE_REPOSITORY, 'tnt_wave_%s.xls' % hash), 'r')
        return file

class ParticipantController(AbstractController):
    """
    Tnt CLM Participant Response controller
    """
    def __init__(self, sa_session, invitee_hash=None, participant_session_id=None, participant_id=None):
        self.sa_session = sa_session

        if invitee_hash is None and participant_session_id is None and participant_id is None:
            raise SystemException('invitee_hash and participant session id and participant id not provided')

        self.participant = None
        self.participant_session = None

        # if provided a participant_session_id we can use that to retrieve the invitee_id
        if participant_session_id:

            # if the invitee_hash is given make sure that it belongs to the same session cuz else we need to make a new one
            if invitee_hash is None or invitee_hash == self._get_invitee_from_session_id(participant_session_id).invitee_hash:
                self.participant_session = self._init_participant_session(participant_session_id)
                self.participant = self.participant_session.participant

        # try to retrieve the invitee_id from the hash if given
        if self.participant is None:
            # get invitee
            if invitee_hash is not None:
                try:
                    invitee = self.sa_session.query(TntCLMWaveInvitee).filter(TntCLMWaveInvitee.invitee_hash==invitee_hash).one()
                except NoResultFound:
                    raise PublicUserException(_('Incorrect invitee hash.'))
                self._init_participant(invitee)
                self.participant_session = self._init_participant_session(None)

        if self.participant is None:
            try:
                self.participant = self.sa_session.query(TntCLMWaveParticipant).filter(
                                   TntCLMWaveParticipant.id==participant_id).one()
                self.participant_session = self._init_participant_session(None)
            except NoResultFound:
                raise SystemException('Invalid participant_id')


    def _get_invitee_from_session_id(self, session_id):
        """
        Returns the invitee_id for the given session_id
        """
        try:
            invitee = self.sa_session.query(TntCLMWaveInvitee).join(
                TntCLMWaveParticipant).join(
                TntCLMWaveParticipantSession).filter(
                TntCLMWaveParticipantSession.id == session_id).one()
        except NoResultFound:
            raise SystemException('Unable to retrieve invitee from session_id [%s]' % session_id)

        return invitee

    def _init_participant_session(self, session_id):
        """
        Initialises the participant session object by either continuing or creating a new session
        """

        # start a new session if no session_id is supplied or else continue
        if session_id is None:
            logger.debug('ParticipantController: creating new session for participant [%s]' % self.participant.id)
            session = TntCLMWaveParticipantSession()
            session.participant = self.participant
        else:
            logger.debug('ParticipantController: continuing session [%s]' % session_id)
            session = self.sa_session.query(TntCLMWaveParticipantSession).get(session_id)
            if not session:
                raise SystemException('Unable to continue session_id [%s]: No session found.' % session_id)

            # we should have a session now.. explicit update its time..
            session.last_update = datetime.now()

        self.sa_session.add(session)
        self.sa_session.flush()

        logger.info('ParticipantController: initialized participant session [%s]' % session.id)

        return session

    @transaction_decorator
    def _init_participant(self, invitee):
        # get participant for invitee (if there is one)
        try:
            self.participant = self.sa_session.query(TntCLMWaveParticipant).filter(
                               TntCLMWaveParticipant.invitee==invitee).one()
        except NoResultFound:
            self.participant = TntCLMWaveParticipant()
            self.participant.invitee = invitee
            self.participant.locale = invitee.contact.company.locale
            self.sa_session.add(self.participant)
            self.sa_session.flush()

    def get_wave_state_for_invitee(self, locale_id=None, wave_status_check=True):
        if self.participant is None:
            raise SystemException('No participant')

        wave = self.participant.invitee.wave

        if wave_status_check:
            if wave.status != TntCLMWave.STATUS_ONLINE:
                raise PublicUserException(_('This wave is not available for responding.'))

        lid = locale_id
        if locale_id is None:
            locale_id = self.participant.locale
            lid = locale_id.id

        state = {}
        q = self.sa_session.query(TntCLMWaveLocaleRevision).filter(
            and_(TntCLMWaveLocaleRevision.locale_id==lid, TntCLMWaveLocaleRevision.wave==wave,
                 TntCLMWaveLocaleRevision.status==TntCLMWaveLocaleRevision.STATUS_ACTIVE_APPROVED
                 )).order_by(TntCLMWaveLocaleRevision.id.desc())

        wc = WaveController(self.sa_session)

        revision = q.first()
        if revision is None:
            raise SystemException('Cannot find active approved revision for locale %s, wave %s' % (lid, wave.id))
        if revision.state is None:
            state = wc.get_wave_state(wave, locale_id, revision.id)
            rc = ResponseController(self.sa_session)
            revision = rc.make_static(wave, state)
        else:
            state = simplejson.loads(revision.state, object_hook=json_object_hook)

        state['values'] = self.get_responses()
        state['participant_session_id'] = self.participant_session.id

        # get all available locales for this participant
        locales = [{'id': self.participant.locale.id, 'label': self.participant.locale.label}]

        wave_default_locale = wc.get_wave_default_locale(wave.id)
        if self.participant.locale.id != wave_default_locale.id:
            locales.append({'id': wave_default_locale.id, 'label': wave_default_locale.label})
        state['available_locales'] = locales

        company_logo = None
        lcc = LicenseConfigController(self.sa_session, scope=wave.groups[0].group.license)
        img = lcc.get('basesoft.license.company.logo')
        if img is not None:
            company_logo = self.sa_session.query(ImageFile).get(img).file_hash

        state['wave']['company_logo'] = company_logo

        return state

    def get_responses(self):
        """
        Returns state with participant's responses
        """
        if self.participant is None:
            raise SystemException('No participant')
        wave_responses = self.participant.responses

        responses = {}
        for wr in wave_responses:
            type = wr.wave_attribute.attribute.type
            if not responses.has_key(wr.wave_attribute_id):
                responses[wr.wave_attribute_id] = {'type': type}

            if type == TntCLMAttribute.TYPE_SCALE:
                responses[wr.wave_attribute_id]['value'] = wr.wave_scale_answer.position

            if type in [TntCLMAttribute.TYPE_SINGLE_CHOICE, TntCLMAttribute.TYPE_SINGLE_SCALE, TntCLMAttribute.TYPE_NPS]:
                responses[wr.wave_attribute_id]['value'] = wr.wave_answer_id

            if type == TntCLMAttribute.TYPE_MULTIPLE_CHOICE:
                if not responses[wr.wave_attribute_id].has_key('value'):
                    responses[wr.wave_attribute_id]['value'] = []
                responses[wr.wave_attribute_id]['value'].append(wr.wave_answer_id)

            if type == TntCLMAttribute.TYPE_TEXT:
                value = wr.response_text

                if wr.wave_answer_id is None:
                    responses[wr.wave_attribute_id]['value'] = value
                else:
                    if not responses[wr.wave_attribute_id].has_key('value'):
                        responses[wr.wave_attribute_id]['value'] = {}
                    responses[wr.wave_attribute_id]['value'][wr.wave_answer_id] = value

        return {'responses': responses,
                'opt_out': self.participant.opt_out and True or False
               }

    def _check_response(self, wave_attribute, wave_answer, value, value_type=0):
        """
        Checks if a response was already saved for args. It updates in case found, else insert.
        @param value_type: 0 - scale answer, 1 - unicode, 2 - answer
        """
        filter = and_(TntCLMWaveResponse.participant==self.participant,
                      TntCLMWaveResponse.wave_attribute==wave_attribute)
        if wave_answer is not None:
            filter = and_(filter, TntCLMWaveResponse.wave_answer==wave_answer)

        wave_response = self.sa_session.query(TntCLMWaveResponse).filter(filter).all()
        if len(wave_response) > 0:
            wave_response = wave_response[0]
            # value was removed, so delete the response
            if value is None or unicode(value).strip() == '':
                self.sa_session.delete(wave_response)
                return None
            # don't save anything if values is the same
            if value_type == 0 and int(value) == wave_response.wave_scale_answer_id:
                return None
            if value_type == 1 and unicode(value) == wave_response.response_text:
                return None
            if value_type == 2 and int(value) == wave_response.wave_answer_id:
                return None
            wave_response.participant_session = self.participant_session
        else:
            # new response
            if value is None or unicode(value).strip() == '':
                return None
            wave_response = TntCLMWaveResponse()
            wave_response.participant = self.participant
            wave_response.wave_attribute = wave_attribute
            if wave_answer is not None:
                wave_response.wave_answer = wave_answer
            wave_response.participant_session = self.participant_session

        # participant might have updated responses previously imported from file as paper responses
        wave_response.is_respondent = TntCLMWaveParticipant.RESPONSES_INTERNET

        return wave_response

    @transaction_decorator
    def save_responses(self, responses_state):
        """
        Saves wave responses for provided participant.
        @param participant_id: int
        @param responses_state: dictionary with responses(list) and opt-out(bool)
            Eg.: {'responses': { <wave_attribute_id>: {'type': 1, value: 1},
                                 <wave_attribute_id>: {'type': 3, value: {<wave_answer_id>: 'text', ...}},
                                 ...
                                }
                  'opt-out': False
                }
        """
        if self.participant is None:
            raise SystemException('No participant')

        responses = responses_state['responses']
        w = self.participant.invitee.wave

        for attr_id, values in responses.items():

            wave_attribute = self.sa_session.query(TntCLMWaveAttribute).filter(TntCLMWaveAttribute.id==attr_id).all()
            if len(wave_attribute) == 0:
                raise SecurityException('Invalid wave attribute id %s' % attr_id)
            wave_attribute = wave_attribute[0]

            wave_response = None

            # scale attributes can have value between 1 and 6
            if values['type'] == TntCLMAttribute.TYPE_SCALE:
                # cannot remove an answer for a scale attribute (cannot uncheck radio)
                if values['value'] is None or unicode(values['value']).strip() == '':
                    continue
                v = int(values['value'])
                if v not in range(1, 7):
                    raise SecurityException('Invalid value (%s) for scale attribute id %s' % (v, attr_id))
                # check the correspondent scale answer

                scale_answer = self.sa_session.query(TntCLMWaveScaleAnswer).filter(
                         and_(TntCLMWaveScaleAnswer.wave==w, TntCLMWaveScaleAnswer.position==v)).all()
                if len(scale_answer) == 0:
                    raise SecurityException('Invalid scale answer position (%s) for wave %s' % (v, w.id))
                scale_answer = scale_answer[0]

                wave_response = self._check_response(wave_attribute, None, scale_answer.id)
                if wave_response is None:
                    continue
                wave_response.wave_scale_answer = scale_answer

            # single choice attributes will have wave answer id as value
            if values['type'] in [TntCLMAttribute.TYPE_SINGLE_CHOICE, TntCLMAttribute.TYPE_SINGLE_SCALE,
                                  TntCLMAttribute.TYPE_NPS]:
                wave_response = self._check_response(wave_attribute, None, values['value'], 2)
                if wave_response is None:
                    continue
                v = int(values['value'])
                # check if answer is correct
                answer = self.sa_session.query(TntCLMWaveAnswer).filter(
                         and_(TntCLMWaveAnswer.wave_attribute==wave_attribute, TntCLMWaveAnswer.id==v)).all()
                if len(answer) == 0:
                    raise SecurityException('Invalid answer id (%s) for attribute %s' % (v, attr_id))
                answer = answer[0]
                wave_response.wave_answer = answer

                # for SINGLE_SCALE we need to save the scale_answer_id also - get it by answer position
                if values['type'] == TntCLMAttribute.TYPE_SINGLE_SCALE:
                    scale_answer = self.sa_session.query(TntCLMWaveScaleAnswer).filter(
                             and_(TntCLMWaveScaleAnswer.wave==w, TntCLMWaveScaleAnswer.position==answer.position)).all()
                    if len(scale_answer) == 0:
                        raise SecurityException('Cannot find scale answer for answer position (%s) for wave %s' % (answer.position, w.id))
                    # no need for checking, it was done in the previous step
                    wave_response.wave_scale_answer = scale_answer[0]

            # text attributes can have as value either a string or a dictionary with values for each answer
            if values['type'] == TntCLMAttribute.TYPE_TEXT:
                if not isinstance(values['value'], dict):
                    wave_response = self._check_response(wave_attribute, None, values['value'], 1)
                    if wave_response is None:
                        continue
                    wave_response.response_text = unicode(values['value'])
                else:
                    for ans_id, val in values['value'].items():
                        answer = self.sa_session.query(TntCLMWaveAnswer).filter(
                                 and_(TntCLMWaveAnswer.wave_attribute==wave_attribute,
                                      TntCLMWaveAnswer.id==ans_id)).all()
                        if len(answer) == 0:
                            raise SecurityException('Invalid answer id (%s) for TEXT attribute %s' % (ans_id, attr_id))
                        answer = answer[0]
                        wave_response = self._check_response(wave_attribute, answer, val, 1)
                        if wave_response is None:
                            continue
                        wave_response.response_text = unicode(val)
                        self.sa_session.add(wave_response)
                    continue

            # multiple choice attributes will have a list with answers ids as value
            if values['type'] == TntCLMAttribute.TYPE_MULTIPLE_CHOICE:
                if not isinstance(values['value'], list):
                    raise SecurityException('Invalid value received for attribute %s' % attr_id)

                # gotta check all attribute answers (some might have been unchecked,
                # so we have to delete them from response)
                answers = self.sa_session.query(TntCLMWaveAnswer).filter(
                         and_(TntCLMWaveAnswer.wave_attribute==wave_attribute)).all()
                # check if all answers are correct
                aids = [aa.id for aa in answers]
                vals = [int(aa) for aa in values['value']]
                for ans_id in vals:
                    if ans_id not in aids:
                        raise SecurityException('Invalid answer id (%s) for MULTI attribute %s' % (ans_id, attr_id))

                for answer in answers:
                    val = None
                    if answer.id in vals:
                        val = answer.id
                    wave_response = self._check_response(wave_attribute, answer, val, 2)
                    if wave_response is None:
                        continue
                    wave_response.wave_answer = answer
                    self.sa_session.add(wave_response)
                continue

            if wave_response is None:
                raise SecurityException('Invalid type (%s) for attribute id (%s)' % (values['type'], attr_id))
            self.sa_session.add(wave_response)

        # this method is always used when saving responses from internet
        self.sa_session.flush()
        q = self.sa_session.query(TntCLMWaveResponse).filter(TntCLMWaveResponse.participant==self.participant).all()
        is_respondent = TntCLMWaveParticipant.NO_RESPONSES
        if len(q) > 0:
            is_respondent = TntCLMWaveParticipant.RESPONSES_INTERNET

        self.participant.opt_out = responses_state['opt_out']
        self.participant.is_respondent = is_respondent
        self.sa_session.add(self.participant)


class ManagerParticipantController(AbstractController):
    """
    Contains methods used by the manager to get participant's responses. Needs permissions and privileges checks
    """
    USED_HEADERS = None

    MAX_RECORDS_FOR_AUTOMATICAL_IMPORT = 150
    IMPORT_RESPONSES_HEADERS = ['COUNTRY', 'DEPOT', 'ACCOUNTNR', 'EMAIL', 'REVENUE', 'ANONYMOUS', 'CHANNEL', 'LOCATION',
                        'SICGROUP', 'SIC', 'IMP',
                        'VR1', 'VR2', 'VR3', 'VR4', 'VR5', 'VR6', 'VR7', 'VR8', 'VR9', 'VR10', 'VR11', 'VR12', 'VR13',
                        'VR14', 'VR15', 'VR16', 'VR17', 'VR18', 'VR19', 'VR20', 'VR21', 'VR22', 'VR23', 'VR24', 'VR25',
                        'VR26', 'VR27', 'VR28', 'VR29', 'VR30', 'VRA', 'VRB', 'VRC', 'VRD', 'VRE', 'VRF', 'VRG']

    def get_participant_wave_state(self, participant_id, locale_id):
        self.assert_permissions('view_wave_responses')
        # get latest participant session (if we just pass the participant_id, a new participant session will be created
        # every time manager checks a participant responses and we don't want that)
        s = self.sa_session.query(TntCLMWaveParticipantSession).filter(
            TntCLMWaveParticipantSession.participant_id==participant_id).order_by(
            desc(TntCLMWaveParticipantSession.created)).all()
        if len(s) == 0:
            raise SystemException('Invalid respondent (no session id for supplied participant id %s)' % participant_id)
        participant_controller = ParticipantController(self.sa_session, None, s[0].id)
        wave_id = participant_controller.participant.invitee.wave.id
        wave = self.get_privileged_object_by_id(TntCLMWave, wave_id, TntCLMWave.PRIV_VIEW_RESPONSE)
        if len(wave) == 0:
            raise SecurityException('Insufficient permissions to view wave responses')
        lc = TntCLMLocaleController(self.sa_session, account_id=self.get_account().id)
        locale = lc.get_locales(locale_id)
        if len(locale) == 0:
            raise SecurityException('No permission for locale %s ' % locale_id)
        return participant_controller.get_wave_state_for_invitee(locale_id, wave_status_check=False)

    @transaction_decorator
    def reset_P1_tiff_status(self, participant_ids):
        """
        Resets participant priority triggered status in order for tiff files to be resent
        """
        self.assert_permissions('view_wave_responses')
        participants = self.sa_session.query(TntCLMWavePriorityCSVFileParticipant).filter(and_(
                        TntCLMWavePriorityCSVFileParticipant.participant_id.in_(participant_ids),
                        TntCLMWavePriorityCSVFileParticipant.has_tiff_triggered==TntCLMWavePriorityCSVFileParticipant.TIFF_SENT)).all()
        for p in participants:
            p.has_tiff_triggered = TntCLMWavePriorityCSVFileParticipant.TIFF_NOT_SENT
            self.sa_session.add(p)

    def _save_response_file(self, wave, file, separator, add_nonexistent_invitees=False, add_invitee_if_duplicate=False,
                            props=None):
        """
        Set import responses in the background
        """
        q = self.sa_session.query(ImportFileType).filter(ImportFileType.name=='iqportal.tnt.response').all()
        if len(q) == 0:
            raise SystemException('Cannot find file import type "iqportal.tnt.response".')
        type = q[0]

        fc = ProcessFileController(sa_session=self.sa_session, account_id=self.get_account().id)

        import_settings = {
            'separator': separator,
            'wave_id': wave.id,
            'wave_name': wave.name,
            'account_id': self.get_account().id,
            'add_nonexistent_invitees': add_nonexistent_invitees,
            'add_invitee_if_duplicate': add_invitee_if_duplicate
        }
        settings_str = simplejson.dumps(import_settings, ensure_ascii=False, cls=JsonRpcJSONEncoder)

        fc.file_type_properties = {'settings': settings_str, 'type': type}
        if props is not None:
            for k,v in props.items():
                fc.file_type_properties[k] = v

        category = fc.get_category_by_label("TNT Wave responses Import files")
        if category is None:
            category = fc.save_or_update_file_category(None, "TNT Wave responses Import files")

        fc.save_files_from_temp(file, category, ImportFile)

    def _get_csv_data(self, file, separator):
        """
        Builds data list from csv file
        """
        self.USED_HEADERS = []
        rows = csv.reader(open(file, 'rU'), delimiter=str(separator))
        # build the data
        not_correct = []
        data = []
        i = 0
        for r in rows:
            # first row is with headers
            if i == 0:
                headers = r
                if len(r) == 1: #most likely invalid separator
                    return None
                for h in headers:
                    if h not in self.IMPORT_RESPONSES_HEADERS:
                        not_correct.append(h)
                    self.USED_HEADERS.append(h)
            else:
                d = {}
                for j, h in enumerate(headers):
                    d[h] = r[j]
                data.append(d)
            i = i + 1
        if len(not_correct) > 0:
            raise PublicUserException(_('Invalid respondents excel file headers:') + ' ' + str(not_correct))

        # there should be at least one data row
        if len(data) == 0:
            raise PublicUserException(_('Empty respondents csv file.'))

        return data

    def _get_xls_data(self, file):
        """
        Builds data list from xls file
        """
        import xlrd

        sheet = xlrd.open_workbook(file).sheet_by_index(0)
        rows = sheet.nrows

        # there should be at least 2 rows (headers and data)
        if rows <= 1:
            raise PublicUserException(_('Empty respondents excel file.'))

        self.USED_HEADERS = []

        headers = sheet.row_values(0)
        not_correct = []
        for h in headers:
            if h not in self.IMPORT_RESPONSES_HEADERS:
                not_correct.append(h)
            self.USED_HEADERS.append(h)
        if len(not_correct) > 0:
            raise PublicUserException(_('Invalid respondents excel file headers:') + ' ' + str(not_correct))

        # build the data
        data = []
        for i in range(1, rows):
            d = {}
            rv = sheet.row_values(i)
            for j, h in enumerate(headers):
                d[h] = rv[j]
            data.append(d)

        return data

    def _get_depot_map(self):
        if self._depot_map is not None:
            return self._depot_map

        self._depot_map = {}
        depots = self.sa_session.query(TntDepot).all()
        for d in depots:
            self._depot_map[d.default_code] = d
            for a in d.aliases:
                self._depot_map[a.code] = d
        return self._depot_map

    def _get_locale_map(self):
        if self._locale_map is not None:
            return self._locale_map, self._country_map

        self._locale_map = {}
        self._country_map = {}
        locales = self.sa_session.query(TntCLMLocale).all()
        for l in locales:
            # get the first locale found for this combination
            k = '%s_%s_%s' % (l.country.iso_3166_a2, l.service.label, l.location)
            if not self._locale_map.has_key(k):
                self._locale_map[k] = l
            self._country_map[l.country.iso_3166_a2] = l.country
        return self._locale_map, self._country_map

    def _get_contact_map(self):
        if self._contact_map is not None:
            return self._contact_map

        self._contact_map = {}
        query = """
            SELECT cc.id, customer.account_number, country.iso_3166_a2 AS code
            FROM tblContact as cc
            INNER JOIN tblTntCLMCompany AS c
                ON cc.company_id = c.company_id
            INNER JOIN tblTntCLMCustomer AS customer
                ON c.customer_id = customer.id
            INNER JOIN tblTntCLMLocale AS l
                ON c.locale_id = l.id
            INNER JOIN tblCountry AS country
                ON l.country_id = country.id
        """
        db_engine = create_engine(settings.DB_CONNSTR, echo=False)

        for row in db_engine.execute(query).fetchall():
            # get the first contact found for this combination
            k = 'contact_%s_%s' % ('%010d' % row['account_number'], row['code'])
            if not self._contact_map.has_key(k):
                self._contact_map[k] = row['id']

        return self._contact_map

    def _create_invitee(self, wave, data):
        logger.info('No invitee exists, checking...')
        locale_map, country_map = self._get_locale_map()

        depot_map = self._get_depot_map()
        depot = depot_map[data['DEPOT']]

        # if not, create a new contact -> we are not supplied a language code so we pick the first express one
        # or the one with the service from the depot.
        if depot.service is None:
            service = 'Express'
        else:
            service = depot.service.label

        if not locale_map.has_key('%s_%s_%s' % (data['COUNTRY'], service, data['LOCATION'])):
            raise SystemException('Cannot find locale for country [%s], service [%s], location [%s]'\
                                  % (data['COUNTRY'], service, data['LOCATION']))
        locale = locale_map['%s_%s_%s' % (data['COUNTRY'], service, data['LOCATION'])]

        # check for customer
        country = country_map[data['COUNTRY']]
        companies = []
        try:
            customer = self.sa_session.query(TntCLMCustomer).filter(and_(TntCLMCustomer.account_number==data['ACCOUNTNR'], TntCLMCustomer.country==country)).one()
            companies = self.sa_session.query(TntCLMCompany).filter(TntCLMCompany.customer==customer).all()
        except NoResultFound:
            logger.info(' - Customer does not exist, creating new one')
            customer = TntCLMCustomer()
            customer.country = country
            customer.account_number = data['ACCOUNTNR']
            self.sa_session.add(customer)

        company = None
        for c in companies:
            if c.locale != locale or c.depot != depot or c.revenue_band != data['REVENUE']:
                continue
            company = c
            break

        if company is None:
            logger.info(' - Company does not exist, creating new one')

            company = TntCLMCompany()
            company.customer = customer
            company.locale = locale
            company.depot = depot
            company.revenue_band = data['REVENUE']

            # this also means we create a new contact.
            contact = Contact()
            contact.person = Person() # we dont have person info
            contact.company = company
            self.save_default_privilege_object_group(contact)
            self.sa_session.add(contact)
        else:
            contacts = self.sa_session.query(Contact).filter(Contact.company==company).all()
            if len(contacts) > 0:
                contact = contacts.pop()
            else:
                logger.info(' - Contact does not exist, creating new one')
                contact = Contact()
                contact.person = Person() # we dont have person info
                contact.company = company
                self.save_default_privilege_object_group(contact)
                self.sa_session.add(contact)

        # add a new invitee to the wave for this contact.
        logger.info(' - Creating invitee')
        invitee = TntCLMWaveInvitee()
        invitee.contact = contact
        invitee.wave = wave
        invitee.invitee_hash = sha(str(random.random())).hexdigest()[:8]

        while True:
            try:
                self.sa_session.add(invitee)
                self.sa_session.flush()
                break
            except IntegrityError, e: # hash must be unique
                if e.message.find("Duplicate entry '"+invitee.invitee_hash+"' for key") < 0:
                    raise
                invitee.invitee_hash = sha(str(random.random())).hexdigest()[:8]

        return invitee

    def import_responses(self, wave_id, file_hash, file_location, separator=None, background=False, file_type=None, loop_validator=None,
                         add_nonexistent_invitees=False, add_invitee_if_duplicate=False):

        self._depot_map = None
        self._locale_map = None
        self._country_map = None

        wc = WaveController(self.sa_session, account_id=self.get_account().id)
        wave = wc.get_waves(wave_id)[0]
        if wave.status not in [TntCLMWave.STATUS_ONLINE, TntCLMWave.STATUS_FINISHED]:
            raise PublicUserException(_('Cannot import responses for wave which is not ONLINE or FINISHED'))

        file = '%s%s' % (file_location, file_hash)

        # check file extension
        if file_type is None:
            if not background:
                from basesoft.web.contrib.file.models import FileTmp
                f = self.sa_session.query(FileTmp).filter(FileTmp.file_hash==file_hash).all()
            else:
                from basesoft.web.contrib.file.models import File
                f = self.sa_session.query(File).filter(File.file_hash==file_hash).all()
            fn = f[0].file_name
            ext = fn[fn.rfind('.')+1:len(fn)]
            if ext not in ['xls', 'csv']:
                raise PublicUserException(_('Invalid file type'))
        else:# this is used for tests
            ext = file_type
        data = []
        if ext == 'xls':
            data = self._get_xls_data(file)
        if ext == 'csv':
            data = self._get_csv_data(file, separator)
            if data is None: #invalid separator
                return {'invalid_separator': True}
        if not background and len(data) > self.MAX_RECORDS_FOR_AUTOMATICAL_IMPORT:
            self._save_response_file(wave, f, separator, add_nonexistent_invitees=add_nonexistent_invitees,
                                     add_invitee_if_duplicate=add_invitee_if_duplicate)
            return {'background': True}

        scale_attributes = [a for a in wave.attributes if a.attribute.type==TntCLMAttribute.TYPE_SCALE]
        choice_attributes = [a for a in wave.attributes if a.attribute.type in [TntCLMAttribute.TYPE_SINGLE_CHOICE,
                             TntCLMAttribute.TYPE_SINGLE_SCALE, TntCLMAttribute.TYPE_MULTIPLE_CHOICE,
                             TntCLMAttribute.TYPE_NPS]]
        scale_answers = wave.scale_answers

        def flush(pids, opt_ids, internet_ids):
            """
            flushes 1000 records per time
            """

            sql = """
            INSERT INTO tblTntCLMWaveResponse (wave_attribute_id, wave_answer_id, wave_scale_answer_id,
                                               participant_id, participant_session_id, created, last_updated)
            VALUES
            """

            values = []
            for vals in flush.buffer:
                sql += "(%s, %s, %s, %s, %s, NOW(), NOW()),"

                for value in vals:
                    values.append(value)
            sql = sql[:-1]

            self.sa_session.connection().execute(sql, *values)
            # empty buffer
            flush.buffer = []

            if len(pids) > 0:
                sql = """
                update tblTntCLMWaveParticipant set is_respondent=%s, opt_out=%s, is_processed=%s, last_updated=NOW()
                where id in(
                """
                for id in pids:
                    sql = sql + '%s,'
                sql = sql[:-1] + ')'
                args = [TntCLMWaveParticipant.RESPONSES_PAPER, False, TntCLMWaveParticipant.DONT_PROCESS] +\
                        pids
                self.sa_session.connection().execute(sql, *args)

            if len(opt_ids) > 0:
                sql = """
                update tblTntCLMWaveParticipant set opt_out=%s where id in(
                """
                for id in opt_ids:
                    sql = sql + '%s,'
                sql = sql[:-1] + ')'
                args = [True] + opt_ids
                self.sa_session.connection().execute(sql, *args)

            if len(internet_ids) > 0:
                sql = """
                update tblTntCLMWaveParticipant set is_respondent=%s where id in(
                """
                for id in internet_ids:
                    sql = sql + '%s,'
                sql = sql[:-1] + ')'
                args = [TntCLMWaveParticipant.RESPONSES_INTERNET] + internet_ids
                self.sa_session.connection().execute(sql, *args)

        self._get_depot_map()
        self._get_locale_map()

        flush.buffer = []

        participants = []
        opt_out_participants = []
        internet_participants = []

        imported = 0
        invalidRows = []
        for d in data:
            a = str(d['ACCOUNTNR']).zfill(10)
            if not self._country_map.has_key(d['COUNTRY']):
                d['ERROR'] = _('Invalid country code %s' % d['COUNTRY'])
                invalidRows.append(d)
                continue
            if not self._depot_map.has_key(d['DEPOT']):
                d['ERROR'] = _('Invalid depot %s' % d['DEPOT'])
                invalidRows.append(d)
                continue
            fo = join(TntCLMWaveInvitee, Contact).join(TntCLMCompany,
                Contact.company_id==TntCLMCompany.company_id).join(TntCLMCustomer,
                TntCLMCompany.customer_id==TntCLMCustomer.id)

            q = self.sa_session.query(TntCLMWaveInvitee).select_from(fo).filter(and_(
                TntCLMWaveInvitee.wave==wave, TntCLMCustomer.account_number==a,
                TntCLMCustomer.country==self._country_map[d['COUNTRY']],
                TntCLMCompany.depot==self._depot_map[d['DEPOT']],
                TntCLMCompany.revenue_band==d['REVENUE'])).all()

            if len(q) == 0:
                if not add_nonexistent_invitees:
                    d['ERROR'] = _('Cannot find invitee for account number/country/revenue/depot, wave %s' % wave_id)
                    invalidRows.append(d)
                    continue
                try:
                    invitee = self._create_invitee(wave, d)
                except SystemException, e:
                    d['ERROR'] = e.message
                    invalidRows.append(d)
                    continue
            else:
                invitee = q[0]

            pc = ParticipantController(self.sa_session, invitee.invitee_hash)
            if pc.participant.is_respondent > 0 or pc.participant.id in participants:
                if not add_invitee_if_duplicate:
                    logger.info(' - Already has responses saved.')
                    d['ERROR'] = _('Already has responses saved')
                    invalidRows.append(d)
                    continue
                try:
                    invitee = self._create_invitee(wave, d)
                except SystemException, e:
                    d['ERROR'] = e.message
                    invalidRows.append(d)
                    continue
                pc = ParticipantController(self.sa_session, invitee.invitee_hash)

            try:
                responses = self._build_responses(scale_attributes, choice_attributes, scale_answers, d)
            except Exception, e:
                d['ERROR'] = e.message
                invalidRows.append(d)
                continue
            if not isinstance(responses, list):
                d['ERROR'] = responses
                invalidRows.append(d)
                continue

            if len(responses) == 0:
                d['ERROR'] = _('Participant has no responses in the file')
                invalidRows.append(d)
                continue

            participants.append(pc.participant.id)

            ch = TntCLMWaveParticipant.RESPONSES_PAPER
            if d.has_key('CHANNEL') and d['CHANNEL'].lower() in ['internet', 'i']:
                internet_participants.append(pc.participant.id)
                ch = TntCLMWaveParticipant.RESPONSES_INTERNET

            if d.has_key('ANONYMOUS') and d['ANONYMOUS'].lower() in ['yes', 'y']:
                opt_out_participants.append(pc.participant.id)

            for r in responses:
                flush.buffer.append(r+[pc.participant.id, pc.participant_session.id])

            imported = imported + 1

            if len(flush.buffer) >= 1000:
                flush(participants, opt_out_participants, internet_participants)
                participants = []
                opt_out_participants = []
                internet_participants = []

            # lets the task abort if stop signal was sent
            if loop_validator is not None:
                loop_validator()

        if len(flush.buffer) > 0:
            flush(participants, opt_out_participants, internet_participants)

        # save response and failures files for future reference
        failures_hash = None
        if not background and file_type is None: # this is not a background import and this method is not called from tests
            props = {'status': ImportFile.STATUS_DO_NOT_PROCESS, 'result': imported}
            if len(invalidRows) > 0:
                failures_file = self.export_failures_csv(invalidRows, fn)
                failures_hash = failures_file.file_hash
                props['failure_file'] = failures_file
            self._save_response_file(wave, f, separator, add_nonexistent_invitees=add_nonexistent_invitees,
                                     add_invitee_if_duplicate=add_invitee_if_duplicate, props=props)

        return {'imported': imported, 'failures': len(invalidRows), 'hash': failures_hash, 'failures_rows': invalidRows}

    def _build_responses(self, scale_attributes, choice_attributes, scale_answers, state):

        response_values = []

        scale_answers_by_position = {}
        for sa in scale_answers:
            scale_answers_by_position[sa.position] = sa

        # scale attributes
        wave_sa = scale_attributes
        for i, attr in enumerate(wave_sa):
            index = i+1
            v = int(state['VR%s' % index])
            if v == 88: # N/A = position 6
                v = 6;
            if v != 99:
                if not scale_answers_by_position.has_key(v):
                    return _('Cannot find wave scale answer for VR%s position' % index)
                response_values.append([attr.id, None, scale_answers_by_position[v].id])

        #general attributes
        letters = ['A', 'B', 'C', 'D', 'E', 'F', 'G']
        wave_ga = choice_attributes
        for i, attr in enumerate(wave_ga):
            v = state['VR%s' % letters[i]]
            if str(v.strip()) in ['99', '']:
                continue

            state_answers = []
            if hasattr(v, 'find') and v.find(',') >= 0:
                if attr.attribute.type != TntCLMAttribute.TYPE_MULTIPLE_CHOICE:
                    return _('Found multiple responses for a single choice attribute (VR%s)' % letters[i])
                state_answers = [int(x) for x in v.strip().split(',') if x != '']
            else:
                state_answers = [int(v)]

            answers = attr.answers
            found = 0
            for a in answers:
                if a.position in state_answers:
                    # check if this choice attr needs to also be saved as scale
                    if attr.attribute.type == TntCLMAttribute.TYPE_SINGLE_SCALE:
                        if not scale_answers_by_position.has_key(a.position):
                            return _('Cannot find wave scale answer for VR%(i)s position %(p)s' % {'i': letters[i], 'p': a.position})
                        s = [attr.id, a.id, scale_answers_by_position[a.position].id]
                    else:
                        s = [attr.id, a.id, None]
                    response_values.append(s)
                    found = found + 1
            if found != len(state_answers):
                return _('Cannot find wave answers for all VR%s positions' % letters[i])

        return response_values

    def export_failures_csv(self, data, response_filename):
        from cStringIO import StringIO
        import os
        # first line with headers
        headers = [h for h in self.USED_HEADERS]
        headers.append('ERROR')
        lines = [headers]
        for d in data:
            line = [d[h] for h in headers]
            lines.append(line)

        stream = StringIO()
        CSV_data = csv.writer(stream, delimiter=',', quoting=csv.QUOTE_ALL)
        CSV_data.writerows(lines)

        name, extension = os.path.splitext(response_filename)
        filename = "%s_%s_%s%s" % (name, 'failures', datetime.now().strftime("%Y%m%d_%H%M"), extension)

        fc = ProcessFileController(self.sa_session, account_id=self.get_account().id)
        category = fc.get_category_by_label(u'TNT Wave responses import failures')
        if category is None:
            category = fc.save_or_update_file_category(None, u'TNT Wave responses import failures')
        file = fc.save_or_update_file_from_content(None, filename, None, category, stream.getvalue())

        stream.close()

        return file

class BackgroundImportResponseController(AbstractController):
    def import_responses_from_task(self, file, loop_validator=None):
        """
        Method called by the task that does the responses import in the background
        """
        assert isinstance(file, ImportFile)

        import_settings = simplejson.loads(file.settings, object_hook=json_object_hook)
        controller = ManagerParticipantController(self.sa_session, account_id=import_settings['account_id'])
        ret = controller.import_responses(import_settings['wave_id'], file.file_hash,
                                    settings.FILE_REPOSITORY, import_settings['separator'], True, loop_validator=loop_validator,
                                    add_nonexistent_invitees=import_settings['add_nonexistent_invitees'],
                                    add_invitee_if_duplicate=import_settings['add_invitee_if_duplicate'])

        f = None
        if len(ret['failures_rows']) > 0:
            f = controller.export_failures_csv(ret['failures_rows'], file.file_name)

        return (ret['imported'], f)

class PriorityController(AbstractController):
    """
    The priority controller is used to trigger exports when participants give priority answers.
    """

    MAX_TIFF_PER_BATCH = 5 # how many TIFFs can be attached to a mail

    def _get_online_waves(self):
        """
        Method to retrieve all the CLM waves that are currently online
        """
        return self.sa_session.query(TntCLMWave).filter(TntCLMWave.status==TntCLMWave.STATUS_ONLINE).all()

    def _get_unprocessed_participants(self, wave):
        """
        Retrieve participants which have not been checked for priorities.
        """
        return self.sa_session.query(TntCLMWaveParticipant).join(TntCLMWaveInvitee).filter(
            and_(TntCLMWaveParticipant.is_processed==TntCLMWaveParticipant.NOT_PROCESSED,
                 TntCLMWaveInvitee.wave==wave)).all()

    def _get_priority_responses(self, participant, default_state):
        """
        Get the responses of questions/answers which have their priority triggered.

        Contact information:
        Format:
         * H
         * Country code
         * Depot code
         * Wave code + Quarter & Year of publication (W38Q22010)
         * Company name (or Name unknown)
         * Account number
         * Revenue band
         * Contact person name (or Contact unknown)
         * Gender (or Gender unknown)
         * D/I (Domestic or International)

        Attribute information:
        Format:
         * P
         * Attribute number (position)
         * Attribute description (for the default locale)
         * Attribute answer description (for the default locale)
         * ... 6 empty fields to match with the number of fields of contact information
        """
        wc = WaveController(self.sa_session)
        state = wc.get_wave_state(participant.invitee.wave, participant.locale)

        priority_attr = {}
        for attr in state['wave']['attributes']:
            if attr['priority_trigger_at'] is None:
                continue

            priority_attr['attribute_%s' % attr['id']] = attr['priority_trigger_at']

        attributes = default_state['wave']['attributes']
        scale_answers = {}
        for sa in default_state['wave']['scale_answers']:
            scale_answers[sa['id']] = sa

        rv = []
        for response in participant.responses:
            if not priority_attr.has_key('attribute_%s' % response.wave_attribute.id):
                continue

            # we are dealing with single choice attributes
            if response.wave_answer is not None:
                if response.wave_answer.position <= priority_attr['attribute_%s' % response.wave_attribute.id]:
                    rv.append([
                        'P',
                        response.wave_attribute.position,
                        attributes[response.wave_attribute.position-1]['label'].encode('utf-8'),
                        attributes[response.wave_attribute.position-1]['answers'][response.wave_answer.position-1]['label'].encode('utf-8'),
                        '', '', '', '', '', ''
                    ])
                continue

            # we are dealing with scale attributes
            if response.wave_scale_answer is not None:
                if response.wave_scale_answer.position <= priority_attr['attribute_%s' % response.wave_attribute.id]:
                    rv.append([
                        'P',
                        response.wave_attribute.position,
                        attributes[response.wave_attribute.position-1]['label'].encode('utf-8'),
                        scale_answers[response.wave_scale_answer_id]['label'].encode('utf-8'),
                        '', '', '', '', '', ''
                    ])

        # no priority answers were given by this participant.
        if len(rv) == 0:
            return None

        company = participant.invitee.contact.company
        person = participant.invitee.contact.person
        # we have priority answers, lets get the contact information
        depot = 'Depot unknown'
        if company.depot is not None:
            depot = company.depot.default_code

        companyname = company.name or 'Name unknown'
        companyname = companyname.encode('utf-8')

        personname = person.full_name or 'Contact unknown'
        personname = personname.encode('utf-8')

        countrycode = company.customer.country.iso_3166_a2
        wavecode = participant.invitee.wave.code

        # Silly exception for GBSPS, see ticket #1864
        if company.locale.label == 'GBSpS':
            countrycode = company.locale.label.upper()
            wavecode = '%s%s' % (wavecode, countrycode)

        ci = [
            'H',
            countrycode,
            depot,
            wavecode,
            companyname,
            company.customer.account_number,
            company.revenue_band or 'Revenue band unknown',
            personname,
            person.gender or 'Gender unknown',
            company.locale.location or 'Location unknown',
        ]
        rv.insert(0, ci)
        return rv

    @transaction_decorator
    def write_CSV(self):
        """
        Write the data to a CSV file.
        """
        wc = WaveController(self.sa_session)

        waves = self._get_online_waves()
        for wave in waves:
            participants = self._get_unprocessed_participants(wave)
            if len(participants) == 0:
                continue

            lines = []
            default_state = wc.get_wave_state(wave, wc.get_wave_default_locale(wave.id))
            priority_participants = []

            for participant in participants:
                responses = self._get_priority_responses(participant, default_state)
                if responses is not None:
                    lines.extend(responses)
                    priority_participants.append(participant)
                    participant.is_processed = TntCLMWaveParticipant.PROCESSED_P1
                else:
                    participant.is_processed = TntCLMWaveParticipant.PROCESSED_NO_P1
                self.sa_session.add(participant)

            if len(lines) == 0:
                continue
            self._write_lines_to_file(wave, lines, priority_participants)
        self.sa_session.flush()

    @transaction_decorator
    def _write_lines_to_file(self, wave, lines, participants):
        """
        Create a new TNT CLM Priority file including the lines.
        """
        from cStringIO import StringIO

        target_group = wave.groups[0].group.license.domain.default_license.default_group

        stream = StringIO()

        afc = AnonymousFileController(self.sa_session)

        priorityCSV = csv.writer(stream, delimiter=';', quoting=csv.QUOTE_ALL)
        [priorityCSV.writerow(row) for row in lines]

        dt = datetime.now()
        file_name = 'priority_%s_%s.csv' % (wave.code, dt.strftime("%Y%m%d_%H%M"))

        file = afc.save_file_for_target_group(stream.getvalue(), file_name, 'Priority %s' % wave.code, target_group, TntCLMWavePriorityCSVFile, props={'wave': wave}, privileges=(TntCLMWavePriorityCSVFile.PRIV_VIEW, ))

        stream.close()

        file.participants = participants
        self.sa_session.add(file)

    @transaction_decorator
    def send_CSV(self, loop_validator=None):
        """
        Sends all previously unsend CSV files.
        """
        fc = ProcessFileController(self.sa_session)

        files = self.sa_session.query(TntCLMWavePriorityCSVFile).filter(TntCLMWavePriorityCSVFile.is_sent==TntCLMWavePriorityCSVFile.EMAIL_NOT_SEND).all()

        s = smtplib.SMTP(settings.EMAIL_HOST, settings.EMAIL_PORT)
        for file in files:
            licensee = file.groups[0].group.license.domain.default_license
            lcc = LicenseConfigController(self.sa_session, licensee)
            to = lcc.get('tnt.response.priority.email.address')

            message = MIMEMultipart()
            message['From'] = settings.EMAIL_FROM
            message['To'] = to
            message['Date'] = formatdate(localtime=True)
            message['Subject'] = '[tnt-iqclm] Priority export'

            message.attach( MIMEText('-') )

            content = fc.get_file_content(file.file_hash)

            part = MIMEBase('application', "octet-stream")
            part.set_payload( content['content'] )
            Encoders.encode_base64(part)
            part.add_header('Content-Disposition', 'attachment; filename="%s"' % file.file_name)
            message.attach(part)

            file.is_sent = True
            self.sa_session.add(file)

            s.sendmail(settings.EMAIL_FROM, to, message.as_string())

            # lets the task abort if stop signal was sent
            if loop_validator is not None:
                loop_validator()

        s.quit()
        self.sa_session.flush()

    def get_tiff_content(self, path):
        """
        Return the content of a tiff file
        """
        f = open(path, 'r')
        rv = {'content': f.read(), 'mimetype': 'image/tiff', 'filename': os.path.basename(path)}
        f.close()
        return rv

    @transaction_decorator
    def _send_tiff_batch(self, response_controller, csv_file, csv_file_participants):
        """
        Send an email with a batch of files for all the participants in the list
        @param response_controller: the controller that generates the pdf file
        @param smtp: the smtp lib used for sending the mails
        @param csv_file: the csv_file that has the TntCLMWavePriorityCSVFileParticipant objects (used to get the TO email)
        @param csv_file_participants: list of TntCLMWavePriorityCSVFileParticipant objects
        """

        tmp_dir = '/tmp/%s/' % settings.SETTINGS_MODULE
        if not os.path.isdir(tmp_dir):
            os.mkdir(tmp_dir)

        filenames = [] # list of filenames (without extensions) that were converted
        converted = [] # list of TntCLMWavePriorityCSVFileParticipant objects that had tifs triggered

        licensee = csv_file.groups[0].group.license.domain.default_license
        lcc = LicenseConfigController(self.sa_session, licensee)
        to = lcc.get('tnt.response.priority.email.address')

        # create and configure the mail
        message = MIMEMultipart()
        message['From'] = settings.EMAIL_FROM
        message['To'] = to
        message['Date'] = formatdate(localtime=True)
        message['Subject'] = '[tnt-iqclm] Priority tiff export'
        message.attach( MIMEText('-') )

        for p in csv_file_participants:
            participant = p.participant
            company = participant.invitee.contact.company
            if company.depot is not None:
                depot = company.depot.default_code
            else:
                depot = ''

            countrycode = company.locale.country.iso_3166_a2
            wavecode = participant.invitee.wave.code

            # Silly exception for GBSPS, see ticket #1864
            if company.locale.label == 'GBSpS':
                countrycode = company.locale.label.upper()
                wavecode = '%s%s' % (wavecode, countrycode)

            filename = "%s_%s_%s_%s_%s_%s" % (
                countrycode, # country code
                company.locale.location or '', # Domestic/International
                depot, # depot code
                wavecode, # wave name
                company.customer.account_number, # account
                company.revenue_band or '',
            )
            pdf_path = '%s%s.pdf' % (tmp_dir, filename)
            tiff_path = '%s%s.tif' % (tmp_dir, filename)
            response_controller.export_response_pdf(pdf_path, p.participant_id)

            # convert the pdf to tiff
            code = call(["convert", "-adjoin", "-compress", "RLE", "-density", "75", pdf_path, tiff_path])
            if code == 0:
                # conversion was successful
                filenames.append(filename) # keep the filenames so they can be discarded after the mail is sent
                converted.append(p)

            # attach the tiff file
            content = self.get_tiff_content(tiff_path)
            part = MIMEBase('application', "octet-stream")
            part.set_payload( content['content'] )
            Encoders.encode_base64(part)
            part.add_header('Content-Disposition', 'attachment; filename="%s"' % content['filename'])
            message.attach(part)

        # send the mail with the batch of tif files attached
        smtp = smtplib.SMTP(settings.EMAIL_HOST, settings.EMAIL_PORT)
        smtp.sendmail(settings.EMAIL_FROM, to, message.as_string())
        smtp.quit()

        # at this point the mail is sent, discard the created files
        for f in filenames:
            if os.path.exists('%s%s.pdf' % (tmp_dir, f)):
                os.remove('%s%s.pdf' % (tmp_dir, f))
            if os.path.exists('%s%s.tif' % (tmp_dir, f)):
                os.remove('%s%s.tif' % (tmp_dir, f))

        # update the TntCLMWavePriorityCSVFileParticipant objects
        for cfp in converted:
            cfp.has_tiff_triggered = TntCLMWavePriorityCSVFileParticipant.TIFF_SENT
            self.sa_session.add(cfp)
        self.sa_session.flush()

    @transaction_decorator
    def _set_file_particpants_status(self, file_participants, status):
        for fp in file_participants:
            fp.has_tiff_triggered = status
            self.sa_session.add(fp)

    def send_TIFF(self, loop_validator=None):
        """
        Send the response TIFF files for each TntCLMWavePriorityCSVFileParticipant object with has_tiff_triggered=false,
        discard files after sending
        """
        fo = join(TntCLMWavePriorityCSVFile, TntCLMWavePriorityCSVFileParticipant,
                  TntCLMWavePriorityCSVFile.file_id==TntCLMWavePriorityCSVFileParticipant.priority_file_id)
        filter = TntCLMWavePriorityCSVFileParticipant.has_tiff_triggered==TntCLMWavePriorityCSVFileParticipant.TIFF_NOT_SENT
        csv_files = self.sa_session.query(TntCLMWavePriorityCSVFile).select_from(fo).filter(filter).all()

        if len(csv_files) == 0:
            return

        participants = []
        for csvf in csv_files:
            for fp in csvf.participants:
                if fp.csv_file_participant[0].has_tiff_triggered == TntCLMWavePriorityCSVFileParticipant.TIFF_NOT_SENT:
                    participants.append(fp.csv_file_participant[0])
        self._set_file_particpants_status(participants, TntCLMWavePriorityCSVFileParticipant.TIFF_SENDING)

        rc = ResponseController(self.sa_session) # create this just once

        for csvf in csv_files:

            if len(csvf.participants) == 0:
                continue

            csv_file_participants = []
            for fp in csvf.participants:
                if fp.csv_file_participant[0].has_tiff_triggered != TntCLMWavePriorityCSVFileParticipant.TIFF_SENDING:
                    continue
                csv_file_participants.append(fp.csv_file_participant[0])
                if len(csv_file_participants) == self.MAX_TIFF_PER_BATCH:
                    self._send_tiff_batch(rc, csvf, csv_file_participants)
                    csv_file_participants = []

            # check if we have any leftovers, and send a mail with those files too
            if len(csv_file_participants) > 0:
                self._send_tiff_batch(rc, csvf, csv_file_participants)

            # lets the task abort if stop signal was sent
            if loop_validator is not None:
                loop_validator()

class ResultsController(AbstractController):
    """
    Tnt CLM Response Results controller (generate charts, etc)
    """

    def __init__(self, sa_session=None, account_id=None, request=None):
        AbstractController.__init__(self, sa_session, account_id=account_id, request=request)
        self.wave_controller = WaveController(sa_session, account_id=account_id)

    def _get_default_waves(self):
        """
        Return the default waves needed for generating respondents data
        @return: the waves the user has PRIV_VIEW_RESPONSE on, that are online. If there are no online waves, get the
            last finished wave. if there are no finished waves, return []
        """
        accessible_waves = self.wave_controller.get_privileged_object_by_id(TntCLMWave, None, TntCLMWave.PRIV_VIEW_RESPONSE)
        online_waves = [e for e in accessible_waves if e.status == TntCLMWave.STATUS_ONLINE]
        if len(online_waves) > 0:
            return online_waves
        else:
            finished_waves = [e for e in accessible_waves if e.status == TntCLMWave.STATUS_FINISHED]
            if len(finished_waves) > 0:
                return finished_waves[-1:] # use only the last one
            else:
                return [] # no default waves to generate data for

    def _prepare_waves(self, waves):
        """
        Return a list of wave objects based on what the user passes in waves
        @param waves: can be a list of waves, wave ids, a wave, or a wave id
        """
        wave_objects = []

        if not isinstance(waves, list):
            waves = [waves]
        for w in waves:
            if type(w) != TntCLMWave:
                w = self.wave_controller.get_privileged_object_by_id(TntCLMWave, int(w), TntCLMWave.PRIV_VIEW_RESPONSE)[0]
            else:
                self.assert_privileges_on_object(w, TntCLMWave.PRIV_VIEW_RESPONSE)
            wave_objects.append(w)

        return wave_objects

    def _prepare_geolevels(self, geolevels):
        """
        Return a list of geolevel objects based on what the user passes in geolevels
        @param geolevels: can be a list of geolevels, geolevel ids, a geolevel or geolevel id
        """
        geolevel_objects = []
        account = self.get_account()
        gc = TntController(self.sa_session, account_id=account.id)

        if not isinstance(geolevels, list):
            geolevels = [geolevels]
        for g in geolevels:
            if type(g) != TntGeolevel:
                g = gc.get_geolevels(int(g))[0] # already checks for tnt accounts if they have access to this geolevel
            else:
                self.assert_privileges_on_object(g, TntGeolevel.PRIV_VIEW)
                if hasattr(account, 'geolevels'):
                    if g.id not in [geo.id for geo in account.geolevels]:
                        raise SecurityException('Trying to access geolevel without the proper rights [id=%s]' % g.id)
            geolevel_objects.append(g)

        return geolevel_objects

    def _get_wave_days(self, wave, begin_day, end_day, last_days):
        """
        Return the dates and days needed for generating queries for this wave
        @param wave: the wave object for which we want to get the data
        @param last_days: the param that indicates whether we want the last days or a regular interval
        @return: a tuple as follows: (
            ds, # date object that indicates the start date for which we want results
            de, # date object that indicates the end date for which we want results
            begin_day, # the begin day as a number (1 - first day of the wave) - same as param if not reset
            end_day, # the end day as a number (1 - first day of the wave) - same as param if not reset
        )
        """

        begin_date = datetime.date(wave.begin_date)
        dt = datetime.date(datetime.now()) # date of today

        # check if we want the last X days
        if last_days is not None:
            # need to calculate begin_day and end_day based on this, ignore the params for those

            # check if we have an end_date for the wave and if it's in the past. if so, use that, else today
            if wave.end_date is not None and datetime.date(wave.end_date) < dt:
                de = datetime.date(wave.end_date)
            else:
                de = dt
            end_day = (de - begin_date).days # reset the param based on the calculated date

            ds = de - timedelta(days=last_days) # start date goes last_days in the past
            begin_day = (ds - begin_date).days + 1 # reset the param based on the calculated date

        else:
            ds = begin_date + timedelta(days=begin_day-1) # start date for the respondents

            if end_day is not None:
                de = begin_date + timedelta(days=end_day-1)
                # check if the wave has an end date, and if that's earlier than the end_day, use that
                if wave.end_date is not None and datetime.date(wave.end_date) < de:
                    de = datetime.date(wave.end_date)
                # also check if the end day is not in the future
                if de > dt:
                    de = dt
            else:
                # didn't specify an end_day, so use the wave end date if that's not in the future (and is set), or today
                if wave.end_date is not None and datetime.date(wave.end_date) < dt:
                    de = datetime.date(wave.end_date)
                else:
                    # no wave end day set, or date is in the future, so we use today
                    de = dt

        return (ds, de, begin_day, end_day)

    def generate_respondents_result(self, wave=None, geolevel=None, begin_day=1, end_day=None, last_days=None, zoom_day=None):
        """
        Generate the result that is to be used for generating the chart with the number of responses for the given
        geolevel/wave(s)/day(s).
        @param wave: wave object, id, or list of waves for which we want the number of responses
            if None, use all the waves that have status STATUS_ONLINE. if there are only waves with STATUS_FINISHED, use
            the last finished one
        @param geolevel: geolevel object, geolevel id or list of geolevels for which we want the number of responses
            if None, show the data for all the geolevels
        @param begin_day: the start of the period for which we want the results. 1 is the begin_date of the wave
        @param end_day: the end of the period for which we want the results. if None, the current day is used, or the
            end_date of the wave if that's in the past
        @param last_days: this is mutually exclusive with begin_day, end_day and zoom_day. if given, use it to show
            the last days of results for the wave. This can only be used with one wave !
            If there are more than one waves, this param is ignored. If this is given, the begin_day and end_day are
            calculated automatically based on this value.
            If the wave doesn't have an end_day, the last days until the current day are used.
        @param zoom_day: if not None, then the result will show the responses per hour for the given params
            The zoom_day (if given) is an int (1 for begin_day). If given, the rest of the params are ignored
        @return: the result that's needed by the SimpleChartWrapper, similar to what execute_mdx returns, or None if
            there are no waves
        """
        data = [ [u''] ]
        geolevel_objects = []

        # each key will be an int >= 1. using this because multiple waves can have days or hours that others don't
        intervals = {}

        if wave is None:
            wave = self._get_default_waves()
            if not wave:
                return False
        wave_objects = self._prepare_waves(wave)

        # check if the geolevel param is ok
        if geolevel is not None:
            geolevel_objects = self._prepare_geolevels(geolevel)

        n_waves = len(wave_objects)

        if last_days is not None and n_waves > 1:
            # can't use the last_days param if we have more waves. just ignore it
            last_days = None

        i = 0 # wave index
        max_interval = 0
        min_interval = 0

        for w in wave_objects:
            if w.begin_date is None:
                continue
            begin_date = datetime.date(w.begin_date)

            # build the query for getting the responses
            fo = join(TntCLMWaveResponse, TntCLMWaveParticipant, TntCLMWaveResponse.participant_id==TntCLMWaveParticipant.id).join(
                 TntCLMWaveInvitee, TntCLMWaveParticipant.invitee_id==TntCLMWaveInvitee.id).join(
                 TntCLMWave, TntCLMWaveInvitee.wave_id==TntCLMWave.id)

            filter = TntCLMWave.id==w.id

            if geolevel is not None:
                fo = fo.join(TntCLMLocale, TntCLMWaveParticipant.locale_id==TntCLMLocale.id).join(
                    TntGeolevelCountry, TntCLMLocale.country_id==TntGeolevelCountry.country_id).join(
                    TntGeolevel, TntGeolevelCountry.geolevel_id==TntGeolevel.id)
                filter = and_(filter, TntGeolevel.id.in_([g.id for g in geolevel_objects]))

            # what to select when we want days or hours
            selection = [func.count(TntCLMWaveResponse.participant_id.distinct())]
            dr = func.date(TntCLMWaveResponse.created) # shortcut for the date of the response creation date

            if zoom_day is not None:
                # we want hours
                d = begin_date + timedelta(days=zoom_day-1)
                filter = and_(filter, dr==d)
                gr_by = func.hour(TntCLMWaveResponse.created)
                selection.append(func.hour(TntCLMWaveResponse.created))
            else:
                # we want days
                ds, de, begin_day, end_day = self._get_wave_days(w, begin_day, end_day, last_days)

                filter = and_(filter, dr>=ds) # start date condition
                filter = and_(filter, dr<=de)
                gr_by = dr
                selection.append(dr) # end date condition

            q = self.sa_session.query(*selection).select_from(fo).filter(filter).group_by(gr_by).all()

            data.append([w.name])
            for r in q:
                # add each day. r is [number_of_respondents, date/hour]
                if zoom_day is not None:
                    # r[1] is the hour [0, 1, ..., 23]
                    d = r[1]
                else:
                    # r[1] is the date (not datetime)
                    d = (r[1] - begin_date).days + 1 # 1 for the begin_date, 2 for the second date, etc.

                    if d > max_interval:
                        max_interval = d

                    if d < min_interval or min_interval == 0: # check which is the smallest day we have to show that has data
                        min_interval = d

                if not intervals.has_key(d):
                    # if it's a day we dind't have, fill it in with empty values for the previous waves
                    # if it's the first wave, it's gonna be an empty list
                    intervals[d] = [None] * i

                # check if we have values for all the previous waves, if not, fill in with None
                if len(intervals[d]) < i:
                    intervals[d].extend([None] * (i - len(intervals[d])))

                intervals[d].append(r[0])

            i = i + 1

        # intervals should now contain as keys the days/hours (as ints) and as values a list of all responses for all wawes
        # eg. if we had two waves, one had 4 responses on the first day, none the second day, and second wave 2 responses on the third day,
        # intervals would be: {1: [4, None], 3: [None, 2]}

        if zoom_day is not None:
            # hour zoom
            min_interval = 0
            max_interval = 23
        else:
            # day view . check that we also have any empty days shown in the beginning and end (if they were requested)
            if min_interval > begin_day:
                min_interval = begin_day
            if end_day is not None and max_interval < end_day:
                max_interval = end_day

        # loop and create missing interval periods if none of the waves had results then
        for i in range(min_interval, max_interval + 1):
            # make sure we have the data for an interval period, even if none of the waves had respondents
            if not intervals.has_key(i):
                intervals[i] = [None] * n_waves

            # make sure all waves have data for the given interval, even if it's just empty
            if len(intervals[i]) < n_waves:
                intervals[i].extend([None] * (n_waves - len(intervals[i])))

            data[0].append(u"%s" % i)
            for j in range(n_waves):
                data[j + 1].append(intervals[i][j]) # data[0] has the intervals, 1 is the first wave data, 2 second

        return data
