# Copyright (c) 2011 BaseSoft B.V.
# See LICENSE for details.

from django.utils.translation import ugettext as _
from sqlalchemy import func, case
from sqlalchemy.sql.expression import cast
from sqlalchemy.types import Integer

from basesoft.web.core.util.custom_query import CustomQuery, CustomQueryField, CustomQueryModel, DATA_TYPE_DATE
from basesoft.web.contrib.contact.models import Contact, Person
from basesoft.web.contrib.i18n.models import Country
from basesoft.web.contrib.file.models import File, ImportFile, ImportFileType
from basesoft.web.contrib.i18n.custom_query import DomainCountryCustomQueryModel, DomainCountryCustomQueryField

from iqportal.tnt.clm.models import TntCLMWave, TntCLMLocale, TntCLMCompany, TntCLMCustomer
from iqportal.tnt.response.models import TntCLMWaveResponse, TntCLMWaveParticipant, TntCLMWaveInvitee
from iqportal.tnt.clm.controllers import TntCLMLocaleController
from iqportal.tnt.core.models import TntDepot, TntGeolevel, TntGeolevelCountry

class TntCLMWaveRespondentsByCountryCustomQuery(CustomQuery):
    """
    """

    def tables(self):
        return (
            CustomQueryModel(TntCLMWave),
            CustomQueryModel(TntCLMWaveInvitee, onclause=TntCLMWaveInvitee.wave_id==TntCLMWave.id),
            CustomQueryModel(TntCLMWaveParticipant, onclause=TntCLMWaveInvitee.id==TntCLMWaveParticipant.invitee_id, use_outerjoin=True),
            CustomQueryModel(TntCLMLocale,
                             onclause=TntCLMLocale.id==TntCLMWaveParticipant.locale_id),
            DomainCountryCustomQueryModel(onclause=Country.id==TntCLMLocale.country_id),

        )

    def fields(self):
        return (
            CustomQueryField(TntCLMWave.name, _('Wave'), key='wave_name', is_filter=True, is_group_by=True),
            DomainCountryCustomQueryField(None, _('Country'), is_filter=True, is_sorted=True, key='country'),
            CustomQueryField(Country.id, 'Country_', is_filter=False),
            #CustomQueryField(func.count(TntCLMWaveInvitee.id), _('Invitees'), key='invitees'),
            CustomQueryField(cast(func.sum(case([(TntCLMWaveParticipant.is_respondent>0, 1)], else_=0)), Integer),
                             _('Respondents'), key='respondents'),
            CustomQueryField(TntCLMWave.id, _('Export CSV'), key='export_csv'),
            CustomQueryField(TntCLMWaveParticipant.is_respondent, _('Channel'), key='channel', is_filter=True, in_select=False,
                             filter_options=[{'value': TntCLMWaveParticipant.RESPONSES_INTERNET,
                                             'label': _("Internet")
                                            },
                                            {'value': TntCLMWaveParticipant.RESPONSES_PAPER,
                                             'label': _("Paper")
                                            }])
        )

    group_by = (Country.id, TntCLMWave.id)

class TntCLMWaveRespondentsCustomQuery(CustomQuery):
    #permissions = ['wave_manager']

    def tables(self):
        return (
            CustomQueryModel(TntCLMWaveInvitee),
            CustomQueryModel(TntCLMWave,
                             onclause=TntCLMWaveInvitee.wave_id==TntCLMWave.id),
            CustomQueryModel(TntCLMWaveParticipant,
                             onclause=TntCLMWaveInvitee.id==TntCLMWaveParticipant.invitee_id),

            CustomQueryModel(TntCLMLocale,
                             onclause=TntCLMLocale.id==TntCLMWaveParticipant.locale_id, use_in_count=False),
            DomainCountryCustomQueryModel(onclause=Country.id==TntCLMLocale.country_id, use_in_count=False),
            CustomQueryModel(Contact,
                             onclause=TntCLMWaveInvitee.contact_id==Contact.id, priv=None, use_in_count=False),
            CustomQueryModel(Person,
                             onclause=Contact.person_id==Person.id, use_in_count=False),
            CustomQueryModel(TntCLMCompany,
                             onclause=Contact.company_id==TntCLMCompany.company_id, use_outerjoin=True, use_in_count=False),
            CustomQueryModel(TntCLMCustomer,
                             onclause=TntCLMCompany.customer_id==TntCLMCustomer.id, use_outerjoin=True, use_in_count=False),
            CustomQueryModel(TntDepot,
                             onclause=TntCLMCompany.depot_id==TntDepot.id, use_outerjoin=True, priv=None, use_in_count=False),
            CustomQueryModel(TntGeolevelCountry,
                             onclause=Country.id==TntGeolevelCountry.country_id, use_outerjoin=True, use_in_count=False),
            CustomQueryModel(TntGeolevel,
                             onclause=TntGeolevelCountry.geolevel_id==TntGeolevel.id, use_outerjoin=True, priv=None, use_in_count=False),
        )
    def fields(self):
        return (
            # latest created waves on top
            CustomQueryField(TntCLMWave.id, _('ID'), key='wave_id'),
            CustomQueryField(TntCLMWave.name, _('Wave'), key='wave_name', is_filter=True),
            CustomQueryField(TntCLMWaveParticipant.locale_id, _('Locale Id'), key='locale_id'),
            CustomQueryField(TntCLMWaveParticipant.id, _('Participant Id'), key='participant_id'),
            DomainCountryCustomQueryField(None, _('Country'), key='country', is_filter=True, is_group_by=True),
            CustomQueryField(Country.id, _('Country'), key='country_id', in_select=False), # used by 'wave respondents by country'
            CustomQueryField(TntCLMCompany.name, _('Name'), key='name'),
            CustomQueryField(TntCLMCustomer.account_number, _('Account no.'), key='account_number', is_filter=True),
            CustomQueryField(Person.email, _('Email')),
            CustomQueryField(TntDepot.default_code, _('Depot'), key='depot_code'),
            CustomQueryField(TntCLMCompany.revenue_band, _('Revenue'), key='revenue'),
            CustomQueryField(TntCLMLocale.location, _('Location'), key='location'),
            CustomQueryField(TntCLMWaveParticipant.is_processed, _('P1'), key='p1_triggered', is_filter=True,
                             filter_options=[{'value': TntCLMWaveParticipant.NOT_PROCESSED,
                                             'label': _("Not processed")
                                            },
                                            {'value': TntCLMWaveParticipant.PROCESSED_NO_P1,
                                             'label': _("No")
                                            },
                                            {'value': TntCLMWaveParticipant.PROCESSED_P1,
                                             'label': _("Yes")
                                            },
                                            {'value': TntCLMWaveParticipant.DONT_PROCESS,
                                             'label': _("Will not be processed")
                                            }]
            ),
            CustomQueryField(TntCLMWaveInvitee.id, _('Responses'), key="responses"),
            CustomQueryField(TntCLMWaveParticipant.is_respondent, _('Channel'), key='channel', is_filter=True,
                             filter_options=[{'value': TntCLMWaveParticipant.RESPONSES_INTERNET,
                                             'label': _("Internet")
                                            },
                                            {'value': TntCLMWaveParticipant.RESPONSES_PAPER,
                                             'label': _("Paper")
                                            }]),
            CustomQueryField(TntCLMWave.id, _('PDF'), key='response_pdf'),

            CustomQueryField(func.max(TntCLMWaveParticipant.last_updated), _('Last response'), key='last_response',
                             type=DATA_TYPE_DATE, is_sorted=True, sorted_asc=False),

            CustomQueryField(TntGeolevel.name, _('Geolevel'), in_select=False, is_filter=True),
        )
    def set_where(self):
        controller = TntCLMLocaleController(self.sa_session, account_id=self.get_account().id)
        self.where = controller.get_query_filter_on_locales(False)

    group_by = (TntCLMWave.id, TntCLMWaveParticipant.id, )


class TntCLMWaveRespondentsParticipantsCustomQuery(TntCLMWaveRespondentsCustomQuery):
    """
    override the TntCLMWaveRespondentsCustomQuery that doesnt use the set_where for the participant drilldown
    """
    def set_where(self):
        pass

class FileImportCustomQuery(CustomQuery):

    from sqlalchemy.orm import aliased
    validationFile = aliased(File, name='validationFile')
    def tables(self):
        return (
            CustomQueryModel(ImportFile),
            CustomQueryModel(ImportFileType),
            CustomQueryModel(self.validationFile,
                             onclause=ImportFile.failure_file_id==self.validationFile.id, use_outerjoin=True),
        )
    def fields(self):
        return (
            CustomQueryField(ImportFile.id, _('ID'), key='id'),
            CustomQueryField(ImportFile.settings, _('Wave'), key='wave'),
            CustomQueryField(ImportFile.file_name, _('File name'), is_filter=True, key='file_name'),
            CustomQueryField(ImportFile.file_hash, _('File hash'), key='file_hash'),
            CustomQueryField(ImportFile.status, _('Status'), is_filter=True, key='status'),
            CustomQueryField(ImportFile.result, _('Comment'), key='result'),
            CustomQueryField(self.validationFile.file_name, _('Failures file'), is_filter=True, key='validation_file'),
            CustomQueryField(self.validationFile.file_hash, _('Failures file hash'), key='validation_file_hash'),
            CustomQueryField(ImportFile.created, _('Created'), key='created'),
            CustomQueryField(ImportFile.last_updated, _('Last updated'), key='last_updated'),
        )
    def set_where(self):
        self.where = ImportFileType.name=='iqportal.tnt.response'

custom_queries = (TntCLMWaveRespondentsByCountryCustomQuery, TntCLMWaveRespondentsCustomQuery, FileImportCustomQuery, TntCLMWaveRespondentsParticipantsCustomQuery)
