# Copyright (c) 2011 BaseSoft B.V.
# See LICENSE for details.

import xmlrpclib
import os

from django.conf import settings
from django.utils.translation import ugettext as _

from basesoft.web.contrib.i18n.models import Country
from basesoft.web.core.exceptions import PublicUserException
from basesoft.web.contrib.graph.controllers import GraphController
from basesoft.web.core.util import custom_query as cq

from iqportal.tnt.core.controllers import TntController
from iqportal.tnt.clm.controllers import WaveController
from iqportal.tnt.clm.models import TntCLMWave
from iqportal.tnt.response.controllers import ResponseController, ParticipantController, ManagerParticipantController, \
    ResultsController
from twisted.web.xmlrpc import Fault
from basesoft.web.contrib.license.models import Domain
from basesoft.web.contrib.file.models import ImageFile
from basesoft.web.core.util.controllers import LicenseConfigController

class ResponseRPCHandler:
    def __init__(self, request, **kwargs):
        self.sa_session = request.Session()
        self.request = request

    def login(self, account_no, country_id):
        controller = ResponseController(self.sa_session)
        invitee = controller.get_invitee_by_account_number(account_no, country_id)
        if invitee is None:
            raise PublicUserException(_('Incorrect account number and country match'))
        return invitee.invitee_hash

    def get_countries(self):
        rv = []
        code = None if not self.request.session.has_key('default_country_code') else self.request.session['default_country_code']
        for country in self.sa_session.query(Country).all():
            rv.append({'label': country.name, 'value': country.id, 'selected': country.iso_3166_a2==code})
        return rv

    def export_wave_csv(self, wave_id, country_id=None, no_texts=False):
        """
        """
        client = xmlrpclib.Server(settings.SCHEDULER_SERVER_XMLRPC, use_datetime=True, allow_none=True)
        try:
            client.start_task('iqportal.tnt.response.schedule.ExportResponsesTask', {'wave_id': wave_id,
                'country_id': country_id, 'account_id': self.request.session['user_id'], 'no_texts': no_texts})
        except Fault, e:
            raise PublicUserException(e.faultString)

    def get_logo(self):
        domain = self.sa_session.query(Domain).filter(Domain.name=='TNT').first()
        lcc = LicenseConfigController(self.sa_session, scope=domain.default_license)
        img = lcc.get('basesoft.license.company.logo')
        if img is not None:
            return self.sa_session.query(ImageFile).get(img).file_hash
        return None

class ParticipantResponseRPCHandler:
    """
    """
    def __init__(self, request, **kwargs):
        self.sa_session = request.Session()

    def save_responses(self, session_id, state):
        controller = ParticipantController(self.sa_session, None, session_id)
        controller.save_responses(state)

    def get_wave_state(self, session_id, locale_id=None):
        controller = ParticipantController(self.sa_session, None, session_id)
        return controller.get_wave_state_for_invitee(locale_id)

class ManagerParticipantResponseRPCHandler:
    """
    """
    def __init__(self, request, **kwargs):
        self.controller = ManagerParticipantController(request.Session(), account_id=request.session['user_id'])

    def get_participant_wave_state(self, participant_id, locale_id=None):
        return self.controller.get_participant_wave_state(participant_id, locale_id)

    def import_responses(self, wave_id, file_hash, separator=None, add_nonexistent_invitees=False, add_invitee_if_duplicate=False):
        ret = self.controller.import_responses(wave_id, file_hash, settings.TMP_FILE_REPOSITORY, separator,
                                               add_nonexistent_invitees=add_nonexistent_invitees, add_invitee_if_duplicate=add_invitee_if_duplicate)
        # not important for client
        if ret.has_key('failures_rows'):
            del ret['failures_rows']
        return ret

    def reset_P1_tiff_status(self, participant_ids):
        return self.controller.reset_P1_tiff_status(participant_ids)

class RespondentsRPCHandler:
    """
    handler used by the respondents widget
    """
    def __init__(self, request, **kwargs):
        self.sa_session = request.Session()
        self.core_controller = TntController(self.sa_session, account_id=request.session['user_id'])
        self.wave_controller = WaveController(self.sa_session, account_id=request.session['user_id'])
        self.results_controller = ResultsController(self.sa_session, account_id=request.session['user_id'])
        self.gc = GraphController(self.sa_session, account_id=request.session['user_id'], request=request)

    def get_filters_and_group_by(self, static_filters=None):
        """
        Provide the widget with the proper filters
        """

        geolevel_options = [{'value': e.id, 'label': e.name} for e in self.core_controller.get_geolevels()]
        wave_options = [{'value': e.id, 'label': e.name} for e in self.wave_controller.get_waves() \
                        if e.status in [TntCLMWave.STATUS_ONLINE, TntCLMWave.STATUS_FINISHED]]

        return {
            'filters':  {
                'wave': {
                    'type': cq.DATA_TYPE_INT,
                    'label': _('Wave'),
                    'options': wave_options
                },
                'geolevel': {
                    'type': cq.DATA_TYPE_INT,
                    'label': _('Geolevel'),
                    'options': geolevel_options
                },
                'begin_day': {
                    'type': cq.DATA_TYPE_INT,
                    'label': _('Begin day'),
                    'unique': True, # can only be used once
                },
                'end_day': {
                    'type': cq.DATA_TYPE_INT,
                    'label': _('End day'),
                    'unique': True, # can only be used once
                },
                'last_days': {
                    'type': cq.DATA_TYPE_INT,
                    'label': _('Last days'),
                    'unique': True, # can only be used once
                },
            },
            'group_by': {},
            # can't use last_days together with begin_day or end_day
            'mutual_exclusive': [['begin_day', 'last_days'], ['end_day', 'last_days']]
        }

    def generate_respondents_chart(self, params):
        """
        Generate the respondents chart based on what filters we have (waves, geolevels, etc)
        @param width: the available width we have for the chart
        @param height: the available height we have for the chart. if None, calculate automatically based on days
        @param params: a dictionary of params that are used for generating the respondents data
            Can contain (none are mandatory): wave, geolevel, begin_day, end_day, last_days, zoom_day
        @param title: optional title for the chart
        """
        return self.results_controller.generate_respondents_result(**params)

handlers = (ResponseRPCHandler, ParticipantResponseRPCHandler, ManagerParticipantResponseRPCHandler, \
            RespondentsRPCHandler)
