# Copyright (c) 2011 BaseSoft B.V.
# See LICENSE for details.

import datetime

from sqlalchemy import Column, Integer, ForeignKey, String, DateTime, Unicode, UnicodeText, SmallInteger
from django.utils.translation import ugettext as _

from basesoft.sqlalchemy.table import IQTable
from basesoft.web.contrib.license import models as license_models
from sqlalchemy.types import Enum

__mapping_defined = False
__tables_defined = False
tables = {}

def reset():
    global __mapping_defined
    global __tables_defined

    __mapping_defined = False
    __tables_defined = False

def get_tables():
    return tables

def do_tables(metadata):
    global __tables_defined
    if not __tables_defined:
        # created on moment of publication
        tblTntCLMWaveInvitee = IQTable('tblTntCLMWaveInvitee', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),

            Column('wave_id', Integer, ForeignKey('tblTntCLMWave.id'), nullable=False),
            Column('contact_id', Integer, ForeignKey('tblContact.id'), nullable=True),
            Column('invitee_hash', String(8), unique=True), # contains a random part of a sha1 hash used to send out to participants

            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now),
        )

        tblTntCLMWaveRespondentType = IQTable('tblTntCLMWaveRespondentType', metadata,
            Column('id', SmallInteger, primary_key=True, autoincrement=False),
            Column('label', Unicode(255), nullable=False),
        )

        # created first time the invitee uses their link/login
        tblTntCLMWaveParticipant = IQTable('tblTntCLMWaveParticipant', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('invitee_id', Integer, ForeignKey('tblTntCLMWaveInvitee.id'), nullable=False, unique=True),
            Column('locale_id', Integer, ForeignKey('tblTntCLMLocale.id'), nullable=False),
            Column('opt_out', SmallInteger, nullable=False, default=0),
            Column('is_processed', SmallInteger, nullable=False, default=0),
            Column('is_respondent', SmallInteger, ForeignKey('tblTntCLMWaveRespondentType.id', ondelete='RESTRICT'), default=0),
            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now)
        )

        # created everytime the participant logs in / clicks link
        tblTntCLMWaveParticipantSession = IQTable('tblTntCLMWaveParticipantSession', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('participant_id', Integer, ForeignKey('tblTntCLMWaveParticipant.id'), nullable=False),
            Column('created', DateTime(), nullable=False, default=datetime.datetime.now, index=True),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now)
        )

        tblTntCLMWaveResponse = IQTable('tblTntCLMWaveResponse', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('participant_id', Integer, ForeignKey('tblTntCLMWaveParticipant.id'), nullable=False),
            Column('wave_attribute_id', Integer, ForeignKey('tblTntCLMWaveAttribute.id'), nullable=False),
            Column('wave_answer_id', Integer, ForeignKey('tblTntCLMWaveAnswer.id'), nullable=True),
            Column('wave_scale_answer_id', Integer, ForeignKey('tblTntCLMWaveScaleAnswer.id'), nullable=True),
            Column('participant_session_id', Integer, ForeignKey('tblTntCLMWaveParticipantSession.id'), nullable=False),
            Column('response_text', UnicodeText, nullable=True),

            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now),

        )

        # used to store csv files of priority triggers per wave.
        tblTntCLMWavePriorityCSVFile = IQTable('tblTntCLMWavePriorityCSVFile', metadata,
            Column('file_id', Integer, ForeignKey('tblFile.id', ondelete='CASCADE'), primary_key=True),
            Column('wave_id', Integer, ForeignKey('tblTntCLMWave.id'), nullable=False),
            Column('is_sent', SmallInteger, nullable=False, default=0), # indication if the file was emailed to TNT
        )

        tblTntCLMWavePriorityCSVFileParticipant = IQTable('tblTntCLMWavePriorityCSVFileParticipant', metadata,
            Column('priority_file_id', Integer, ForeignKey('tblTntCLMWavePriorityCSVFile.file_id', ondelete='CASCADE'), primary_key=True),
            Column('participant_id', Integer, ForeignKey('tblTntCLMWaveParticipant.id'), primary_key=True),
            Column('has_tiff_triggered', SmallInteger, nullable=False, default=0), # indication if the tif was generated/sent
        )

        tables['tblTntCLMWaveRespondentType'] = tblTntCLMWaveRespondentType
        tables['tblTntCLMWaveInvitee'] = tblTntCLMWaveInvitee
        tables['tblTntCLMWaveParticipant'] = tblTntCLMWaveParticipant
        tables['tblTntCLMWaveParticipantSession'] = tblTntCLMWaveParticipantSession
        tables['tblTntCLMWaveResponse'] = tblTntCLMWaveResponse
        tables['tblTntCLMWavePriorityCSVFile'] = tblTntCLMWavePriorityCSVFile
        tables['tblTntCLMWavePriorityCSVFileParticipant'] = tblTntCLMWavePriorityCSVFileParticipant

        __tables_defined = True

def do_mappings():
    """
    This function is called by Django middleware when needed.
    """
    global __mapping_defined
    if not __mapping_defined:
        from sqlalchemy.orm import mapper, relation
        from basesoft.web.contrib.contact import models as contact_models
        from basesoft.web.contrib.i18n import models as i18n_models
        from basesoft.web.contrib.file import models as file_models
        from iqportal.tnt.response import models as response_models
        from iqportal.tnt.clm import models as clm_models
        from iqportal.tnt.core import models as core_models
        from iqportal.tnt.response.hooks import TntCLMWaveResponseHook

        mapper(response_models.TntCLMWaveRespondentType, tables['tblTntCLMWaveRespondentType'])

        mapper(response_models.TntCLMWaveInvitee, tables['tblTntCLMWaveInvitee'],
            properties = {
                'wave': relation(clm_models.TntCLMWave),
                'contact': relation(contact_models.Contact)
            }
        )

        mapper(response_models.TntCLMWaveParticipant, tables['tblTntCLMWaveParticipant'],
            properties = {
                'invitee': relation(response_models.TntCLMWaveInvitee),
                'locale': relation(clm_models.TntCLMLocale),
            }
        )

        mapper(response_models.TntCLMWaveParticipantSession, tables['tblTntCLMWaveParticipantSession'],
            properties = {
                'participant': relation(response_models.TntCLMWaveParticipant)
            }
        )

        mapper(response_models.TntCLMWaveResponse, tables['tblTntCLMWaveResponse'],
            extension = TntCLMWaveResponseHook(),
            properties = {
                'participant': relation(response_models.TntCLMWaveParticipant, backref='responses'),
                'wave_attribute': relation(clm_models.TntCLMWaveAttribute),
                'wave_answer': relation(clm_models.TntCLMWaveAnswer),
                'wave_scale_answer': relation(clm_models.TntCLMWaveScaleAnswer),
                'participant_session': relation(response_models.TntCLMWaveParticipantSession),
            }
        )

        mapper(response_models.TntCLMWavePriorityCSVFile, tables['tblTntCLMWavePriorityCSVFile'],
            properties = {
                'wave': relation(clm_models.TntCLMWave),
                'participants': relation(response_models.TntCLMWaveParticipant, secondary=tables['tblTntCLMWavePriorityCSVFileParticipant']),
            },
            polymorphic_identity='TntCLMWavePriorityCSVFile',
            inherits=file_models.File
        )

        mapper(response_models.TntCLMWavePriorityCSVFileParticipant, tables['tblTntCLMWavePriorityCSVFileParticipant'],
            properties = {
                'participant': relation(response_models.TntCLMWaveParticipant, backref='csv_file_participant'),
            }
        )

        __mapping_defined = True

def app_permissions_category():
    """
    Returns a category for all permissions defined for this app
    """
    return license_models.PermissionCategory(u'TNT-CLM Response')

def app_permissions():
    """
    The permissions used within this application. Can be sync'd with the `iqsyncpermissions` command.

    Titles are with gettext calls, but in the database they appear in English. We need it like this so we get the
    titles in the language files. Later, in the js interfaces we show them translated.
    """
    return [

        # Wave
        license_models.Permission(u'tnt_clm_view_priority_export_status', u"""
            This gives the right to view the priority export status.
            """, _("View priority export status")),

    ]

def app_configs():
    """
    The account configs available for this application. This can be sync'd with the
    `iqsyncpermissions` command.
    """
    from basesoft.web.contrib.account import models as account_models
    configs = []

    config = account_models.Config()
    config.name = 'tnt.response.priority.email.address'
    config.type = 'string'
    config.default = 'priority@mail-test.basesoft.com'
    config.description = u'Email address to send priority triggers to (CSV & TIFF)'
    configs.append(config)

    return configs

