# Copyright (c) 2011 BaseSoft B.V.
# See LICENSE for details.

from sqlalchemy import and_

from basesoft.scheduler.task import ScheduleTask
from iqportal.tnt.response.controllers import PriorityController,\
    ResponseController

from iqportal.tnt.response.controllers import BackgroundImportResponseController
from basesoft.web.contrib.file.models import ImportFile, ImportFileType
from basesoft.web.contrib.file.controllers import ImportFileController

from basesoft import logging
from types import NoneType
from basesoft.web.contrib.notification.controllers import NotificationController

from django.utils.translation import ugettext as _
from basesoft.web.contrib.account.models import Account
from iqportal.tnt.clm.controllers import WaveController
from basesoft.web.contrib.i18n.controllers import I18nController

logger = logging.getLogger("iqportal.tnt.response")

class PriorityCSVTask(ScheduleTask):
    """
    Task checks any new (non-updated) responses for priority triggered answers
    This compiles a list of participants + priority triggered answers and writes them to a CSV file to be mailed.
    """
    INTERVAL = 28800 # 8 hours

    def execute(self):
        pc = PriorityController(self.sa_session)
        pc.write_CSV()
        pc.send_CSV(loop_validator=self.loop_validator)

class PriorityTIFFTask(ScheduleTask):
    """
    Creates fully filled out TIFF files of the wave with the responses
    for the participants listed in the previously generated CSV.
    """
    INTERVAL = 21600 # 6 hours

    def execute(self):
        pc = PriorityController(self.sa_session)
        pc.send_TIFF(loop_validator=self.loop_validator)

class ImportResponsesTask(ScheduleTask):

    INTERVAL = 63

    def execute(self):
        """
        Imports responses from large files saved in ImportFile
        """
        results = self.sa_session.query(ImportFile).join(ImportFileType).filter(and_(
                    ImportFileType.name=='iqportal.tnt.response',
                    ImportFile.status==ImportFile.STATUS_NEW,
                    ImportFile.deleted==False,
                    ImportFile.archived==False)).all()
        if len(results) == 0:
            return

        bc = BackgroundImportResponseController(self.sa_session)
        impc = ImportFileController(self.sa_session)

        for file in results:
            impc.set_data(file, ImportFile.STATUS_STARTED)

        for file in results:
            try:
                resp, failures_file = bc.import_responses_from_task(file, loop_validator=self.loop_validator)
                impc.set_data(file, ImportFile.STATUS_SUCCESS, resp, failures_file)
            except Exception, e:
                impc.set_data(file, ImportFile.STATUS_FAILURE, e.message)
                raise

class ExportResponsesTask(ScheduleTask):

    ALLOW_DUPLICATES = True

    # we dont want to run multiple exports at the same time, see ticket #3735
    QUEUE_IF_EXISTS = True

    # task disabled, can only be triggered from interface
    INTERVAL = 0
    MAX_COMPARABLE_TASKS_IN_QUEUE = 10

    def execute(self):
        logger.info('ExportResponsesTask: starting export responses')

        wave_id = self.params['wave_id']
        assert isinstance(wave_id, int)

        country_id = self.params['country_id']
        assert isinstance(country_id, (int, NoneType))

        account_id = self.params['account_id']
        assert isinstance(account_id, int)

        if self.params.has_key('no_texts'):
            no_texts = self.params['no_texts']
        else:
            no_texts = False

        c = ResponseController(self.sa_session, account_id=account_id)
        hash = c.export_response_csv(wave_id, country_id, no_texts)

        # send a notificatiopn
        controller = NotificationController(self.sa_session, account_id=account_id)

        url = '/iqportal.tnt.clm/export_wave_csv/' + str(wave_id) + '/' + hash
        if country_id is not None:
            url = url + '/' + str(country_id)

        link = '<a href="%s" target="_blank">download</a>' % url

        notif = controller.set_notification(
            _("TNT response export"),
            _("The TNT reponse export is ready and can be downloaded here: %s") % link,
            callback={
                'href': url
            })
        controller.create_system_notification(notif, controller.get_default_group())

        logger.info('ExportResponsesTask: done export responses')

    def queue_full_notification(self):
        """
        The maximum number of queued exports has been reached, notify the user.
        """
        logger.info('ExportResponsesTask: queue is full, sending notification')

        wave_id = self.params['wave_id']
        assert isinstance(wave_id, int)

        country_id = self.params['country_id']
        assert isinstance(country_id, (int, NoneType))

        account_id = self.params['account_id']
        assert isinstance(account_id, int)

        if self.params.has_key('no_texts'):
            no_texts = self.params['no_texts']
        else:
            no_texts = False

        wc = WaveController(self.sa_session, account_id=account_id)
        wave = wc.get_waves(wave_id).pop()

        extra = ''

        if country_id is not None:
            lc = I18nController(self.sa_session, account_id=account_id)
            country = lc.get_domain_countries(country_id).pop()
            extra = ' (%s)' % country.name
        if no_texts:
            extra = ', without text responses, '

        message = 'The export for responses of %(wave)s%(extra)s could not be added to the queue as the queue is full. Please try again later.'
        message = message % ({'wave': wave.name, 'extra': extra})

        controller = NotificationController(self.sa_session, account_id=account_id)
        notif = controller.set_notification(_("TNT response export - queue full"), message)
        controller.create_system_notification(notif, controller.get_default_group(), True)

        logger.info('ExportResponsesTask: notification sent, %s' % message)

tasks = (PriorityCSVTask, PriorityTIFFTask, ImportResponsesTask, ExportResponsesTask)
