# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.

import math

from sqlalchemy import and_, desc, func

from basesoft.web.core.util.controllers import AbstractController, transaction_decorator
from basesoft.web.contrib.i18n.controllers import I18nController

from iqportal.tnt.core.controllers import TntController
from iqportal.tnt.sr.models import TntSRQualityAuditResponse, TntSRQualityAuditResponseGroup
from iqportal.tnt.core.models import TntDepot, TntDepotGroup

# same as the values used in the widget
CHART_TYPE_BOXPLOT = 'boxplot'
CHART_TYPE_SCATTER = 'scatter'

class ServiceRecoveryController(AbstractController):
    """
    Service Recovery Controller
    """

    def get_audit_responses(self, id=None, priv=TntSRQualityAuditResponse.PRIV_VIEW):
        """
        Gets audit responses from the database
        """
        self.assert_permissions('view_tnt_sr_audit_form')

        return self.get_privileged_object_by_id(TntSRQualityAuditResponse, id, priv)

    @transaction_decorator
    def save_or_update_audit_response_from_state(self, id, state, group=None,
                                       group_privileges=TntSRQualityAuditResponse.all_privileges,
                                       license_privileges=TntSRQualityAuditResponse.no_privileges,
                                       domain_privileges=TntSRQualityAuditResponse.no_privileges):
        """
        Saves or updates an audit response from a state dictionary
        """
        if id is None:
            self.assert_permissions('create_tnt_sr_audit_form')
            audit = TntSRQualityAuditResponse()
        else:
            self.assert_permissions('update_tnt_sr_audit_form')
            audit = self.get_audit_responses(id, TntSRQualityAuditResponse.PRIV_UPDATE).pop()

        if group is None:
            group = self._get_default_group()

        i18n_controller = I18nController(self.sa_session, account_id=self.account_id)
        core_controller = TntController(self.sa_session, account_id=self.account_id)

        audit.account = self._get_account()
        audit.group = group
        audit.group_privileges = group_privileges
        audit.license_privileges = license_privileges
        audit.domain_privileges = domain_privileges

        # Contact Information
        audit.supervisor_name = state['supervisor_name']
        audit.auditee = state['auditee']
        audit.date_of_audit = state['date_of_audit'].date()
        audit.country = i18n_controller.get_countries(state['country_id']).pop()

        # TODO change when BASESOFT.util.form.Combobox is refactored and is sending ids properly
        try:
            depot_id = int(state['depot'])
            depot = core_controller.get_depots(depot_id).pop()
        except ValueError:
            # we received the name from the Combo, so we need to get the id by using the name
            self.assert_permissions('view_depot')
            filter, fo = self.get_query_filter_on_privileges(TntDepot, TntDepotGroup, TntDepot.PRIV_VIEW)
            filter = and_(filter, TntDepot.default_code==state['depot'])
            depots = self.sa_session.query(TntDepot).select_from(fo).filter(filter).all()
            if len(depots)==0:
                # this is a new depot, so we need to create it
                depot = core_controller.save_depot_state({
                    'dbID': None, 'default_name': state['depot'], 'country': audit.country.id, 'service': None,  'new_aliases': [], 'default_code': '', 'label': '',
                })
            else:
                depot = depots.pop()

        audit.depot = depot

        audit.date_case_opened = state['date_case_opened'].date()
        audit.contact_number = state['contact_number']
        audit.contact_name = state['contact_name']
        audit.telephone_number = state['telephone_number']

        # Section I
        audit.customer_perspective_question1 = int(state['customer_perspective']['customer_perspective_question1'])
        audit.customer_perspective_question2 = int(state['customer_perspective']['customer_perspective_question2'])
        audit.customer_perspective_question3 = int(state['customer_perspective']['customer_perspective_question3'])
        audit.customer_perspective_question4 = int(state['customer_perspective']['customer_perspective_question4'])
        audit.customer_perspective_question5 = int(state['customer_perspective']['customer_perspective_question5'])
        audit.customer_perspective_question6 = int(state['customer_perspective']['customer_perspective_question6'])
        if state.has_key('customer_perspective_remark'):
            audit.customer_perspective_remark = state['customer_perspective_remark']

        # Section II
        audit.data_capture_question1 = state['data_capture_question1']
        audit.data_capture_question2_a = int(state['data_capture_question2']['data_capture_question2_a'])
        audit.data_capture_question2_b = int(state['data_capture_question2']['data_capture_question2_b'])
        audit.data_capture_question3_a = int(state['data_capture_question3']['data_capture_question3_a'])
        audit.data_capture_question3_b = int(state['data_capture_question3']['data_capture_question3_b'])
        audit.data_capture_question3_c = int(state['data_capture_question3']['data_capture_question3_c'])
        audit.data_capture_question3_d = int(state['data_capture_question3']['data_capture_question3_d'])

        # Section III
        audit.supervisor_perspective_question1 = state['supervisor_perspective_question1']
        audit.supervisor_perspective_question2 = state['supervisor_perspective_question2']
        audit.supervisor_perspective_question3 = state['supervisor_perspective_question3']

        self.sa_session.add(audit)
        return audit

    def load_audit_state(self, id):
        """
        Load the state for the answer with the given id
        """
        audits = self.get_audit_responses(id)
        if len(audits) == 0:
            return {}

        a = audits[0]

        state = {
            'supervisor_name': a.supervisor_name,
            'auditee': a.auditee,
            'date_of_audit': a.date_of_audit,
            'country_id': a.country.id,
            'depot': a.depot.id,
            'depot_name': a.depot.default_name, # we send this too, in case the user has a combobox (that doesn't use ids)
            'date_case_opened': a.date_case_opened,
            'contact_number': a.contact_number,
            'contact_name': a.contact_name,
            'telephone_number': a.telephone_number,

            'customer_perspective': {
                'customer_perspective_question1': a.customer_perspective_question1,
                'customer_perspective_question2': a.customer_perspective_question2,
                'customer_perspective_question3': a.customer_perspective_question3,
                'customer_perspective_question4': a.customer_perspective_question4,
                'customer_perspective_question5': a.customer_perspective_question5,
                'customer_perspective_question6': a.customer_perspective_question6
            },
            'customer_perspective_remark': a.customer_perspective_remark,

            'data_capture_question1': a.data_capture_question1,
            'data_capture_question2': {
                'data_capture_question2_a': a.data_capture_question2_a,
                'data_capture_question2_b': a.data_capture_question2_b
            },
            'data_capture_question3': {
                'data_capture_question3_a': a.data_capture_question3_a,
                'data_capture_question3_b': a.data_capture_question3_b,
                'data_capture_question3_c': a.data_capture_question3_c,
                'data_capture_question3_d': a.data_capture_question3_d
            },
            'supervisor_perspective_question1': a.supervisor_perspective_question1,
            'supervisor_perspective_question2': a.supervisor_perspective_question2,
            'supervisor_perspective_question3': a.supervisor_perspective_question3
        }

        return state

    @transaction_decorator
    def delete_audit_responses(self, id):
        """
        Delete the specified audit response(s)

        @param id: {int} or {List} of ints that represent the ids of the responses we want to delete
        """
        return self.mark_as_deleted(TntSRQualityAuditResponse, id)

    def get_country_boxplot_data(self, country_id, period, depot_id=None):
        """
        Return a list of the 6 values that are needed to build a boxplot for the given country(ies)
        @param country_id: None (in case of world), id or list of id's of countries
        @param period: datetime object (if we have a single day) or a list of 2 datetime objects if we need data for the
            range of dates between the 2 given dates
        @param depot_id: optional - id of the depot we want the data for. if it's given, then the depot is used as
            filter, not the countries
        @return: a list of 6 values representing [min, Q1, Q2, Q3, max, n], where values are calculated according to:
            http://en.wikipedia.org/wiki/Quartile , method 1. if there is no data for this countries the values will be
            None values (not represented on the chart). n is the number of records used for the given boxplot
        """
        self.assert_permissions('view_tnt_sr_audit_form')
        filter, fo = self.get_query_filter_on_privileges(TntSRQualityAuditResponse,
                                                         TntSRQualityAuditResponseGroup,
                                                         TntSRQualityAuditResponse.PRIV_VIEW)

        if depot_id is not None:
            # if we have a depot, we use that as filter
            filter = and_(filter, TntSRQualityAuditResponse.depot_id==depot_id)
        else:
            # if not, use the countries
            if country_id is not None:
                if not isinstance(country_id, list):
                    country_id = [country_id]
                filter = and_(filter, TntSRQualityAuditResponse.country_id.in_(country_id))

        # add the date restrictions
        if not isinstance(period, list):
            filter = and_(filter, TntSRQualityAuditResponse.date_of_audit == period)
        else:
            filter = and_(filter, TntSRQualityAuditResponse.date_of_audit >= period[0])
            filter = and_(filter, TntSRQualityAuditResponse.date_of_audit < period[1])

        q_base = self.sa_session.query(TntSRQualityAuditResponse).select_from(fo)
        q_ordered_asc = q_base.order_by(TntSRQualityAuditResponse.total_score, TntSRQualityAuditResponse.id)
        q_ordered_asc_filtered = q_ordered_asc.filter(filter)
        q_ordered_desc = q_base.order_by(desc(TntSRQualityAuditResponse.total_score), TntSRQualityAuditResponse.id)

        # count of the results for world
        n = q_base.filter(filter).count()

        # if we don't have any results, nothing should be shown on the chart
        if n == 0:
            return [None, None, None, None, None, n]

        mi = int(math.ceil(float(n)/2)) # position of the median

        # total score of the median value
        obj_m = q_ordered_asc.filter(filter).offset(mi - 1).limit(1).one() # offset is zero based
        m = obj_m.total_score

        # if we have only one result, all values are that, no need for extra calculations
        if n == 1:
            return [m, m, m, m, m, n]

        # min and max
        min = q_ordered_asc_filtered.limit(1).one().total_score
        max = q_ordered_desc.filter(filter).limit(1).one().total_score

        # if we have only 3 results, we can skip some calculations again
        if n == 3:
            return [min, min, m, max, max, n]

        # Q1
        filter_lower = and_(filter, TntSRQualityAuditResponse.id <= obj_m.id)
        n_lower = q_ordered_asc.filter(filter_lower).count()
        i_lower = int(math.ceil(float(n_lower)/2))
        q1 = q_ordered_asc_filtered.offset(i_lower - 1).limit(1).one().total_score

        # Q3
        filter_upper = and_(filter, TntSRQualityAuditResponse.id >= obj_m.id)
        n_upper = q_ordered_asc.filter(filter_upper).count()
        i_upper = int(math.ceil(float(n_upper)/2)) + mi # we're starting from the middle, so add the first half too
        q3 = q_ordered_asc_filtered.offset(i_upper - 1).limit(1).one().total_score

        return [min, q1, m, q3, max, n]

    def get_geolevel_boxplot_data(self, geolevel_id, period):
        """
        Return a list of the 5 values that are needed to build a boxplot for the given geolevel
        @param geolevel_id: id of the geolevel that we're interested in
        @param period: datetime object (if we have a single day) or a list of 2 datetime objects if we need data for the
            range of dates between the 2 given dates
        @return: a list of 5 values representing [min, Q1, Q2, Q3, max], where values are calculated according to:
            http://en.wikipedia.org/wiki/Quartile , method 1. if there is no data for this countries the values will be
            None values (not represented on the chart)
        """
        core_controller = TntController(self.sa_session, account_id=self.account_id)
        geolevel = core_controller.get_geolevels(geolevel_id)[0]

        country_ids = [country.id for country in geolevel.countries]
        return self.get_country_boxplot_data(country_ids, period)

    def get_depot_boxplot_data(self, depot_id, period):
        """
        Return a list of the 5 values that are needed to build a boxplot for the given depot
        @param depot_id: id of the depot that we're interested in
        @param period: datetime object (if we have a single day) or a list of 2 datetime objects if we need data for the
            range of dates between the 2 given dates
        @return: a list of 5 values representing [min, Q1, Q2, Q3, max], where values are calculated according to:
            http://en.wikipedia.org/wiki/Quartile , method 1. if there is no data for this countries the values will be
            None values (not represented on the chart)
        """
        core_controller = TntController(self.sa_session, account_id=self.account_id)
        core_controller.get_depots(depot_id)[0]

        return self.get_country_boxplot_data(None, period, depot_id=depot_id)

    def get_country_scatter_data(self, country_id, period, depot_id=None):
        """
        Returns data for the desired countries so it can be used for the scatter chart
        @param country_id: can be None (for world), an int or a list of int. the ids of the countries we want data for
        @param period: datetime object (if we have a single day) or a list of 2 datetime objects if we need data for the
            range of dates between the 2 given dates
        @param depot_id: optional - id of the depot we want the data for. if it's given, then the depot is used as
            filter, not the countries

        @return: a list of values for the requested data, each value being a list that has the score, a datetime
            object for the date of audit and the number of results that have that score
        """

        self.assert_permissions('view_tnt_sr_audit_form')
        filter, fo = self.get_query_filter_on_privileges(TntSRQualityAuditResponse,
                                                         TntSRQualityAuditResponseGroup,
                                                         TntSRQualityAuditResponse.PRIV_VIEW)

        if depot_id is not None:
            # if we have a depot, we use that as filter
            filter = and_(filter, TntSRQualityAuditResponse.depot_id==depot_id)
        else:
            # if not, use the countries
            if country_id is not None:
                if not isinstance(country_id, list):
                    country_id = [country_id]
                filter = and_(filter, TntSRQualityAuditResponse.country_id.in_(country_id))

        # add the date restrictions
        if not isinstance(period, list):
            filter = and_(filter, TntSRQualityAuditResponse.date_of_audit == period)
        else:
            filter = and_(filter, TntSRQualityAuditResponse.date_of_audit >= period[0])
            filter = and_(filter, TntSRQualityAuditResponse.date_of_audit < period[1])

        q = self.sa_session.query(
            TntSRQualityAuditResponse.total_score, TntSRQualityAuditResponse.date_of_audit, func.count()).select_from(
            fo).filter(filter).group_by(TntSRQualityAuditResponse.total_score, TntSRQualityAuditResponse.date_of_audit).all()

        return q

    def get_geolevel_scatter_data(self, geolevel_id, period):
        """
        Returns data for the desired geolevel so it can be used for the scatter chart
        @param geolevel_id: id of the desired geolevel
        @param period: datetime object (if we have a single day) or a list of 2 datetime objects if we need data for the
            range of dates between the 2 given dates

        @return: a list of values for the requested data, each value being a list that has the score, and a datetime
            object for the date of audit
        """
        core_controller = TntController(self.sa_session, account_id=self.account_id)
        geolevel = core_controller.get_geolevels(geolevel_id)[0]

        country_ids = [country.id for country in geolevel.countries]
        return self.get_country_scatter_data(country_ids, period)

    def get_depot_scatter_data(self, depot_id, period):
        """
        Returns data for the desired geolevel so it can be used for the scatter chart
        @param depot_id: id of the desired depot
        @param period: datetime object (if we have a single day) or a list of 2 datetime objects if we need data for the
            range of dates between the 2 given dates

        @return: a list of values for the requested data, each value being a list that has the score, and a datetime
            object for the date of audit
        """
        core_controller = TntController(self.sa_session, account_id=self.account_id)
        core_controller.get_depots(depot_id)[0]
        return self.get_country_scatter_data(None, period, depot_id=depot_id)

from basesoft.web.dev.util.controllers import IQuestPHP4AbstractConversionController

class TntServiceRecoveryPHP4ConversionController(IQuestPHP4AbstractConversionController):
    """
    """
    SR_PUBLICATION_ID = 5

    def __init__(self, sa_session=None, account_id=None):
        AbstractController.__init__(self, sa_session, account_id=account_id)
        self.assert_permissions('develop_tools')

    @transaction_decorator
    def convert_user_country(self, user_map):
        """
        Converts the users country link from SR to the IQPortal setup
        """
        from basesoft.web.contrib.contact.models import Address

        country_map = self.get_country_map()

        sql = """
            SELECT USER_ID, COUNTRY_ID FROM TBL_MANAGER
        """
        result = self.db_edit.execute(sql)
        for row in result:
            try:
                account = user_map['user_%s' % row['USER_ID']]
                country = country_map['sr_country_%s' % row['COUNTRY_ID']]
            except:
                continue
            account.person.address = Address()
            account.person.address.country = country
            self.session.add(account.person.address)

    @transaction_decorator
    def _convert_depot(self, target_group, country, depot_code):
        """
        """
        q = self.sa_session.query(TntDepot).filter(and_(and_(
            TntDepot.group==target_group, TntDepot.country==country),
            TntDepot.default_code==depot_code
        ))
        try:
            depot = q.one()
            return depot
        except:
            pass

        depot = TntDepot()
        depot.default_code = depot_code
        depot.group = target_group
        depot.group_privileges = TntDepot.all_privileges
        depot.license_privileges = TntDepot.no_privileges
        depot.domain_privileges = TntDepot.no_privileges
        depot.country = country
        self.sa_session.add(depot)
        return depot

    @transaction_decorator
    def convert_geolevels(self, target_group):
        """
        Queries the Regions from TNT SR database and inserts them as new ones in the IQPortal db.
        """
        pass

    def get_country_map(self, map_col = 'COUNTRY_ID'):
        """
        We need a map from TNT SR Country IDs to IQPortal Country
        """
        from basesoft.web.contrib.i18n.models import Country

        map = {}

        sql = """
            SELECT COUNTRY_ID, COUNTRY_CODE, COUNTRY_NAME FROM TBL_COUNTRY
        """

        result = self.db_edit.execute(sql)
        for row in result:
            if row['COUNTRY_CODE'] == '':
                continue
            try:
                c = self.sa_session.query(Country).filter(Country.iso_3166_a2==row['COUNTRY_CODE']).one()
            except:
                # TODO: log this country exception
                print 'failed: %s' % row['COUNTRY_CODE']
                continue
            map['sr_country_%s' % row[map_col]] = c

        return map

    @transaction_decorator
    def convert_responses(self, target_group, user_map):
        """
        Convert current responses for SR_PUBLICATION_ID to TntSRQualityAuditResponses
        """
        qa_map = self._get_qa_map()
        country_map = self.get_country_map()

        # select the results
        sql = """
            SELECT R.LOGIN_ID, DATE(L.LOGIN_DATE) as date_of_audit, L.USER_ID, QUESTION_ID, SUBQUESTION_ID, ANSWER_ID, RESPONSE_OPENANSWER
            FROM TBL_RESPONSE as R
            INNER JOIN TBL_LOGIN as L
                ON R.LOGIN_ID = L.LOGIN_ID
            WHERE PUBLISH_ID = %s
            ORDER BY R.LOGIN_ID, ANSWER_ID
        """

        # this comes from the edit db so we cannot do a join.
        country_sql = """
            SELECT COUNTRY_ID FROM TBL_MANAGER WHERE USER_ID = %s
        """

        result = self.db_publish.execute(sql % self.SR_PUBLICATION_ID)
        login_id = None
        audit = None

        answer_map = {
            'q77': {
                'a1065': 5,
                'a1066': 4,
                'a1067': 3,
                'a1068': 2,
                'a1069': 1,
                'a1070': 5,
                'a1071': 4,
                'a1072': 3,
                'a1073': 2,
                'a1074': 1,
                'a1075': 5,
                'a1076': 4,
                'a1077': 3,
                'a1078': 2,
                'a1079': 1,
                'a1080': 5,
                'a1081': 4,
                'a1082': 3,
                'a1083': 2,
                'a1084': 1,
                'a1085': 5,
                'a1086': 4,
                'a1087': 3,
                'a1088': 2,
                'a1089': 1,
                'a1090': 5,
                'a1091': 4,
                'a1092': 3,
                'a1093': 2,
                'a1094': 1,
            },
            'q80': {
                'a1095': 2,
                'a1096': 1,
                'a1097': 0,
                'a1098': 2,
                'a1099': 1,
                'a1100': 0,
            },
            'q81': {
                'a1101': 2,
                'a1102': 0,
                'a1103': 2,
                'a1104': 0,
                'a1105': 2,
                'a1106': 0,
                'a1107': 2,
                'a1108': 0,
            },
            'q83': {
                'a1109': 'yes',
                'a1110': 'no'
            }
        }

        for row in result:
            if login_id != row['LOGIN_ID']:
                if audit is not None and audit.depot is not None:
                        self.sa_session.add(audit)

                audit = TntSRQualityAuditResponse()
                audit.group = target_group
                audit.group_privileges = TntSRQualityAuditResponse.all_privileges
                audit.license_privileges = TntSRQualityAuditResponse.no_privileges
                audit.domain_privileges = TntSRQualityAuditResponse.no_privileges
                audit.date_of_audit = row['date_of_audit']
                audit.customer_perspective_question1 = 0
                audit.customer_perspective_question2 = 0
                audit.customer_perspective_question3 = 0
                audit.customer_perspective_question4 = 0
                audit.customer_perspective_question5 = 0
                audit.customer_perspective_question6 = 0
                audit.data_capture_question2_a = 0
                audit.data_capture_question2_b = 0
                audit.data_capture_question3_a = 0
                audit.data_capture_question3_b = 0
                audit.data_capture_question3_c = 0
                audit.data_capture_question3_d = 0

                country_row = self.db_edit.execute(country_sql % row['USER_ID']).fetchone()
                if country_row is None:
                    audit = None
                    continue
                country_id = country_row['COUNTRY_ID']
                try:
                    audit.country = country_map['sr_country_%s' % country_id]
                except KeyError:
                    audit = None
                    continue

                try:
                    audit.account = user_map['user_%s' % row['USER_ID']]
                except KeyError:
                    audit = None
                    continue

            login_id = row['LOGIN_ID']
            if audit is None:
                continue

            q = row['QUESTION_ID']
            sq = row['SUBQUESTION_ID']
            a = row['ANSWER_ID']
            try:
                response = unicode(row['RESPONSE_OPENANSWER'], "utf-8")
            except:
                response = row['RESPONSE_OPENANSWER']
            if sq is None:
                sq = 0

            # we are dealing with open questions.
            if q not in [77, 80, 81, 83]:
                # country comes from manager table.
                if a == 1059:
                    continue
                # depot field
                if a == 1058:
                    audit.depot = self._convert_depot(target_group, audit.country, response)
                    continue
                # audit date comes from the login table.
                if a == 1057:
                    continue

                col = qa_map['qa%s_%s_%s' % (q, sq, a)]
                setattr(audit, col, response)
                continue

            col = qa_map['qa%s_%s_%s' % (q, sq, 0)]
            value = answer_map['q%s' % q]['a%s' % a]
            setattr(audit, col, value)

    def _get_qa_map(self):
        """
        Map question/answers to columns in tblTntSRQualityAuditResponse
        """

        return {
            'qa76_0_1111': 'supervisor_name',
            'qa76_0_1057': 'date_of_audit', # audit date comes from the login table.
            'qa76_0_1059': 'country_id',
            'qa76_0_1058': 'depot_id',
            'qa76_0_1060': 'auditee',
            'qa76_0_1061': 'date_case_opened',
            'qa76_0_1062': 'contact_number',
            'qa76_0_1063': 'contact_name',
            'qa76_0_1064': 'telephone_number',

            # Section two: Customer Perspective
            'qa77_121_0': 'customer_perspective_question1',
            'qa77_122_0': 'customer_perspective_question2',
            'qa77_123_0': 'customer_perspective_question3',
            'qa77_124_0': 'customer_perspective_question4',
            'qa77_125_0': 'customer_perspective_question5',
            'qa77_126_0': 'customer_perspective_question6',
            'qa78_0_0': 'customer_perspective_remark',

            # Section three: Data Capture Section
            'qa79_0_0': 'data_capture_question1',
            'qa80_127_0': 'data_capture_question2_a',
            'qa80_128_0': 'data_capture_question2_b',
            'qa81_129_0': 'data_capture_question3_a',
            'qa81_130_0': 'data_capture_question3_b',
            'qa81_131_0': 'data_capture_question3_c',
            'qa81_132_0': 'data_capture_question3_d',

            # Section four: CS Manager / Supervisor Perspective
            'qa82_0_0': 'supervisor_perspective_question1',
            'qa83_0_0': 'supervisor_perspective_question2',
            'qa84_0_0': 'supervisor_perspective_question3'
        }

