# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.

import datetime
from sqlalchemy.orm import mapper, relation
from sqlalchemy import Column, Integer, String, ForeignKey, DateTime, Date, UnicodeText, SmallInteger, Enum
from basesoft.sqlalchemy.table import IQTable
from django.utils.translation import ugettext as _
from basesoft.web.contrib.license import models as license_models

__mapping_defined = False
__tables_defined = False
tables = {}

def reset():
    global __mapping_defined
    global __tables_defined

    __mapping_defined = False
    __tables_defined = False

def get_tables():
    """
    """
    return tables

def do_tables(metadata):
    global __tables_defined
    if not __tables_defined:

        tblTntSRQualityAuditResponse = IQTable('tblTntSRQualityAuditResponse', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('account_id', Integer, ForeignKey('tblAccount.id'), nullable=False),

            # Section one: Contact Details
            Column('supervisor_name', UnicodeText, nullable=True),
            Column('date_of_audit', Date(), nullable=False),
            Column('country_id', Integer, ForeignKey('tblCountry.id'), nullable=False),
            Column('depot_id', Integer, ForeignKey('tblTntDepot.id'), nullable=False),
            Column('auditee', UnicodeText, nullable=False),
            Column('date_case_opened', Date(), nullable=True),
            Column('contact_number', Integer, nullable=True),
            Column('contact_name', UnicodeText, nullable=True),
            Column('telephone_number', String(32), nullable=True),

            # Section two: Customer Perspective
            Column('customer_perspective_question1', SmallInteger, nullable=False, default=0),
            Column('customer_perspective_question2', SmallInteger, nullable=False, default=0),
            Column('customer_perspective_question3', SmallInteger, nullable=False, default=0),
            Column('customer_perspective_question4', SmallInteger, nullable=False, default=0),
            Column('customer_perspective_question5', SmallInteger, nullable=False, default=0),
            Column('customer_perspective_question6', SmallInteger, nullable=False, default=0),
            Column('customer_perspective_score', Integer, nullable=False), # read-only value, see hooks.py
            Column('customer_perspective_remark', UnicodeText, nullable=True),

            # Section three: Data Capture Section
            Column('data_capture_question1', UnicodeText, nullable=True),
            Column('data_capture_question2_a', SmallInteger, nullable=False, default=0),
            Column('data_capture_question2_b', SmallInteger, nullable=False, default=0),
            Column('data_capture_question3_a', SmallInteger, nullable=False, default=0),
            Column('data_capture_question3_b', SmallInteger, nullable=False, default=0),
            Column('data_capture_question3_c', SmallInteger, nullable=False, default=0),
            Column('data_capture_question3_d', SmallInteger, nullable=False, default=0),
            Column('data_capture_score', Integer, nullable=False), # read-only value, see hooks.py

            Column('total_score', Integer, nullable=False), # read-only value, see hooks.py

            # Section four: CS Manager / Supervisor Perspective
            Column('supervisor_perspective_question1', UnicodeText, nullable=True),
            Column('supervisor_perspective_question2', Enum('yes', 'no'), nullable=True),
            Column('supervisor_perspective_question3', UnicodeText, nullable=True),

            Column('deleted', SmallInteger, nullable=False, default=0),
            Column('archived', SmallInteger, nullable=False, default=0),

            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now)
        )

        tblTntSRQualityAuditResponseGroup = IQTable('tblTntSRQualityAuditResponseGroup', metadata,
            Column('response_id', Integer, ForeignKey('tblTntSRQualityAuditResponse.id', ondelete="CASCADE"), primary_key=True),
            Column('group_id', Integer, ForeignKey('tblGroup.id'), primary_key=True),
            Column('privilege', Integer, nullable=False, primary_key=True, autoincrement=False),
            Column('group_has_privilege', SmallInteger, default=0),
            Column('license_has_privilege', SmallInteger, default=0),
            Column('domain_has_privilege', SmallInteger, default=0),
            Column('world_has_privilege', SmallInteger, default=0),
            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now)
        )

        tables['tblTntSRQualityAuditResponse'] = tblTntSRQualityAuditResponse
        tables['tblTntSRQualityAuditResponseGroup'] = tblTntSRQualityAuditResponseGroup

        __tables_defined = True


def do_mappings():
    """
    This function is only called by the django middleware
    when it is needed
    """
    global __mapping_defined
    if not __mapping_defined:
        from basesoft.web.contrib.license.hooks import PrivilegeObjectMapper
        from basesoft.web.contrib.license import models as license_models
        from basesoft.web.contrib.i18n import models as i18n_models
        from basesoft.web.contrib.account import models as account_models
        from iqportal.tnt.sr import models as sr_models
        from iqportal.tnt.core import models as core_models
        from iqportal.tnt.sr.hooks import TntSRQualityAuditResponseHook

        mapper(sr_models.TntSRQualityAuditResponse, tables['tblTntSRQualityAuditResponse'],
            extension = [TntSRQualityAuditResponseHook(), PrivilegeObjectMapper()],
            properties = {
                'account': relation(account_models.Account),
                'groups': relation(sr_models.TntSRQualityAuditResponseGroup, backref='response', cascade="all", passive_deletes=True),
                'country': relation(i18n_models.Country),
                'depot': relation(core_models.TntDepot)
            }
        )

        mapper(sr_models.TntSRQualityAuditResponseGroup, tables['tblTntSRQualityAuditResponseGroup'],
            properties = {
                'group': relation(license_models.Group)
            }
        )

        __mapping_defined = True

def fill_data(conn):
    pass

def app_permissions_category():
    """
    Returns a category for all permissions defined for this app
    """
    return license_models.PermissionCategory(u'TNT-SR')

def app_permissions():
    """
    The permissions used within this application. Can be sync'd with the
    `iqsyncpermissions` command.

    Titles are with gettext calls, but in the database they appear in English. We need it like this so we get the
    titles in the language files. Later, in the js interfaces we show them translated.
    """
    return [
        license_models.Permission(u'create_tnt_sr_audit_form', u"""
            This gives the right to create a Quality Audit.
            """, _("Create quality audit")),
        license_models.Permission(u'update_tnt_sr_audit_form', u"""
            This gives the right to update a Quality Audit.
            """, _("Update quality audit")),
        license_models.Permission(u'delete_tnt_sr_audit_form', u"""
            This gives the right to delete a Quality Audit.
            """, _("Delete quality audit")),
        license_models.Permission(u'view_tnt_sr_audit_form', u"""
            This gives the right to view a Quality Audit.
            """, _("View quality audit")),
    ]
