# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.

from basesoft.web.core.jsonrpc.dispatcher import JsonRpcDispatcher

from iqportal.tnt.sr.controllers import ServiceRecoveryController, CHART_TYPE_BOXPLOT, CHART_TYPE_SCATTER
from iqportal.tnt.core.controllers import TntController
from basesoft.web.contrib.i18n.controllers import I18nController
from basesoft.web.contrib.account.decorators import login_required
from basesoft.web.core.util import custom_query as cq
from django.utils.translation import ugettext as _
from datetime import datetime
from basesoft.web.core.util.custom_query import FILTER_EQUAL, FILTER_BETWEEN
import math

@login_required
def audit_form_validation_rpc(request):

    from basesoft.web.core.forms.handler import FormValidator
    from basesoft.web.core.input.validation import plugins

    form_handler = FormValidator({
        'supervisor_name': [(plugins.Required, {}), (plugins.IQPortalName, {})],
        'date_of_audit': [(plugins.Required, {})],
        'date_case_opened': [(plugins.Required, {})],
        'depot': [(plugins.Required, {})],
    })

    dispatcher = JsonRpcDispatcher()
    dispatcher.register_instance(form_handler)
    return dispatcher.handle_request(request)

@login_required
def audit_form_view(request):

    class Handler:
        def __init__(self, sa_session, account_id):
            self.controller = ServiceRecoveryController(sa_session, account_id=account_id)

        def save_or_update_audit_response_from_state(self, id, state):
            audit = self.controller.save_or_update_audit_response_from_state(id, state)
            return audit.id

        def load_audit_response_state(self, id):
            audit_state = self.controller.load_audit_state(id)
            return audit_state

        def delete_audit_responses(self, id):
            self.controller.delete_audit_responses(id)

    dispatcher = JsonRpcDispatcher()
    dispatcher.register_instance(Handler(request.Session(), request.session['user_id']))
    return dispatcher.handle_request(request)

@login_required
def audit_form_widget_rpc(request):
    user_id = request.session['user_id']
    sa_session = request.Session()
    sr_controller = ServiceRecoveryController(sa_session, account_id=user_id)
    core_controller = TntController(sa_session, account_id=user_id)
    i18n_controller = I18nController(sa_session, account_id=user_id)

    class TntSrWidgetRpc:
        """
        Class responsible for handling the rpc calls from the widget
        """
        def get_filters_and_group_by(self, static_filters=None):
            """
            Provide the widget with the proper filters
            """
            geolevel_options = [{'value': e.id, 'label': e.name} for e in core_controller.get_geolevels()]
            country_options = [{'value': e.id, 'label': e.iso_3166_a2} for e in i18n_controller.get_countries()]
            depot_options = [{'value': e.id, 'label': e.default_name} for e in core_controller.get_depots()]

            return {
                'filters':  {
                    'world': {
                        'type': cq.DATA_TYPE_BOOLEAN,
                        'label': _('World')
                    },
                    'geolevel': {
                        'type': cq.DATA_TYPE_INT,
                        'label': _('Geolevel'),
                        'options': geolevel_options
                    },
                    'country': {
                        'type': cq.DATA_TYPE_INT,
                        'label': _('Country'),
                        'options': country_options
                    },
                    'depot': {
                        'type': cq.DATA_TYPE_INT,
                        'label': _('Depot'),
                        'options': depot_options
                    },
                    'date_of_audit': {
                        'type': cq.DATA_TYPE_DATE,
                        'label': _('Date of audit')
                    },
                },
                'group_by': {}
            }

        def _get_default_period(self):
            """
            Return the default period in case we haven't received filters for the date of audit.
            @return: a list that contains a list of start and end dates for the current month
            """
            now = datetime.now()
            y1 = now.year
            m1 = now.month
            y2 = y1
            m2 = m1 + 1
            if m2 == 13:
                m2 = 1
                y2 = y2 + 1
            return [[datetime(y1, m1, 1), datetime(y2, m2, 1)]]

        def _get_formated_data(self, d):
            """
            TODO: use the user preferences to return the data
            Format a date acording to the locale
            @param d: a datetime object or a list of 2 datetime objects (in case of a period)
            @return: a string that represents the date
            """
            if not isinstance(d, list):
                return d.strftime('%d/%m/%y')
            else:
                return "%s - %s" % (d[0].strftime('%d/%m/%y'), d[1].strftime('%d/%m/%y'))

        def _get_named_data(self, period):
            """
            Show the name of the period (Q1, Month name, Full year, etc)
            @param period: a list of 2 datetime objects
            TODO: make this nicer maybe, because currently we're replicating the method from form-manager, js-land
            """
            quarter_labels = {
                'q1': _('Q1'),
                'q2': _('Q2'),
                'q3': _('Q3'),
                'q4': _('Q4')
            }

            month_labels = [
                _('January'),
                _('February'),
                _('March'),
                _('April'),
                _('May'),
                _('June'),
                _('July'),
                _('August'),
                _('September'),
                _('October'),
                _('November'),
                _('December')
            ]

            d1 = period[0]
            d2 = period[1]

            if (d2.month - d1.month == 3) or ((d2.year == d1.year + 1) and (d1.month == 10) and (d2.month == 1)):
                # We are dealing with a quarter.
                q = int(math.floor((d1.month - 1) / 3)) + 1
                handle = 'q%s' % q
                text = quarter_labels[handle]
                text = '%s %s' % (text, d1.year)
            elif (d2.month - d1.month == 1) or ((d2.year == d1.year + 1) and (d1.month == 12) and (d2.month == 1)):
                # dealing with a month
                text = month_labels[d1.month - 1];
                text = '%s %s' % (text, d1.year)
            else:
                # dealing with a full year
                text = '%s' % d1.year

            return text

        def _get_title(self, title_base, period=None, figure=None, new_line=False):
            """
            Create the title for the chart based on the base title and the period. If None, we only show the period and
                figure, because we're showing the title as the chart title, not label
            @param title_base: the string that's used for the title
            @param period: optional. a datetime object or a list of 2 datetime objects (in case of range)
            @param figure: optional. a number that represents the figure for the given entity
            @param new_line: a boolean that indicates if the period should be written on a new line
            @return: a string of form 'Title (formated_date)' or 'Title (first_date_of_range - last_date_of_range)'
                or just the title if we don't specify any periods. if new_line=True, the result will be
                'Title\nformated_date' or 'Title\nfirst_date_of_range - last_date_of_range'
            """
            if title_base is not None:
                title = title_base
            else:
                title = ''

            if figure is not None:
                format_str = ' #%s'
                if title_base is None:
                    format_str = '#%s' # no need for the separating space if we don't show the unit name in the label
                title = title + format_str % figure

            # TODO: format the date according to the locale
            if period is not None:
                period_formatted = self._get_formated_data(period)
                period_named = self._get_named_data(period)

                if period_formatted in self.period_mapping.values():
                    # we already have this date in the legend, so use the abbreviation it has
                    period_abbreviation = [k for k in self.period_mapping.keys() \
                                           if self.period_mapping[k] == period_formatted][0]
                else:
                    # new date, so we calculate an abbreviation for it
                    index = 1
                    period_abbreviation = 'P' + str(index)
                    while self.period_mapping.has_key(period_abbreviation):
                        index = index + 1
                        period_abbreviation = 'P' + str(index)

                    # now let's also add the date to the legend
                    self.period_mapping[period_abbreviation] = period_formatted
                    self.period_order.append(period_abbreviation)

                format_str = ' %s'
                if new_line:
                    format_str = "\n%s"

                #title = title + format_str % period_abbreviation # shows abbreviation, not the period
                title = title + format_str % period_named # shows the period, not the abbreviation

            return title

        def get_chart_data(self, chart_type, filters):
            """
            Return the data based on what kind of chart we have and what filters are selected
            @param chart_type:
            """
            # data to be shown, in the format expected by each chart_type
            # - 'boxplot' expects a list of lists, each list containing 6 values, for min, min standard deviation,
            # average, max standard deviation, max and the number of responses for each boxplot
            # - 'scatter' expects a list of lists, each list containing all the scores for that unit
            data = []
            self.period_mapping = {} # a mapping for the periods, that has periods as values and their abbreviations
                # (P1, P2, etc) as keys this will also be used for generating the periods legend

            chart_title = '' # in case we have a single unit, we show that as title and we remove if from the labels
                # to avoid redundancy
            labels = [] # labels shown for the chart

            messages = [] # messages to be shown to the user (as growls) in case some filter were incorrect, etc
            legend_mapping = {} # a mapping for entity names (used for legends in case of scatter), so we don't send
                # the name of the entity (world, country, depot, geolevel) with each dot
            date_order = [] # the cronological order for the dates in case we have a scatter chart
            self.period_order = [] # the correct order for the periods in the legend

            filter_keys = filters.keys()

            # a list containing dates or lists with 2 dates (in case of between) for the periods that we need data for
            periods = []

            # check if we have the date set
            if 'date_of_audit' in filter_keys:
                filter_keys.remove('date_of_audit')

                date_filter = filters['date_of_audit']
                # check if we received proper data
                if date_filter['operator'] == FILTER_EQUAL:
                    for d in date_filter['values']: # each d should be a datetime object
                        if isinstance(d, datetime):
                            periods.append(d)
                elif date_filter['operator'] == FILTER_BETWEEN:
                    for di in date_filter['values']: # each di should be a list with two datetime objects
                        if isinstance(di[0], datetime) and isinstance(di[1], datetime):
                            periods.append(di)
                else:
                    messages.append(_("You must specify the operator for the date of audit."))
                    periods = self._get_default_period()

                    if chart_type == CHART_TYPE_BOXPLOT:
                        self._get_title('', periods[0]) # for setting the period & it's abbreviation in the legend

                # check if we have any valid dates
                if len(periods) == 0:
                    # the data we received was not all valid, so we use the default interval and show a message
                    messages.append(_("You didn't specify any valid date for the date of audit filter."))
                    periods = self._get_default_period()

                    if chart_type == CHART_TYPE_BOXPLOT:
                        self._get_title('', periods[0]) # for setting the period & it's abbreviation in the legend

                elif len(periods) == 1:
                    # we have only one date, or range, so we add it to the legend
                    if chart_type == CHART_TYPE_BOXPLOT:
                        self._get_title('', periods[0]) # for setting the period & it's abbreviation in the legend

                # if we have more dates, we show them appended to the units (world, countries, etc)

            else:
                # we didn't receive and date, so we use the last month as an interval
                periods = self._get_default_period()
                if chart_type == CHART_TYPE_BOXPLOT:
                    self._get_title('', periods[0]) # for setting the period & it's abbreviation in the legend

            # check if we have anything else set (world, geolevel, countries, depots)
            if len(filter_keys) == 0:
                # we don't have anything, so let's show the data for world
                filter_keys.append('world')

            units = 0
            for fk in filter_keys: # we might have filters with more values, so that's not just one unit
                if filters.has_key(fk):
                    units = units + len(filters[fk]['values'])

            if filters.has_key('world') and filters['world']['values'][0] == False:
                units = units - 1 # we received world in the filters, but it was unchecked, so we don't show that

            # boxplots return a list of lists, each list representing one unit (world, geolevel, country or depot) and
            # contains 5 values: min, first quartile, median (second quartile), third quartile, max
            # Q1, Q2, Q3 (the quartiles) are calculated acording to http://en.wikipedia.org/wiki/Quartile , Method 1
            if chart_type == CHART_TYPE_BOXPLOT:
                # for each period we make a set of units. so if we have 4 units and 2 periods, we'll show 8 boxplots
                for period in periods:
                    # need to analyze every period

                    # we have some entities, so we show data for them
                    for entity in filter_keys:

                        if entity == 'world':
                            # we want the world in the chart

                            # need to check if we have world because it was added by default (filters were empty) or we
                            # received it in the filters (in which case it might be with the False value, so we need to
                            # check that)
                            if filters.has_key('world') and filters['world']['values'][0] == False:
                                continue

                            d = sr_controller.get_country_boxplot_data(None, period)
                            # last value of d is the number of results for this boxplot

                            if units == 1:
                                label = None # we show the name of the unit in the chart title
                                chart_title = _('World')
                            else:
                                label = _('World')

                            labels.append(self._get_title(label, period, d.pop(), True)) # always show the period for now
                            #if len(periods) > 1:
                            #    # we have more periods, so we need to add the period to the unit
                            #    labels.append(self._get_title(label, period, d.pop(), True))
                            #else:
                            #    labels.append(self._get_title(label, None, d.pop(), True))

                            data.append(d)

                        elif entity == 'country':
                            # we want countries (we might have more than one), and country ids arrive as strings
                            country_ids = filters['country']['values']
                            for c in country_ids:
                                # let's get the country object
                                country = i18n_controller.get_countries(int(c))[0]

                                d = sr_controller.get_country_boxplot_data(country.id, period)

                                if units == 1:
                                    label = None # we show the name of the unit in the chart title
                                    chart_title = country.iso_3166_a2
                                else:
                                    label = country.iso_3166_a2


                                labels.append(self._get_title(label, period, d.pop(), True)) # always show the period for now
                                #if len(periods) > 1:
                                #    # we have more periods, so we need to add the period to the unit
                                #    labels.append(self._get_title(label, period, d.pop(), True))
                                #else:
                                #    labels.append(self._get_title(label, None, d.pop(), True))

                                data.append(d)

                        elif entity == 'geolevel':
                            # we want geolevels (we might have more than one), and geolevel ids arrive as strings
                            geolevel_ids = filters['geolevel']['values']
                            for g in geolevel_ids:
                                # let's get the geolevel object
                                geolevel = core_controller.get_geolevels(int(g))[0]

                                d = sr_controller.get_geolevel_boxplot_data(geolevel.id, period)

                                if units == 1:
                                    label = None # we show the name of the unit in the chart title
                                    chart_title = geolevel.name
                                else:
                                    label = geolevel.name

                                labels.append(self._get_title(label, period, d.pop(), True)) # always show the period for now
                                #if len(periods) > 1:
                                #    # we have more periods, so we need to add the period to the unit
                                #    labels.append(self._get_title(label, period, d.pop(), True))
                                #else:
                                #    labels.append(self._get_title(label, None, d.pop(), True))

                                data.append(d)

                        elif entity == 'depot':
                            # we want depots (we might have more than one), and depot ids arrive as strings
                            depot_ids = filters['depot']['values']
                            for d in depot_ids:
                                # let's get the geolevel object
                                depot = core_controller.get_depots(int(d))[0]

                                d = sr_controller.get_depot_boxplot_data(depot.id, period)

                                if units == 1:
                                    label = None # we show the name of the unit in the chart title
                                    chart_title = depot.name
                                else:
                                    label = depot.name

                                labels.append(self._get_title(label, period, d.pop(), True)) # always show the period for now
                                #if len(periods) > 1:
                                #    # we have more periods, so we need to add the period to the unit
                                #    labels.append(self._get_title(label, period, d.pop(), True))
                                #else:
                                #    labels.append(self._get_title(label, None, d.pop(), True))

                                data.append(d)

            if chart_type == CHART_TYPE_SCATTER:
                # for each day we generate a list of values (from all units. units will have different colors)
                data = {} # keys are the formatted dates
                entities = {} # a mapping so we can check what ids we associate with each region
                entity_id = 0 # a counter so we can keep track of what id we associate with a new entity

                for period in periods:
                    # need to analyze every period

                    # we have some entities, so we show data for them
                    for entity in filter_keys:

                        if entity == 'world':
                            # we want the world in the chart

                            # need to check if we have world because it was added by default (filters were empty) or we
                            # received it in the filters (in which case it might be with the False value, so we need to
                            # check that)
                            if filters.has_key('world') and filters['world']['values'][0] == False:
                                continue

                            world_data = sr_controller.get_country_scatter_data(None, period)

                            # get the id of 'world' so we can send it so we can generate the legend
                            if entities.has_key('world'):
                                id = entities['world']
                            else:
                                entity_id = entity_id + 1
                                id = entity_id
                                entities['world'] = id
                                legend_mapping[id] = _('World') # all data that has this id is from world

                            for d in world_data: # each d is a list like [score, date_of_audit, count]
                                dat = self._get_formated_data(d[1])
                                if not data.has_key(dat):
                                    data[dat] = []
                                    labels.append(dat)
                                    date_order.append(d[1])
                                data[dat].append([d[0], id, d[2]])

                        elif entity == 'country':
                            # we want countries (we might have more than one), and country ids arrive as strings
                            country_ids = filters['country']['values']
                            for c in country_ids:
                                # let's get the country object
                                country = i18n_controller.get_countries(int(c))[0]

                                country_data = sr_controller.get_country_scatter_data(country.id, period)

                                # get the id of this country so we can generate the legend
                                if entities.has_key(country.name):
                                    id = entities[country.name]
                                else:
                                    entity_id = entity_id + 1
                                    id = entity_id
                                    entities[country.name] = id
                                    legend_mapping[id] = country.name

                                for d in country_data: # each d is a list like [score, date_of_audit, count]
                                    dat = self._get_formated_data(d[1])
                                    if not data.has_key(dat):
                                        data[dat] = []
                                        labels.append(dat)
                                        date_order.append(d[1])
                                    data[dat].append([d[0], id, d[2]])

                        elif entity == 'geolevel':
                            # we want geolevels (we might have more than one), and geolevel ids arrive as strings
                            geolevel_ids = filters['geolevel']['values']
                            for g in geolevel_ids:
                                # let's get the geolevel object
                                geolevel = core_controller.get_geolevels(int(g))[0]

                                geolevel_data = sr_controller.get_geolevel_scatter_data(geolevel.id, period)

                                # get the id of this geolevel so we can generate the legend
                                if entities.has_key(geolevel.name):
                                    id = entities[geolevel.name]
                                else:
                                    entity_id = entity_id + 1
                                    id = entity_id
                                    entities[geolevel.name] = id
                                    legend_mapping[id] = geolevel.name

                                for d in geolevel_data: # each d is a list like [score, date_of_audit, count]
                                    dat = self._get_formated_data(d[1])
                                    if not data.has_key(dat):
                                        data[dat] = []
                                        labels.append(dat)
                                        date_order.append(d[1])
                                    data[dat].append([d[0], id, d[2]])

                        elif entity == 'depot':
                            # we want depots (we might have more than one), and depot ids arrive as strings
                            depot_ids = filters['depot']['values']
                            for d in depot_ids:
                                # let's get the geolevel object
                                depot = core_controller.get_depots(int(d))[0]

                                depot_data = sr_controller.get_depot_scatter_data(depot.id, period)

                                # get the id of this depot so we can generate the legend
                                if entities.has_key(depot.name):
                                    id = entities[depot.name]
                                else:
                                    entity_id = entity_id + 1
                                    id = entity_id
                                    entities[depot.name] = id
                                    legend_mapping[id] = depot.name

                                for d in depot_data: # each d is a list like [score, date_of_audit, count]
                                    dat = self._get_formated_data(d[1])
                                    if not data.has_key(dat):
                                        data[dat] = []
                                        labels.append(dat)
                                        date_order.append(d[1])
                                    data[dat].append([d[0], id, d[2]])

            date_order.sort()
            return {
                'labels': labels,
                'period_mapping': self.period_mapping,
                'data': data,
                'messages': messages,
                'legend_mapping': legend_mapping,
                'chart_title': chart_title,

                # formatted dates, sorted, so we can have the dates shown in the chronological order
                'date_order': [self._get_formated_data(d) for d in date_order],

                # periods from the legend, sorted, so we can have the legend items in the correct order
                'period_order': self.period_order
            }

    dispatcher = JsonRpcDispatcher()
    dispatcher.register_instance(TntSrWidgetRpc())
    return dispatcher.handle_request(request)
